/* -*- c-basic-offset: 3 -*-
 *
 * Copyright (c) 2013, GREYC.
 * All rights reserved
 *
 * You may use this file under the terms of the BSD license as follows:
 *
 * "Redistribution and use in source and binary forms, with or without
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, are
 * met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of the GREYC, nor the name of its
 *     contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
 * 
 *   
 * For more information, refer to:
 * https://clouard.users.greyc.fr/Pandore
 */

package fr.greyc.pantheon.ariane;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.Frame;
import java.awt.Image;
import java.awt.Toolkit;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.io.File;
import java.io.IOException;
import java.lang.reflect.Field;
import java.net.MalformedURLException;
import java.net.URI;
import java.net.URISyntaxException;
import java.net.URL;
import java.nio.charset.Charset;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Locale;
import java.util.ResourceBundle;

import javax.imageio.spi.IIORegistry;
import javax.swing.ImageIcon;
import javax.swing.JFrame;
import javax.swing.UIManager;
import javax.swing.UIManager.LookAndFeelInfo;
import javax.swing.UnsupportedLookAndFeelException;
import javax.swing.WindowConstants;
import javax.xml.parsers.ParserConfigurationException;

import org.xml.sax.SAXException;

import fr.greyc.pantheon.ariane.graph.Graph;
import fr.greyc.pantheon.ariane.ui.PreferenceChooser;
import fr.greyc.pantheon.ariane.ui.SplashScreen;
import fr.greyc.pantheon.ariane.ui.Workspace;
import fr.greyc.pantheon.ariane.utils.Tree;
import fr.greyc.pantheon.ariane.xml.XMLElement;
import fr.greyc.pantheon.ariane.xml.XMLFileInputStream;

/**
 * <code>Ariane</code> is the main class of the software.
 * It reads the configuration file and initializes the application.
 * It uses a Splash Screen to pass the time.
 * 
 * @author Régis CLOUARD - Oct 01, 1998
 * @version Sébastien LEFÈVRE, Richard LOBRÉAU - Feb 21, 1999
 * @version Adrien SCHRÈQUE, Hicham AMADI - Apr 11, 2000
 * @version Mathieu JOETS - Sep 07, 2001
 * @version Alban DOUBLET, Guillaume CHATELET - Oct 06, 2002
 * @version Régis CLOUARD - Oct 16, 2001
 * @version Régis CLOUARD - Dec 10, 2004 (Look&Feel Revision)
 * @version Régis CLOUARD - Jan 16, 2007 (add icon)
 * @version Frédéric LAIR - Sep 04 , 2007
 */
public class Ariane {
	/** The base name of the software ... en français dans le texte. */
	public static final String TITLE_BASE = "Au fil d'Ariane ... @ GREYC (ENSICAEN & Université de Caen Basse-Normandie)";

	/** The current version identifier. */
	public static final String VERSION_ID = "3.3.4-1 (Xxx ##, 2014)";

	/** The large list of contributors. */
	public static final String[] AUTHORS = { "Régis CLOUARD",
		"François RIOULT",
		"Sébastien LEFÈVRE (1999)",
		"Richard LOBRÉAU (1999)",
		"Adrien SCHRÈQUE (2000)",
		"Hicham AMADI (2000)",
		"Mathieu JOETS (2001)",
		"Guillaume CHATELET (2002)",
		"Alban DOUBLET (2002)",
		"Frédéric LAIR (2007)",
		"Look and Feel: Nilo J. GONZALEZ (see http://personales.ya.com/nimrod)"
	};

	///// 
	///// C O N S T A N T S
	/////

	/** The default extension for Ariane's filenames (aws::Ariane WorkSpace). */
	public static final String DEFAULT_FILE_EXT = ".aws";

	/** The name of operator library configuration file. */
	public static final String LIBRARY_CONFIG_FILENAME = "ariane.cfg";

	/** The name of the static configuration path: $HOME/.greyc/ariane. */
	private static final String CONFIG_STATIC_DIRNAME = ".greyc" + File.separator + "ariane";

	/** The name of the dynamic configuration path: $JARPATH/.ariane */
	private static final String CONFIG_DYNAMIC_DIRNAME = ".ariane";
	/** The name of the dynamic configuration path: $JARPATH/ariane */
	private static final String EX_CONFIG_DYNAMIC_DIRNAME = "ariane";

	/** The name of the user preferences file. */
	private static final String PREFS_FILENAME = "ariane.ini";

	/** The name of the subroutines file. */
	private static final String SUBROUTINE_FILENAME = "subroutines.cfg";

	/** The name of the operator file. */
	private static final String OPERATOR_FILENAME = "operators.cfg";

	/** The operators library path. */
	private static String _libraryPath = ".";

	/** The configuration path name. */
	private static String _configurePathName; 

	/** 
	 * The documentation path. Each document is
	 * supposed located in _docPath/fr/ or _docPath/en/ or ...
	 */
	private static String _docPath = "";

	/** The HOME path. */
	private static String _homePath;

	/** The jar path. */
	private static String _jarPath;

	/** List of paths to search when loading dynamic libraries. */
	private static String _javaLibraryPath;

	/** The history size for do/undo management. */
	public static final int HISTORYSIZE = 10;

	/** The directory where are stored all icons. */
	private static final String ICON_DIRECTORY = "fr/greyc/pantheon/ariane/properties/images/";

	/** The directory where are stored all resources. */
	private static final String RESOURCE_DIRECTORY = "fr/greyc/pantheon/ariane/properties/locale/";

	/** The name of the resource file. */
	private static final String RESOURCE_FILE = "messages";

	/** The ResourceBundle to load properties. */
	private static ResourceBundle _language;

	/** The URL for manual files. */
	private static String _manualURL;

	/** The URL for properties files. */
	private static String _url;

	/** The main frame that support all other frames. */
	private static Workspace _workspace;

	/** The frame that supports all the Ariane interface. */
	private static GFrame _frame;

	public static Workspace getWorkspace() {
		return _workspace;
	}

	public static GFrame getFrame( ) {
		return _frame;
	}

	public static ImageIcon getIcon( String filename ) {
		URL location = Ariane.class.getClassLoader().getResource(Ariane.ICON_DIRECTORY + filename);
		if (location == null) {
			return null;
		}
		return new ImageIcon(Toolkit.getDefaultToolkit().getImage(location));
	}

	public static String getString( final String name ) {
		return _language.getString(name);
	}

	public static ResourceBundle getLanguage() {
		return _language;
	}

	/**
	 * Sets the global context variables:
	 * homePath for the user directory,
	 * javaLibraryPath : used for setting adding libraries Java can used. 
	 * jarPath : path of the jar file.
	 */
	private void setGlobalContextVariables() {
		_homePath = System.getProperty("user.home", ".");
		_javaLibraryPath = System.getProperty("java.library.path");
		// Tips: getting the absolute file name of the jar.
		Class<?> arianeClass = Ariane.class;
		URL tmp = arianeClass.getResource(arianeClass.getSimpleName() + ".class");
		String filename = tmp.getFile();
		int jarFilenameLength = filename.indexOf('!'); // Jar file uses the ! meta character
		if (jarFilenameLength > 0) {
			filename = filename.substring(0, jarFilenameLength);
		}

		try {
			URL url = new URL(filename);
			_jarPath = new File(url.toURI()).getParent();
		} catch (URISyntaxException e1) {
			_jarPath = new File(filename).getParent();
		} catch (MalformedURLException e2) {
			_jarPath = new File(filename).getParent();
		}
	}

	// For tests
	public Ariane() {
		setGlobalContextVariables();
	}

	public Ariane( String[] args ) {
		String operatorLibraryFilename = null;
		String configDirName = null;
		final List<String> graphFilenames = new ArrayList<String>();

		setGlobalContextVariables();
		// Parse arguments
		boolean doSplash = true;
		for (int i = 0; i < args.length; i++) {
			if (args[i].startsWith("-")) {
				if (args[i].equalsIgnoreCase("-version") || args[i].equalsIgnoreCase("-v")) {
					System.err.println("Ariane \"" + VERSION_ID  + "\"");
					System.exit(0);
				} else if (args[i].equalsIgnoreCase("-nosplash")) {
					doSplash = false;
				} else if (args[i].equalsIgnoreCase("-config")) {
					i++;
					if (i == args.length) {
						System.err.println("No configuration directory specified ... option -config ignored.");
					} else {
						configDirName = args[i];
					}
				} else if (args[i].equalsIgnoreCase("-lib")) {
					i++;
					if (i == args.length) {
						System.err.println("No operator library configuration file specified... option -lib ignored.");
					} else {
						operatorLibraryFilename = args[i];
					}
				} else {
					if (args[i].equalsIgnoreCase("-help") || args[i].equalsIgnoreCase("-h")) {
						System.err.println("Usage: ariane [options] [fichier-graphe ..]");
						System.err.println("Options include: ");
						System.err.println("  -help -h -?	display this message");
						System.err.println("  -version -v	print product version and exit");
						System.err.println("  -nosplash	launch Ariane without the splash screen");
						System.err.println("  -lib <file>	use<fichier> instead of $PANDOREHOME/lib/pandore.cfg");
						System.exit(0);
					} else {
						System.err.println("Option '" + args[i] + "' unknown option, ignored...");
					}
				}
			} else {
				graphFilenames.add(args[i]);
			}
		}
		initializeInterface(doSplash, operatorLibraryFilename, graphFilenames, configDirName);
	}

	/**
	 * Returns the name of the file that stores the preferences.
	 * 
	 * @return the preference file name.
	 */
	public static final String getPreferencesFilename() {
		return _configurePathName + File.separator + PREFS_FILENAME;
	}

	/**
	 * Returns the name of the file that stores the subroutines.
	 * 
	 * @return the absolute file name.
	 */
	public static final String getSubroutineFilename() {
		return _configurePathName + File.separator + SUBROUTINE_FILENAME;
	}

	/**
	 * Returns the name of the file that stores the personal operators.
	 * 
	 * @return the absolute file name.
	 */
	public static final String getPersonalOperatorFilename() {
		return _configurePathName + File.separator + OPERATOR_FILENAME;
	}

	/**
	 * Sets the new path to the library of operators
	 * and modify the related execution environment.
	 *
	 * @param path the new path.
	 */
	public static final void setLibraryPath( String path ) {
		_libraryPath = path;
		System.setProperty("java.library.path", _javaLibraryPath + File.pathSeparatorChar + _libraryPath + File.separatorChar + "lib");
	}

	public static final String getLibraryPath( ) {
		return _libraryPath;
	}

	/**
	 * Sets the new doc path.
	 * 
	 * @param path the new path.
	 */
	public static final void setDocPath( final String path ) {
		_docPath = path;
	}

	/**
	 * Gets the current doc path.
	 * 
	 * @return the current doc path.
	 */
	public static final String getDocPath( ) {
		return _docPath + "/" + PreferenceChooser.getCurrentLocale().getLanguage();
	}

	/**
	 * Sets the current language to the specified language.
	 * Loads related resources form the resource files.
	 * 
	 * @param locale a locale (eg., Fr_fr, En_us).
	 */
	public static final void setLocale( Locale locale ) {
		_url = RESOURCE_DIRECTORY + locale.getLanguage() + "/";
		_language = ResourceBundle.getBundle(_url + RESOURCE_FILE, locale);
		_manualURL = _url + "manual/";
	}

	public static URI getDocumentationURI( String page ) throws URISyntaxException {
		return Ariane._workspace.getClass().getClassLoader().getResource(_manualURL + page).toURI();
	}

	/**
	 * Loads the configuration file.
	 * 
	 * @param configFile the file that contains the configuration.
	 * @return the XML tree of or null.
	 */
	protected static final List<XMLElement> readConfigFile( File configFile ) {
		try {
			return readXMLFile(configFile);
		} catch (IOException e) {
			System.err.println(Ariane.getString("error") + e.getLocalizedMessage());
			return null;
		}
	}

	/**
	 * Loads an XML file.
	 * 
	 * @param configFile the file that contains the configuration.
	 * @return the XML tree of or null.
	 * @throws IOException bad file format.
	 */
	public static final List<XMLElement> readXMLFile( File configFile ) throws IOException {
		try {
			XMLFileInputStream convert = new XMLFileInputStream(configFile);
			convert.parse();
			convert.close();
			return convert.getElements();
		} catch (SAXException | IOException | ParserConfigurationException e) {
			throw new IOException(e);
		}
	}

	/**
	 * Locates the configuration file.
	 * The sequence of configuration files is:
	 * 0: the specified directory.
	 * 1: jarPath/ariane/ariane.ini
	 * 2: $HOME/.greyc/ariane/ariane.ini
	 * and for sake of backwards compatibility
	 * 4: jarPath/.ariane
	 * 4: jarPath/ariane.cfg
	 * 3: $HOME/.greyc/ariane.ini
	 * 5: $HOME/.greyc/ariane.cfg
	 * 6: $HOME/.ariane
	 *
	 * @param dirname a specified configuration directory or null.
	 * @return the configuration file or null.
	 */
	protected final File locateConfigFile( String dirname ) {
		File configurationFile;

		// 0. specified dir name
		if (dirname != null) {
			configurationFile = new File(dirname + File.separator + PREFS_FILENAME);
			if (configurationFile.exists()) {
				_configurePathName = configurationFile.getAbsoluteFile().getParent();
				return configurationFile;
			}
		}

		// 1. jarPath/ariane/ariane.ini (for compatibility)
		configurationFile = new File(_jarPath + File.separator + CONFIG_DYNAMIC_DIRNAME + File.separator + PREFS_FILENAME);
		if (configurationFile.exists()) {
			_configurePathName = configurationFile.getParent();
			return configurationFile;
		}

		// 2. jarPath/ariane/ariane.ini (for compatibility)
		configurationFile = new File(_jarPath + File.separator + EX_CONFIG_DYNAMIC_DIRNAME + File.separator + PREFS_FILENAME);
		if (configurationFile.exists()) {
			_configurePathName = configurationFile.getParent();
			return configurationFile;
		}

		// 3. jarPath/ariane.cfg (for compatibility)
		configurationFile = new File(_jarPath + File.separator + "ariane.cfg");
		if (configurationFile.exists()) {
			_configurePathName = _jarPath + File.separator + CONFIG_DYNAMIC_DIRNAME;
			return configurationFile;
		}

		// 4. jarPath/.ariane
		configurationFile = new File(_jarPath + File.separator + ".ariane");
		if (configurationFile.exists()) {
			_configurePathName = _jarPath + File.separator + CONFIG_DYNAMIC_DIRNAME;
			return configurationFile;
		}

		// 5. $HOME/.greyc/ariane/ariane.ini
		configurationFile = new File(_homePath + File.separator + CONFIG_STATIC_DIRNAME + File.separator + PREFS_FILENAME);
		if (configurationFile.exists()) {
			_configurePathName = configurationFile.getParent();
			return configurationFile;
		}

		// 6. $HOME/.greyc/ariane.cfg
		configurationFile = new File(_homePath + File.separator + ".greyc"  + File.separator + "ariane.cfg");
		if (configurationFile.exists()) {
			_configurePathName = _homePath + File.separator + CONFIG_STATIC_DIRNAME;
			return configurationFile;
		}

		// 7. $HOME/.ariane
		configurationFile = new File(_homePath + File.separator + ".ariane");
		if (configurationFile.exists()) {
			_configurePathName = _homePath + File.separator + CONFIG_STATIC_DIRNAME;
			return configurationFile;
		}

		// 8. default configuration file
		_configurePathName = _homePath + File.separator + CONFIG_STATIC_DIRNAME;
		return null;
	}

	/**
	 * Initializes the interface.
	 * 
	 * @param doSplash does the splash screen?
	 * @param operatorLibraryFilename an operators library configuration filename or null.
	 * @param graphFilenames a set of input graph filenames or null.
	 * @param configDirName a specified configuration directory.
	 */
	private void initializeInterface( boolean doSplash, String operatorLibraryFilename, List<String> graphFilenames, String configDirName ) {
		_frame = new GFrame();
		SplashScreen splash = null;
		installUI();
		if (doSplash) {
			splash = new SplashScreen(ICON_DIRECTORY + "splash.png");
			splash.setVisible(true);
			splash.setAlwaysOnTop(true);
			splash.showProgress("Loading Ariane settings...", 0);
			splash.showProgress("Reading preferences file ...", 10);
		}
		PreferenceChooser.initPreferences();

		File configFile = locateConfigFile(configDirName);
		List<XMLElement> preferenceHandler = null;
		if (configFile != null) {
			preferenceHandler = readConfigFile(configFile);
		}

		String defaultLibraryFilename = null;
		if (preferenceHandler != null) {
			PreferenceChooser.load(preferenceHandler);
			defaultLibraryFilename = getLibraryPath() + File.separator + "lib" + File.separator + LIBRARY_CONFIG_FILENAME;
			File file = new File(defaultLibraryFilename);
			if (!file.exists()) {
				// Keep this for sake of compatibility
				defaultLibraryFilename = getLibraryPath() + File.separator + "lib" + File.separator + "pandore.cfg";
				File file1 = new File(defaultLibraryFilename);
				if (!file1.exists()) {
					defaultLibraryFilename = null;
				}
			}
		}

		// Build the tree of operators
		if (doSplash) {
			splash.showProgress("Reading common operators...", 30);
		}
		List<XMLElement> libraryHandler;
		if (operatorLibraryFilename != null) {
			libraryHandler = readConfigFile(new File(operatorLibraryFilename));
		} else if (defaultLibraryFilename != null && PreferenceChooser.getBoolean(PreferenceChooser.USE_OPERATOR_LIBRARY)) {
			libraryHandler = readConfigFile(new File(defaultLibraryFilename));
		} else {
			libraryHandler = null;
		}
		Tree itemTree = new Tree(null);
		if (libraryHandler != null) {
			itemTree.setCommonOperators(libraryHandler);
		}
		libraryHandler = null;
		// Add personal operators.

		if (doSplash) {
			splash.showProgress("Reading personal operators ...", 40);
		}
		// new version
		List<XMLElement> personalOperatorHandler = null;
		File personalOperatorFile = new File(Ariane.getPersonalOperatorFilename());
		if (personalOperatorFile.exists()) {
			personalOperatorHandler = readConfigFile(personalOperatorFile);
			itemTree.addUserOperators(personalOperatorHandler);
		} else {
			// Keep this for sake of compatibility
			if (preferenceHandler != null) {
				itemTree.addUserOperators(preferenceHandler);
				File directory = personalOperatorFile.getParentFile();
				directory.mkdirs();
				configFile.renameTo(personalOperatorFile);
			}
		}

		// Add the subroutines stored in the init file.
		if (doSplash) {
			splash.showProgress("Reading subroutines ...", 50);
		}
		File subRoutineFile = new File(Ariane.getSubroutineFilename());
		if (subRoutineFile.exists()) {
			List<XMLElement> subRoutineHandler = readConfigFile(subRoutineFile);
			itemTree.addSubroutines(subRoutineHandler);
		}

		if (doSplash) {
			splash.showProgress("Building graphical interfaces ...", 60);
		}
		_workspace = new Workspace(itemTree);

		if (doSplash) {
			splash.showProgress("Initializing windows contents ...", 80);
		}

		// The icon :: should be a gif image.
		ImageIcon logo64 = Ariane.getIcon("logo64.png");
		ImageIcon logo32 = Ariane.getIcon("logo32.png");
		_frame.setIconImages(Arrays.asList(new Image[] {logo64.getImage(), logo32.getImage()}));
		_frame.setJMenuBar(_workspace.buildMenuBar()); 
		_workspace.initContent();
		_frame.setContentPane(_workspace);
		_frame.initFrame();

		Graph[] graphs = null;
		if (!graphFilenames.isEmpty()) {
			if (doSplash) {
				splash.showProgress("Reading the input graph(s) ...", 90);
			}
			graphs = new Graph[graphFilenames.size()];
			for ( int i = 0; i < graphFilenames.size(); i++) {
				graphs[i] = _workspace.loadGraph(new File(graphFilenames.get(i)));
			}
			File file = new File(graphFilenames.get(0));
			Workspace.setCurrentDir(file.getAbsoluteFile().getParent());
		}

		if (doSplash) {
			splash.showProgress("Starting session ...", 100);
		}

		// Close the SplashScreen.
		if (doSplash) {
			splash.setVisible(false);
			splash.dispose();
		}

		if (graphs != null) {
			_workspace.display(graphs);
		}
		_frame.restoreFrameSize();
		_workspace.restoreSplitPanesLocation();
		_frame.setVisible(true);

		// First use of the software: configure directories.
		if (getLibraryPath().equalsIgnoreCase(".") && PreferenceChooser.getBoolean(PreferenceChooser.USE_OPERATOR_LIBRARY)) {
			_workspace.configurationAction();
		}
	}

	/**
	 * Installs the L&F preferences.
	 */
	private void installUI() {
		try {
			UIManager.put("control", new Color(215, 214, 223));
			UIManager.put("nimbusBlueGrey", new Color(167, 167, 190));
			UIManager.put("nimbusBase", new Color(42, 42, 140));
			for (LookAndFeelInfo info : UIManager.getInstalledLookAndFeels()) {
				if ("Nimbus".equals(info.getName())) {
					UIManager.setLookAndFeel(info.getClassName());
					break;
				}
			}
		} catch ( ReflectiveOperationException | UnsupportedLookAndFeelException e) {
		}

		UIManager.put("OptionPane.yesIcon", Ariane.getIcon("gtk-yes.png"));
		UIManager.put("OptionPane.noIcon", Ariane.getIcon("gtk-no.png"));
		UIManager.put("OptionPane.cancelIcon", Ariane.getIcon("gtk-cancel.png"));
		UIManager.put("OptionPane.okIcon", Ariane.getIcon("gtk-apply.png"));
	}

	private static void changeDefaultCharsetSet( String  charsetEncoder ) throws IllegalAccessException, NoSuchFieldException {
		Field defaultCharsetField = Charset.class.getDeclaredField("defaultCharset");
		defaultCharsetField.setAccessible(true);
		defaultCharsetField.set(null, Charset.forName(charsetEncoder));
	}

	/**
	 * Creates the application (i.e., the workspace)
	 * and initializes all the components.
	 * Displays a <code>SplashScreen</code> during initialization process.
	 * 
	 * @param args the list of input arguments.
	 */
	public static void main( String[] args ) {
		try {
			// install imageio.
			changeDefaultCharsetSet("utf-8");
			IIORegistry registry = IIORegistry.getDefaultInstance();
			registry.registerServiceProvider(new com.sun.media.imageioimpl.plugins.tiff.TIFFImageReaderSpi());
		} catch (ReflectiveOperationException e) { 
//			e.printStackTrace();
		}
		new Ariane(args);
	}

	/**
	 * The List of all error codes used in the project.
	 */
	public enum ErrorCode {
		/** Bottom of the stack error. */
		NOERROR,
		/** Missing parameter value. */
		MISSINGPARAMETER,
		/** Missing input data file. */
		MISSINGINPUT,
		/** Command runtime error during execution. */
		RUNTIME,
		/** Input or Output errors during execution. */
		INPUTOUTPUT,
		/** Command execution is interrupted. */
		INTERRUPTEDEXECUTION,
		/** Missing file. */
		MISSINGFILE,
		/** Bad file name. */
		UNKNOWNFILE,
		/** Arithmetics error. */
		MISSINGFORMULA,
		/** Arithmetics error. */
		ARITHMETICS,
	}

	public final class GFrame extends JFrame {
		private Dimension _size;

		public void initFrame( ) {
			setTitle(Ariane.TITLE_BASE);
			_frame.addWindowListener(new WindowAdapter() {
				@Override
				public void windowClosing( WindowEvent event ) {
					if (_workspace.quitConfirmed()) {
						((JFrame)(event.getWindow())).setDefaultCloseOperation(WindowConstants.DISPOSE_ON_CLOSE);
						System.exit(0);
					} else {
						((JFrame)(event.getWindow())).setDefaultCloseOperation(WindowConstants.DO_NOTHING_ON_CLOSE);
					}
				}
			});
		}

		public Dimension getFinalSize() {
			if (getExtendedState() ==  Frame.MAXIMIZED_BOTH) {
				return _size;
			} else {
				return super.getSize();
			}
		}

		private void restoreFrameSize( ) {
			_size = PreferenceChooser.getDimension(PreferenceChooser.FRAME_SIZE);
			if (_size != null) {
				setPreferredSize(_size);
				setSize(_size);
			}
			if (PreferenceChooser.getBoolean(PreferenceChooser.FRAME_MAXIMIZED)) {
				setExtendedState(Frame.MAXIMIZED_BOTH);
			}
			setLocationByPlatform(true);
		}
	}
}
