/* -*- c-basic-offset: 3 -*-
 *
 * Copyright (c) 2013, GREYC.
 * All rights reserved
 *
 * You may use this file under the terms of the BSD license as follows:
 *
 * "Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of the GREYC, nor the name of its
 *     contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
 * 
 *   
 * For more information, refer to:
 * https://clouard.users.greyc.fr/Pandore
 */

package fr.greyc.pantheon.ariane.generator;

import fr.greyc.pantheon.ariane.Ariane;
import fr.greyc.pantheon.ariane.graph.Graph;
import fr.greyc.pantheon.ariane.graph.SubGraph;
import fr.greyc.pantheon.ariane.graph.item.CompositeItem;
import fr.greyc.pantheon.ariane.graph.item.ConstantItem;
import fr.greyc.pantheon.ariane.graph.item.EvalItem;
import fr.greyc.pantheon.ariane.graph.item.ControlStructureForItem;
import fr.greyc.pantheon.ariane.graph.item.ControlStructureIfItem;
import fr.greyc.pantheon.ariane.graph.item.ControlStructureMergeItem;
import fr.greyc.pantheon.ariane.graph.item.ControlStructureSelectItem;
import fr.greyc.pantheon.ariane.graph.item.ControlStructureWhileItem;
import fr.greyc.pantheon.ariane.graph.item.FileNameItem;
import fr.greyc.pantheon.ariane.graph.item.FormatItem;
import fr.greyc.pantheon.ariane.graph.item.GraphItem;
import fr.greyc.pantheon.ariane.graph.item.LoaderItem;
import fr.greyc.pantheon.ariane.graph.item.MacroOperatorItem;
import fr.greyc.pantheon.ariane.graph.item.NoteItem;
import fr.greyc.pantheon.ariane.graph.item.OperatorItem;
import fr.greyc.pantheon.ariane.graph.item.SaverItem;
import fr.greyc.pantheon.ariane.graph.item.ScriptItem;
import fr.greyc.pantheon.ariane.graph.item.SystemItem;
import fr.greyc.pantheon.ariane.graph.item.ViewerItem;
import fr.greyc.pantheon.ariane.graph.item.VirtualItem;
import fr.greyc.pantheon.ariane.graph.slot.InputSlot;
import fr.greyc.pantheon.ariane.graph.slot.InputDataSlot;
import fr.greyc.pantheon.ariane.graph.slot.InputMaskSlot;
import fr.greyc.pantheon.ariane.graph.slot.InputParameterSlot;
import fr.greyc.pantheon.ariane.ui.PreferenceChooser;

import java.text.DateFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.LinkedList;
import java.util.List;
import java.util.StringTokenizer;

/**
 * The <code>ShellGenerator</code> generates bash Shell code
 * for the specified graph.
 * It operates in three steps:<ol>
 * <li>transforms the graph into a list of operators usable for the topological sort;
 * <li>executes the topological sort;
 * <li>generates the related code.
 * </ol>
 *
 * @version Régis CLOUARD - dec 04, 2007
 * @version Régis CLOUARD - Jul 03, 2010 (fixed bugs)
 * @version Régis CLOUARD - Nov 19, 2011 (fixed bugs)
 * @version Régis CLOUARD - Aug 25, 2012 (add system + script items)
 */ 
public final class BashGenerator extends CodeGenerator {
	/** The list of generated output images. */
	private final List<String> _outputImages = new ArrayList<String>();

	protected BashGenerator() {
		// Only created by the code generator.
	}

	/**
	 * Generates the header code of the Shell file and
	 * the code that checks the argument line consistency.
	 * 
	 * @param fp the output stream.
	 * @param filename the name of the code file.
	 * @param arguments the list of arguments defined in the graph.
	 * @param inputs the list of loaders defined in the graph.
	 * @param outputs the list of savers defined in the graph.
	 */
	@Override
	protected void generateScriptHeader( FileOutputCode file, String filename, List<GraphItem> arguments, List<GraphItem> inputs, List<GraphItem> outputs) {
		// The current date.
		Date creationDate = new Date();
		DateFormat dateFormat = DateFormat.getDateInstance(DateFormat.MEDIUM);
		file.println("#!/bin/bash");
		file.println("#");
		file.println("# @(#)" + filename + "				" + dateFormat.format(creationDate));
		file.println("#");
		file.println("# This file was automatically generated by Ariane.");
		file.println("# Ariane is part of the Pantheon Project.");
		file.println("#\n# GREYC - ENSICAEN & University of Caen Basse-Normandie, France");
		file.println("");
		
		file.println("IFS=$(echo -en \"\\n\\b\")");
		file.println("");

		file.println("# Check argument line consistency.");
		file.println("if [ $# -ge 1 -a \"$1\" = \"-h\" -o $# -ne " + (arguments.size() + inputs.size() + outputs.size()) + " ]");
		file.println("then");

		file.print("echo \"Usage: $0 ", 1);
		for (int i = 0; i < arguments.size(); i++) {
			file.print(" " + arguments.get(i).getNickName());
		}
		for (int i = 0; i < inputs.size();) {
			file.print(" input-file" + (++i));
		}
		for (int i = 0; i < outputs.size();) {
			file.print(" output-file" + (++i));
		}
		file.println("\"");
		file.print("echo \"  example: $0", 1);
		for (int i = 0; i < arguments.size(); i++) {
			if (arguments.get(i).getFormula() != null && arguments.get(i).getFormula().length() > 0) {
				file.print(" " + arguments.get(i).getFormula());
			} else {
				file.print(" " + arguments.get(i).getNickName());
			}
		}
		for (int i = 0; i < inputs.size(); i++) {
			file.print(" " + inputs.get(i).getNickName());
		}
		for (int i = 0; i < outputs.size(); i++) {
			String name = outputs.get(i).getNickName();
			if ("saver".equals(name)) {
				name += outputs.get(i).getId();
			}
			file.print(" " + name);
		}
		file.println("\"");

		file.println("exit", 1);
		file.println("fi");
	}

	/**
	 * Generates the footer code of the script file.
	 * 
	 * @param fp the output stream.
	 */
	@Override
	protected void generateScriptFooter( FileOutputCode file ) {
		file.println("\nmain $*");
	}

	/**
	 * Generates utility functions required to execute the script.
	 * 
	 * @param fp the output stream.
	 */
	@Override
	protected void generateUtilitySubroutinesDefinition( FileOutputCode file ) { 
		// Do nothing
	}

	/**
	 * Creates the name of the image from the specified slot number,
	 * item and graph identifier.
	 * 
	 * @param slotIndex the identifier of the slot.
	 * @param itemIndex the identifier of the item.
	 * @param graphNumber the graph identifier
	 * @return a name that is supposed unique.
	 */
	private String createImageName( int slotIndex, int itemIndex, int graphNumber ) {
		String name = PreferenceChooser.createTemporaryName(graphNumber, itemIndex, slotIndex);
		_outputImages.add(name);
		return name;
	}

	/**
	 * Returns the command line for getting the
	 * input images for the specified item slot of the specified
	 * graph identifier.
	 * 
	 * @param slot the related slot.
	 * @param graphNumber the graph identifier
	 * @return a name that is supposed to unique.
	 */
	private String getInputImage( InputSlot slot, int graphNumber ) {
		return PreferenceChooser.createTemporaryName(graphNumber, slot.getConnectedSlot().getGraphItem().getId(), slot.getConnectedSlot().getRank());
	}

	/**
	 * Creates a  name with the specified arguments.
	 * 
	 * @param graph the corresponding graph.
	 * @param item the corresponding item.
	 * @param slot the corresponding slot.
	 * @return the name of the variable.
	 */
	private String variable( int graph, int item, int slot ) {
		return "var" + graph + "_" + item + "_" + slot;
	}

	/**
	 * Returns the parameters line command of the specified item.
	 * 
	 * @param item the related item.
	 * @param slotIndex the slot index.
	 * @param graphId the current graph identifier.
	 * @return a value or a command to get the value.
	 */
	private String getParameter( GraphItem item, int slotIndex, int graphId ) {
		InputParameterSlot isp = (InputParameterSlot) item.getSlot("InSlotParameter", slotIndex);
		if (isp.getConnectedSlot() == null) {
			return "'" + ((InputParameterSlot)item.getSlot("InSlotParameter", slotIndex)).getValue() + "'";
		} else {
			return "${" + variable(graphId, isp.getConnectedSlot().getGraphItem().getId(), isp.getConnectedSlot().getRank()) + "}";
		}
	}

	/**
	 * Returns the parameters line command of the specified item.
	 * 
	 * @param item the related item.
	 * @param slotIndex the slot index.
	 * @param graphId the current graph identifier.
	 * @return a value or a command to get the value.
	 */
	private String getNumericalParameter( GraphItem item, int slotIndex, int graphId ) {
		InputParameterSlot isp = (InputParameterSlot) item.getSlot("InSlotParameter", slotIndex);
		if (isp.getConnectedSlot() == null) {
			return ((InputParameterSlot)item.getSlot("InSlotParameter", slotIndex)).getValue();
		} else {
			return "${" + variable(graphId, isp.getConnectedSlot().getGraphItem().getId(), isp.getConnectedSlot().getRank()) + "}";
		}
	}

	/**
	 * Translates the specified logical formula into Shell.
	 * 
	 * @param formula the logical formula.
	 * @param parameter0 the value of the parameter0.
	 * @param parameter1 the value of the parameter1.
	 * @return the Shell code for the formula.
	 */
	private String translateLogicalFormula( String formula, String parameter0, String parameter1 ) {
		// Parse the boolean expression into Shell.
		char[] formulaCharacters = formula.toCharArray();
		String commandShell = "";
		int index = 0;
		while (index < formulaCharacters.length) {
			if (formulaCharacters[index] == '&') {
				commandShell += " && ";
				index++;
			} else if (formulaCharacters[index] == '|') {
				commandShell += " || ";
				index++;
			} else if (formulaCharacters[index] == '<') {
				if (formulaCharacters[index + 1] == '=') {
					commandShell += " <= ";
					index += 2;
				} else {
					commandShell += " < ";
					index++;
				}
			} else if (formulaCharacters[index] == '>') {
				if (formulaCharacters[index + 1] == '=') {
					commandShell += " >= ";
					index += 2;
				} else {
					commandShell += " > ";
					index++;
				}
			} else if (formulaCharacters[index] == '=') {
				if (formulaCharacters[index + 1] == '=') {
					commandShell += " == ";
					index += 2;
				} else {
					index++; // degenerated case!
				}
			} else if (formulaCharacters[index] == '!') {
				if (formulaCharacters[index + 1] == '=') {
					commandShell += " != ";
					index += 2;
				} else {
					commandShell += " ! ";
					index++;
				}
			} else if (formulaCharacters[index] == '(') {
				commandShell += " ";
				index++;
			} else if (formulaCharacters[index] == ')') {
				commandShell += " ";
				index++;
			} else if (formulaCharacters[index] == '$') {
				if (formulaCharacters[index + 1] == '0') {
					commandShell += parameter0;
				} else {
					commandShell += parameter1;
				}
				index += 2;
			} else {
				commandShell += formulaCharacters[index++];
			}
		}
		return commandShell;
	}

	/**
	 * Generates code for the main.
	 * 
	 * @param graph the graph to encode.
	 * @param fp the output stream.
	 * @param arguments the list of argument.
	 * @param inputs the list of inputs.
	 */
	@Override
	protected void generateMainPrologue( Graph graph, FileOutputCode file, List<GraphItem> arguments, List<GraphItem> inputs, List<GraphItem> outputs ) {
		file.println("\n#\n# The main program\n#");
		file.println("function main() {");
		file.println("rm -f " + PreferenceChooser.getTemporaryFilePrefix() + "?*"  + PreferenceChooser.getTemporaryFileExtension(), 1);
		for ( GraphItem item : arguments ) {
			file.println(variable(1, item.getId(), 0) + "=$1; shift", 1);
		}
		for ( GraphItem item : inputs ) {
			file.println("ln -sf $1 " + createImageName(0, item.getId(), 1) + "; shift", 1);
		}
	}

	/**
	 * Generates code to erase all temporary files.
	 * 
	 * @param graph the graph to encode.
	 * @param fp the output stream.
	 */
	@Override
	protected void generateMainEpilogue( Graph graph, FileOutputCode file ) {
		//		file.println("rm -f "+PreferenceChooser.TEMPORARY_FILE_PREFIX+"*"+PreferenceChooser.TEMPORARY_FILE_EXT, 1);
		file.println("}");
	}

	/**
	 * Generates the code for loading an image.
	 * 
	 * @param item the item to encode.
	 * @param graphNumber the identifier number of the graph.
	 * @param fp the output stream.
	 * @param indent the indentation level.
	 * @throws Exception Incomplete graph.
	 */
	@Override
	protected void generateItem( LoaderItem item, int graphNumber, FileOutputCode file, int indent ) throws BadGraphException {
		if (item.getInputParameterSlots().size() > 0 && item.getInputParameterSlots().get(0).isConnected()) {
			String parameters = getParameter(item, 0 , graphNumber);
			file.println("cp " + parameters + " " + createImageName(0, item.getId(), graphNumber), indent);
		}
	}

	/**
	 * Generates the code for saving an image.
	 * 
	 * @param item the item to encode.
	 * @param graphNumber the identifier number of the graph.
	 * @param fp the output stream.
	 * @param indent the indentation level.
	 * @throws Exception Incomplete graph.
	 */
	@Override
	protected void generateItem( SaverItem item, int graphNumber, FileOutputCode file, int indent ) throws BadGraphException {
		if (item.getInputDataSlots().get(0).getConnectedSlot() != null) {
			if (item.getInputParameterSlots().size() > 0 && item.getInputParameterSlots().get(0).isConnected()) {
				String parameters = getParameter(item, 0, graphNumber);
				file.println("cp " + getInputImage(item.getInputDataSlots().get(0), graphNumber) + " " + parameters, indent);
			} else {
				file.println("cp " + getInputImage(item.getInputDataSlots().get(0), graphNumber) + " $1; shift", indent);
			}
			file.println("ln -sf "
					+ getInputImage(item.getInputDataSlots().get(0), graphNumber)
					+ " "
					+ createImageName(0, item.getId(), graphNumber)
					+ ";", indent);
		}
	}

	/**
	 * Generates the code for viewer item.
	 * 
	 * @param item the item to encode.
	 * @param graphNumber the identifier number of the graph.
	 * @param fp the output stream.
	 * @param indent the indentation level.
	 * @throws Exception Incomplete graph.
	 */
	@Override
	protected void generateItem( ViewerItem item, int graphNumber, FileOutputCode file, int indent ) throws BadGraphException {
		String commandShell = item.getViewerCommandName();
		if (item.getInputDataSlots().get(0).getConnectedSlot() != null) {
			file.println(commandShell + " " + getInputImage(item.getInputDataSlots().get(0), graphNumber), indent);
			file.println("ln -sf "
					+ getInputImage(item.getInputDataSlots().get(0), graphNumber)
					+ " "
					+ createImageName(0, item.getId(), graphNumber)
					+ ";", indent);
		}
	}

	/**
	 * Generates the code for executing an operator.
	 * 
	 * @param item the operator to encode.
	 * @param graphNumber the identifier number of the graph.
	 * @param fp the output stream.
	 * @param indent the indentation level.
	 * @throws Exception Incomplete graph.
	 */
	@Override
	protected void generateItem( OperatorItem item, int graphNumber, FileOutputCode file, int indent ) throws BadGraphException {
		String codeName = OperatorItem.getCommandPart(item.getBinary());
		if (codeName == null) {
			throw new BadGraphException(Ariane.getString("Generator.error2") + item.getName());
		}

		// Get parameters
		List<String> parameterLine = new LinkedList<String>();
		for (int i = 0; i < item.getInputParameterSlots().size(); i++) {
			parameterLine.add(getParameter(item, i, graphNumber));
		}

		// Get mask
		List<String> maskLine = new LinkedList<String>();
		InputMaskSlot mask = item.getInputMaskSlot();
		if (mask != null && mask.isConnected() && _outputImages.contains(getInputImage(mask, graphNumber))) {
			maskLine.add(getInputImage(mask, graphNumber));
		}

		// Get input images.
		List<String> inputLine = new LinkedList<String>();
		for (InputDataSlot image : item.getInputDataSlots() ) {
			if (image.getConnectedSlot() == null) {
				throw new BadGraphException(item.getNickName() + ": " + Ariane.getString("Generator.error.incomplete"));
			}
			inputLine.add(getInputImage(image, graphNumber));
		}

		// Get output images.
		List<String> outputLine = new LinkedList<String>();
		for (int i = 0; i < item.getOutputDataSlots().size(); i++) {
			outputLine.add(createImageName(i, item.getId(), graphNumber));
		}

		String code = "";
		List<String> commandLine = item.buildAbstractCommandLine(maskLine, parameterLine, inputLine, outputLine);
		for (String command : commandLine) {
			code += command + " ";
		}

		// Get output result.
		file.print(code, indent);
		file.println("; " + variable(graphNumber, item.getId(), 0) + "=`" + GraphItem.getCommandStatus() + "`");
	}

	/**
	 * Generates the code for executing an operating system command.
	 * 
	 * @param item the operator to encode.
	 * @param graphNumber the identifier number of the graph.
	 * @param fp the output stream.
	 * @param indent the indentation level.
	 * @throws Exception Incomplete graph.
	 */
	@Override
	protected void generateItem( SystemItem item, int graphNumber, FileOutputCode file, int indent ) throws BadGraphException {
		String codeName = OperatorItem.getCommandPart(item.getBinary());
		if (codeName == null) {
			throw new BadGraphException(Ariane.getString("Generator.error2") + item.getName());
		}

		// Get parameters
		List<String> parameterLine = new LinkedList<String>();
		for (int i = 0; i < item.getInputParameterSlots().size(); i++) {
			parameterLine.add(getParameter(item, i, graphNumber));
		}

		// Get mask
		List<String> maskLine = new LinkedList<String>();
		InputMaskSlot mask = item.getInputMaskSlot();
		if (mask != null && mask.isConnected() && _outputImages.contains(getInputImage(mask, graphNumber))) {
			maskLine.add(getInputImage(mask, graphNumber));
		}

		// Get input images.
		List<String> inputLine = new LinkedList<String>();
		for (InputDataSlot image : item.getInputDataSlots() ) {
			if (image.getConnectedSlot() == null) {
				throw new BadGraphException(item.getNickName() + ": " + Ariane.getString("Generator.error.incomplete"));
			}
			inputLine.add(getInputImage(image, graphNumber));
		}

		// Get output images.
		List<String> outputLine = new LinkedList<String>();
		for (int i = 0; i < item.getOutputDataSlots().size(); i++) {
			outputLine.add(createImageName(i, item.getId(), graphNumber));
		}

		String code = "";
		List<String> commandLine = item.buildAbstractCommandLine(maskLine, parameterLine, inputLine, outputLine);
		for ( String command : commandLine) {
			code += command + " ";
		}

		// Get output result.
		file.print(code, indent);
		file.println("; " + variable(graphNumber, item.getId(), 0) + "=`" + GraphItem.getCommandStatus() + "`");
	}

	/**
	 * Parses the give formula into shell code.
	 * 
	 * @param formula the function.
	 * @param parameters the list of parameters.
	 * @return the shell code.
	 */
	private String parseFormula( String formula, String[] parameters ) {
		String code = "";
		StringTokenizer stingTokenizer = new StringTokenizer(formula, "$ -+()*/^0123456789.E", true);
		String etape;
		while (stingTokenizer.hasMoreTokens()) {
			etape = stingTokenizer.nextToken();
			if ("e".equals(etape) || "E".equals(etape)) {
				code += "*10^";
			} else if ("cos".equals(etape)) {
				code += "c";
			} else if ("sin".equals(etape)) {
				code += "s";
			} else if ("ln".equals(etape)) {
				code += "l";
			} else if ("exp".equals(etape)) {
				code += "e";
			} else if ("atan".equals(etape)) {
				code += "a";
			} else if ("$".equals(etape)) {
				int numberOfParameters = Integer.parseInt(stingTokenizer.nextToken());
				code += parameters[numberOfParameters];
			} else {
				code += etape;
			}
		}
		return code;
	}

	/**
	 * Parses the give formula into shell code.
	 * 
	 * @param formula the function.
	 * @return the shell code.
	 */
	private String setFunction( String formula ) {
		String code = "";
		StringTokenizer stringTokenizer = new StringTokenizer(formula, "$ -+()*/^0123456789.E", true);
		String etape;
		boolean abs = false;
		boolean max = false;
		boolean min = false;
		boolean tan = false;
		boolean log2 = false;
		boolean log10 = false;
		boolean acos = false;
		boolean asin = false;

		while (stringTokenizer.hasMoreTokens()) {
			etape = stringTokenizer.nextToken();
			if ("abs".equals(etape) && !abs) {
				code += " define abs(x){; if (x>=0) return(x) else return (-x)} ";
				abs = true;
			} else if ("max".equals(etape) && !max) {
				code += " define max(x,y){; if (x>y) return(x) else return(y)} ";
				max = true;
			} else if ("min".equals(etape) && !min) {
				code += " define min(x,y){; if (x<y) return(x) else return(y)} ";
				min = true;
			} else if ("atan".equals(etape) && !tan) {
				code += " define tan(x){; return (s(x)/c(x)} ";
				tan = true;
			} else if ("acos".equals(etape) && !acos) {
				code += " define acos(x){; if (x==1) return(0) if (x==-1) return (4*a(1)) return (2*a(1)-a(x/sqrt(1-(x^2))))} ";
				acos = true;
			} else if ("asin".equals(etape) && !asin) {
				code += " define asin(x){; if (x==1) return(2*a(1)) if (x==-1) return (-2*a(1)) return (a(x/sqrt(1-(x^2))))} ";
				asin = true;
			} else if ("log".equals(etape)) {
				etape = stringTokenizer.nextToken();
				if ("2".equals(etape) && !log2) {
					code += " define log2(x){; return l(x)/l(2)} ";
					log2 = true;
				} else if ("1".equals(etape) && !log10) {
					code += " define log10(x){; return l(x)/l(10)} ";
					log10 = true;
				}
			}
		}
		return code;
	}

	/**
	 * Generates the code for an Eval item.
	 * 
	 * @param item the item to encode.
	 * @param graphNumber the identifier number of the graph.
	 * @param fp the output stream.
	 * @param indent the indentation level.
	 * @throws Exception Incomplete graph.
	 */
	@Override
	protected void generateItem( EvalItem item, int graphNumber, FileOutputCode file, int indent ) throws BadGraphException {
		String[] parameters = new String[item.getInputParameterSlots().size()];
		// Get parameters
		for (int i = 0; i < item.getInputParameterSlots().size(); i++) {
			parameters[i] = getParameter(item, i, graphNumber);
		}

		file.println(variable(graphNumber, item.getId(), 0) + "=" + " $(echo \""
				+ setFunction(item.getFormula())
				+ parseFormula(item.getFormula(), parameters)
				+ "\"|bc -l);", indent);
	}

	@Override
	protected void generateItem( FormatItem item, int graphNumber, FileOutputCode file, int indent ) throws BadGraphException {
		String[] parameters = new String[item.getInputParameterSlots().size()];
		// Get parameters
		for (int i = 0; i < item.getInputParameterSlots().size(); i++) {
			parameters[i] = getParameter(item, i, graphNumber);
		}

		String code = "";
		StringTokenizer stringTokenizer = new StringTokenizer(item.getFormula(), " $()+'\"", true);
		String etape;
		while (stringTokenizer.hasMoreTokens()) {
			etape = stringTokenizer.nextToken();
			if ("+".equals(etape)) {
				code += "";
			} else if ("quote".equals(etape)) {
				code += "";
			} else if ("\"".equals(etape)) {
				code += "";
			} else if ("(".equals(etape)) {
				code += "\'";
			} else if (")".equals(etape)) {
				code += "\'";
			} else if ("$".equals(etape)) {
				int numberOfParameters = Integer.parseInt(stringTokenizer.nextToken());
				// Remove the quotes.
				String parameter = parameters[numberOfParameters].substring(1, parameters[numberOfParameters].length() - 1);
				code += parameter;
			} else {
				code += etape;
			}
		}
		file.println(variable(graphNumber, item.getId(), 0) + "=\"" + code + "\"", indent);
	}

	/**
	 * Generates the code for the control structure Select.
	 * 
	 * @param item the select to encode.
	 * @param graphNumber the identifier number of the graph.
	 * @param fp the output stream.
	 * @param indent the indentation level.
	 * @throws Exception Incomplete graph.
	 */
	@Override
	protected void generateItem( ControlStructureSelectItem item, int graphNumber, FileOutputCode file, int indent ) throws BadGraphException {
		String value = getParameter(item, 0, graphNumber);
		file.println("# Select", indent);
		if (value == null) {
			value = variable(graphNumber, item.getId(), 0);
			file.println("read " + value, indent);
			value = "$" + value;
		}
		for (int i = 0; i < item.getInputDataSlots().size(); i++) {
			file.println("if [ " + value + " -eq " + i + " ]", indent);
			file.println("then", indent);
			file.println("ln -sf "
					+ getInputImage(item.getInputDataSlots().get(i), graphNumber) + " "
					+ createImageName(0, item.getId(), graphNumber), indent + 1);
			file.println(variable(graphNumber, item.getId(), 0) + "=" + i + ";", indent + 1);
			file.println("fi", indent);
		}
	}

	/**
	 * Generates the code for the control structure If.
	 * 
	 * @param item the item to encode.
	 * @param graphNumber the identifier number of the graph.
	 * @param fp the output stream.
	 * @param indent the indentation level.
	 * @throws Exception Incomplete graph.
	 */
	@Override
	protected void generateItem( ControlStructureIfItem item, int graphNumber, FileOutputCode file, int indent ) throws BadGraphException {
		String[] parameters = new String[2];
		for (int i = 0; i < item.getInputParameterSlots().size(); i++) {
			parameters[i] = getNumericalParameter(item, i, graphNumber);
		}
		String commandShell = translateLogicalFormula(item.getFormula(), parameters[0], parameters[1]);
		file.println("if [ $( echo \"" + commandShell + "\" |bc -l) -gt 0 ]", indent);
	}

	/**
	 * Generates the code for the 'then' part.
	 * 
	 * @param item the item to encode.
	 * @param graphNumber the identifier number of the graph.
	 * @param fp the output stream.
	 * @param indent the indentation level.
	 * @throws Exception Incomplete graph.
	 */
	@Override
	protected void generateThenPart( ControlStructureIfItem item, int graphNumber, FileOutputCode file, int indent) throws BadGraphException {
		file.println("then", indent);
		file.print("ln -sf " + getInputImage(item.getInputDataSlots().get(0), graphNumber) + " "
				+ createImageName(0, item.getId(), graphNumber), indent + 1);
		file.println("; " + variable(graphNumber, item.getId(), 0) + "=0");
	}

	/**
	 * Generates the code for the 'else' part.
	 * 
	 * @param item the item to encode.
	 * @param graphNumber the identifier number of the graph.
	 * @param fp the output stream.
	 * @param indent the indentation level.
	 * @throws Exception Incomplete graph.
	 */
	@Override
	protected void generateElsePart( ControlStructureIfItem item, int graphNumber, FileOutputCode file, int indent) throws BadGraphException {
		file.println("else", indent);
		file.print("ln -sf " + getInputImage(item.getInputDataSlots().get(0), graphNumber) + " "
				+ createImageName(1, item.getId(), graphNumber), indent + 1);
		file.println("; " + variable(graphNumber, item.getId(), 0) + "=1");
	}

	/**
	 * Generates the code for the end of a if control structure.
	 * 
	 * @param item the item to encode.
	 * @param graphNumber the identifier number of the graph.
	 * @param fp the output stream.
	 * @param indent the indentation level.
	 * @throws Exception Incomplete graph.
	 */
	@Override
	protected void generateFiPart( ControlStructureIfItem item, int graphNumber, FileOutputCode file, int indent) throws BadGraphException {
		file.println("fi", indent);
	}

	/**
	 * Generates the code for the control structure merge.
	 * 
	 * @param item the item to encode.
	 * @param graphNumber the identifier number of the graph.
	 * @param fp the output stream.
	 * @param indent the indentation level.
	 * @throws Exception Incomplete graph.
	 */
	@Override
	protected void generateItem( ControlStructureMergeItem item, int graphNumber, FileOutputCode file, int indent ) throws BadGraphException {
		// input image
		InputDataSlot in0;
		in0 = item.getInputDataSlots().get(0);
		int id0;
		id0 = in0.getConnectedSlot().getGraphItem().getId();
		file.print("if [ -n \"${" + variable(graphNumber, id0, 0) + "}\" -a ", indent);
		file.println("-r " + getInputImage(item.getInputDataSlots().get(0), graphNumber) + " ]");
		file.println("then", indent);
		file.print("cp "
				+ getInputImage(item.getInputDataSlots().get(0), graphNumber)
				+ " "
				+ createImageName(0, item.getId(), graphNumber), indent + 1);
		file.println("; " + variable(graphNumber, item.getId(), 0) + "=0;");
		file.println("rm " + getInputImage(item.getInputDataSlots().get(0), graphNumber), indent + 1);
		file.println("else", indent);
		file.print("cp "
				+ getInputImage(item.getInputDataSlots().get(1), graphNumber) + " "
				+ createImageName(0, item.getId(), graphNumber), indent + 1);
		file.println("; " + variable(graphNumber, item.getId(), 0) + "=1;");
		file.println("rm " + getInputImage(item.getInputDataSlots().get(1), graphNumber), indent + 1);
		file.println("fi", indent);
	}

	/**
	 * Generates the code for a note. Nothing to do.
	 * 
	 * @param item the note to encode.
	 * @param graphNumber the identifier number of the graph.
	 * @param fp the output stream.
	 * @param indent the indentation level.
	 */
	@Override
	protected void generateItem( NoteItem item, int graphNumber, FileOutputCode file, int indent ) {
		// Nothing to do
	}

	/**
	 * Generates the code for a <code>Filename</code> item.
	 * 
	 * @param item the item to encode.
	 * @param graphNumber the identifier number of the graph.
	 * @param fp the output stream.
	 * @param indent the indentation level.
	 * @throws Exception Incomplete graph.
	 */
	@Override
	protected void generateItem( FileNameItem item, int graphNumber, FileOutputCode file, int indent ) throws BadGraphException {
		if (!item.isArgument()) {
			file.println(variable(graphNumber, item.getId(), 0) + "='" + item.getFormula() + "'", indent);
		}
	}

	/**
	 * Generates the code for a <code>Constant</code>item.
	 * 
	 * @param item the item to encode.
	 * @param graphNumber the identifier number of the graph.
	 * @param fp the output stream.
	 * @param indent the indentation level.
	 * @throws Exception Incomplete graph.
	 */
	@Override
	protected void generateItem( ConstantItem item, int graphNumber, FileOutputCode file, int indent ) throws BadGraphException {
		if (!item.isArgument() ) {
			String value = item.getFormula().replaceAll("\"", "\\\\\"");
			file.println(variable(graphNumber, item.getId(), 0) + "=\"" + value + "\"", indent);
		}
	}

	/**
	 * Generates the code for a <code>Script</code>item.
	 * 
	 * @param item the item to encode.
	 * @param graphNumber the identifier number of the graph.
	 * @param fp the output stream.
	 * @param indent the indentation level.
	 * @throws Exception Incomplete graph.
	 */
	@Override
	protected void generateItem( ScriptItem item, int graphNumber, FileOutputCode file, int indent ) throws BadGraphException {
		String value = item.getFormula().replaceAll("\"", "\\\\\"");
		file.println(variable(graphNumber, item.getId(), 0) + "=\"" + value + "\"", indent);
	}

	/**
	 * Generates the code for the specified MacroOperator,
	 * for, or while call.
	 * 
	 * @param item the item to encode.
	 * @param graphNumber the identifier number of the graph.
	 * @param fp the output stream.
	 * @param indent the indentation level.
	 * @throws Exception in case of incomplete graph.
	 */
	@Override
	protected void generateItem( CompositeItem item, int graphNumber, FileOutputCode file, int indent ) throws BadGraphException {
		String code = "macro" + item.getId() + "_" + item.getGraph().getIdentifier();

		// Get parameters
		for (int i = 0; i < item.getInputParameterSlots().size(); i++) {
			code += " " + getParameter(item, i, graphNumber);
		}

		// Get mask
		InputMaskSlot mask = item.getInputMaskSlot();
		if (mask != null && mask.isConnected() && _outputImages.contains(getInputImage(mask, graphNumber))) {
			code += " " + getInputImage(mask, graphNumber);
		}

		// Get input images.
		for ( InputDataSlot image : item.getInputDataSlots() ) {
			if (image.getConnectedSlot() == null) {
				throw new BadGraphException(item.getNickName() + ": " + Ariane.getString("Generator.error.incomplete"));
			}
			code += " " + getInputImage(image, graphNumber);
		}

		// Get output images.
		for (int i = 0; i < item.getOutputDataSlots().size(); i++) {
			code += " " + createImageName(i, item.getId(), graphNumber);
		}
		// Get output result.
		file.println(code + "; " + variable(graphNumber, item.getId(), 0) + "=`" + GraphItem.getCommandStatus() + "`", indent);
	}

	/**
	 * Parses the comment in order to add a sharp after each newlines.
	 * 
	 * @param comment a string that contains the comment.
	 * @return a new string.
	 */
	private String parseComment( String comment ) {
		return comment.replaceAll("\n", "\n# ");
	}

	/**
	 * Generates the code for the specified control structure For.
	 * 
	 * @param item the item to encode.
	 * @param fp the output stream.
	 * @throws Exception Incomplete graph.
	 */
	@Override
	protected void generateSubGraph( ControlStructureForItem item, FileOutputCode file ) throws BadGraphException {
		SubGraph subGraph = item.getSubGraph();
		List<GraphItem> litems = buildItemsSequence(subGraph);

		// Get and discard the virtual operator
		Object virtualOperator = getVirtualOperator(litems);
		litems.remove(virtualOperator);
		if (litems.isEmpty()) {
			throw new BadGraphException(item.getNickName() + ": " + Ariane.getString("Generator.error.incomplete"));
		}

		file.println("\n#\n# " + item.getNickName());
		for (GraphItem note : subGraph.getGraphItems()) {
			if (note instanceof NoteItem) {
				file.println("# " + parseComment(note.getFormula()));
			}
		}
		file.println("#");

		file.println("function macro" + item.getId() + "_" + item.getGraph().getIdentifier() + "() {");

		// Get parameters
		file.println("# Parameters", 1);
		file.println("begin=$1; shift", 1);
		file.println("end=$1; shift", 1);
		file.println("step=$1; shift", 1);
		
		int subgraphNumber = subGraph.getIdentifier();
		for (int i = 3; i < item.getInputParameterSlots().size(); i++) {
			file.println(variable(subgraphNumber, 0, i - 2) + "=$1; shift", 1);
		}

		// Get mask
		file.println("# Input data files", 1);
		InputMaskSlot mask = item.getInputMaskSlot();
		if (mask != null && mask.isConnected() && _outputImages.contains(getInputImage(mask, item.getGraph().getIdentifier()))) {
			int maskIndex = subGraph.getVirtualOperator().getVirtualInputMaskSlot().getRank();
			file.println("ln -fs $1 " + createImageName(maskIndex, 0, subgraphNumber) + "; shift", 1);
		}

		// Get input images
		List<InputDataSlot> inputImages = item.getInputDataSlots();
		for (int i = 0; i < inputImages.size(); i++) {
			InputDataSlot isi = inputImages.get(i);
			// Unlinked input!
			if (isi.getConnectedSlot() == null) {
				throw new BadGraphException(item.getNickName() + ": " + Ariane.getString("Generator.error.incomplete"));
			}
			file.println("cp $1 " + createImageName(i, 0, subgraphNumber) + "; shift", 1);
		}
		file.println("");
		// Create the output images in case of no iteration
		VirtualItem virtualItem = (VirtualItem)virtualOperator;
		for (int i = 0; i < item.getOutputDataSlots().size(); i++) {
			InputDataSlot isi = virtualItem.getVirtualOutputDataSlots().get(i);
			file.println("touch " + getInputImage(isi, subgraphNumber), 1);
		}

		// Initialize the return variable in case of no iteration
		{
			InputParameterSlot isp = (InputParameterSlot)virtualItem.getVirtualOutputResultSlot();
			file.println(variable(subgraphNumber, isp.getConnectedSlot().getGraphItem().getId(), isp.getConnectedSlot().getRank()) + "=SUCCESS", 1);
		}

		file.println("");
		file.println("local i", 1);
		file.println("for ((i=$begin; i<=$end; i+=$step ))", 1);
		file.println("do", 1);
		file.println(variable(subgraphNumber, 0, 0) + "=$i;", 2);
		generateSubroutineBody(litems, subgraphNumber, file, 2);

		// Set intermediate images
		for (int i = 0; i < item.getInputDataSlots().size(); i++) {
			InputDataSlot isi = virtualItem.getVirtualIntermediateDataSlots(i);
			// Unlinked input!
			if (isi.getConnectedSlot() == null) {
				throw new BadGraphException(item.getNickName() + ": " + Ariane.getString("Generator.error.incomplete"));
			}
			String input = getInputImage(isi, subgraphNumber);
			String output = createImageName(i, 0, subgraphNumber);
			if (!input.equals(output)) {
				file.println("cp " + input + " " + output, 2);
			}
		}

		file.println("done", 1);

		file.println("");
		// Get output images
		for (int i = 0; i < item.getOutputDataSlots().size(); i++) {
			InputDataSlot isi = virtualItem.getVirtualOutputDataSlots().get(i);
			file.println("ln -fs " + getInputImage(isi, subgraphNumber) + " $1; shift", 1);
		}
		// Output result
		InputParameterSlot isp = (InputParameterSlot)virtualItem.getVirtualOutputResultSlot();
		if (isp.getConnectedSlot() == null) {
			file.println("psetstatus SUCCESS", 1);
		} else {
			file.println("psetstatus ${" + variable(subgraphNumber, isp.getConnectedSlot().getGraphItem().getId(), isp.getConnectedSlot().getRank()) + "}", 1);
		}
		file.println("}");

		generateSubroutinesDefinition(litems, file);
	}

	/**
	 * Generates the code for the specified control structure While.
	 * 
	 * @param item the item to encode.
	 * @param fp the output stream.
	 * @throws Exception Incomplete graph.
	 */
	@Override
	protected void generateSubGraph( ControlStructureWhileItem item, FileOutputCode file ) throws BadGraphException {
		SubGraph subGraph = item.getSubGraph();
		List<GraphItem> litems = buildItemsSequence(subGraph);

		// Get and discard the virtual operator
		Object virtualOperator = getVirtualOperator(litems);
		litems.remove(virtualOperator);
		if (litems.isEmpty()) {
			throw new BadGraphException(item.getNickName() + ": " + Ariane.getString("Generator.error.incomplete"));
		}

		file.println("\n#\n# " + item.getNickName());
		for (GraphItem note : subGraph.getGraphItems()) {
			if (note instanceof NoteItem) {
				file.println("# " + parseComment(note.getFormula()));
			}
		}
		file.println("#");

		file.println("function macro" + item.getId() + "_" + item.getGraph().getIdentifier() + "() {");

		// Get parameters
		file.println("# Parameters", 1);
		int subgraphNumber = subGraph.getIdentifier();
		String[] parameters = new String[item.getInputParameterSlots().size()];
		for (int i = 0; i < item.getInputParameterSlots().size(); i++) {
			file.println(variable(subgraphNumber, 0, i) + "=$1; shift", 1);
			parameters[i] = "${" + variable(subgraphNumber, 0, i) + "}";
		}

		// Get mask.
		file.println("# Input data files", 1);
		InputMaskSlot mask = item.getInputMaskSlot();
		if (mask != null && mask.isConnected() && _outputImages.contains(getInputImage(mask, item.getGraph().getIdentifier()))) {
			int maskIndex = subGraph.getVirtualOperator().getVirtualInputMaskSlot().getRank();
			file.println("ln -fs $1 " + createImageName(maskIndex, 0, subgraphNumber) + "; shift", 1);
		}

		// Get input images.
		List<InputDataSlot> inputImages = item.getInputDataSlots();
		for (int i = 0; i < inputImages.size(); i++) {
			InputDataSlot isi = inputImages.get(i);
			// Unlinked input!
			if (isi.getConnectedSlot() == null) {
				throw new BadGraphException(item.getNickName() + ": " + Ariane.getString("Generator.error.incomplete"));
			}
			file.println("cp $1 " + createImageName(i, 0, subgraphNumber) + "; shift", 1);
		}

		// Parse the boolean expression.
		String testFormula  = translateLogicalFormula(item.getFormula(), parameters[0], parameters[1]);
		VirtualItem virtualItem = (VirtualItem)virtualOperator;
		file.println("");
		// Create output images in case of no repetition
		for (int i = 0; i < item.getOutputDataSlots().size(); i++) {
			InputDataSlot isi = virtualItem.getVirtualOutputDataSlots().get(i);
			file.println("touch " + getInputImage(isi, subgraphNumber), 1);
		}

		// Initialize the return variable in case of no iteration
		{
			InputParameterSlot isp = (InputParameterSlot)virtualItem.getVirtualOutputResultSlot();
			file.println(variable(subgraphNumber, isp.getConnectedSlot().getGraphItem().getId(), isp.getConnectedSlot().getRank()) + "=SUCCESS", 1);
		}

		file.println("");
		file.println("while [ $( echo \"" + testFormula + "\" |bc -l) -gt 0 ]", 1);
		file.println("do", 1);
		generateSubroutineBody(litems, subgraphNumber, file, 2);

		// Get intermediate images
		for (int i = 0; i < item.getInputDataSlots().size(); i++) {
			InputDataSlot isi = virtualItem.getVirtualIntermediateDataSlots(i);
			// Unlinked input!
			if (isi.getConnectedSlot() == null) {
				throw new BadGraphException(item.getNickName() + ": " + Ariane.getString("Generator.error.incomplete"));
			}
			// Discard cp image image (it happens with the output = input)
			String input = getInputImage(isi, subgraphNumber);
			String output = createImageName(i, 0, subgraphNumber);
			if (!input.equals(output)) {
				file.println("cp " + input + " " + output, 2);
			}
		}

		// Get intermediate parameters
		for (int i = 0; i < 2; i++) {
			InputParameterSlot isp = virtualItem.getVirtualIntermediateResultSlot(i);
			if (isp.getConnectedSlot() != null) {
				file.println(variable(subgraphNumber, 0, i)
						+ "=${"
						+ variable(subgraphNumber, isp.getConnectedSlot().getGraphItem().getId(), isp.getConnectedSlot().getRank())
						+ "}", 2);
			}
		}
		file.println("done", 1);
		// Get output images
		for (int i = 0; i < item.getOutputDataSlots().size(); i++) {
			InputDataSlot isi = virtualItem.getVirtualOutputDataSlots().get(i);
			file.println("ln -sf " + getInputImage(isi, subgraphNumber) + " $1; shift", 1);
		}

		// Output result.
		InputParameterSlot isp = (InputParameterSlot)virtualItem.getVirtualOutputResultSlot();
		if (isp.getConnectedSlot() == null) {
			file.println("psetstatus SUCCESS", 1);
		} else {
			file.println("psetstatus ${" + variable(subgraphNumber, isp.getConnectedSlot().getGraphItem().getId(), isp.getConnectedSlot().getRank()) + "}", 1);
		}
		file.println("}");

		generateSubroutinesDefinition(litems, file);
	}

	/**
	 * Generates the code for the specified MacroOperator body.
	 * 
	 * @param item the item to encode.
	 * @param fp the output stream.
	 * @throws Exception In case of incomplete graph.
	 */
	@Override
	protected void generateSubGraph( MacroOperatorItem item, FileOutputCode file ) throws BadGraphException {
		SubGraph subGraph = item.getSubGraph();
		List<GraphItem> litems = buildItemsSequence(subGraph);

		// Get and discard the virtual operator.
		Object virtualOperator = getVirtualOperator(litems);
		litems.remove(virtualOperator);
		if (litems.isEmpty()) {
			throw new BadGraphException(item.getNickName() + ": " + Ariane.getString("Generator.error.incomplete"));
		}

		file.println("\n#\n# " + item.getNickName());
		for (GraphItem note : subGraph.getGraphItems()) {
			if (note instanceof NoteItem) {
				file.println("# " + parseComment(note.getFormula()));
			}
		}
		file.println("#");

		file.println("function macro" + item.getId() + "_" + item.getGraph().getIdentifier() + "() {");

		// Get parameters.
		file.println("# arguments", 1);
		int subgraphNumber = subGraph.getIdentifier();
		for (int i = 0; i < item.getInputParameterSlots().size(); i++) {
			file.println(variable(subgraphNumber, 0, i) + "=$1; shift", 1);
		}

		// Get mask.
		file.println("# Input data files", 1);
		InputMaskSlot mask = item.getInputMaskSlot();
		if (mask != null && mask.isConnected() && _outputImages.contains(getInputImage(mask, item.getGraph().getIdentifier()))) {
			int maskIndex = subGraph.getVirtualOperator().getVirtualInputMaskSlot().getRank();
			file.println("ln -fs $1 " + createImageName(maskIndex, 0, subgraphNumber) + "; shift", 1);
		}

		// Get input images.
		List<InputDataSlot> inputImages = item.getInputDataSlots();
		for (int i = 0; i < inputImages.size(); i++) {
			InputDataSlot isi = inputImages.get(i);
			// Unlinked input!
			if (isi.getConnectedSlot() == null) {
				throw new BadGraphException(item.getNickName() + ": " + Ariane.getString("Generator.error.incomplete"));
			}
			file.println("ln -fs $1 " + createImageName(i, 0, subgraphNumber) + "; shift", 1);
		}

		file.println();

		generateSubroutineBody(litems, subgraphNumber, file, 1);

		VirtualItem virtualItem = (VirtualItem)virtualOperator;
		// Get output images.
		for (int i = 0; i < item.getOutputDataSlots().size(); i++) {
			InputDataSlot isi = virtualItem.getVirtualOutputDataSlots().get(i);
			file.println("ln -fs " + getInputImage(isi, subgraphNumber) + " $1; shift", 1);
		}

		file.println("");
		// Output result.
		InputParameterSlot isp = (InputParameterSlot)virtualItem.getVirtualOutputResultSlot();
		if (isp.getConnectedSlot() == null) {
			file.println("psetstatus SUCCESS", 1);
		} else {
			file.println("psetstatus ${" + variable(subgraphNumber, isp.getConnectedSlot().getGraphItem().getId(), isp.getConnectedSlot().getRank()) + "}", 1);
		}
		file.println("}");

		generateSubroutinesDefinition(litems, file);
	}
}