/* -*- c-basic-offset: 3 -*-
 *
 * Copyright (c) 2013, GREYC.
 * All rights reserved
 *
 * You may use this file under the terms of the BSD license as follows:
 *
 * "Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of the GREYC, nor the name of its
 *     contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
 * 
 *   
 * For more information, refer to:
 * https://clouard.users.greyc.fr/Pandore
 */

package fr.greyc.pantheon.ariane.generator;

import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;
import java.util.ListIterator;

import javax.swing.JFileChooser;
import javax.swing.JOptionPane;
import javax.swing.filechooser.FileNameExtensionFilter;
import javax.swing.plaf.basic.BasicFileChooserUI;

import fr.greyc.pantheon.ariane.Ariane;
import fr.greyc.pantheon.ariane.graph.Graph;
import fr.greyc.pantheon.ariane.graph.item.CompositeItem;
import fr.greyc.pantheon.ariane.graph.item.ConstantItem;
import fr.greyc.pantheon.ariane.graph.item.EvalItem;
import fr.greyc.pantheon.ariane.graph.item.ControlStructureForItem;
import fr.greyc.pantheon.ariane.graph.item.ControlStructureIfItem;
import fr.greyc.pantheon.ariane.graph.item.ControlStructureMergeItem;
import fr.greyc.pantheon.ariane.graph.item.ControlStructureSelectItem;
import fr.greyc.pantheon.ariane.graph.item.ControlStructureWhileItem;
import fr.greyc.pantheon.ariane.graph.item.FileNameItem;
import fr.greyc.pantheon.ariane.graph.item.FormatItem;
import fr.greyc.pantheon.ariane.graph.item.GraphItem;
import fr.greyc.pantheon.ariane.graph.item.IScriptArgument;
import fr.greyc.pantheon.ariane.graph.item.LoaderItem;
import fr.greyc.pantheon.ariane.graph.item.MacroOperatorItem;
import fr.greyc.pantheon.ariane.graph.item.NoteItem;
import fr.greyc.pantheon.ariane.graph.item.OperatorItem;
import fr.greyc.pantheon.ariane.graph.item.SaverItem;
import fr.greyc.pantheon.ariane.graph.item.ScriptItem;
import fr.greyc.pantheon.ariane.graph.item.SystemItem;
import fr.greyc.pantheon.ariane.graph.item.ViewerItem;
import fr.greyc.pantheon.ariane.graph.item.VirtualItem;
import fr.greyc.pantheon.ariane.graph.slot.InputSlot;
import fr.greyc.pantheon.ariane.graph.slot.InputDataSlot;
import fr.greyc.pantheon.ariane.graph.slot.InputParameterSlot;
import fr.greyc.pantheon.ariane.graph.slot.OutputDataSlot;
import fr.greyc.pantheon.ariane.graph.slot.OutputResultSlot;
import fr.greyc.pantheon.ariane.ui.PreferenceChooser;
import fr.greyc.pantheon.ariane.ui.Workspace;

/**
 * The <code>CodeGenerator</code> is the base class for 
 * all the language generators.
 * A <code>CodeGenerator</code> generates a script file
 * that contains code of a specified language
 * for encoding a specified graph.
 * 
 * It implements the <b>Template Method Design Pattern</b>.
 * 
 * @version Régis CLOUARD - Dec 04, 2007
 */
public abstract class CodeGenerator {
	/** The list of all available languages. */
	private static final String[][] LANGUAGES = {
		{"fr.greyc.pantheon.ariane.generator.BashGenerator", "Shell", "sh"},
		{"fr.greyc.pantheon.ariane.generator.PythonGenerator", "Python", "py"},
		{"fr.greyc.pantheon.ariane.generator.PerlGenerator", "Perl", "pl"},
		{"fr.greyc.pantheon.ariane.generator.MsdosGenerator", "MsDos", "bat"}
	};

	/** Correspondence table : index i -> item identifier. */
	private GraphItem[] _correspondenceTable;

	private File _scriptFile;

	/** The graph to save. */
	private Graph _graph;

	protected CodeGenerator( ) {
		// empty
	}

	/**
	 * @param graph the graph to save.
	 * @return the selected generator from the file extension.
	 */
	public static CodeGenerator selectCodeGenerator( Graph graph ) {
		CodeGenerator generator = null;
		String[] chosenFile = chooseFilename(graph.getName());
		if (chosenFile.length > 0) {
			String filename = chosenFile[0];
			final String language = chosenFile[1];
			File file = new File(filename);
			for (int i = 0; i < LANGUAGES.length; i++) {
				if (language.equals(LANGUAGES[i][1])) {
					try {
						generator = (CodeGenerator)Class.forName(LANGUAGES[i][0]).newInstance();
						generator.init(file, graph);
					} catch (Exception e) {
//						e.printStackTrace();
					}
				}
			}
			if (file.exists() && JOptionPane.showConfirmDialog(Ariane.getFrame(),
					Ariane.getString("Generator.dialog.message.part1")
					+ filename
					+ Ariane.getString("Generator.dialog.message.part2"),
					Ariane.getString("Generator.dialog.title"),
					JOptionPane.YES_NO_OPTION) == JOptionPane.NO_OPTION) {
				Ariane.getWorkspace().printErrorMessage(Ariane.getString("Generator.dialog.error1"));
				return null;
			}
		}
		return generator;
	}

	private void init( File chosenFile, Graph graph ) {
		_scriptFile = chosenFile;
		_graph = graph;
		_correspondenceTable = null;
	}

	/**
	 * Launches the code generation for the specified graph
	 * and related subgraphs.
	 * Creates the file named: graph name +extension. If this file already
	 * exists, asks for overwriting permission.
	 * 
	 * @return true if the code has been successfully generated.
	 */
	public final boolean generate( ) {
		if (_scriptFile != null) {
			FileOutputCode file = null;
			try {
				file = new FileOutputCode(_scriptFile);
				generateCode(_graph, file, _scriptFile.getName());
				Ariane.getWorkspace().printSystemMessage(Ariane.getString("Generator.success") + " " + _scriptFile);
				return true;
			} catch (BadGraphException e) {
//				e.printStackTrace();
				Ariane.getWorkspace().printErrorMessage(Ariane.getString("Generator.failure") + Ariane.getString("Generator.failure.text"));
			} catch (IOException e) {
//				e.printStackTrace();
				Ariane.getWorkspace().printErrorMessage(Ariane.getString("Generator.failure") + Ariane.getString("Generator.failure.text"));
			} finally {
				if (file != null) {
					file.close();
				}
			}
		}
		return false;
	}

	/**
	 * Generates the code for the specified graph.
	 * Implements the <b>Template Method Design Pattern</b>.
	 * 
	 * @param graph the graph.
	 * @param file the file descriptor where to save the code.
	 * @param filename the name of the file (for comments)
	 * @throws Exception in case of malformed code.
	 */
	private void generateCode( Graph graph, FileOutputCode file, String filename ) throws BadGraphException {
		List<GraphItem> litems = buildItemsSequence(graph);
		List<GraphItem> args = getArguments(litems);
		List<GraphItem> inputs = getLoaderItems(litems);
		List<GraphItem> outputs = getSaverItems(litems);

		generateScriptHeader(file, filename, args, inputs, outputs);
		generateSubroutinesDefinition(litems, file);
		generateUtilitySubroutinesDefinition(file);
		generateMainPrologue(graph, file, args, inputs, outputs);
		generateSubroutineBody(litems, 1, file, 1);
		generateMainEpilogue(graph, file);
		generateScriptFooter(file);
	}

	/**
	 * Returns the list of all the loaders in the graph.
	 * The list keeps the topological order.
	 * 
	 * @param litems the list of graph items.
	 * @return the list of input images.
	 */
	private List<GraphItem> getLoaderItems( List<GraphItem> litems ) {
		List<GraphItem> items = new ArrayList<GraphItem>();
		// In practice: it is better to return a reverse list. 
		ListIterator<GraphItem> iterator = litems.listIterator(litems.size());
		while (iterator.hasPrevious()) {
			GraphItem item = iterator.previous();
			if (item instanceof LoaderItem) {
				LoaderItem loaderItem = (LoaderItem)item;
				if (!(loaderItem.getInputParameterSlots().size() > 0 && loaderItem.getInputParameterSlots().get(0).isConnected())) {
					items.add(item);
				}
			}
		}
		return items;
	}

	/**
	 * Returns the list of all the loaders in the graph.
	 * The list keeps the topological order.
	 * 
	 * @param litems the list of graph items.
	 * @return the list of input images.
	 */
	private List<GraphItem> getArguments( List<GraphItem> litems ) {
		List<GraphItem> items = new ArrayList<GraphItem>();
		// In practice: it is better to return a reverse list.
		ListIterator<GraphItem> iterator = litems.listIterator(litems.size());
		while (iterator.hasPrevious()) {
			GraphItem item = iterator.previous();
			if (item instanceof IScriptArgument && ((IScriptArgument) item).isArgument()) {
				items.add(item);
			}
		}
		return items;
	}

	/**
	 * Returns the list of all the savers in the graph.
	 * The list keeps the topological order.
	 * 
	 * @param litems the list of graph items.
	 * @return the list of output images.
	 */
	private List<GraphItem> getSaverItems( List<GraphItem> litems ) {
		ArrayList<GraphItem> items = new ArrayList<GraphItem>();
		// In practice: it is better to return a reverse list.
		ListIterator<GraphItem> iterator = litems.listIterator(litems.size());
		while (iterator.hasPrevious()) {
			GraphItem item = iterator.previous();
			if (item instanceof SaverItem) {
				SaverItem saverItem = (SaverItem)item;
				if (!(saverItem.getInputParameterSlots().size() > 0 && saverItem.getInputParameterSlots().get(0).isConnected())) {
					items.add(item);
				}
			}
		}
		return items;
	}

	/**
	 * Generates the header of the script file and
	 * the code to check the argument line consistency.
	 * 
	 * @param file the output stream.
	 * @param filename the name of the code file.
	 * @param arguments the list of arguments defined in the graph.
	 * @param inputs the list of loaders defined in the graph.
	 * @param outputs the list of savers defined in the graph.
	 */
	protected abstract void generateScriptHeader( FileOutputCode file, String filename, List<GraphItem> arguments, List<GraphItem> inputs, List<GraphItem> outputs );

	/**
	 * Generates the footer of the script file.
	 * 
	 * @param file the output stream.
	 */
	protected abstract void generateScriptFooter( FileOutputCode file );

	/**
	 * Generates utility subroutines required to execute the script.
	 * 
	 * @param file the output stream.
	 */
	protected abstract void generateUtilitySubroutinesDefinition( FileOutputCode file );

	/**
	 * Generates code for the main subroutine.
	 * 
	 * @param graph the graph to encode.
	 * @param file the output stream.
	 * @param arguments the list of argument.
	 * @param inputs the list of inputs.
	 */
	protected abstract void generateMainPrologue( Graph graph, FileOutputCode file, List<GraphItem> arguments, List<GraphItem> inputs, List<GraphItem> outputs );

	/**
	 * Generates code to conclude the program code.
	 * At least, erases all temporary files.
	 * 
	 * @param graph the graph to encode.
	 * @param file the output stream. Generates the prologue and epilogue code.
	 */
	protected abstract void generateMainEpilogue( Graph graph, FileOutputCode file );

	/**
	 * Opens a dialog box for selecting the output filename.
	 * 
	 * @param filebasename the presupposed file base name.
	 * @return the chosen file name or null when canceled.
	 */
	private static String[] chooseFilename( String filebasename ) {
		File currentDirName = new File(Workspace.getCurrentDir());
		final JFileChooser fileChooser = new JFileChooser();
		fileChooser.setDialogType(JFileChooser.SAVE_DIALOG);
		fileChooser.setCurrentDirectory(currentDirName);
		fileChooser.setDialogTitle(Ariane.getString("Generator.dialogchooser.title"));
		fileChooser.setFileSelectionMode(JFileChooser.FILES_ONLY);
		fileChooser.addPropertyChangeListener(new PropertyChangeListener() {
			@Override
			public void propertyChange(PropertyChangeEvent evt) {
				if (JFileChooser.FILE_FILTER_CHANGED_PROPERTY.equals(evt.getPropertyName())) {
					JFileChooser chooser = (JFileChooser)evt.getSource();
					FileNameExtensionFilter filename = ((FileNameExtensionFilter)chooser.getFileFilter());
					if (filename != null) {
						String currentName = ((BasicFileChooserUI)fileChooser.getUI()).getFileName();
						if (currentName.lastIndexOf('.') > 0) {
							currentName = currentName.substring(0, currentName .lastIndexOf('.'));
						}
						final String extension = ((FileNameExtensionFilter)fileChooser.getFileFilter()).getExtensions()[0];
						final String fullFileName = currentName + "." + extension;
						fileChooser.setSelectedFile(new File(fullFileName));
					}
				}
			}
		});

		// Use file type preference
		FileNameExtensionFilter defaultFilter = null;
		int typePreference = PreferenceChooser.getInt(PreferenceChooser.CODE_GENERATOR_PREFERENCE);
		fileChooser.setAcceptAllFileFilterUsed(false);
		for (int i = 0; i < LANGUAGES.length; i++) {
			FileNameExtensionFilter filter = new FileNameExtensionFilter(LANGUAGES[i][1], LANGUAGES[i][2]);
			fileChooser.addChoosableFileFilter(filter);
			if (i == typePreference ) {
				defaultFilter = filter;
			}
		}
		fileChooser.setFileFilter(defaultFilter);
		// Set the first name using default extension
		String extension = defaultFilter.getExtensions()[0];
		File currentFileName = new File(filebasename + "." + extension);
		fileChooser.setSelectedFile(currentFileName);

		if (fileChooser.showSaveDialog(Ariane.getFrame()) == JFileChooser.APPROVE_OPTION) {
			// Update file type preference.
			for (int i = 0; i < LANGUAGES.length; i++ ) {
				if (LANGUAGES[i][1].equals(fileChooser.getFileFilter().getDescription())) {
					PreferenceChooser.putInt("CodeGenerator.Preference", i);
				}
			}

			File theFile = fileChooser.getSelectedFile();
			if (theFile != null) {
				String fullFileName = fileChooser.getSelectedFile().getPath();
				Workspace.setCurrentDir(fileChooser.getSelectedFile().getParent());
				String description = ((FileNameExtensionFilter)fileChooser.getFileFilter()).getDescription();
				return new String[]{ fullFileName, description};
			}
		}
		return new String[0];
	}

	/**
	 * Builds a representation of the graph as a sequence
	 * of items sorted according to the topological order.
	 * 
	 * @param graph the graph to be transformed.
	 * @return the ordered list of items.
	 */
	protected final List<GraphItem> buildItemsSequence( Graph graph ) {
		return topologicalSort(buildCorrespondenceTable(graph));
	}

	/**
	 * Builds a representation of the graph as a representation
	 * that is use for the topological sort.<br>
	 * Builds the correspondence table: item i -> item identifier,
	 * and the list of output slots for each item.
	 * Caution: keep also the order defined of the output rank!
	 * 
	 * @param graph the graph to be transformed.
	 * @return the list of adjacency between items in the graph.
	 */
	private List<Integer>[] buildCorrespondenceTable( Graph graph ) {
		List<GraphItem> items = graph.getGraphItems();
		final int itemCount = items.size();

		// The node table.
		@SuppressWarnings("unchecked")
		ArrayList<Integer>[] adjacenceList = new ArrayList[itemCount];
		for (int i = 0; i < itemCount; i++) {
			adjacenceList[i] = new ArrayList<Integer>();
		}

		_correspondenceTable = new GraphItem[itemCount]; // The correspondence table: [num] -> id
		int max = 0; // The maximum id
		for (int i = 0; i < itemCount; i++) {
			_correspondenceTable[i] = items.get(i);
			if (_correspondenceTable[i].getId() > max) {
				max = _correspondenceTable[i].getId();
			}
		}

		int[] inverseTab = new int[max + 1]; // Inverse correspondence table: id -> num
		for (int i = 0; i < itemCount; i++) {
			inverseTab[items.get(i).getId()] = i;
		}

		// For each item -> get the output slots (images + results).
		for (int i = 0; i < itemCount; i++) {
			List<OutputDataSlot> out = items.get(i).getOutputDataSlots();
			List<InputSlot> next;
			// Keep this inverse loop because we use add item at the end of the list!
			for (int j = 0; j < out.size(); j++) {
				next = out.get(j).getConnectedSlots();
				for (int k = 0; k < next.size(); k++) {
					GraphItem graphItem = next.get(k).getGraphItem();
					adjacenceList[i].add(Integer.valueOf(inverseTab[graphItem.getId()]));
				}
			}
		}
		for (int i = 0; i < itemCount; i++) {
			OutputResultSlot out = items.get(i).getOutputResultSlot();
			if (out == null) {
				continue;
			}
			List<InputSlot> next;
			// Keep this inverse loop because we use add item in the head of the list!
			next = out.getConnectedSlots();
			for (int k = 0; k < next.size(); k++) {
				GraphItem graphItem = next.get(k).getGraphItem();
				adjacenceList[i].add(Integer.valueOf(inverseTab[graphItem.getId()]));
			}
		}
		return adjacenceList;
	}

	/**
	 * Sorts the items according to inputs and outputs
	 * order since later items use inputs of the former ones.
	 * Uses the the topological sort on the adjacency list
	 * of items.
	 * Note, all virtual operator are removed from the list.
	 * 
	 * @param adjacenceList the list of adjacent items.
	 * @return the ordered list of items.
	 */
	protected final List<GraphItem> topologicalSort( List<Integer>[] adjacenceList) {
		int size = adjacenceList.length;
		int[] items = new int[size];
		ArrayList<GraphItem> ordered = new ArrayList<GraphItem>();

		// Creates the "inside half-degree table".
		List<Integer> successors;
		for (int i = 0; i < size; i++) {
			successors = adjacenceList[i];
			for (int j = 0; j < successors.size(); j++) {
				items[successors.get(j).intValue()]++;
			}
		}

		// Linking index i such as d[i] = 0.
		int stackTop = 0;
		for (int i = 0; i < size; i++) {
			if (items[i] == 0) {
				items[i] = stackTop;
				stackTop = i;
			}
		}

		// Build the result list.
		for (int i = 0; i < size; i++) {
			ordered.add(_correspondenceTable[stackTop]);
			int item = stackTop;
			stackTop = items[stackTop];
			items[item] = -1;
			successors = adjacenceList[item];
			for (int successor = successors.size() - 1; successor >= 0; successor--) {
				int successorValue = successors.get(successor).intValue();
				items[successorValue]--;
				if (items[successorValue] == 0) {
					items[successorValue] = stackTop;
					stackTop = successorValue;
				}
			}
		}
		return ordered;
	}

	/**
	 * Generates the body of the specified graph.
	 * 
	 * @param litems the list of items in the body.
	 * @param graphNumber the identifier of the graph.
	 * @param file the output stream.
	 * @param indent the current code indentation
	 * @throws Exception in case of IO exception or malformed graph.
	 */
	protected final void generateSubroutineBody( List<GraphItem> litems, int graphNumber, FileOutputCode file, int indent ) throws BadGraphException {
		// 'IF' generation is more complex that is seems
		// when there is item that are shared by operators in
		// both IF and ELSE  parts.
		ArrayList<GraphItem> ifItems = new ArrayList<GraphItem>();
		ArrayList<List<GraphItem>> elseItems = new ArrayList<List<GraphItem>>();
		ArrayList<GraphItem> generatedItems = new ArrayList<GraphItem>();
		int indentation = indent;
		for (GraphItem graphItem: litems) {
			generatedItems.add(graphItem);
			if (!ifItems.isEmpty() && !elseItems.isEmpty()) {
				List<GraphItem> items = elseItems.get(0);
				if (items.contains(graphItem)) {
					generateElsePart((ControlStructureIfItem)ifItems.get(0), graphNumber, file, indentation - 1);
					ifItems.remove(0);
				}
			}
			if (graphItem instanceof LoaderItem) {
				generateItem((LoaderItem)graphItem, graphNumber, file, indentation);
			} else if (graphItem instanceof SaverItem) {
				generateItem((SaverItem)graphItem, graphNumber, file, indentation);
			} else if (graphItem instanceof ViewerItem) {
				generateItem((ViewerItem)graphItem, graphNumber, file, indentation);
			} else if (graphItem instanceof EvalItem) {
				generateItem((EvalItem)graphItem, graphNumber, file, indentation);
			} else if (graphItem instanceof FileNameItem) {
				generateItem((FileNameItem)graphItem, graphNumber, file, indentation);
			} else if (graphItem instanceof FormatItem) {
				generateItem((FormatItem)graphItem, graphNumber, file, indentation);
			} else if (graphItem instanceof ConstantItem) {
				generateItem((ConstantItem)graphItem, graphNumber, file, indentation);
			} else if (graphItem instanceof ScriptItem) {
				generateItem((ScriptItem)graphItem, graphNumber, file, indentation);
			} else if (graphItem instanceof SystemItem) {
				generateItem((SystemItem)graphItem, graphNumber, file, indentation);
			} else if (graphItem instanceof NoteItem) {
				generateItem((NoteItem)graphItem, graphNumber, file, indentation);
			} else if (graphItem instanceof OperatorItem) {
				generateItem((OperatorItem)graphItem, graphNumber, file, indentation);
			} else if (graphItem instanceof ControlStructureSelectItem) {
				generateItem((ControlStructureSelectItem)graphItem, graphNumber, file, indentation);
			} else if (graphItem instanceof ControlStructureIfItem) {
				ifItems.add(0, graphItem);
				generateItem((ControlStructureIfItem)graphItem, graphNumber, file, indentation);
				generateThenPart((ControlStructureIfItem)graphItem, graphNumber, file, indentation++);
				elseItems.add(0, searchForElsePartFirstItem(graphItem, generatedItems));
			} else if (graphItem instanceof ControlStructureMergeItem) {
				if (!ifItems.isEmpty()) {
					ifItems.remove(0);
				}
				if (!elseItems.isEmpty()) {
					elseItems.remove(0);
				}
				indentation--;
				generateFiPart(null, graphNumber, file, indentation);
				generateItem((ControlStructureMergeItem)graphItem, graphNumber, file, indentation);
			} else if (graphItem instanceof MacroOperatorItem) {
				generateItem((MacroOperatorItem) graphItem, graphNumber, file, indentation);
			} else if (graphItem instanceof ControlStructureWhileItem) {
				generateItem((ControlStructureWhileItem)graphItem, graphNumber, file, indentation);
			} else if (graphItem instanceof ControlStructureForItem) {
				generateItem((ControlStructureForItem)graphItem, graphNumber, file, indentation);
			}
		}
		int size = ifItems.size();
		for (int i = 0; i < size; i++) {
			if (!ifItems.isEmpty()) {
				ifItems.remove(0);
			}
			if (!elseItems.isEmpty()) {
				elseItems.remove(0);
			}
			generateFiPart(null, graphNumber, file, --indentation);
		}
	}

	private List <GraphItem> searchForElsePartFirstItem( GraphItem ifItem, List<GraphItem> generatedItems ) {
		OutputDataSlot elseSlot = ifItem.getOutputDataSlots().get(1);
		List<InputSlot> connectedSlots = elseSlot.getConnectedSlots();
		ArrayList <GraphItem> graphItems = new ArrayList <GraphItem>();

		// slots connected to the else part.
		for ( InputSlot slot : connectedSlots ) {
			graphItems.add(slot.getGraphItem());
		}

		// add operators that computes the parameters
		// that are not already generated.
		ArrayList <GraphItem> graphItemsBis = new ArrayList <GraphItem>();
		for ( GraphItem item : graphItems ) {
			for ( InputParameterSlot parameter : item.getInputParameterSlots()) {
				if (parameter.isConnected()) {
					GraphItem connectedItem = parameter.getConnectedSlot().getGraphItem();
					if (!generatedItems.contains(connectedItem)) {
						graphItemsBis.add(connectedItem);
					}
				}
			}
		}

		// add operators that computes the input images
		// that are not already generated.
		for ( GraphItem item : graphItems ) {
			for ( InputDataSlot slot : item.getInputDataSlots()) {
				GraphItem connectedItem = slot.getConnectedSlot().getGraphItem();
				if (slot.isConnected() && !generatedItems.contains(connectedItem)) {
					graphItemsBis.add(connectedItem);
				}
			}
		}
		graphItems.addAll(graphItemsBis);
		return graphItems;
	}

	/**
	 * Generates the code of sub-functions called in the graph body.
	 * 
	 * @param litems the list of item in the body.
	 * @param file the output stream.
	 * @throws Exception incomplete graph.
	 */
	protected final void generateSubroutinesDefinition( List<GraphItem> litems, FileOutputCode file ) throws BadGraphException {
		for (int i = 0; i < litems.size(); i++) {
			GraphItem graphItem = litems.get(i);
			if (graphItem instanceof ControlStructureWhileItem) {
				generateSubGraph((ControlStructureWhileItem)graphItem, file);
			} else if (graphItem instanceof ControlStructureForItem) {
				generateSubGraph((ControlStructureForItem)graphItem, file);
			} else if (graphItem instanceof MacroOperatorItem) {
				generateSubGraph((MacroOperatorItem)graphItem, file);
			}
		}
	}

	protected final Object getVirtualOperator( List<GraphItem> litems ) {
		for ( GraphItem graphItem : litems ) {
			if (graphItem instanceof VirtualItem) {
				return graphItem;
			}
		}
		return null;
	}

	/**
	 * Generates the code for loading an image.
	 * 
	 * @param item the item to encode.
	 * @param graphNumber the identifier number of the graph.
	 * @param file the output stream.
	 * @param indent the indentation level.
	 * @throws Exception Incomplete graph.
	 */
	protected abstract void generateItem( LoaderItem item, int graphNumber, FileOutputCode file, int indent ) throws BadGraphException;

	/**
	 * Generates the code for saving an image.
	 * 
	 * @param item the item to encode.
	 * @param graphNumber the identifier number of the graph.
	 * @param file the output stream.
	 * @param indent the indentation level.
	 * @throws Exception Incomplete graph.
	 */
	protected abstract void generateItem( SaverItem item, int graphNumber, FileOutputCode file, int indent ) throws BadGraphException;

	/**
	 * Generates the code for viewer item.
	 * 
	 * @param item the item to encode.
	 * @param graphNumber the identifier number of the graph.
	 * @param file the output stream.
	 * @param indent the indentation level.
	 * @throws Exception Incomplete graph.
	 */
	protected abstract void generateItem( ViewerItem item, int graphNumber, FileOutputCode file, int indent ) throws BadGraphException;

	/**
	 * Generates the code for executing an operator.
	 * 
	 * @param item the operator to encode.
	 * @param graphNumber the identifier number of the graph.
	 * @param file the output stream.
	 * @param indent the indentation level.
	 * @throws Exception Incomplete graph.
	 */
	protected abstract void generateItem( OperatorItem item, int graphNumber, FileOutputCode file, int indent ) throws BadGraphException;

	/**
	 * Generates the code for an Eval item.
	 * 
	 * @param item the item to encode.
	 * @param graphNumber the identifier number of the graph.
	 * @param file the output stream.
	 * @param indent the indentation level.
	 * @throws Exception Incomplete graph.
	 */
	protected abstract void generateItem( EvalItem item, int graphNumber, FileOutputCode file, int indent ) throws BadGraphException;

	/**
	 * Generates the code for a <code>FilenameItem</code> item.
	 * 
	 * @param item the item to encode.
	 * @param graphNumber the identifier number of the graph.
	 * @param file the output stream.
	 * @param indent the indentation level.
	 * @throws Exception Incomplete graph.
	 */
	protected abstract void generateItem( FileNameItem item, int graphNumber, FileOutputCode file, int indent ) throws BadGraphException;

	/**
	 * Generates the code for a <code>FormatItem</code> item.
	 * 
	 * @param item the item to encode.
	 * @param graphNumber the identifier number of the graph.
	 * @param file the output stream.
	 * @param indent the indentation level.
	 * @throws Exception Incomplete graph.
	 */
	protected abstract void generateItem( FormatItem item, int graphNumber, FileOutputCode file, int indent ) throws BadGraphException;

	/**
	 * Generates the code for an <code>ConstantItem</code>item.
	 * 
	 * @param item the item to encode.
	 * @param graphNumber the identifier number of the graph.
	 * @param file the output stream.
	 * @param indent the indentation level.
	 * @throws Exception Incomplete graph.
	 */
	protected abstract void generateItem( ConstantItem item, int graphNumber, FileOutputCode file, int indent ) throws BadGraphException;

	/**
	 * Generates the code for an <code>ScriptItem</code>item.
	 * 
	 * @param item the item to encode.
	 * @param graphNumber the identifier number of the graph.
	 * @param file the output stream.
	 * @param indent the indentation level.
	 * @throws Exception Incomplete graph.
	 */
	protected abstract void generateItem( ScriptItem item, int graphNumber, FileOutputCode file, int indent ) throws BadGraphException;

	/**
	 * Generates the code for an <code>SystemtItem</code>item.
	 * 
	 * @param item the item to encode.
	 * @param graphNumber the identifier number of the graph.
	 * @param file the output stream.
	 * @param indent the indentation level.
	 * @throws Exception Incomplete graph.
	 */
	protected abstract void generateItem( SystemItem item, int graphNumber, FileOutputCode file, int indent ) throws BadGraphException;

	/**
	 * Generates the code for the control structure Select.
	 * 
	 * @param item the select to encode.
	 * @param graphNumber the identifier number of the graph.
	 * @param file the output stream.
	 * @param indent the indentation level.
	 * @throws Exception Incomplete graph.
	 */
	protected abstract void generateItem( ControlStructureSelectItem item, int graphNumber, FileOutputCode file, int indent ) throws BadGraphException;

	/**
	 * Generates the code for the 'then' part of a if structure.
	 * 
	 * @param item the item to encode.
	 * @param graphNumber the identifier number of the graph.
	 * @param file the output stream.
	 * @param indent the indentation level.
	 * @throws Exception Incomplete graph.
	 */
	protected abstract void generateThenPart( ControlStructureIfItem item, int graphNumber, FileOutputCode file, int indent) throws BadGraphException;

	/**
	 * Generates the code for the 'else' part of a if structure.
	 * 
	 * @param item the item to encode.
	 * @param graphNumber the identifier number of the graph.
	 * @param file the output stream.
	 * @param indent the indentation level.
	 * @throws Exception Incomplete graph.
	 */
	protected abstract void generateElsePart( ControlStructureIfItem item, int graphNumber, FileOutputCode file, int indent) throws BadGraphException;

	/**
	 * Generates the code for the 'end' part of a if structure.
	 * 
	 * @param item the item to encode.
	 * @param graphNumber the identifier number of the graph.
	 * @param file the output stream.
	 * @param indent the indentation level.
	 * @throws Exception Incomplete graph.
	 */
	protected abstract void generateFiPart( ControlStructureIfItem item, int graphNumber, FileOutputCode file, int indent) throws BadGraphException;

	/**
	 * Generates the code for the control structure If.
	 * 
	 * @param item the item to encode.
	 * @param graphNumber the identifier number of the graph.
	 * @param file the output stream.
	 * @param indent the indentation level.
	 * @throws Exception Incomplete graph.
	 */
	protected abstract void generateItem( ControlStructureIfItem item, int graphNumber, FileOutputCode file, int indent ) throws BadGraphException;

	/**
	 * Generates the code for the control structure merge.
	 * 
	 * @param item the item to encode.
	 * @param graphNumber the identifier number of the graph.
	 * @param file the output stream.
	 * @param indent the indentation level.
	 * @throws Exception Incomplete graph.
	 */
	protected abstract void generateItem( ControlStructureMergeItem item, int graphNumber, FileOutputCode file, int indent ) throws BadGraphException;

	/**
	 * Generates the code for a note. Nothing to do.
	 * 
	 * @param item the note to encode.
	 * @param graphNumber the identifier number of the graph.
	 * @param file the output stream.
	 * @param indent the indentation level.
	 */
	protected abstract void generateItem( NoteItem item, int graphNumber, FileOutputCode file, int indent );

	/**
	 * Generates the code for the specified MacroOperator,
	 * for, or while call.
	 * 
	 * @param item the item to encode.
	 * @param graphNumber the identifier number of the graph.
	 * @param file the output stream.
	 * @param indent the indentation level.
	 * @throws Exception in case of incomplete graph.
	 */
	protected abstract void generateItem( CompositeItem item, int graphNumber, FileOutputCode file, int indent ) throws BadGraphException;

	/**
	 * Generates the code for the specified MacroOperator.
	 * 
	 * @param item the item to encode.
	 * @param file the output stream.
	 * @throws Exception In case of incomplete graph.
	 */
	protected abstract void generateSubGraph( MacroOperatorItem item, FileOutputCode file ) throws BadGraphException;

	/**
	 * Generates the code for the specified control structure For.
	 * 
	 * @param item the item to encode.
	 * @param file the output stream.
	 * @throws Exception Incomplete graph.
	 */
	protected abstract void generateSubGraph( ControlStructureForItem item, FileOutputCode file ) throws BadGraphException;

	/**
	 * Generates the code for the specified control structure While.
	 * 
	 * @param item the item to encode.
	 * @param file the output stream.
	 * @throws Exception Incomplete graph.
	 */
	protected abstract void generateSubGraph( ControlStructureWhileItem item, FileOutputCode file ) throws BadGraphException;
}

class BadGraphException extends Exception {

	public BadGraphException( String string ) {
		super(string);
	}
}
