/* -*- c-basic-offset: 3 -*-
 *
 * Copyright (c) 2013, GREYC.
 * All rights reserved
 *
 * You may use this file under the terms of the BSD license as follows:
 *
 * "Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of the GREYC, nor the name of its
 *     contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
 * 
 *   
 * For more information, refer to:
 * https://clouard.users.greyc.fr/Pandore
 */

package fr.greyc.pantheon.ariane.generator;

import java.text.DateFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.LinkedList;
import java.util.List;
import java.util.StringTokenizer;

import fr.greyc.pantheon.ariane.Ariane;
import fr.greyc.pantheon.ariane.graph.Graph;
import fr.greyc.pantheon.ariane.graph.SubGraph;
import fr.greyc.pantheon.ariane.graph.item.CompositeItem;
import fr.greyc.pantheon.ariane.graph.item.ConstantItem;
import fr.greyc.pantheon.ariane.graph.item.ControlStructureForItem;
import fr.greyc.pantheon.ariane.graph.item.ControlStructureIfItem;
import fr.greyc.pantheon.ariane.graph.item.ControlStructureMergeItem;
import fr.greyc.pantheon.ariane.graph.item.ControlStructureSelectItem;
import fr.greyc.pantheon.ariane.graph.item.ControlStructureWhileItem;
import fr.greyc.pantheon.ariane.graph.item.EvalItem;
import fr.greyc.pantheon.ariane.graph.item.FileNameItem;
import fr.greyc.pantheon.ariane.graph.item.FormatItem;
import fr.greyc.pantheon.ariane.graph.item.GraphItem;
import fr.greyc.pantheon.ariane.graph.item.LoaderItem;
import fr.greyc.pantheon.ariane.graph.item.MacroOperatorItem;
import fr.greyc.pantheon.ariane.graph.item.NoteItem;
import fr.greyc.pantheon.ariane.graph.item.OperatorItem;
import fr.greyc.pantheon.ariane.graph.item.SaverItem;
import fr.greyc.pantheon.ariane.graph.item.ScriptItem;
import fr.greyc.pantheon.ariane.graph.item.SystemItem;
import fr.greyc.pantheon.ariane.graph.item.ViewerItem;
import fr.greyc.pantheon.ariane.graph.item.VirtualItem;
import fr.greyc.pantheon.ariane.graph.slot.InputSlot;
import fr.greyc.pantheon.ariane.graph.slot.InputDataSlot;
import fr.greyc.pantheon.ariane.graph.slot.InputMaskSlot;
import fr.greyc.pantheon.ariane.graph.slot.InputParameterSlot;
import fr.greyc.pantheon.ariane.ui.PreferenceChooser;

/**
 * The <code>PerlGenerator</code> generates Perl code
 * for the specified graph.
 * It operates in three steps:<ol>
 * <li>transforms the graph into a list of operators usable for the topological sort;
 * <li>executes the topological sort;
 * <li>generates the related code.
 * </ol>
 * 
 * @author DOUBLET Alban - Jun 10, 2002
 * @version Régis CLOUARD - Jan 10, 2003
 * @version Régis CLOUARD - Jan 07, 2007
 * @version Régis CLOUARD - Jul 03, 2010 (fix bugs)
 * @version Régis CLOUARD - Aug 25, 2012 (add system + script items)
 */
public final class PerlGenerator extends CodeGenerator {
	/** The list of generated output images. */
	private final List<String> _outputImages = new ArrayList<String>();

	protected PerlGenerator( ) {
		// Only created by the code generator.
	}

	/**
	 * Generates the header of the code file and
	 * the argument line consistency checking.
	 * 
	 * @param fp the output stream.
	 * @param filename the name of the code file.
	 * @param inputs the list of loaders defined in the graph.
	 * @param outputs the list of savers defined in the graph.
	 */
	@Override
	protected void generateScriptHeader( FileOutputCode file, String filename, List<GraphItem> arguments, List<GraphItem> inputs, List<GraphItem> outputs) {
		// The current date.
		Date creationDate = new Date();
		DateFormat dateFormat = DateFormat.getDateInstance(DateFormat.MEDIUM);

		file.println("#!/usr/bin/env perl");
		file.println("#");
		file.println("# @(#)" + filename + "				" + dateFormat.format(creationDate));
		file.println("#");
		file.println("# This file was automatically generated by Ariane.");
		file.println("# Ariane is part of the Pantheon Project.");
		file.println("# GREYC - ENSICAEN & University of Caen Basse-Normandie, France");

		file.println("use Getopt::Std;");
		file.println("use File::Copy;");
		file.println("use Math::Trig;\n");
		file.println("#\n# Displays the program usage\n#");
		file.println("sub displayUsage {");
		file.print("print \"Usage: $0 [-v]", 1);
		for (int i = 0; i < arguments.size(); i++) {
			file.print(" " + arguments.get(i).getNickName());
		}
		for (int i = 0; i < inputs.size();) {
			file.print(" input-file" + (++i));
		}
		for (int i = 0; i < outputs.size();) {
			file.print(" output-file" + (++i));
		}
		file.println("\\n\\nOptions:\\n\";");
		file.println("print \"  -v verbose	verbose\\n\";", 1);
		file.print("print \"Example:\\n  $0", 1);
		for (int i = 0; i < arguments.size(); i++) {
			if (arguments.get(i).getFormula() != null && arguments.get(i).getFormula().length() > 0) {
				file.print(" " + arguments.get(i).getFormula());
			} else {
				file.print(" " + arguments.get(i).getNickName());
			}
		}
		for (int i = 0; i < inputs.size(); i++) {
			file.print(" " + inputs.get(i).getNickName());
		}
		for (int i = 0; i < outputs.size(); i++) {
			String name = outputs.get(i).getNickName();
			if ("saver".equals(name)) {
				name += outputs.get(i).getId();
			}
			file.print(" " + name);
		}

		file.println("\\n\";");
		file.println("exit;\n}", 1);

		file.println("\n#\n# Reads the command line\n#");
		file.println("getopts(\"hv\");");
		file.println("$verbose =0;");
		file.println("if ($opt_h) { displayUsage }");
		// Verbose mode.
		file.println("if ($opt_v) { $verbose =1; }");
		file.println("print \"Bad argument number\\n\\n\"; displayUsage;", 1);
		file.println("}\nmain();");
	}

	/**
	 * Generates the footer of the script file.
	 * 
	 * @param fp the output stream.
	 */
	@Override
	protected void generateScriptFooter( FileOutputCode file ) {
		// Do nothing
	}

	/**
	 * Generates utility functions required to execute the script.
	 * 
	 * @param fp the output stream.
	 */
	@Override
	protected void generateUtilitySubroutinesDefinition( FileOutputCode file ) {
		file.println("\n#\n# Executes an operator.\n#");
		file.println("sub execute {");
		file.println("my ($operator) = @_;", 1);
		file.println("print \"$operator\\n\" if $verbose;", 1);
		file.println("system($operator);", 1);
		file.println("$result = `" + GraphItem.getCommandStatus() + "`;", 1); 
		file.println("chomp($result);", 1);
		file.println("return $result;", 1);
		file.println("}\n");
		file.println("\n# Some math complements.\n#");
		file.println("sub max {");
		file.println("my ($a, $b)= @_;", 1);
		file.println("print \"$operator\\n\" if $verbose;", 1);
		file.println("return ($a>$b)?$a:$b;", 1);
		file.println("}\n");
		file.println("sub min {");
		file.println("my ($a, $b)= @_;", 1);
		file.println("print \"$operator\\n\" if $verbose;", 1);
		file.println("return ($a<$b)?$a:$b;", 1);
		file.println("}\n");
		file.println("sub log10{");
		file.println("my $n=shift;", 1);
		file.println("return log($n)/log(10);", 1);
		file.println("}\n");
		file.println("sub log2{");
		file.println("my $n=shift;", 1);
		file.println("return log($n)/log(10);", 1);
		file.println("}\n");
	}

	/**
	 * Creates the name of the image from the specified slot number,
	 * item and graph identifier.
	 * 
	 * @param slotIndex the identifier of the slot.
	 * @param itemIndex the identifier of the item.
	 * @param numGraph the graph identifier
	 * @return a name that is supposed unique.
	 */
	private String createImageName( int slotIndex, int itemIndex, int numGraph ) {
		String name = PreferenceChooser.createTemporaryName(numGraph, itemIndex, slotIndex);
		_outputImages.add(name);
		return name;
	}

	/**
	 * Returns the command line for getting the
	 * input images for the specified item slot of the specified
	 * graph identifier.
	 * 
	 * @param slot the related slot.
	 * @param numGraph the graph identifier
	 * @return a name that is supposed to unique.
	 */
	private String getInputImage( InputSlot slot, int numGraph ) {
		return PreferenceChooser.createTemporaryName(numGraph, slot.getConnectedSlot().getGraphItem().getId(), slot.getConnectedSlot().getRank());
	}

	/**
	 * Returns the parameters line command of the specified item.
	 * 
	 * @param item the related item.
	 * @param slotIndex the slot index.
	 * @param graphId the current graph identifier.
	 * @return a value or a command to get the value.
	 */
	private String getParameter( GraphItem item, int slotIndex, int graphId ) {
		InputParameterSlot isp = (InputParameterSlot) item.getSlot("InSlotParameter", slotIndex);
		if (isp.getConnectedSlot() == null) {
			return ((InputParameterSlot)item.getSlot("InSlotParameter", slotIndex)).getValue();
		} else {
			return "$result" + graphId + "[" + isp.getConnectedSlot().getGraphItem().getId() + "][" + isp.getConnectedSlot().getRank() + "]";
		}
	}

	/**
	 * Generates code for the main.
	 * 
	 * @param graph the graph to encode.
	 * @param fp the output stream.
	 * @param arguments the list of argument.
	 * @param inputs the list of inputs.
	 */
	@Override
	protected void generateMainPrologue( Graph graph, FileOutputCode file, List<GraphItem> arguments, List<GraphItem> inputs, List<GraphItem> outputs ) {
		file.println("\n### The main program");
		file.println("sub main {");
		file.println("unlink(<" + PreferenceChooser.getTemporaryFilePrefix() + "*" + PreferenceChooser.getTemporaryFileExtension() + ">);", 1);
		for ( GraphItem item : arguments ) {
			file.println("$result1[" + item.getId() + "][0] =shift(@ARGV);", 1);
		}
		for ( GraphItem item : inputs ) {
			file.println("symlink(shift(@ARGV), \"" + createImageName(0, item.getId(), 1) + "\");", 1);
		}
	}

	/**
	 * Generates code to erase all temporary files.
	 * 
	 * @param graph the graph to encode.
	 * @param fp the output stream.
	 */
	@Override
	protected void generateMainEpilogue( Graph graph, FileOutputCode file ) {
		//		file.println("unlink(<" +PreferenceChooser.TEMPORARY_FILE_PREFIX+ "*" +PreferenceChooser.TEMPORARY_FILE_EXT+ ">);", 1);
		file.println("}");
	}

	/**
	 * Generates the code for loading an image.
	 * 
	 * @param item the item to encode.
	 * @param numGraph the identifier number of the graph.
	 * @param fp the output stream.
	 * @param indent the indentation level.
	 * @throws Exception Incomplete graph.
	 */
	@Override
	protected void generateItem( LoaderItem item, int numGraph, FileOutputCode file, int indent ) throws BadGraphException {
		if (item.getInputParameterSlots().size() > 0 && item.getInputParameterSlots().get(0).isConnected()) {
			String parameters = getParameter(item, 0, numGraph);
			file.println("copy(" + parameters + ", \"" + createImageName(0, item.getId(), numGraph) + "\");", indent);
		}
	}

	/**
	 * Generates the code for saving an image.
	 * 
	 * @param item the item to encode.
	 * @param numGraph the identifier number of the graph.
	 * @param fp the output stream.
	 * @param indent the indentation level.
	 * @throws Exception Incomplete graph.
	 */
	@Override
	protected void generateItem( SaverItem item, int numGraph, FileOutputCode file, int indent ) throws BadGraphException {
		if (item.getInputDataSlots().get(0).getConnectedSlot() != null) {
			if (item.getInputParameterSlots().size() > 0 && item.getInputParameterSlots().get(0).isConnected()) {
				String parameters = getParameter(item, 0, numGraph);
				file.println("copy(\"" + getInputImage(item.getInputDataSlots().get(0), numGraph)
						+ "\", " + parameters + ");", indent);
			} else {
				file.println("copy(\"" + getInputImage(item.getInputDataSlots().get(0), numGraph)
						+ "\", shift(@ARGV));", indent);
			}
			file.println("symlink(\""
					+ getInputImage(item.getInputDataSlots().get(0), numGraph)
					+ "\", \""
					+ createImageName(0, item.getId(), numGraph)
					+ "\");", indent);
		}
	}

	/**
	 * Generates the code for viewer item.
	 * 
	 * @param item the item to encode.
	 * @param numGraph the identifier number of the graph.
	 * @param fp the output stream.
	 * @param indent the indentation level.
	 * @throws Exception Incomplete graph.
	 */
	@Override
	protected void generateItem( ViewerItem item, int numGraph, FileOutputCode file, int indent ) throws BadGraphException {
		String commandShell = item.getViewerCommandName();
		if (item.getInputDataSlots().get(0).getConnectedSlot() != null) {
			file.println("system(\"" + commandShell + " " + getInputImage(item.getInputDataSlots().get(0), numGraph) + "\");", indent);
			file.println("symlink (\""
					+ getInputImage(item.getInputDataSlots().get(0), numGraph)
					+ "\", \""
					+ createImageName(0, item.getId(), numGraph)
					+ "\");", indent);
		}
	}

	/**
	 * Generates the code for executing an operator.
	 * 
	 * @param item the operator to encode.
	 * @param numGraph the identifier number of the graph.
	 * @param fp the output stream.
	 * @param indent the indentation level.
	 * @throws Exception Incomplete graph.
	 */
	@Override
	protected void generateItem( OperatorItem item, int numGraph, FileOutputCode file, int indent ) throws BadGraphException {

		String codeName = OperatorItem.getCommandPart(item.getBinary());
		if (codeName == null) {
			throw new BadGraphException(Ariane.getString("Generator.error2") + item.getName());
		}

		// Get parameters
		List<String> parameterLine = new LinkedList<String>();
		for (int i = 0; i < item.getInputParameterSlots().size(); i++) {
			parameterLine.add(getParameter(item, i, numGraph));
		}

		// Get mask
		List<String> maskLine = new LinkedList<String>();
		InputMaskSlot mask = item.getInputMaskSlot();
		if (mask != null && mask.isConnected() && _outputImages.contains(getInputImage(mask, numGraph))) {
			maskLine.add(getInputImage(mask, numGraph));
		}

		// Get input images.
		List<String> inputLine = new LinkedList<String>();
		for (InputDataSlot image : item.getInputDataSlots() ) {
			if (image.getConnectedSlot() == null) {
				throw new BadGraphException(item.getNickName() + ": " + Ariane.getString("Generator.error.incomplete"));
			}
			inputLine.add(getInputImage(image, numGraph));
		}


		// Get output images.
		List<String> outputLine = new LinkedList<String>();
		for (int i = 0; i < item.getOutputDataSlots().size(); i++) {
			outputLine.add(createImageName(i, item.getId(), numGraph));
		}

		String code = null;
		List<String> commandLine = item.buildAbstractCommandLine(maskLine, parameterLine, inputLine, outputLine);
		for (String command : commandLine) {
			if (code == null) {
				code = command;
			} else {
				code += " " + command;
			}
		}

		// Get output result.
		file.println("$result" + numGraph + "[" + item.getId() + "][0] =" + "&execute(\"" + code + "\");", indent);
	}

	/**
	 * Generates the code for executing an operating system command.
	 * 
	 * @param item the operator to encode.
	 * @param numGraph the identifier number of the graph.
	 * @param fp the output stream.
	 * @param indent the indentation level.
	 * @throws Exception Incomplete graph.
	 */
	@Override
	protected void generateItem( SystemItem item, int numGraph, FileOutputCode file, int indent ) throws BadGraphException {
		String codeName = OperatorItem.getCommandPart(item.getBinary());
		if (codeName == null) {
			throw new BadGraphException(Ariane.getString("Generator.error2") + item.getName());
		}

		// Get parameters
		List<String> parameterLine = new LinkedList<String>();
		for (int i = 0; i < item.getInputParameterSlots().size(); i++) {
			parameterLine.add(getParameter(item, i, numGraph));
		}

		// Get mask
		List<String> maskLine = new LinkedList<String>();
		InputMaskSlot mask = item.getInputMaskSlot();
		if (mask != null && mask.isConnected() && _outputImages.contains(getInputImage(mask, numGraph))) {
			maskLine.add(getInputImage(mask, numGraph));
		}

		// Get input images.
		List<String> inputLine = new LinkedList<String>();
		for (InputDataSlot image : item.getInputDataSlots() ) {
			if (image.getConnectedSlot() == null) {
				throw new BadGraphException(item.getNickName() + ": " + Ariane.getString("Generator.error.incomplete"));
			}
			inputLine.add(getInputImage(image, numGraph));
		}


		// Get output images.
		List<String> outputLine = new LinkedList<String>();
		for (int i = 0; i < item.getOutputDataSlots().size(); i++) {
			outputLine.add(createImageName(i, item.getId(), numGraph));
		}

		String code = null;
		List<String> commandLine = item.buildAbstractCommandLine(maskLine, parameterLine, inputLine, outputLine);
		for ( String command : commandLine) {
			if (code == null) {
				code = command;
			} else {
				code += " " + command;
			}
		}

		// Get output result.
		file.println("$result" + numGraph + "[" + item.getId() + "][0] =" + "&execute(\"" + code + "\");", indent);
	}

	/**
	 * Parses the give formula into Perl code.
	 * 
	 * @param formula the function.
	 * @param parameters the list of parameters.
	 * @return the Perl code.
	 */
	private String parseFormula( String formula, String[] parameters ) {
		// Parse the formula.
		StringTokenizer stringTokenizer = new StringTokenizer(formula, "$ -+ ()*/^0123456789.E", true);
		String etape;
		String code = "";
		while (stringTokenizer.hasMoreTokens()) {
			etape = stringTokenizer.nextToken();
			if ("^".equals(etape)) {
				code += "**";
			} else if ("e".equals(etape) || "E".equals(etape)) {
				code += "*10^";
			} else if ("ln".equals(etape)) {
				code += "log";
			} else if ("atan".equals(etape)) {
				code += "a";
			} else if ("$".equals(etape)) {
				int nnumberOfParameters = Integer.parseInt(stringTokenizer.nextToken());
				code += parameters[nnumberOfParameters];
			} else {
				code += etape;
			}
		}
		return code;
	}

	/**
	 * Generates the code for an Eval item.
	 * 
	 * @param item the item to encode.
	 * @param numGraph the identifier number of the graph.
	 * @param fp the output stream.
	 * @param indent the indentation level.
	 * @throws Exception Incomplete graph.
	 */
	@Override
	protected void generateItem( EvalItem item, int numGraph, FileOutputCode file, int indent ) throws BadGraphException {
		String[] parameters = new String[item.getInputParameterSlots().size()];

		// Get parameters
		for (int i = 0; i < item.getInputParameterSlots().size(); i++) {
			parameters[i] = getParameter(item, i, numGraph);
		}

		file.println("$result" + numGraph + "[" + item.getId() + "][0] =" 
				+ parseFormula(item.getFormula(), parameters)
				+ ";", indent);
	}

	@Override
	protected void generateItem( FormatItem item, int numGraph, FileOutputCode file, int indent ) throws BadGraphException {
		String[] parameters = new String[item.getInputParameterSlots().size()];
		// Get parameters
		for (int i = 0; i < item.getInputParameterSlots().size(); i++) {
			parameters[i] = getParameter(item, i, numGraph);
		}

		String code = "";
		StringTokenizer stringTokenizer = new StringTokenizer(item.getFormula(), " $() +'\"", true);
		String etape;
		while (stringTokenizer.hasMoreTokens()) {
			etape = stringTokenizer.nextToken();
			if ("+".equals(etape)) {
				code += " + ";
//			} else if ("quote".equals(etape)) {
//				code = code + "";
//			} else if ("\"".equals(etape)) {
//				code = code + "";
			} else if ("(".equals(etape)) { 
				code += "\"'\" + ";
			} else if (")".equals(etape)) {
				code += " +\"'\"";
			} else if ("$".equals(etape)) {
				int numberOfParameters = Integer.parseInt(stringTokenizer.nextToken());
				code += parameters[numberOfParameters];
			} else {
				code += etape;
			}
		}
		file.println("$result" + numGraph + "[" + item.getId() + "][0] =\"" + code + "\";", indent);
	}

	/**
	 * Generates the code for the control structure Select.
	 * 
	 * @param item the select to encode.
	 * @param numGraph the identifier number of the graph.
	 * @param fp the output stream.
	 * @param indent the indentation level.
	 * @throws Exception Incomplete graph.
	 */
	@Override
	protected void generateItem( ControlStructureSelectItem item, int numGraph, FileOutputCode file, int indent ) throws BadGraphException {
		String value = getParameter(item, 0, numGraph);

		file.println("# Select", indent);
		for (int i = 0; i < item.getInputDataSlots().size(); i++) {
			file.println("if (" + value + "==" + i + "){", indent);
			file.println("symlink(\""
					+ getInputImage(item.getInputDataSlots().get(i), numGraph) + "\", \""
					+ createImageName(0, item.getId(), numGraph) + "\");", indent + 1);
			file.println("$result" + numGraph + "[" + item.getId() + "][0] =" + i + ";", indent + 1);
			file.println("}", indent);
		}
	}

	/**
	 * Generates the code for the control structure If.
	 * 
	 * @param item the item to encode.
	 * @param numGraph the identifier number of the graph.
	 * @param fp the output stream.
	 * @param indent the indentation level.
	 * @throws Exception Incomplete graph.
	 */
	@Override
	protected void generateItem( ControlStructureIfItem item, int numGraph, FileOutputCode file, int indent ) throws BadGraphException {
		String[] parameters = new String[2];
		for (int i = 0; i < item.getInputParameterSlots().size(); i++) {
			parameters[i] = getParameter(item, i, numGraph);
		}

		// Parse the boolean expression.
		StringTokenizer stringTokenizer = new StringTokenizer(item.getFormula(), "&|$^!0123456789.E><=() +*-/", true);
		String etape;
		String commandShell = "";
		while (stringTokenizer.hasMoreTokens()) {
			etape = stringTokenizer.nextToken();
			if ("&".equals(etape)) {
				commandShell += " && ";
			} else if ("|".equals(etape)) {
				commandShell += " || ";
			} else if ("$".equals(etape)) {
				if (stringTokenizer.nextToken().equals("0")) {
					commandShell += parameters[0];
				} else {
					commandShell += parameters[1];
				}
			} else {
				commandShell += etape;
			}
		}

		file.println("if (" + commandShell + ") {", indent);
	}

	/**
	 * Generates the code for the 'then' part.
	 * 
	 * @param item the item to encode.
	 * @param numGraph the identifier number of the graph.
	 * @param fp the output stream.
	 * @param indent the indentation level.
	 * @throws Exception Incomplete graph.
	 */
	@Override
	protected void generateThenPart( ControlStructureIfItem item, int numGraph, FileOutputCode file, int indent) throws BadGraphException {
		file.println("symlink(\""
				+ getInputImage(item.getInputDataSlots().get(0), numGraph) + "\", \""
				+ createImageName(0, item.getId(), numGraph) + "\");", indent + 1);
		file.println("$result" + numGraph + "[" + item.getId() + "][0] = 0;", indent + 1);
	}

	/**
	 * Generates the code for the 'else' part.
	 * 
	 * @param item the item to encode.
	 * @param numGraph the identifier number of the graph.
	 * @param fp the output stream.
	 * @param indent the indentation level.
	 * @throws Exception Incomplete graph.
	 */
	@Override
	protected void generateElsePart( ControlStructureIfItem item, int numGraph, FileOutputCode file, int indent) throws BadGraphException {
		file.println("} else {", indent);
		file.println("symlink(\""
				+ getInputImage(item.getInputDataSlots().get(0), numGraph) + "\", \""
				+ createImageName(1, item.getId(), numGraph) + "\");", indent + 1);
		file.println("$result" + numGraph + "[" + item.getId() + "][0] = 1;", indent + 1);
	}

	/**
	 * Generates the code for the end of a if control structure.
	 * 
	 * @param item the item to encode.
	 * @param numGraph the identifier number of the graph.
	 * @param fp the output stream.
	 * @param indent the indentation level.
	 * @throws Exception Incomplete graph.
	 */
	@Override
	protected void generateFiPart( ControlStructureIfItem item, int numGraph, FileOutputCode file, int indent) throws BadGraphException {
		file.println("}", indent);
	}

	/**
	 * Generates the code for the control structure merge.
	 * 
	 * @param item the item to encode.
	 * @param numGraph the identifier number of the graph.
	 * @param fp the output stream.
	 * @param indent the indentation level.
	 * @throws Exception Incomplete graph.
	 */
	@Override
	protected void generateItem( ControlStructureMergeItem item, int numGraph, FileOutputCode file, int indent ) throws BadGraphException {
		InputDataSlot in0 = item.getInputDataSlots().get(0);
		int id0 = in0.getConnectedSlot().getGraphItem().getId();

		file.print("if (defined($result" + numGraph + "[" + id0 + "][0]) && ", indent);
		file.print("$result" + numGraph + "[" + id0 + "][0] ne \"FAILURE\" && ");
		file.println("-e \"" + getInputImage(item.getInputDataSlots().get(0), numGraph) + "\"){");
		file.println("symlink(\""
				+ getInputImage(item.getInputDataSlots().get(0), numGraph)
				+ "\", \""
				+ createImageName(0, item.getId(), numGraph) + "\");", indent + 1);
		file.println("$result" + numGraph + "[" + item.getId() + "][0] =0;", indent + 1);
		file.print("} else {");
		file.println("symlink(\""
				+ getInputImage(item.getInputDataSlots().get(1), numGraph) + "\", \""
				+ createImageName(0, item.getId(), numGraph) + "\");", indent + 1);
		file.println("$result" + numGraph + "[" + item.getId() + "][0] =1;", indent + 1);
		file.println("}", indent);
	}

	/**
	 * Generates the code for a note. Nothing to do.
	 * 
	 * @param item the note to encode.
	 * @param numGraph the identifier number of the graph.
	 * @param fp the output stream.
	 * @param indent the indentation level.
	 */
	@Override
	protected void generateItem( NoteItem item, int numGraph, FileOutputCode file, int indent ) {
		// Do nothing
	}

	/**
	 * Generates the code for a <code>Filename</code> item.
	 * 
	 * @param item the item to encode.
	 * @param numGraph the identifier number of the graph.
	 * @param fp the output stream.
	 * @param indent the indentation level.
	 * @throws Exception Incomplete graph.
	 */
	@Override
	protected void generateItem( FileNameItem item, int numGraph, FileOutputCode file, int indent ) throws BadGraphException {
		if (!item.isArgument()) {
			file.println("$result" + numGraph + "[" + item.getId() + "][0] =\"" + item.getFormula() + "\";", indent);
		}
	}

	/**
	 * Generates the code for an <code>Constant</code>item.
	 * 
	 * @param item the item to encode.
	 * @param numGraph the identifier number of the graph.
	 * @param fp the output stream.
	 * @param indent the indentation level.
	 * @throws Exception Incomplete graph.
	 */
	@Override
	protected void generateItem( ConstantItem item, int numGraph, FileOutputCode file, int indent ) throws BadGraphException {
		if (!item.isArgument()) {
			String value = item.getFormula().replaceAll("\"", "\\\\\"");
			file.println("$result" + numGraph + "[" + item.getId() + "][0] =\"" + value + "\";", indent);
		}
	}

	/**
	 * Generates the code for an <code>Script</code>item.
	 * 
	 * @param item the item to encode.
	 * @param numGraph the identifier number of the graph.
	 * @param fp the output stream.
	 * @param indent the indentation level.
	 * @throws Exception Incomplete graph.
	 */
	@Override
	protected void generateItem( ScriptItem item, int numGraph, FileOutputCode file, int indent ) throws BadGraphException {
		String value = item.getFormula().replaceAll("\"", "\\\\\"");
		file.println("$result" + numGraph + "[" + item.getId() + "][0] =\"\'" + value + "\'\";", indent);
	}

	/**
	 * Generates the code for the specified MacroOperator,
	 * FFor, or While call.
	 * 
	 * @param item the item to encode.
	 * @param numGraph the identifier number of the graph.
	 * @param fp the output stream.
	 * @param indent the indentation level.
	 * @throws Exception in case of incomplete graph.
	 */
	@Override
	protected void generateItem( CompositeItem item, int numGraph, FileOutputCode file, int indent ) throws BadGraphException {
		String code = "&macro" + item.getId() + "_" + item.getGraph().getIdentifier() + "(";
		boolean comma = false;

		// Get parameters
		for (int i = 0; i < item.getInputParameterSlots().size(); i++) {
			if (comma) {
				code += ", ";
			} else {
				comma = true;
			}
			code += "\"" + getParameter(item, i, numGraph) + "\"";
		}

		// Get mask
		InputMaskSlot mask = item.getInputMaskSlot();
		if (mask != null && mask.isConnected() && _outputImages.contains(getInputImage(mask, numGraph))) {
			if (comma) {
				code += ", ";
			} else {
				comma = true;
			}
			code += "\"" + getInputImage(mask, numGraph) + "\"";
		}

		// Get input images.
		for (InputDataSlot image : item.getInputDataSlots()) {
			if (image.getConnectedSlot() == null) {
				throw new BadGraphException(Ariane.getString("Generator.error.incomplete"));
			}
			if (comma) {
				code += ", ";
			} else {
				comma = true;
			}
			code += "\"" + getInputImage(image, numGraph) + "\"";
		}

		// Get output images.
		for (int i = 0; i < item.getOutputDataSlots().size(); i++) {
			if (comma) {
				code += ", ";
			} else {
				comma = true;
			}
			code += "\"" + createImageName(i, item.getId(), numGraph) + "\"";
		}
		// Get output result.
		file.println("$result" + numGraph + "[" + item.getId() + "][0] =" + code + ");", indent);
	}

	/**
	 * Parses the comment in order to add a sharp after each newlines.
	 * 
	 * @param comment a string that constains the comment.
	 * @return a new string.
	 */
	private String parseComment( String comment ) {
		return comment.replaceAll("\n", "\n# ");
	}

	/**
	 * Generates the code for the specified control structure For.
	 * 
	 * @param item the item to encode.
	 * @param fp the output stream.
	 * @throws Exception Incomplete graph.
	 */
	@Override
	protected void generateSubGraph( ControlStructureForItem item, FileOutputCode file ) throws BadGraphException {
		SubGraph subGraph = item.getSubGraph();
		List<GraphItem> litems = buildItemsSequence(subGraph);

		// Get and discard the virtual operator
		Object virtualOperator = getVirtualOperator(litems);
		litems.remove(virtualOperator);
		if (litems.isEmpty()) {
			throw new BadGraphException(item.getNickName() + ": " + Ariane.getString("Generator.error.incomplete"));
		}

		file.println("\n#\n# " + item.getNickName());
		for (GraphItem note : subGraph.getGraphItems()) {
			if (note instanceof NoteItem) {
				file.println("# " + parseComment(note.getFormula()));
			}
		}
		file.println(" ");

		file.println("sub macro" + item.getId() + "_" + item.getGraph().getIdentifier() + " {");

		// Get parameters
		int subgraphNumber = subGraph.getIdentifier();
		file.println("# Input parameters", 1);
		file.println("my $result" + subgraphNumber + ";", 1);
		file.println("$begin=shift(@_);", 1);
		file.println("$end=shift(@_);", 1);
		file.println("$step=shift(@_);", 1);
		for (int i = 3; i < item.getInputParameterSlots().size(); i++) {
			file.println("$result" + subgraphNumber + "[0][" + (i - 2) + "] =shift(@_);\n", 1);
		}
		file.println("$result" + subgraphNumber + "[0][0] =$begin;", 1);

		// Get mask
		file.println("# Input data files", 1);

		InputMaskSlot mask = item.getInputMaskSlot();
		if (mask != null && mask.isConnected() && _outputImages.contains(getInputImage(mask, item.getGraph().getIdentifier()))) {
			int maskIndex = subGraph.getVirtualOperator().getVirtualInputMaskSlot().getRank();
			file.println("symlink(shift(@_), \"" + createImageName(maskIndex, 0, subgraphNumber) + "\");", 1);
		}

		// Get input images
		int rank = 0;
		for (InputDataSlot isi : item.getInputDataSlots() ) {
			// Unlinked input!
			if (isi.getConnectedSlot() == null) {
				throw new BadGraphException("Generator.error.incomplete");
			}
			file.println("copy(shift(@_), \"" + createImageName(rank++, 0, subgraphNumber) + "\");", 1);
		}

		// Create the output images in case of no iteration
		VirtualItem virtualItem = (VirtualItem)virtualOperator;
		for (int i = 0; i < item.getOutputDataSlots().size(); i++) {
			InputDataSlot isi = virtualItem.getVirtualOutputDataSlots().get(i);
			file.println("$f=open(\"" + getInputImage(isi, subgraphNumber) + "\"); close($f);", 1);
		}

		// Initialize the return variable in case of no iteration
		{
			InputParameterSlot isp = (InputParameterSlot)virtualItem.getVirtualOutputResultSlot();
			file.println("$result" + subgraphNumber + "[" + isp.getConnectedSlot().getGraphItem().getId() + "][" + isp.getConnectedSlot().getRank() + "] =\"SUCCESS\";", 1);
		}

		file.println();
		file.println("for ( $i=$begin ; $i <= $end ; $i += $step) {", 1);
		file.println("$result" + subgraphNumber + "[0][0] = $i;", 2);

		generateSubroutineBody(litems, subgraphNumber, file, 2);

		// Get intermediate images
		for (int i = 0; i < item.getInputDataSlots().size(); i++) {
			InputDataSlot isi = virtualItem.getVirtualIntermediateDataSlots(i);
			// Unlinked input!
			if (isi.getConnectedSlot() == null) {
				throw new BadGraphException("Generator.error.incomplete");
			}
			String input = getInputImage(isi, subgraphNumber);
			String output = createImageName(i, 0, subgraphNumber);
			if (!input.equals(output)) {
				file.println("copy(\"" + input + "\",\"" + output + "\");", 2);
			}
		}

		file.println("}", 1);
		// Get output images
		file.println();
		file.println("# Output images", 1);
		for (int i = 0; i < item.getOutputDataSlots().size(); i++) {
			InputDataSlot isi = virtualItem.getVirtualOutputDataSlots().get(i);
			file.println("symlink(\"" + getInputImage(isi, subgraphNumber) + "\",shift(@_));", 1);
		}

		// Output result
		InputParameterSlot isp = (InputParameterSlot)virtualItem.getVirtualOutputResultSlot();
		if (isp.getConnectedSlot() == null) {
			file.println("return 0;", 1);
		} else {
			file.println("return $result" + subgraphNumber + "[" + isp.getConnectedSlot().getGraphItem().getId() + "][" + isp.getConnectedSlot().getRank() + "];", 1);
		}
		file.println("}");

		generateSubroutinesDefinition(litems, file);
	}

	/**
	 * Generates the code for the specified control structure While.
	 * 
	 * @param item the item to encode.
	 * @param fp the output stream.
	 * @throws Exception Incomplete graph.
	 */
	@Override
	protected void generateSubGraph( ControlStructureWhileItem item, FileOutputCode file ) throws BadGraphException {
		SubGraph subGraph = item.getSubGraph();
		List<GraphItem> litems = buildItemsSequence(subGraph);

		// Get and discard the virtual operator
		Object virtualOperator = getVirtualOperator(litems);
		litems.remove(virtualOperator);
		if (litems.isEmpty()) {
			throw new BadGraphException(item.getNickName() + ": " + Ariane.getString("Generator.error.incomplete"));
		}

		file.println("\n#\n# " + item.getNickName());
		for (GraphItem note : subGraph.getGraphItems()) {
			if (note instanceof NoteItem) {
				file.println("# " + parseComment(note.getFormula()));
			}
		}
		file.println("#");

		file.println("sub macro" + item.getId() + "_" + item.getGraph().getIdentifier() + " {");

		// Get parameters
		String[] parameters = new String[item.getInputParameterSlots().size()];
		int subgraphNumber = subGraph.getIdentifier();
		file.println("# Input parameters", 1);
		file.println("my $result" + subgraphNumber + ";", 1);
		for (int i = 0; i < item.getInputParameterSlots().size(); i++) {
			file.println("$result" + subgraphNumber + "[0][" + i + "] =shift(@_);", 1);
			parameters[i] = "$result" + subgraphNumber + "[0][" + i + "]";
		}

		// Get mask
		file.println("# Input data files", 1);
		InputMaskSlot mask = item.getInputMaskSlot();
		if (mask != null && mask.isConnected() && _outputImages.contains(getInputImage(mask, item.getGraph().getIdentifier()))) {
			int maskIndex = subGraph.getVirtualOperator().getVirtualInputMaskSlot().getRank();
			file.println("symlink(shift(@_), \"" + createImageName(maskIndex, 0, subgraphNumber) + "\");", 1);
		}

		// Get input images
		int rank = 0;
		for (InputDataSlot isi : item.getInputDataSlots() ) {
			// Unlinked input!
			if (isi.getConnectedSlot() == null) {
				throw new BadGraphException("Generator.error.incomplete");
			}
			file.println("copy(shift(@_), \"" + createImageName(rank++, 0, subgraphNumber) + "\");", 1);
		}

		// Parse the boolean expression.
		StringTokenizer stringTokenizer = new StringTokenizer(item.getFormula(), "&|$^!0123456789.E><=() +*-/", true);
		String etape;
		String testFormula = "";
		while (stringTokenizer.hasMoreTokens()) {
			etape = stringTokenizer.nextToken();
			if ("&".equals(etape)) {
				testFormula += " && ";
			} else if ("|".equals(etape)) {
				testFormula += " || ";
			} else if ("$".equals(etape)) {
				if (stringTokenizer.nextToken().equals("0")) {
					testFormula += parameters[0];
				} else {
					testFormula += parameters[1];
				}
			} else {
				testFormula += etape;
			}
		}

		VirtualItem virtualItem = (VirtualItem)virtualOperator;
		file.println();
		file.println("while (" + testFormula + ") {", 1);

		generateSubroutineBody(litems, subgraphNumber, file, 2);

		// Get intermediate images
		for (int i = 0; i < item.getInputDataSlots().size(); i++) {
			InputDataSlot isi = virtualItem.getVirtualIntermediateDataSlots(i);
			// Unlinked input!
			if (isi.getConnectedSlot() == null) {
				throw new BadGraphException("Generator.error.incomplete");
			}
			String input = getInputImage(isi, subgraphNumber);
			String output = createImageName(i, 0, subgraphNumber);
			if (!input.equals(output)) {
				file.println("copy(\"" + input + "\",\"" + output + "\");", 2);
			}
		}

		// Get intermediate parameters
		for (int i = 0; i < 2; i++) {
			InputParameterSlot isp = virtualItem.getVirtualIntermediateResultSlot(i);
			if (isp.getConnectedSlot() != null) {
				file.println(parameters[i] + "=$result" + subgraphNumber + "[" + isp.getConnectedSlot().getGraphItem().getId() + "][" + isp.getConnectedSlot().getRank() + "];", 2);
			}
		}
		file.println("}", 1);

		// Get output images
		file.println("# Output images", 1);
		for (int i = 0; i < item.getOutputDataSlots().size(); i++) {
			InputDataSlot isi = virtualItem.getVirtualOutputDataSlots().get(i);
			file.println("symlink(\"" + getInputImage(isi, subgraphNumber) + "\",shift(@_));", 1);
		}

		// Output result
		InputParameterSlot isp = (InputParameterSlot)virtualItem.getVirtualOutputResultSlot();
		if (isp.getConnectedSlot() == null) {
			file.println("return 0;", 1);
		} else {
			file.println("return $result" + subgraphNumber + "[" + isp.getConnectedSlot().getGraphItem().getId() + "][" + isp.getConnectedSlot().getRank() + "];", 1);
		}
		file.println("}");

		generateSubroutinesDefinition(litems, file);
	}

	/**
	 * Generates the code for the specified MacroOperator body.
	 * 
	 * @param item the item to encode.
	 * @param fp the output stream.
	 * @throws Exception In case of incomplete graph.
	 */
	@Override
	protected void generateSubGraph( MacroOperatorItem item, FileOutputCode file ) throws BadGraphException {
		SubGraph subGraph = item.getSubGraph();
		List<GraphItem> litems = buildItemsSequence(subGraph);

		// Get and discard the virtual operator.
		Object virtualOperator = getVirtualOperator(litems);
		litems.remove(virtualOperator);
		if (litems.isEmpty()) {
			throw new BadGraphException(item.getNickName() + ": " + Ariane.getString("Generator.error.incomplete"));
		}

		file.println("\n#\n# " + item.getNickName());
		for (GraphItem note : subGraph.getGraphItems()) {
			if (note instanceof NoteItem) {
				file.println("# " + parseComment(note.getFormula()));
			}
		}
		file.println(" ");

		file.println("sub macro" + item.getId() + "_" + item.getGraph().getIdentifier() + " {");
		// Get parameters.
		int numSubgraph = subGraph.getIdentifier();
		file.println("# Input parameters", 1);
		file.println("my $result" + numSubgraph + ";", 1);
		for (int i = 0; i < item.getInputParameterSlots().size(); i++) {
			file.println("$result" + numSubgraph + "[0][" + i + "] =shift(@_);", 1);
		}

		// Get mask.
		file.println("# Input data files", 1);
		InputMaskSlot mask = item.getInputMaskSlot();
		if (mask != null && mask.isConnected() && _outputImages.contains(getInputImage(mask, item.getGraph().getIdentifier()))) {
			int maskIndex = subGraph.getVirtualOperator().getVirtualInputMaskSlot().getRank();
			file.println("symlink(shift(@_), \"" + createImageName(maskIndex, 0, numSubgraph) + "\");", 1);
		}

		// Get input images.
		int rank = 0;
		for (InputDataSlot isi : item.getInputDataSlots() ) {
			// Unlinked input!
			if (isi.getConnectedSlot() == null) {
				throw new BadGraphException("Generator.error.incomplete");
			}
			file.println("symlink(shift(@_), \"" + createImageName(rank++, 0, numSubgraph) + "\");", 1);
		}
		file.println();

		generateSubroutineBody(litems, numSubgraph, file, 1);

		VirtualItem virtualItem = (VirtualItem)virtualOperator;
		// Get output images.
		file.println();
		file.println("# Output images", 1);
		for (int i = 0; i < item.getOutputDataSlots().size(); i++) {
			InputDataSlot isi = virtualItem.getVirtualOutputDataSlots().get(i);
			file.println("symlink(\"" + getInputImage(isi, numSubgraph) + "\",shift(@_));", 1);
		}
		// Output result.
		InputParameterSlot isp = (InputParameterSlot)virtualItem.getVirtualOutputResultSlot();
		if (isp.getConnectedSlot() == null) {
			file.println("return 0;", 1);
		} else {
			file.println("return $result" + numSubgraph + "[" + isp.getConnectedSlot().getGraphItem().getId() + "][" + isp.getConnectedSlot().getRank() + "];", 1);
		}
		file.println("}");

		generateSubroutinesDefinition(litems, file);
	}
}
