/* -*- c-basic-offset: 3 -*-
 *
 * Copyright (c) 2013, GREYC.
 * All rights reserved
 *
 * You may use this file under the terms of the BSD license as follows:
 *
 * "Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of the GREYC, nor the name of its
 *     contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
 * 
 *   
 * For more information, refer to:
 * https://clouard.users.greyc.fr/Pandore
 */

package fr.greyc.pantheon.ariane.generator;

import fr.greyc.pantheon.ariane.Ariane;
import fr.greyc.pantheon.ariane.graph.Graph;
import fr.greyc.pantheon.ariane.graph.SubGraph;
import fr.greyc.pantheon.ariane.graph.item.CompositeItem;
import fr.greyc.pantheon.ariane.graph.item.ConstantItem;
import fr.greyc.pantheon.ariane.graph.item.EvalItem;
import fr.greyc.pantheon.ariane.graph.item.ControlStructureForItem;
import fr.greyc.pantheon.ariane.graph.item.ControlStructureIfItem;
import fr.greyc.pantheon.ariane.graph.item.ControlStructureMergeItem;
import fr.greyc.pantheon.ariane.graph.item.ControlStructureSelectItem;
import fr.greyc.pantheon.ariane.graph.item.ControlStructureWhileItem;
import fr.greyc.pantheon.ariane.graph.item.FileNameItem;
import fr.greyc.pantheon.ariane.graph.item.FormatItem;
import fr.greyc.pantheon.ariane.graph.item.GraphItem;
import fr.greyc.pantheon.ariane.graph.item.LoaderItem;
import fr.greyc.pantheon.ariane.graph.item.MacroOperatorItem;
import fr.greyc.pantheon.ariane.graph.item.NoteItem;
import fr.greyc.pantheon.ariane.graph.item.OperatorItem;
import fr.greyc.pantheon.ariane.graph.item.SaverItem;
import fr.greyc.pantheon.ariane.graph.item.ScriptItem;
import fr.greyc.pantheon.ariane.graph.item.SystemItem;
import fr.greyc.pantheon.ariane.graph.item.ViewerItem;
import fr.greyc.pantheon.ariane.graph.item.VirtualItem;
import fr.greyc.pantheon.ariane.graph.slot.InputSlot;
import fr.greyc.pantheon.ariane.graph.slot.InputDataSlot;
import fr.greyc.pantheon.ariane.graph.slot.InputMaskSlot;
import fr.greyc.pantheon.ariane.graph.slot.InputParameterSlot;
import fr.greyc.pantheon.ariane.graph.slot.OutputDataSlot;
import fr.greyc.pantheon.ariane.ui.PreferenceChooser;

import java.text.DateFormat;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Date;
import java.util.LinkedList;
import java.util.List;
import java.util.StringTokenizer;

/**
 * The <code>PerlGenerator</code> generates Python code
 * for the specified graph.
 * It operates in three steps:<ol>
 * <li>transforms the graph into a list of operators usable for the topological sort;
 * <li>executes the topological sort;
 * <li>generates the related code.
 * </ol>
 *
 * @version Régis CLOUARD - Dec 28, 2011
 */ 
public final class PythonGenerator extends CodeGenerator {
	/** The list of generated output images. */
	private final List<String> _outputImages = new ArrayList<String>();

	protected PythonGenerator( ) {
		// Only created by the code generator.
	}

	/**
	 * Generates the header code of the Python file and
	 * the code that checks the argument line consistency.
	 * 
	 * @param fp the output stream.
	 * @param filename the name of the code file.
	 * @param arguments the list of arguments defined in the graph.
	 * @param inputs the list of loaders defined in the graph.
	 * @param outputs the list of savers defined in the graph.
	 */
	@Override
	protected void generateScriptHeader( FileOutputCode file, String filename, List<GraphItem> arguments, List<GraphItem> inputs, List<GraphItem> outputs) {
		// The current date.
		Date creationDate = new Date();
		DateFormat dateFormat = DateFormat.getDateInstance(DateFormat.MEDIUM);
		file.println("# -*- coding: utf-8 -*-");
		file.println("#");
		file.println("# @(#)" + filename + "				" + dateFormat.format(creationDate));
		file.println("#");
		file.println("# This file was automatically generated by Ariane.");
		file.println("# Ariane is part of the Pantheon Project.");
		file.println("#\n# GREYC - ENSICAEN & University of Caen Basse-Normandie, France");

		file.println("import sys, getopt");
		file.println("import subprocess");
		file.println("import os");
		file.println("import os.path");
		file.println("import shutil");
		file.println("import glob");
		file.println("from math import *\n");

		file.println("verbose=False\n", 0);
		file.println("def printUsage( ):", 0);
		file.print("print \"Usage: $0 [-hv]", 1);
		for (int i = 0; i < arguments.size(); i++) {
			file.print(" " + arguments.get(i).getNickName());
		}
		for (int i = 0; i < inputs.size();) {
			file.print(" input-file" + (++i));
		}
		for (int i = 0; i < outputs.size();) {
			file.print(" output-file" + (++i));
		}
		file.println("\"");
		file.println("print \"  -h help\"", 1);
		file.println("print \"  -v verbose\"", 1);
		file.print("print \"  example: $0", 1);
		for (int i = 0; i < arguments.size(); i++) {
			file.print(" " + arguments.get(i).getNickName());
		}
		for (int i = 0; i < inputs.size(); i++) {
			file.print(" " + inputs.get(i).getNickName());
		}
		for (int i = 0; i < outputs.size(); i++) {
			String name = outputs.get(i).getNickName();
			if ("saver".equals(name)) {
				name += outputs.get(i).getId();
			}
			file.print(" " + name);
		}
		file.println("\"");

		file.println("sys.exit(1)\n", 1);
	}

	/**
	 * Generates the footer of the script file.
	 * 
	 * @param fp the output stream.
	 */
	@Override
	protected void generateScriptFooter( FileOutputCode file ) {
		// Do nothing
	}

	/**
	 * Generates utility functions required to execute the script.
	 * 
	 * @param fp the output stream.
	 */
	@Override
	protected void generateUtilitySubroutinesDefinition( FileOutputCode file ) {
		file.println("def execute( *command ):");
		file.println("''' Executes an operator. '''", 1);
		file.println("if (verbose): print command", 1);
		file.println("print subprocess.Popen(command, stdout=subprocess.PIPE).communicate()[0]", 1);
		file.println("return subprocess.Popen(\"" + GraphItem.getCommandStatus() + "\", stdout=subprocess.PIPE).communicate()[0]", 1);

		file.println("\ndef link( src, dest ):");
		file.println("''' Creates a symbolic link between the two files. '''", 1);
		file.println("try:", 1);
		file.println("os.symlink(src,dest)", 2);
		file.println("except OSError:", 1);
		file.println("pass", 2);
		file.println("", 1);
}

	/**
	 * Creates the name of the image from the specified slot number,
	 * item and graph identifier.
	 * 
	 * @param slotIndex the identifier of the slot.
	 * @param itemIndex the identifier of the item.
	 * @param graphNumber the graph identifier
	 * @return a name that is supposed unique.
	 */
	private String createImageName( int slotIndex, int itemIndex, int graphNumber ) {
		String name = PreferenceChooser.createTemporaryName(graphNumber, itemIndex, slotIndex);
		_outputImages.add(name);
		return name;
	}

	/**
	 * Returns the command line for getting the
	 * input images for the specified item slot of the specified
	 * graph identifier.
	 * 
	 * @param slot the related slot.
	 * @param graphNumber the graph identifier
	 * @return a name that is supposed to unique.
	 */
	private String getInputImage( InputSlot slot, int graphNumber ) {
		return PreferenceChooser.createTemporaryName(graphNumber, slot.getConnectedSlot().getGraphItem().getId(), slot.getConnectedSlot().getRank());
	}

	/**
	 * Creates a variable name with the specified arguments.
	 * 
	 * @param graph the corresponding graph.
	 * @param item the corresponding item.
	 * @param slot the corresponding slot.
	 * @return the name of the variable.
	 */
	private String variable( int graph, int item, int slot ) {
		return "var" + graph + "_" + item + "_" + slot;
	}

	/**
	 * Returns the parameters line command of the specified item.
	 * 
	 * @param item the related item.
	 * @param slotIndex the slot index.
	 * @param graphId the current graph identifier.
	 * @return a value or a command to get the value.
	 */
	private String getParameter( GraphItem item, int slotIndex, int graphId ) {
		InputParameterSlot isp = (InputParameterSlot) item.getSlot("InSlotParameter", slotIndex);
		if (isp.getConnectedSlot() == null) {
			return "\"" + ((InputParameterSlot)item.getSlot("InSlotParameter", slotIndex)).getValue() + "\"";
		} else {
			return variable(graphId, isp.getConnectedSlot().getGraphItem().getId(), isp.getConnectedSlot().getRank());
		}
	}

	/**
	 * Generates code for the main.
	 * 
	 * @param graph the graph to encode.
	 * @param fp the output stream.
	 * @param arguments the list of argument.
	 * @param inputs the list of inputs.
	 */
	@Override
	protected void generateMainPrologue( Graph graph, FileOutputCode file, List<GraphItem> arguments, List<GraphItem> inputs, List<GraphItem> outputs ) {
		file.println("\n### The main program"); 
		file.println("def main( args ):");

		// Comments
		file.println("''' main ", 1);
		for (GraphItem note : graph.getGraphItems()) {
			if (note instanceof NoteItem) {
				file.println(parseComment(note.getFormula()), 1);
			}
		}
		file.println("'''", 1);

		file.println("if (len(args) != " + (arguments.size() + inputs.size() + outputs.size()) + " ):", 1);
		file.println("printUsage()", 2);

		file.println("fileList= glob.glob(\"" + PreferenceChooser.getTemporaryFilePrefix() + "*" + PreferenceChooser.getTemporaryFileExtension() + "\")", 1);		
		file.println("for file in fileList: os.remove(file)", 1);
		file.println("argc = 0", 1);
		for ( GraphItem item : arguments ) {
			file.println(variable(1, item.getId(), 0) + " = args[argc]; argc += 1", 1);
		}
		for ( GraphItem item : inputs ) {
			file.println("link(args[argc], \"" + createImageName(0, item.getId(), 1) + "\"); argc += 1", 1);
		}
	}

	/**
	 * Generates code to erase all temporary files.
	 * 
	 * @param graph the graph to encode.
	 * @param fp the output stream.
	 */
	@Override
	protected void generateMainEpilogue( Graph graph, FileOutputCode file ) {
		file.println("\nif __name__ == \"__main__\":", 0);
		file.println("try:", 1);
		file.println("opts, args = getopt.getopt(sys.argv[1:], \"hv\")", 2);
		file.println("except getopt.GetoptError:", 1);
		file.println("printUsage()", 2);
		file.println("for opt, arg in opts:", 1);
		file.println("if opt=='-h':", 2);
		file.println("printUsage()", 3);
		file.println("if opt=='-v':", 2);
		file.println("verbose =True", 3);
		file.println("main(args)", 1);
	}

	/**
	 * Generates the code for loading an image.
	 * 
	 * @param item the item to encode.
	 * @param graphNumber the identifier number of the graph.
	 * @param fp the output stream.
	 * @param indent the indentation level.
	 * @throws Exception Incomplete graph.
	 */
	@Override
	protected void generateItem( LoaderItem item, int graphNumber, FileOutputCode file, int indent ) throws BadGraphException {
		if (item.getInputParameterSlots().size() > 0 && item.getInputParameterSlots().get(0).isConnected()) {
			String parameter = getParameter(item, 0, graphNumber);
			file.println("shutil.copy(" + parameter + ", "
			+ "\"" + createImageName(0, item.getId(), graphNumber) + "\")", indent);
		}
	}

	/**
	 * Generates the code for saving an image.
	 * 
	 * @param item the item to encode.
	 * @param graphNumber the identifier number of the graph.
	 * @param fp the output stream.
	 * @param indent the indentation level.
	 * @throws Exception Incomplete graph.
	 */
	@Override
	protected void generateItem( SaverItem item, int graphNumber, FileOutputCode file, int indent ) throws BadGraphException {
		if (item.getInputDataSlots().get(0).getConnectedSlot() != null) {
			if (item.getInputParameterSlots().size() > 0 && item.getInputParameterSlots().get(0).isConnected()) {
				String parameters = getParameter(item, 0, graphNumber);
				file.println("shutil.copy(\"" + getInputImage(item.getInputDataSlots().get(0), graphNumber)
						+ "\", " + parameters + ")", indent);
			} else {
				file.println("shutil.copy(\"" + getInputImage(item.getInputDataSlots().get(0), graphNumber)
					+ "\", args[argc]); argc+=1", indent);
			}
			file.println("link(\""
					+ getInputImage(item.getInputDataSlots().get(0), graphNumber)
					+ "\", \""
					+ createImageName(0, item.getId(), graphNumber)
					+ "\")", indent);
		}
	}

	/**
	 * Generates the code for viewer item.
	 * 
	 * @param item the item to encode.
	 * @param graphNumber the identifier number of the graph.
	 * @param fp the output stream.
	 * @param indent the indentation level.
	 * @throws Exception Incomplete graph.
	 */
	@Override
	protected void generateItem( ViewerItem item, int graphNumber, FileOutputCode file, int indent ) throws BadGraphException {
		String commandShell = item.getViewerCommandName();
		if (item.getInputDataSlots().get(0).getConnectedSlot() != null) {
			file.println("subprocess.call([\"" + commandShell + "\", \"" + getInputImage(item.getInputDataSlots().get(0), graphNumber) + "\"])", indent);
			file.println("link (\""
					+ getInputImage(item.getInputDataSlots().get(0), graphNumber)
					+ "\", \""
					+ createImageName(0, item.getId(), graphNumber)
					+ "\")", indent);
		}
	}

	/**
	 * Generates the code for executing an operator.
	 * 
	 * @param item the operator to encode.
	 * @param graphNumber the identifier number of the graph.
	 * @param fp the output stream.
	 * @param indent the indentation level.
	 * @throws Exception Incomplete graph.
	 */
	@Override
	protected void generateItem( OperatorItem item, int graphNumber, FileOutputCode file, int indent ) throws BadGraphException {
		String codeName = OperatorItem.getCommandPart(item.getBinary());
		if (codeName == null) {
			throw new BadGraphException(Ariane.getString("Generator.error2") + item.getName());
		}

		// Get parameters
		List<String> parameterLine = new LinkedList<String>();
		for (int i = 0; i < item.getInputParameterSlots().size(); i++) {
			parameterLine.add(getParameter(item, i, graphNumber));
		}

		// Get mask
		List<String> maskLine = new LinkedList<String>();
		InputMaskSlot mask = item.getInputMaskSlot();
		if (mask != null && mask.isConnected() && _outputImages.contains(getInputImage(mask, graphNumber))) {
			maskLine.add(getInputImage(mask, graphNumber));
		}

		// Get input images.
		List<String> inputLine = new LinkedList<String>();
		for (InputDataSlot image : item.getInputDataSlots() ) {
			if (image.getConnectedSlot() == null) {
				throw new BadGraphException(item.getNickName() + ": " + Ariane.getString("Generator.error.incomplete"));
			}
			inputLine.add("\"" + getInputImage(image, graphNumber) + "\"");
		}


		// Get output images.
		List<String> outputLine = new LinkedList<String>();
		List<OutputDataSlot> outputDataSlots = item.getOutputDataSlots();
		for (int i = 0; i < outputDataSlots.size(); i++) {
			outputLine.add("\"" + createImageName(i, item.getId(), graphNumber) + "\"");
		}

		String code = null;
		List<String> commandLine = item.buildAbstractCommandLine(maskLine, parameterLine, inputLine, outputLine);
		for (String command : commandLine) {
			if (code == null) {
				code = "\"" + command + "\"";
			} else {
				code += ", " + command;
			}
		}

		// Get output result.
		file.println(variable(graphNumber, item.getId(), 0) + " = " + "execute(" + code + ")", indent);
	}

	/**
	 * Generates the code for executing an operating system command.
	 * 
	 * @param item the operator to encode.
	 * @param graphNumber the identifier number of the graph.
	 * @param fp the output stream.
	 * @param indent the indentation level.
	 * @throws Exception Incomplete graph.
	 */
	@Override
	protected void generateItem( SystemItem item, int graphNumber, FileOutputCode file, int indent ) throws BadGraphException {
		String codeName = OperatorItem.getCommandPart(item.getBinary());
		if (codeName == null) {
			throw new BadGraphException(Ariane.getString("Generator.error2") + item.getName());
		}

		// Get parameters
		List<String> parameterLine = new LinkedList<String>();
		for (int i = 0; i < item.getInputParameterSlots().size(); i++) {
			parameterLine.add(getParameter(item, i, graphNumber));
		}

		// Get mask
		List<String> maskLine = new LinkedList<String>();
		InputMaskSlot mask = item.getInputMaskSlot();
		if (mask != null && mask.isConnected() && _outputImages.contains(getInputImage(mask, graphNumber))) {
			maskLine.add(getInputImage(mask, graphNumber));
		}

		// Get input images.
		List<String> inputLine = new LinkedList<String>();
		for (InputDataSlot image : item.getInputDataSlots() ) {
			if (image.getConnectedSlot() == null) {
				throw new BadGraphException(item.getNickName() + ": " + Ariane.getString("Generator.error.incomplete"));
			}
			inputLine.add("\"" + getInputImage(image, graphNumber) + "\"");
		}

		// Get output images.
		List<OutputDataSlot> outputDataSlots = item.getOutputDataSlots();
		List<String> outputLine = new LinkedList<String>();
		for (int i = 0; i < outputDataSlots.size(); i++) {
			outputLine.add("\"" + createImageName(i, item.getId(), graphNumber) + "\"");
		}

		List<String> commandLine = item.buildAbstractCommandLine(maskLine, parameterLine, inputLine, outputLine);
		String code = null;
		for ( String command : commandLine) {
			if (code == null) {
				code = "\"" + command + "\"";
			} else if (command.startsWith("var")) {
				code += ", " + command;
			} else {
				code += ", \"" + command  + "\"";
			}
		}

		// Get output result.
		file.println(variable(graphNumber, item.getId(), 0) + " = " + "execute(" + code + ")", indent);
	}

	/**
	 * Parses the give formula into Perl code.
	 * 
	 * @param formula the function.
	 * @param parameters the list of parameters.
	 * @return the Perl code.
	 */
	private String parseFormula( String formula, String[] parameters ) {
		// Parse the formula.
		StringTokenizer stringTokenizer = new StringTokenizer(formula, "$ -+()*/^0123456789.E", true);
		String etape;
		String code = "";
		while (stringTokenizer.hasMoreTokens()) {
			etape = stringTokenizer.nextToken();
			if ("^".equals(etape)) {
				code += "**";
			} else if ("ln".equals(etape)) {
				code += "log";
			} else if ("atan".equals(etape)) {
				code += "a";
			} else if ("$".equals(etape)) {
				int nnumberOfParameters = Integer.parseInt(stringTokenizer.nextToken());
				code += "float(" + parameters[nnumberOfParameters] + ")";
			} else {
				code += etape;
			}
		}
		return code;
	}

	/**
	 * Generates the code for an Eval item.
	 * 
	 * @param item the item to encode.
	 * @param graphNumber the identifier number of the graph.
	 * @param fp the output stream.
	 * @param indent the indentation level.
	 * @throws Exception Incomplete graph.
	 */
	@Override
	protected void generateItem( EvalItem item, int graphNumber, FileOutputCode file, int indent ) throws BadGraphException {
		String[] parameters = new String[item.getInputParameterSlots().size()];

		// Get parameters
		for (int i = 0; i < item.getInputParameterSlots().size(); i++) {
			parameters[i] = getParameter(item, i, graphNumber);
		}

		file.println(variable(graphNumber, item.getId(), 0) + " = str("
				+ parseFormula(item.getFormula(), parameters)
				+ ")", indent);
	}

	@Override
	protected void generateItem( FormatItem item, int graphNumber, FileOutputCode file, int indent ) throws BadGraphException {
		String[] parameters = new String[item.getInputParameterSlots().size()];
		// Get parameters
		for (int i = 0; i < item.getInputParameterSlots().size(); i++) {
			parameters[i] = getParameter(item, i, graphNumber);
		}

		String code = "";
		StringTokenizer stringTokenizer = new StringTokenizer(item.getFormula(), " $()+'\"", true);
		String etape;
		while (stringTokenizer.hasMoreTokens()) {
			etape = stringTokenizer.nextToken();
			if ("+".equals(etape)) {
				code += "+";
			} else if ("quote".equals(etape)) {
				code += "";
			} else if ("\"".equals(etape)) {
				code += "";
			} else if ("(".equals(etape)) {
				code += "\"'\"+";
			} else if (")".equals(etape)) {
				code += "+\"'\"";
			} else if ("$".equals(etape)) {
				int numberOfParameters = Integer.parseInt(stringTokenizer.nextToken());
				code += parameters[numberOfParameters];
			} else {
				code += etape;
			}
		}
		file.println(variable(graphNumber, item.getId(), 0) + " = " + code, indent);
	}

	/**
	 * Generates the code for the control structure Select.
	 * 
	 * @param item the select to encode.
	 * @param graphNumber the identifier number of the graph.
	 * @param fp the output stream.
	 * @param indent the indentation level.
	 * @throws Exception Incomplete graph.
	 */
	@Override
	protected void generateItem( ControlStructureSelectItem item, int graphNumber, FileOutputCode file, int indent ) throws BadGraphException {
		String value = getParameter(item, 0, graphNumber);

		file.println("# Select", indent);
		for (int i = 0; i < item.getInputDataSlots().size(); i++) {
			file.println("if (int(" + value + ") == " + i + "):", indent);
			file.println("link(\""
					+ getInputImage(item.getInputDataSlots().get(i), graphNumber) + "\", \""
					+ createImageName(0, item.getId(), graphNumber) + "\")", indent + 1);
			file.println(variable(graphNumber, item.getId(), 0) + " = \"" + i + "\"", indent + 1);
		}
	}

	/**
	 * Generates the code for the control structure If.
	 * 
	 * @param item the item to encode.
	 * @param graphNumber the identifier number of the graph.
	 * @param fp the output stream.
	 * @param indent the indentation level.
	 * @throws Exception Incomplete graph.
	 */
	@Override
	protected void generateItem( ControlStructureIfItem item, int graphNumber, FileOutputCode file, int indent ) throws BadGraphException {
		String[] parameters = new String[2];
		for (int i = 0; i < item.getInputParameterSlots().size(); i++) {
			parameters[i] = getParameter(item, i, graphNumber);
		}

		// Parse the boolean expression.
		StringTokenizer stringTokenizer = new StringTokenizer(item.getFormula(), "&|$^!0123456789.E><=() +*-/", true);
		String etape;
		String code = "";
		while (stringTokenizer.hasMoreTokens()) {
			etape = stringTokenizer.nextToken();
			if ("&".equals(etape)) {
				code += " and ";
			} else if ("|".equals(etape)) {
				code += " or ";
			} else if ("$".equals(etape)) {
				if (stringTokenizer.nextToken().equals("0")) {
					code += "float(" + parameters[0] + ")";
				} else {
					code += "float(" + parameters[0] + ")";
				}
			} else {
				code += etape;
			}
		}

		file.println("if (" + code + "):", indent);
	}

	/**
	 * Generates the code for the 'then' part.
	 * 
	 * @param item the item to encode.
	 * @param graphNumber the identifier number of the graph.
	 * @param fp the output stream.
	 * @param indent the indentation level.
	 * @throws Exception Incomplete graph.
	 */
	@Override
	protected void generateThenPart( ControlStructureIfItem item, int graphNumber, FileOutputCode file, int indent) throws BadGraphException {
		file.println("link(\""
				+ getInputImage(item.getInputDataSlots().get(0), graphNumber) + "\", \""
				+ createImageName(0, item.getId(), graphNumber) + "\")", indent + 1);
		file.println(variable(graphNumber, item.getId(), 0) + " = 0", indent + 1);
	}

	/**
	 * Generates the code for the 'else' part.
	 * 
	 * @param item the item to encode.
	 * @param graphNumber the identifier number of the graph.
	 * @param fp the output stream.
	 * @param indent the indentation level.
	 * @throws Exception Incomplete graph.
	 */
	@Override
	protected void generateElsePart( ControlStructureIfItem item, int graphNumber, FileOutputCode file, int indent) throws BadGraphException {
		file.println("else:", indent);
		file.println("link(\""
				+ getInputImage(item.getInputDataSlots().get(0), graphNumber) + "\", \""
				+ createImageName(1, item.getId(), graphNumber) + "\")", indent + 1);
		file.println(variable(graphNumber, item.getId(), 0) + " = 1", indent + 1);
	}

	/**
	 * Generates the code for the end of a if control structure.
	 * 
	 * @param item the item to encode.
	 * @param graphNumber the identifier number of the graph.
	 * @param fp the output stream.
	 * @param indent the indentation level.
	 * @throws Exception Incomplete graph.
	 */
	@Override
	protected void generateFiPart( ControlStructureIfItem item, int graphNumber, FileOutputCode file, int indent) throws BadGraphException {
		// Nothing to do
	}


	/**
	 * Generates the code for the control structure merge.
	 * 
	 * @param item the item to encode.
	 * @param graphNumber the identifier number of the graph.
	 * @param fp the output stream.
	 * @param indent the indentation level.
	 * @throws Exception Incomplete graph.
	 */
	@Override
	protected void generateItem( ControlStructureMergeItem item, int graphNumber, FileOutputCode file, int indent ) throws BadGraphException {
		InputDataSlot in0 = item.getInputDataSlots().get(0);

		int id0 = in0.getConnectedSlot().getGraphItem().getId();
		
		file.println("# Merge" , indent);
		file.println("try: " + variable(graphNumber, id0, 0) + "; input= 0", indent);
		file.println("except UnboundLocalError: input=1" , indent);
		file.print("if (input== 0) and ", indent);
		file.println("os.path.isfile(\"" + getInputImage(item.getInputDataSlots().get(0), graphNumber) + "\"):");
		file.println("link(\""
				+ getInputImage(item.getInputDataSlots().get(0), graphNumber)
				+ "\", \""
				+ createImageName(0, item.getId(), graphNumber) + "\")", indent + 1);
		file.println(variable(graphNumber, item.getId(), 0) + " = 0", indent + 1);
		file.println("else:", indent);
		file.println("link(\""
				+ getInputImage(item.getInputDataSlots().get(1), graphNumber) + "\", \""
				+ createImageName(0, item.getId(), graphNumber) + "\")", indent + 1);
		file.println(variable(graphNumber, item.getId(), 0) + " = 1", indent + 1);
	}

	/**
	 * Generates the code for a note. Nothing to do.
	 * 
	 * @param item the note to encode.
	 * @param graphNumber the identifier number of the graph.
	 * @param fp the output stream.
	 * @param indent the indentation level.
	 */
	@Override
	protected void generateItem( NoteItem item, int graphNumber, FileOutputCode file, int indent ) {
		// Do nothing
	}

	/**
	 * Generates the code for a <code>Filename</code> item.
	 * 
	 * @param item the item to encode.
	 * @param graphNumber the identifier number of the graph.
	 * @param fp the output stream.
	 * @param indent the indentation level.
	 * @throws Exception Incomplete graph.
	 */
	@Override
	protected void generateItem( FileNameItem item, int graphNumber, FileOutputCode file, int indent ) throws BadGraphException {
		if (!item.isArgument()) {
			file.println(variable(graphNumber, item.getId(), 0) + " = '" + item.getFormula() + "'", indent);
		}
	}

	/**
	 * Generates the code for an <code>Constant</code>item.
	 * 
	 * @param item the item to encode.
	 * @param graphNumber the identifier number of the graph.
	 * @param fp the output stream.
	 * @param indent the indentation level.
	 * @throws Exception Incomplete graph.
	 */
	@Override
	protected void generateItem( ConstantItem item, int graphNumber, FileOutputCode file, int indent ) throws BadGraphException {
		if (!item.isArgument() ) {
			String value = item.getFormula().replaceAll("\"", "\\\\\"");
			file.println(variable(graphNumber, item.getId(), 0) + " = \"" + value + "\"", indent);
		}
	}

	/**
	 * Generates the code for an <code>Script</code>item.
	 * 
	 * @param item the item to encode.
	 * @param graphNumber the identifier number of the graph.
	 * @param fp the output stream.
	 * @param indent the indentation level.
	 * @throws Exception Incomplete graph.
	 */
	@Override
	protected void generateItem( ScriptItem item, int graphNumber, FileOutputCode file, int indent ) throws BadGraphException {
		String value = item.getFormula().replaceAll("\"", "\\\\\"").replaceAll("\n", "\\\\\n");
		file.println(variable(graphNumber, item.getId(), 0) + " = \"" + value + "\"", indent);
	}

	/**
	 * Generates the code for the specified MacroOperator,
	 * FFor, or While call.
	 * 
	 * @param item the item to encode.
	 * @param graphNumber the identifier number of the graph.
	 * @param fp the output stream.
	 * @param indent the indentation level.
	 * @throws Exception in case of incomplete graph.
	 */
	@Override
	protected void generateItem( CompositeItem item, int graphNumber, FileOutputCode file, int indent ) throws BadGraphException {
		String code = "macro" + item.getId() + "_" + item.getGraph().getIdentifier() + "(";
		boolean comma = false;

		// Get parameters
		for (int i = 0; i < item.getInputParameterSlots().size(); i++) {
			if (comma) {
				code += ", ";
			} else {
				comma = true;
			}
			code += getParameter(item, i, graphNumber);
		}

		// Get mask
		InputMaskSlot mask = item.getInputMaskSlot();
		if (mask != null && mask.isConnected() && _outputImages.contains(getInputImage(mask, graphNumber))) {
			if (comma) {
				code += ", ";
			} else {
				comma = true;
			}
			code += "\"" + getInputImage(mask, graphNumber) + "\"";
		}

		// Get input images.
		for ( InputDataSlot image : item.getInputDataSlots() ) {
			if (image.getConnectedSlot() == null) {
				throw new BadGraphException(Ariane.getString("Generator.error.incomplete"));
			}
			if (comma) {
				code += ", ";
			} else {
				comma = true;
			}
			code += "\"" + getInputImage(image, graphNumber) + "\"";
		}

		// Get output images.
		Collection<OutputDataSlot> outputDataSlots = item.getOutputDataSlots();
		for (int i = 0; i < outputDataSlots.size(); i++) {
			if (comma) {
				code += ", ";
			} else {
				comma = true;
			}
			code += "\""  + createImageName(i, item.getId(), graphNumber) + "\"";
		}
		// Get output result.
		file.println(variable(graphNumber, item.getId(), 0) + " = " + code + ")", indent);
	}

	/**
	 * Parses the comment in order to add a sharp after each newlines.
	 * 
	 * @param comment a string that constrains the comment.
	 * @return a new string.
	 */
	private String parseComment( String comment ) {
		return comment.replaceAll("\n", "\n    ");
	}

	/**
	 * Generates the code for the specified control structure For.
	 * 
	 * @param item the item to encode.
	 * @param fp the output stream.
	 * @throws Exception Incomplete graph.
	 */
	@Override
	protected void generateSubGraph( ControlStructureForItem item, FileOutputCode file ) throws BadGraphException {
		SubGraph subGraph = item.getSubGraph();
		List<GraphItem> litems = buildItemsSequence(subGraph);

		// Get and discard the virtual operator.
		Object virtualOperator = getVirtualOperator(litems);
		litems.remove(virtualOperator);
		if (litems.isEmpty()) {
			throw new BadGraphException(item.getNickName() + ": " + Ariane.getString("Generator.error.incomplete"));
		}

		file.println("def macro" + item.getId() + "_" + item.getGraph().getIdentifier() + "( *args ):");
		// Comments
		file.println("''' " + item.getNickName() + "\n", 1);
		for (GraphItem note : subGraph.getGraphItems()) {
			if (note instanceof NoteItem) {
				file.println(note.getFormula(), 1);
			}
		}
		file.println("'''", 1);
		file.println("argc = 0", 1);

		// Get parameters
		int subgraphNumber = subGraph.getIdentifier();
		file.println("# Input parameters", 1);
		file.println("begin = int(args[argc]); argc += 1", 1);
		file.println("end = int(args[argc]); argc += 1", 1);
		file.println("step = int(args[argc]); argc += 1", 1);
		for (int i = 3; i < item.getInputParameterSlots().size(); i++) {
			file.println(variable(subgraphNumber, 0, i - 2) + " = args[argc]; argc += 1", 1);
		}
		file.println(variable(subgraphNumber, 0, 0) + " = str(begin)", 1);

		// Get mask.
		file.println("# Input data files", 1);
		InputMaskSlot mask = item.getInputMaskSlot();
		if (mask != null && mask.isConnected() && _outputImages.contains(getInputImage(mask, item.getGraph().getIdentifier()))) {
			int maskIndex = subGraph.getVirtualOperator().getVirtualInputMaskSlot().getRank();
			file.println("link(args[argc], \"" + createImageName(maskIndex, 0, subgraphNumber) + "\"); argc += 1", 1);
		}

		// Get input files
		int rank = 0;
		for (InputDataSlot isi : item.getInputDataSlots() ) {
			// Unlinked input!
			if (isi.getConnectedSlot() == null) {
				throw new BadGraphException("Generator.error.incomplete");
			}
			file.println("shutil.copy(args[argc], \"" + createImageName(rank++, 0, subgraphNumber) + "\"); argc += 1", 1);
		}

		// Create the output images in case of no iteration
		VirtualItem virtualItem = (VirtualItem)virtualOperator;
		for (int i = 0; i < item.getOutputDataSlots().size(); i++) {
			InputDataSlot isi = virtualItem.getVirtualOutputDataSlots().get(i);
			file.println("f = open(\"" + getInputImage(isi, subgraphNumber) + "\",\"w\"); f.close()", 1);
		}

		// Initialize the return variable in case of no iteration
		{
			InputParameterSlot isp = (InputParameterSlot)virtualItem.getVirtualOutputResultSlot();
			file.println(variable(subgraphNumber, isp.getConnectedSlot().getGraphItem().getId(), isp.getConnectedSlot().getRank()) + "=\"SUCCESS\"", 1);
		}

		file.println();
		file.println("for i in range(begin, end+1, step):", 1);
		file.println(variable(subgraphNumber, 0, 0) + " = str(i)", 2);

		generateSubroutineBody(litems, subgraphNumber, file, 2);

		// Get intermediate images
		for (int i = 0; i < item.getInputDataSlots().size(); i++) {
			InputDataSlot isi = virtualItem.getVirtualIntermediateDataSlots(i);
			// Unlinked input!
			if (isi.getConnectedSlot() == null) {
				throw new BadGraphException("Generator.error.incomplete");
			}
			String input = getInputImage(isi, subgraphNumber);
			String output = createImageName(i, 0, subgraphNumber);
			if (!input.equals(output)) {
				file.println("shutil.copy(\"" + input + "\",\"" + output + "\")", 2);
			}
		}

		// Get output images.
		file.println();
		file.println("# Output images", 1);
		for (int i = 0; i < item.getOutputDataSlots().size(); i++) {
			InputDataSlot isi = virtualItem.getVirtualOutputDataSlots().get(i);
			file.println("link(\"" + getInputImage(isi, subgraphNumber) + "\",args[argc]); argc += 1", 1);
		}

		// Output result
		InputParameterSlot isp = (InputParameterSlot)virtualItem.getVirtualOutputResultSlot();
		if (isp.getConnectedSlot() == null) {
			file.println("return 0", 1);
		} else {
			file.println("return " + variable(subgraphNumber, isp.getConnectedSlot().getGraphItem().getId(), isp.getConnectedSlot().getRank()), 1);
		}
		file.println("");

		generateSubroutinesDefinition(litems, file);
	}

	/**
	 * Generates the code for the specified control structure While.
	 * 
	 * @param item the item to encode.
	 * @param fp the output stream.
	 * @throws Exception Incomplete graph.
	 */
	@Override
	protected void generateSubGraph( ControlStructureWhileItem item, FileOutputCode file ) throws BadGraphException {
		SubGraph subGraph = item.getSubGraph();
		List<GraphItem> litems = buildItemsSequence(subGraph);

		// Get and discard the virtual operator.
		Object virtualOperator = getVirtualOperator(litems);
		litems.remove(virtualOperator);
		if (litems.isEmpty()) {
			throw new BadGraphException(item.getNickName() + ": " + Ariane.getString("Generator.error.incomplete"));
		}

		file.println("def macro" + item.getId() + "_" + item.getGraph().getIdentifier() + "( *args ):");
		// Comments
		file.println("''' " + item.getNickName() + "\n", 1);
		for (GraphItem note : subGraph.getGraphItems()) {
			if (note instanceof NoteItem) {
				file.println(note.getFormula(), 1);
			}
		}
		file.println("'''", 1);
		file.println("argc = 0", 1);

		// Get parameters
		int subgraphNumber = subGraph.getIdentifier();
		String[] parameters = new String[item.getInputParameterSlots().size()];
		file.println("# Input parameters", 1);
		for (int i = 0; i < item.getInputParameterSlots().size(); i++) {
			file.println(variable(subgraphNumber, 0, i) + " = args[argc]; argc += 1", 1);
			parameters[i] = variable(subgraphNumber, 0, i);
		}

		// Get mask.
		file.println("# Input data files", 1);
		InputMaskSlot mask = item.getInputMaskSlot();
		if (mask != null && mask.isConnected() && _outputImages.contains(getInputImage(mask, item.getGraph().getIdentifier()))) {
			int maskIndex = subGraph.getVirtualOperator().getVirtualInputMaskSlot().getRank();
			file.println("link(args[argc], \"" + createImageName(maskIndex, 0, subgraphNumber) + "\"); argc += 1", 1);
		}

		// Get input files
		int rank = 0;
		for (InputDataSlot isi : item.getInputDataSlots() ) {
			// Unlinked input!
			if (isi.getConnectedSlot() == null) {
				throw new BadGraphException("Generator.error.incomplete");
			}
			file.println("shutil.copy(args[argc], \"" + createImageName(rank++, 0, subgraphNumber) + "\"); argc += 1", 1);
		}

		// Parse the boolean expression.
		StringTokenizer stringTokenizer = new StringTokenizer(item.getFormula(), "&|$^!0123456789.E><=() +*-/", true);
		String etape;
		String testFormula = "";
		while (stringTokenizer.hasMoreTokens()) {
			etape = stringTokenizer.nextToken();
			if ("&".equals(etape)) {
				testFormula += " and ";
			} else if ("|".equals(etape)) {
				testFormula += " or ";
			} else if ("$".equals(etape)) {
				if (stringTokenizer.nextToken().equals("0")) {
					testFormula += "float(" + parameters[0] + ")";
				} else {
					testFormula += "float(" + parameters[1] + ")";
				}
			} else {
				testFormula += etape;
			}
		}

		VirtualItem virtualItem = (VirtualItem)virtualOperator;
		file.println();
		file.println("while (" + testFormula + "):", 1);

		generateSubroutineBody(litems, subgraphNumber, file, 2);

		// Get intermediate images
		for (int i = 0; i <  item.getInputDataSlots().size(); i++) {
			InputDataSlot isi = virtualItem.getVirtualIntermediateDataSlots(i);
			// Unlinked input!
			if (isi.getConnectedSlot() == null) {
				throw new BadGraphException("Generator.error.incomplete");
			}
			String input = getInputImage(isi, subgraphNumber);
			String output = createImageName(i, 0, subgraphNumber);
			if (!input.equals(output)) {
				file.println("shutil.copy(\"" + input + "\",\"" + output + "\")", 2);
			}
		}

		// Get intermediate parameters
		for (int i = 0; i < 2; i++) {
			InputParameterSlot isp = virtualItem.getVirtualIntermediateResultSlot(i);
			if (isp.getConnectedSlot() != null) {
				file.println(parameters[i] + " = " + variable(subgraphNumber, isp.getConnectedSlot().getGraphItem().getId(), isp.getConnectedSlot().getRank()), 2);
			}
		}

		// Get output images.
		file.println();
		file.println("# Output images", 1);
		for (int i = 0; i < item.getOutputDataSlots().size(); i++) {
			InputDataSlot isi = virtualItem.getVirtualOutputDataSlots().get(i);
			file.println("link(\"" + getInputImage(isi, subgraphNumber) + "\",args[argc]); argc += 1", 1);
		}

		// Output result
		InputParameterSlot isp = (InputParameterSlot)virtualItem.getVirtualOutputResultSlot();
		if (isp.getConnectedSlot() == null) {
			file.println("return 0", 1);
		} else {
			file.println("return " + variable(subgraphNumber, isp.getConnectedSlot().getGraphItem().getId(), isp.getConnectedSlot().getRank()), 1);
		}
		file.println("");

		generateSubroutinesDefinition(litems, file);
	}

	/**
	 * Generates the code for the specified MacroOperator body.
	 * 
	 * @param item the item to encode.
	 * @param fp the output stream.
	 * @throws Exception In case of incomplete graph.
	 */
	@Override
	protected void generateSubGraph( MacroOperatorItem item, FileOutputCode file ) throws BadGraphException {
		SubGraph subGraph = item.getSubGraph();
		List<GraphItem> litems = buildItemsSequence(subGraph);

		// Get and discard the virtual operator.
		Object virtualOperator = getVirtualOperator(litems);
		litems.remove(virtualOperator);
		if (litems.isEmpty()) {
			throw new BadGraphException(item.getNickName() + ": " + Ariane.getString("Generator.error.incomplete"));
		}

		file.println("def macro" + item.getId() + "_" + item.getGraph().getIdentifier() + "( *args ):");
		// Comments
		file.println("''' " + item.getNickName() + "\n", 1);
		for (GraphItem note : subGraph.getGraphItems()) {
			if (note instanceof NoteItem) {
				file.println(note.getFormula(), 1);
			}
		}
		file.println("'''", 1);
		file.println("argc = 0", 1);
		// Get parameters.
		int subgraphNumber = subGraph.getIdentifier();
		file.println("# Input parameters", 1);
		for (int i = 0; i < item.getInputParameterSlots().size(); i++) {
			file.println(variable(subgraphNumber, 0, i) + " = args[argc]; argc += 1", 1);
		}

		// Get mask.
		file.println("# Input data files", 1);
		InputMaskSlot mask = item.getInputMaskSlot();
		if (mask != null && mask.isConnected() && _outputImages.contains(getInputImage(mask, item.getGraph().getIdentifier()))) {
			int maskIndex = subGraph.getVirtualOperator().getVirtualInputMaskSlot().getRank();
			file.println("link(args[argc], \"" + createImageName(maskIndex, 0, subgraphNumber) + "\"); argc += 1", 1);
		}

		// Get input images.
		int rank = 0;
		for (InputDataSlot isi : item.getInputDataSlots() ) {
			// Unlinked input!
			if (isi.getConnectedSlot() == null) {
				throw new BadGraphException("Generator.error.incomplete");
			}
			file.println("link(args[argc], \"" + createImageName(rank++, 0, subgraphNumber) + "\"); argc += 1", 1);
		}

		generateSubroutineBody(litems, subgraphNumber, file, 1);

		// Get output images.
		VirtualItem virtualItem = (VirtualItem)virtualOperator;
		file.println();
		file.println("# Output images", 1);
		for (int i = 0; i < item.getOutputDataSlots().size(); i++) {
			InputDataSlot isi = virtualItem.getVirtualOutputDataSlots().get(i);
			file.println("link(\"" + getInputImage(isi, subgraphNumber) + "\",args[argc]); argc += 1", 1);
		}

		// Output result.
		InputParameterSlot isp = (InputParameterSlot)virtualItem.getVirtualOutputResultSlot();
		if (isp.getConnectedSlot() == null) {
			file.println("return 0", 1);
		} else {
			file.println("return " + variable(subgraphNumber, isp.getConnectedSlot().getGraphItem().getId(), isp.getConnectedSlot().getRank()), 1);
		}
		file.println("");

		generateSubroutinesDefinition(litems, file);
	}
}