/* -*- c-basic-offset: 3 -*-
 *
 * Copyright (c) 2013, GREYC.
 * All rights reserved
 *
 * You may use this file under the terms of the BSD license as follows:
 *
 * "Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of the GREYC, nor the name of its
 *     contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
 * 
 *   
 * For more information, refer to:
 * https://clouard.users.greyc.fr/Pandore
 */

package fr.greyc.pantheon.ariane.graph;

import fr.greyc.pantheon.ariane.Ariane;
import fr.greyc.pantheon.ariane.graph.item.GraphItem;
import fr.greyc.pantheon.ariane.graph.item.VirtualItem;

import java.awt.Component;
import java.awt.Point;

/**
 * A <code>SubGraph</code> holds configuration of a complex graph
 * item such as <code>MacroOperator</code> or <code>ControlStructureFor</code>.
 * Its inputs, parameters and outputs are held by a <code>VirtualOperator</code>.
 * 
 * @author Régis CLOUARD - Oct 01, 1998
 * @version Sébastien LEFÈVRE, Richard LOBRÉAU - Feb 21, 1999
 * @version Adrien SCHRÈQUE, Hicham AMADI - Nov 05, 2000
 * @version Mathieu JOETS - Jul 09, 2001
 * @version Régis CLOUARD - Oct 16, 2001
 * @version Régis CLOUARD - Jan 23, 2007 (move to front selected subgraph);
 */
public final class SubGraph extends Graph {
	/** The related graph item. */
	private GraphItem _parent;

	/** The <code>VirtualOperator</code> that supports the graph. */
	private final VirtualItem _virtualOperator;

	/**
	 * Creates a subgraph form the given attributes.
	 * <br>Also creates the related <code>VirtualOperator</code>.
	 */
	public SubGraph( GraphItem parent, String name, int inputDataSlotNumber, int inputParameterSlotNumber,
			int outputDataSlotNumber, int intermediateDataSlotNumber, int intermediateParamaterSlotNumber ) {
		super(name, false);
		_parent = parent;

		// Builds the related VirtualOperator (identifier=0).
		_virtualOperator = new VirtualItem(this, name, inputDataSlotNumber, inputParameterSlotNumber, outputDataSlotNumber , intermediateDataSlotNumber, intermediateParamaterSlotNumber);
		addItem(_virtualOperator, new Point(0, 0), true);
		for (int i = 0; i <  inputDataSlotNumber; i++) {
			add(_virtualOperator.getVirtualInputDataSlots().get(i), XYLayout.Direction.WESTNORTH, i);
		}
		for (int i = 0; i <  inputParameterSlotNumber; i++) {
			add(_virtualOperator.getVirtualInputParameterSlots().get(i), XYLayout.Direction.NORTHWEST, i);
		}
		add(_virtualOperator.getVirtualInputMaskSlot(), XYLayout.Direction.NORTHWEST, inputParameterSlotNumber);
		for (int i = 0; i <  outputDataSlotNumber; i++) {
			add(_virtualOperator.getVirtualOutputDataSlots().get(i), XYLayout.Direction.NORTHEAST, i);
		}
		add(_virtualOperator.getVirtualOutputResultSlot(), XYLayout.Direction.SOUTHEAST, 0);
		for (int i = 0; i <  intermediateDataSlotNumber; i++) {
			add(_virtualOperator.getVirtualIntermediateDataSlots(i), XYLayout.Direction.EASTSOUTH, i);
		}
		for (int i = 0; i <  intermediateParamaterSlotNumber; i++) {
			add(_virtualOperator.getVirtualIntermediateResultSlot(i), XYLayout.Direction.SOUTHWEST, i);
		}
	}

	/**
	 * Disposes the graph item (definitively).
	 * Tries to avoid memory leak.
	 */
	@Override
	public void disposeHook( ) {
		// Note: the virtual operator is disposed
		// with all other items of the subgraph.
		_parent = null;
	}

	/**
	 * Returns the super graph item that contains
	 * this graph.
	 * @return the graph item that encloses this graph.
	 */
	@Override
	public GraphItem getSuperGraphItem() {
		return _parent;
	}

	/**
	 * Returns the root graph of this subgraph.
	 * @return the root graph.
	 */
	@Override
	public Graph getRoot( ) {
		return _parent.getGraph().getRoot();
	}

	/**
	 * Returns the absolute name of the graph including
	 * the parent graphs.
	 * @return the name.
	 */
	@Override
	public String getName( ) {
		if (_parent == null) {
			return "";
		}
		return _parent.getGraph().getName() + "::" + super.getName();
	}

	/**
	 * Returns the related <code>VirtualOperator</code>.
	 * @return the virtualOperator.
	 */
	public VirtualItem getVirtualOperator( ) {
		return _virtualOperator;
	}

	/**
	 * @return true if the graph has been modified, false if not.
	 */
	@Override
	public boolean isModified( ) {
		return getRoot().isModified();
	}

	/**
	 * Sets the state of the subgraph to the given value.
	 * @param value true or false.
	 */
	@Override
	public void setModified( boolean value ) {
		getRoot().setModified(value);
	}

	@Override
	public void add( Component slot, Object alignment, int index ) {
		XYConstraint constraint = new XYConstraint((XYLayout.Direction)alignment, index);
		add(slot, constraint);
	}

	/**
	 * Displays the related graph editor.
	 */
	public void showContent( ) {
		Ariane.getWorkspace().addGraph(this);
	}
}