package fr.greyc.pantheon.ariane.graph;

import javax.swing.undo.CompoundEdit;
import javax.swing.undo.UndoManager;
import javax.swing.undo.UndoableEdit;

import fr.greyc.pantheon.ariane.Ariane;
import fr.greyc.pantheon.ariane.ui.PreferenceChooser;

/**
 * UndoRedoManager manages a list of UndoableEdits, providing a way
 * to undo or redo the appropriate actions.
 */
public final class UndoRedoManager extends UndoManager {
	/** The current compound edit to store several edit in a same one. */
	private CompoundEdit _edit;

	/**
	 * Creates a new UndoRedoManager.
	 */
	public UndoRedoManager() {
		setLimit(PreferenceChooser.getInt(PreferenceChooser.UNDO_HISTORY_SIZE));
	}

	/**
	 * Validates the given modification, it means add
	 * the given modification in the current undo/redo history.
	 * 
	 * @param action the change to be added.
	 */
	public void validateAction( UndoableEdit action ) {
		if (_edit == null) {
			_edit = new CompoundEdit();
		}
		_edit.addEdit(action);
	}

	/**
	 * Starts a new list of actions.
	 * An action is composed of several modifications.
	 */
	public void startActionEdition( ) {
		_edit = new CompoundEdit();
	}

	/**
	 * Ends the current list of actions.
	 */
	public void endActionEdition( Graph graph ) {
		if (_edit != null && _edit.isSignificant()) {
			_edit.end();
			addEdit(_edit);
		}
		graph.setModified(true);
		Ariane.getWorkspace().update();
	}

	/**
	 * Cancels the current edited action.
	 */
	public void cancelCurrentActionEdition() {
		_edit = null;
	}

	/**
	 * Undoes the last do action.
	 * Removes the last undo item from the undo list
	 * and adds it in the redo list. 
	 */
	@Override
	public void undo( ) {
		try {
			super.undo();
			Ariane.getWorkspace().update();
		} catch (javax.swing.undo.CannotUndoException e) {
			Ariane.getWorkspace().printErrorMessage(Ariane.getString("Graph.history.error1"));
		}
	}

	/**
	 * Redoes the last undo action.
	 * Removes the last redo item from the redo list
	 * and adds it in the undo list.
	 */
	@Override
	public void redo( ) {
		try {
			super.redo();
			Ariane.getWorkspace().update();
		} catch (javax.swing.undo.CannotRedoException e) {
			Ariane.getWorkspace().printErrorMessage(Ariane.getString("Graph.history.error1"));
		}
	}
}
