/* -*- c-basic-offset: 3 -*-
 *
 * Copyright (c) 2013, GREYC.
 * All rights reserved
 *
 * You may use this file under the terms of the BSD license as follows:
 *
 * "Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of the GREYC, nor the name of its
 *     contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
 * 
 *   
 * For more information, refer to:
 * https://clouard.users.greyc.fr/Pandore
 */

package fr.greyc.pantheon.ariane.graph;

import java.awt.Point;

/**
 * A <code>XYConstraint</code> defines a constraint that can be
 * specified for locating a component in a XYLayout.
 * A constraints can be either:
 * <ul>
 * <li>an absolute location: x,y
 * <li>an alignment: GridBagConstraints.EAST WEST, NORHT SOUTH.
 * </ul>
 *
 * @author Régis CLOUARD - Oct 01, 1998
 * @version Sébastien LEFÈVRE, Richard LOBRÉAU - Feb 21, 1999
 * @version Régis CLOUARD - Oct 16, 2001
 * @version Régis CLOUARD - Nov 24, 2004
 */
public final class XYConstraint implements Cloneable {
	/** The alignment constraints in GridBagConstraints.{EAST, WEST, NORTH, SOUTH}. */
	private XYLayout.Direction _alignement;
	private final int _rank;
	private Point _location;

	/**
	 * Creates an empty constraint.
	 */
	public XYConstraint( ) {
		_rank = -1;
		_location = new Point(0, 0);
	}

	/**
	 * Creates a bounds constraint with the
	 * specified values.
	 */
	public XYConstraint( Point coordinates ) {
		_rank = -1;
		_location = new Point(coordinates.x, coordinates.y);
	}

	/**
	 * Creates an alignment region constraint with the
	 * specified values.
	 * 
	 * @param alignment type of alignment between
	 *        GridBagConstraints.{EAST WEST, NORHT SOUTH}.
	 * @param rank the rank in the same alignment (>=0).
	 */
	public XYConstraint( XYLayout.Direction alignment, int rank ) {
		_alignement = alignment;
		_rank = rank;
	}

	/**
	 * Returns the rank part of the constraint.
	 * @return the rank of the slot.
	 */
	public int getRank( ) {
		return _rank; 
	}

	/**
	 * Returns the alignment part of the constraint.
	 * @return the alignment.
	 */
	public XYLayout.Direction getAlignement( ) {
		return _alignement;
	}

	/**
	 * Returns the x part of the constraint.
	 * @return the x coordinate.
	 */
	public Point getLocation( ) {
		return _location;
	}

	/**
	 * Defines equality between two constraints.
	 * @param that a constraints.
	 */
	@Override
	public boolean equals( Object that ) {
		if (that instanceof XYConstraint) {
			XYConstraint other = (XYConstraint)that;
			return other._location.equals(_location) && other._rank == _rank && other._alignement == _alignement;
		}
		return false;
	}
	
	@Override
	public int hashCode() {
		int code = 17;
		code = code * 37 + _alignement.ordinal();
		code = code * 37 + _rank;
		code = code * 37 + _location.x;
		code = code * 37 + _location.y;
		return code;
	}
}