/* -*- c-basic-offset: 3 -*-
 *
 * Copyright (c) 2013, GREYC.
 * All rights reserved
 *
 * You may use this file under the terms of the BSD license as follows:
 *
 * "Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of the GREYC, nor the name of its
 *     contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
 * 
 *   
 * For more information, refer to:
 * https://clouard.users.greyc.fr/Pandore
 */

package fr.greyc.pantheon.ariane.graph;

import java.awt.Component;
import java.awt.Container;
import java.awt.Dimension;
import java.awt.LayoutManager2;
import java.awt.Rectangle;
import java.util.Hashtable;
import java.util.Map;

/**
 * Defines a class that knows how to lay out a graph.
 * A component is added with a location constraints.
 *
 * @author Régis CLOUARD - Oct 01, 1998
 * @version Sébastien LEFÈVRE, Richard LOBRÉAU - Feb 21, 1999
 * @version Régis CLOUARD - Oct 16, 2001
 * @version Régis CLOUARD - Jan 07, 2007
 * @version Régis CLOUARD - Jul 20, 2007 (fix bug on slots position)
 */
public final class XYLayout implements LayoutManager2 {
	/** A Hash table for indexing components. */
	private final Map<Component, XYConstraint> _info = new Hashtable<Component, XYConstraint>(); // NOPMD by regis on 7/23/14 7:23 PM
	private static final int ICON_WIDTH = 12;
	private static final int ICON_HEIGHT = 12;
	private static final int OFFSET = 60;

	/**
	 * The various status of a graph item
	 */
	public enum Direction {
		NORTHWEST,
		NORTHEAST,
		SOUTHWEST,
		SOUTHEAST,
		EASTSOUTH,
		WESTNORTH
	}

	@Override
	public void addLayoutComponent( String name, Component component ) {
		// Do nothing
	}

	/**
	 * Adds the specified component to the layout using the specified
	 * constraints.
	 * If the component was previously included, it is changed from the
	 * specified constraints.
	 * 
	 * @param component the component to be added.
	 * @param constraints where/how the component is added to the layout.
	 */
	@Override
	public void addLayoutComponent( Component component, Object constraints ) {
		Object innerConstraints;
		if (constraints == null) {
			innerConstraints = new XYConstraint();
		} else {
			innerConstraints = constraints;
		}
		synchronized (component.getTreeLock()) {
			if (innerConstraints instanceof XYConstraint) {
				XYConstraint constraint = (XYConstraint)innerConstraints;
				_info.put(component, (XYConstraint)innerConstraints);
				if (constraint.getRank() < 0) {
					component.setBounds(constraint.getLocation().x, constraint.getLocation().y, 0, 0);
				}
			}
		}
	}

	/**
	 * Removes the specified component from the layout.
	 * 
	 * @param component the component to be removed.
	 */
	@Override
	public void removeLayoutComponent( Component component ) {
		synchronized (component.getTreeLock()) {
			_info.remove(component);
		}
	}

	/**
	 * Determines the preferred size dimensions for the
	 * specified container, given the components it contains.
	 * Does not take care of insets of the container:
	 * the layout begins at coordinate 0,0.
	 * 
	 * @param container the container in which to do the layout.
	 * @return the preferred dimension.
	 */
	@Override
	public Dimension preferredLayoutSize( Container container ) {
		return minimumLayoutSize(container);
	}

	/**
	 * Calculates the minimum size dimensions for the specified
	 * container, given the components it contains.
	 * 
	 * @param container the component to be laid out.
	* @return the minimum layout size.
	 */
	@Override
	public Dimension minimumLayoutSize( Container container ) {
		synchronized (container.getTreeLock()) {
			Dimension dim = new Dimension(0, 0);
			final Dimension containerDim = container.getSize();
			for (Component component : container.getComponents()) {
				if (component.isVisible() && _info.get(component).getRank() < 0) {
					Rectangle rectangle = getComponentBounds(component, containerDim);
					dim.width  = Math.max(dim.width , rectangle.x + rectangle.width + 50);
					dim.height = Math.max(dim.height, rectangle.y + rectangle.height + 50);
				}
			}
			return dim;
		}
	}

	/**
	 * Returns the maximum dimensions for this layout
	 * given the components in the specified container.
	 * 
	 * @param container the component which needs to be laid out.
	 * @return the maximum layout size.
	 */
	@Override
	public Dimension maximumLayoutSize( Container container ) {
		return new Dimension(Integer.MAX_VALUE, Integer.MAX_VALUE);
	}

	/**
	 * Organizes the components in the container.
	 * Sets the new size of the XYLayout if needed.
	 * Does not take care of insets of the container:
	 * the layout begins at coordinate 0,0.
	 * 
	 * @param container the container to be laid out.
	 */
	@Override
	public void layoutContainer( Container container ) {
		synchronized (container.getTreeLock()) {
			final Dimension containerDim = container.getSize();
			for (Component component : container.getComponents()) {
				if (component.isVisible()) {
					component.setBounds(getComponentBounds(component, containerDim));
				}
			}
		}
	}

	/**
	 * Returns the alignment along the x axis.
	 * This specifies how the component would like to be aligned
	 * relative to other components.
	 * The value should be a number between 0 and 1 where 0 represents
	 * alignment along the origin, 1 is aligned the farest
	 * away from the origin, 0.5 is centered, etc.
	 * 
	 * @param container the container.
	 * @return the X layout alignment.
	 */
	@Override
	public float getLayoutAlignmentX( Container container ) {
		return 0.0f; // Prefer centered alignment.
	}

	/**
	 * Returns the alignment along the y axis.
	 * This specifies how the component would like to be aligned
	 * relative to other components.
	 * The value should be a number between 0 and 1 where 0 represents
	 * alignment along the origin, 1 is aligned the farest
	 * away from the origin, 0.5 is centered, etc.
	 * 
	 * @param container the container.
	 * @return the Y layout alignment.
	 */
	@Override
	public float getLayoutAlignmentY( Container container) {
		return 0.0f; // Prefer centered alignment.
	}

	/**
	 * Invalidates the layout, indicating that if the layout manager 
	 * has cached information it should be discarded.
	 * 
	 * @param container the container.
	 */
	@Override
	public void invalidateLayout( Container container ) {
		// Do nothing
	}

	/**
	 * Calculates the bounds of the component given its constraints
	 * the current size of the container that holds the component.
	 * 
	 * @param component the component to be laid out.
	 * @param size the current size of the container.
	 * @return the bounds rectangle.
	 */
	private Rectangle getComponentBounds( Component component, Dimension size ) {
		XYConstraint constraints = _info.get(component);

		// Constraints by border location.
		if (constraints.getRank() != -1) {
			int rank = constraints.getRank() + 1;
			switch(constraints.getAlignement()) {
			case WESTNORTH:
				return new Rectangle(0, OFFSET + rank * 20, ICON_WIDTH, ICON_HEIGHT);
			case NORTHWEST:
				return new Rectangle(OFFSET + rank * 20, 0, ICON_WIDTH, ICON_HEIGHT);
			case SOUTHEAST:
				return new Rectangle(200 + OFFSET + rank * 20, size.height - ICON_HEIGHT, ICON_WIDTH, ICON_HEIGHT);
			case SOUTHWEST:
				return new Rectangle(OFFSET + rank * 20, size.height - ICON_HEIGHT, ICON_WIDTH, ICON_HEIGHT);
			case NORTHEAST:
				return new Rectangle(size.width - ICON_WIDTH, OFFSET + rank * 20, ICON_WIDTH, ICON_HEIGHT);
			case EASTSOUTH:
				return new Rectangle(size.width - ICON_WIDTH, 200 + OFFSET + rank * 20, ICON_WIDTH, ICON_HEIGHT);
			default:
				break;
			}
		}
		// Constraints by coordinates.
		Rectangle rectangle = component.getBounds();
		Dimension dimension = component.getPreferredSize();
		rectangle.width = dimension.width;
		rectangle.height = dimension.height;
		return rectangle;
	}
}