/* -*- c-basic-offset: 3 -*-
 *
 * Copyright (c) 2013, GREYC.
 * All rights reserved
 *
 * You may use this file under the terms of the BSD license as follows:
 *
 * "Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of the GREYC, nor the name of its
 *     contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
 * 
 *   
 * For more information, refer to:
 * https://clouard.users.greyc.fr/Pandore
 */

package fr.greyc.pantheon.ariane.graph.item;

import fr.greyc.pantheon.ariane.Ariane;
import fr.greyc.pantheon.ariane.graph.Graph;
import fr.greyc.pantheon.ariane.graph.SubGraph;
import fr.greyc.pantheon.ariane.graph.command.ChangeCompositeStatusEdit;
import fr.greyc.pantheon.ariane.graph.slot.InputDataSlot;
import fr.greyc.pantheon.ariane.graph.slot.InputParameterSlot;
import fr.greyc.pantheon.ariane.graph.slot.OutputDataSlot;
import fr.greyc.pantheon.ariane.ui.SubroutineEditor;
import fr.greyc.pantheon.ariane.utils.FadingEffects;
import fr.greyc.pantheon.ariane.utils.SubroutineTreeItem;
import fr.greyc.pantheon.ariane.utils.TreeItem;
import fr.greyc.pantheon.ariane.xml.XMLElement;

import java.awt.Color;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyEvent;
import java.awt.event.MouseEvent;
import java.io.File;
import java.io.IOException;
import java.util.List;

import javax.swing.JMenuItem;
import javax.swing.JPopupMenu;
import javax.swing.KeyStroke;

/**
 * An <code>CompositeItem</code> represents an item with
 * an inner subgraph.
 * 
 * @author Régis CLOUARD - Oct 01, 1998
 * @version Sébastien LEFÈVRE, Richard LOBRÉAU - Feb 21, 1999
 * @version Adrien SCHRÈQUE, Hicham AMADI - Nov 05, 2000
 * @version Mathieu JOETS - Jul 09, 2001
 * @version Régis CLOUARD - Oct 16, 2001
 * @version Régis CLOUARD - July 17, 2010 (storable in archive)
 * @version Régis CLOUARD - Jan 04, 2012 (add editable/uneditable routines)
 */
public abstract class CompositeItem extends ControlStructureItem {
	/** The related subgraph. */
	private SubGraph _subGraph;
	/** The methods that depends on the current state. */
	private State _state;

	/**
	 * Creates a new composite composite item with the given
	 * attributes values.
	 * @param treeItem the reference tree item.
	 * @param graph the parent graph.
	 * @param name the name of the item.
	 */
	public CompositeItem( TreeItem treeItem, Graph graph, String name ) {
		super(treeItem, graph, name);
		_state = new Editable();
		createSlots(0, 0, 1, 0, 0);
	}

	/**
	 * Creates a new composite item with the given
	 * attributes values.
	 * @param graph the parent graph.
	 */
	public CompositeItem( Graph graph ) {
		super(graph);
		_state = new Editable();
		createSlots(0, 0, 1, 0, 0);
	}

	/**
	 * Creates a new subroutine item (uneditable item)
	 * with the given attributes values.
	 * @param graph the parent graph.
	 * @param treeItem the related tree item.
	 */
	public CompositeItem( Graph graph, TreeItem treeItem ) {
		super(graph);
		_treeItem = treeItem;
		createSlots(0, 0, 1, 0, 0);
		_state = new Uneditable();
	}

	/**
	 * Builds or changes the prototype of the macro-operator
	 * from the given values.
	 * Creates and changes also the related subgraph.
	 */
	protected final void setPrototype( String name, int inputImageCount, int imageParameterCount, int outputImageCount, int interImageCount, int interParameterCount ) {
		adjustSlotsNumber(name, inputImageCount, imageParameterCount, outputImageCount);
		createSubGraph(name, inputImageCount, imageParameterCount, outputImageCount);
		draw();
	}

	/**
	 * Creates or removes input and output slots of the macro-operator
	 * from the given number of slots and from the existing number of slots.
	 */
	protected final void adjustSlotsNumber( String name, int inputImageCount, int inputParameterCount, int outputImageCount ) {
		// The name.
		if (!name.equals(getName())) {
			setNickName(name);
		}

		// Input image slots.
		List<InputDataSlot> inputDataSlots = getInputDataSlots();
		if (inputImageCount > inputDataSlots.size()) { // adds slots
			for (int i = inputDataSlots.size(); i < inputImageCount; i++) {
				inputDataSlots.add(new InputDataSlot(this, i));
			}
		} else if (inputImageCount < inputDataSlots.size()) { // removes slots
			for (int i = inputDataSlots.size() - 1; i >= inputImageCount; i--) {
				InputDataSlot slot = inputDataSlots.remove(i);
				slot.disconnect();
				slot.dispose();
			}
		}

		// Input parameters slots.
		if (inputParameterCount > getInputParameterSlots().size()) { // adds slots
			for (int i = getInputParameterSlots().size(); i < inputParameterCount; i++) {
				getInputParameterSlots().add(new InputParameterSlot(this, i));
			}
		} else if (inputParameterCount < getInputParameterSlots().size()) { // removes slots
			for (int i = getInputParameterSlots().size() - 1; i >= inputParameterCount; i--) {
				InputParameterSlot slot = getInputParameterSlots().remove(i);
				slot.disconnect();
				slot.dispose();
			}
		}

		// Output image slots.
		if (outputImageCount > getOutputDataSlots().size()) { // adds slots
			for (int i = getOutputDataSlots().size(); i < outputImageCount; i++) {
				getOutputDataSlots().add(new OutputDataSlot(this, i));
			}
		} else if (outputImageCount < getOutputDataSlots().size()) { // removes slots
			for (int i = getOutputDataSlots().size() - 1; i >= outputImageCount; i--) {
				OutputDataSlot slot = getOutputDataSlots().remove(i);
				slot.disconnectAll();
				slot.dispose();
			}
		}
	}

	/**
	 * Creates the subgraph with the convenient prototype.
	 * @return the subgraph.
	 */
	protected final SubGraph createSubGraph( String name, int inputImageCount, int inputParameterCount, int outputImageCount, int interImageCount, int interParameterCount) {
		_subGraph = new SubGraph(this, name, inputImageCount, inputParameterCount, outputImageCount, interImageCount, interParameterCount);
		return _subGraph;
	}

	/**
	 * Creates the subgraph with the convenient prototype.
	 * @return the subgraph.
	 */
	protected abstract SubGraph createSubGraph( String name, int inputImageCount, int inputParameterCount, int outputImageCount );
	@Override
	public final Color getBackgroundColor( ) {
		if (_state == null) {
			return super.getColor();
		} else {
			return _state.getBackgroundColor();
		}
	}
	
	/**
	 * Sets the item nickname.
	 * 
	 * @param name the new nickname.
	 */
	@Override
	public final void setNickName( String name ) {
		super.setNickName(name);
		Ariane.getWorkspace().renameGraph(_subGraph, name);
	}

	/**
	 * Removes temporary the item from the graph.
	 * Disconnects it from all other item.
	 */
	@Override
	public final void removeHook() {
		_subGraph.closeFrame();
	}

	/** 
	 * Disposes definitively the item.
	 * Tries to avoid memory leak.
	 */
	@Override
	protected final void disposeHook() {
		_subGraph.dispose();
		_subGraph = null;

	}

	protected abstract void linkItemWithVirtualItem( VirtualItem virtualOperator, boolean isFailureFatalError );

	///// 
	///// A C C E S S O R S
	/////

	/**
	 * Returns the related subgraph.
	 * @return the subgraph.
	 */
	@Override
	public final SubGraph getSubGraph() {
		return _subGraph;
	}

	/**
	 * Opens the related subgraph and shows its content.
	 */
	protected final void viewSubgraph() {
		_subGraph.showContent();
	}

	///// 
	///// M E N U
	/////

	/**
	 * Builds the contextual menu as a Popup Menu.
	 */
	@Override
	protected final JPopupMenu initPopup( ) {
		return _state.initPopup();
	}

	///// 
	///// E X E C U T I O N
	/////

	/**
	 * Resets the composite item and its related sub-graph.
	 */
	@Override
	public final void resetAll( ) {
		super.resetAll();
		if (_subGraph != null) {
			_subGraph.resetAll();
		}
	}

	/**
	 * Resets the items.
	 */
	@Override
	protected void reset() {
		super.reset();
		for (GraphItem item: _subGraph.getGraphItems()) {
			item.reset();
		}
	}

	/**
	 * @return the virtual operator of the subgraph.
	 */
	protected final VirtualItem getVirtualOperator() {
		return _subGraph.getVirtualOperator();
	}

	///// 
	///// I N P U T   /   O U T P U T
	/////

	/**
	 * Saves the composite item as an XML tree.
	 * Saves also the subgraph.
	 * 
	 * @param file the current filename of the project.
	 * @return the XML tree containing the subgraph.
	 * @throws IOException any XML errors.
	 */
	@Override
	public final XMLElement save( File file ) throws IOException {
		XMLElement element = new XMLElement(getTagName());
		List<XMLElement> xmlTree = saveProperties(file);
		if (_treeItem instanceof SubroutineTreeItem && _state instanceof Uneditable) {
			xmlTree.add(new XMLElement("SubroutineReference", _treeItem.getName()));
		}
		element.addChildren(xmlTree);
		element.addChildren(_subGraph.getXML(file));
		return element;
	}

	/**
	 * Loads a composite structure from an XML tree.
	 * 
	 * @param xmlTree the XML tree of data.
	 * @throws IOException any XML errors.
	 */
	@Override
	protected final void loadAdditionalProperties( List<XMLElement> xmlTree ) throws IOException {
		boolean subroutine = false;
		for (XMLElement element: xmlTree) {
			if (element.getName().equalsIgnoreCase("SubroutineReference")) {
				subroutine = _treeItem instanceof SubroutineTreeItem;
			}
		}
		SubGraph subGraph = createSubGraph(getName(), getInputDataSlots().size(), getInputParameterSlots().size(), getOutputDataSlots().size());
		if (subroutine) {
			_state = new Uneditable();
			subGraph.setXML(((SubroutineTreeItem)_treeItem).getBody());
		} else {
			subGraph.setXML(xmlTree);
		}
	}

	///// 
	///// P R O P E R T I E S
	/////

	/**
	 * Invoked when the mouse button has been
	 * double clicked (pressed and released) on a graph item.
	 * Shows the dialog box for setting the item's properties.
	 * 
	 * @param e the event attributes.
	 */
	@Override
	protected final void onMouseDoubleClicked( MouseEvent event ) {
		_state.onMouseDoubleClicked(event);
	}

	/**
	 * Asks interactively the prototype of the Macro.
	 */
	protected abstract void askProperties() throws InterruptedException;

	protected abstract TreeItem restoreDefaultTreeItem();

	public final void changeEditableStatus( boolean editable ) {
		if (editable) {
			_state = new Editable();
			_treeItem = restoreDefaultTreeItem();
		} else {
			_state = new Uneditable();
		}
		draw();
	}

	/**
	 * The state design pattern to manage the composite as
	 * an uneditable subroutine or an editable macro.
	 */
	abstract class State {
		public abstract JPopupMenu initPopup( );
		public abstract void onMouseDoubleClicked( MouseEvent event );
		public abstract Color getBackgroundColor();
	}

	class Uneditable extends State {
		@Override
		public final JPopupMenu initPopup( ) {
			JPopupMenu popupMenu = new JPopupMenu();
			JMenuItem item;
			item = new JMenuItem(Ariane.getString("Graph.Menu.copy"));
			item.addActionListener(new ActionListener() {
				@Override
				public void actionPerformed( ActionEvent event ) {
					Ariane.getWorkspace().copy();
				}
			});
			item.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_C, KeyEvent.CTRL_MASK));
			popupMenu.add(item);

			item = new JMenuItem(Ariane.getString("Graph.Menu.cut"));
			item.addActionListener(new ActionListener() {
				@Override
				public void actionPerformed( ActionEvent event ) {
					Ariane.getWorkspace().cut();
				}
			});
			item.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_X, KeyEvent.CTRL_MASK));
			popupMenu.add(item);

			popupMenu.addSeparator();

			item = new JMenuItem(Ariane.getString("PopupMenu.execute"));
			item.addActionListener(new ActionListener() {
				@Override
				public final void actionPerformed( ActionEvent event ) { getGraph().execute(); }
			});
			item.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_R, KeyEvent.CTRL_MASK));
			popupMenu.add(item);

			item = new JMenuItem(Ariane.getString("PopupMenu.reset"));
			item.addActionListener(new ActionListener() {
				@Override
				public final void actionPerformed( ActionEvent event ) { resetAll(); }
			});
			item.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_U, KeyEvent.CTRL_MASK));
			popupMenu.add(item);

			item = new JMenuItem(Ariane.getString("PopupMenu.remove"));
			item.addActionListener(new ActionListener() {
				@Override
				public final void actionPerformed( ActionEvent event ) { delete(); }
			});
			item.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_DELETE, 0));
			popupMenu.add(item);

			popupMenu.addSeparator();
			item = new JMenuItem(Ariane.getString("PopupMenu.convertAsItem"));
			item.addActionListener(new ActionListener() {
				@Override
				public final void actionPerformed( ActionEvent event ) {
					removePropertyViewer();
					changeEditableStatus(true);
					getGraph().setModified(true);
					getGraph().startActionEdition();
					getGraph().validateAction(new ChangeCompositeStatusEdit(CompositeItem.this, true));
					getGraph().endActionEdition();
				}
			});
			popupMenu.add(item);
			return popupMenu;
		}
		@Override
		public void onMouseDoubleClicked( MouseEvent event ) {
			Ariane.getWorkspace().printWarning(getNickName() + ": " + Ariane.getString("CompositeItem.message1"));
		}
		
		@Override
		public Color getBackgroundColor() { 
			return ControlStructureItem.getColor().darker();
		}
	}

	class Editable extends State {
		@Override
		public void onMouseDoubleClicked( MouseEvent event ) {
			viewSubgraph();
		}

		@Override
		public final JPopupMenu initPopup( ) {
			JPopupMenu popupMenu = new JPopupMenu();
			JMenuItem item;
			item = new JMenuItem(Ariane.getString("Graph.Menu.copy"));
			item.addActionListener(new ActionListener() {
				@Override
				public void actionPerformed( ActionEvent event ) {
					Ariane.getWorkspace().copy();
				}
			});
			item.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_C, KeyEvent.CTRL_MASK));
			popupMenu.add(item);

			item = new JMenuItem(Ariane.getString("Graph.Menu.cut"));
			item.addActionListener(new ActionListener() {
				@Override
				public void actionPerformed( ActionEvent event ) {
					Ariane.getWorkspace().cut();
				}
			});
			item.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_X, KeyEvent.CTRL_MASK));
			popupMenu.add(item);

			popupMenu.addSeparator();

			item = new JMenuItem(Ariane.getString("PopupMenu.execute"));
			item.addActionListener(new ActionListener() {
				@Override
				public final void actionPerformed( ActionEvent event ) { getGraph().execute(); }
			});
			item.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_R, KeyEvent.CTRL_MASK));
			popupMenu.add(item);

			item = new JMenuItem(Ariane.getString("PopupMenu.reset"));
			item.addActionListener(new ActionListener() {
				@Override
				public final void actionPerformed( ActionEvent event ) { resetAll(); }
			});
			item.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_U, KeyEvent.CTRL_MASK));
			popupMenu.add(item);

			item = new JMenuItem(Ariane.getString("PopupMenu.remove"));
			item.addActionListener(new ActionListener() {
				@Override
				public final void actionPerformed( ActionEvent event ) { delete(); }
			});
			item.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_DELETE, 0));
			popupMenu.add(item);

			popupMenu.addSeparator();
			item = new JMenuItem(Ariane.getString("PopupMenu.store"));
			item.addActionListener(new ActionListener() {
				@Override
				public final void actionPerformed( ActionEvent event ) {
					storeAsRoutine();
				}
			});
			popupMenu.add(item);

			popupMenu.addSeparator();
			item = new JMenuItem(Ariane.getString("PopupMenu.openSubgraph"));
			item.addActionListener(new ActionListener() {
				@Override
				public final void actionPerformed( ActionEvent event ) { viewSubgraph(); }
			});
			popupMenu.add(item);
			return popupMenu;
		}

		public final void storeAsRoutine() {
			SubroutineEditor subroutineEditor = new SubroutineEditor(CompositeItem.this);
			FadingEffects.fadeIn(Ariane.getFrame());
			subroutineEditor.setVisible(true);
			FadingEffects.fadeOut(Ariane.getFrame());
			if (subroutineEditor.isValidated()) {
				_treeItem = subroutineEditor.getTreeItem();
				setSignature(_treeItem.getSignature());
				setNickName(_treeItem.getName());
				setAbstractName(_treeItem.getName());
				removePropertyViewer();
				changeEditableStatus(false);
				getGraph().setModified(true);
				getGraph().startActionEdition();
				getGraph().validateAction(new ChangeCompositeStatusEdit(CompositeItem.this, false));
				getGraph().endActionEdition();
			}
		}

		@Override
		public Color getBackgroundColor() {
			return ControlStructureItem.getColor();
		}
	}
}