/* -*- c-basic-offset: 3 -*-
 *
 * Copyright (c) 2013, GREYC.
 * All rights reserved
 *
 * You may use this file under the terms of the BSD license as follows:
 *
 * "Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of the GREYC, nor the name of its
 *     contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
 * 
 *   
 * For more information, refer to:
 * https://clouard.users.greyc.fr/Pandore
 */

package fr.greyc.pantheon.ariane.graph.item;

import fr.greyc.pantheon.ariane.Ariane;
import fr.greyc.pantheon.ariane.graph.Graph;
import fr.greyc.pantheon.ariane.graph.command.ChangeArgumentValueEdit;
import fr.greyc.pantheon.ariane.graph.exception.ArianeArithmeticException;
import fr.greyc.pantheon.ariane.graph.item.etc.PropertyViewer;
import fr.greyc.pantheon.ariane.utils.TreeItem;
import fr.greyc.pantheon.ariane.xml.XMLElement;

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyEvent;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

import javax.swing.JCheckBox;
import javax.swing.JMenuItem;
import javax.swing.JPopupMenu;
import javax.swing.KeyStroke;

/**
 * A <code>ConstantItem</code> implements a text value.
 * 
 * @author Régis CLOUARD - Jun 10, 2010
 * @author Régis CLOUARD - Apr 17, 2011 (add rename facility) 
 * @author Régis CLOUARD - Apr 23, 2011
 */
public final class ConstantItem extends ControlItem implements IScriptArgument {
	/** The tag name of the item. */
	public static final String TAGNAME = "Constant";
	/** Is this item a script argument? */
	private boolean _isArgument;
	/** The check box used in the property window. */
	private JCheckBox _checkBox;

	/**
	 * Creates a <code>ConstantItem</code> with the given
	 * attributes.
	 * Initializes the mouse listeners, the popup menu,
	 * and displays it in the graph.
	 * 
	 * @param treeItem the reference tree item.
	 * @param graph the parent graph.
	 * @param name the name of the item.
	 */
	public ConstantItem( TreeItem treeItem, Graph graph, String name ) {
		super(treeItem, graph, name);
		setFormula(""); // IMPERATIVE to signify that it uses formula.
		setComment(getFormula());
		draw();
	}

	/**
	 * Creates a <code>ConstantItem</code> on the given
	 * graph with the given XML data.
	 * 
	 * @param graph the parent graph.
	 * @param elements the XML tree of data.
	 * @throws IOException XML errors.
	 */
	public ConstantItem( Graph graph, List<XMLElement> elements ) throws IOException {
		super(graph);
		setFormula(""); // IMPERATIVE to signify that it uses formula.
		load(elements);
		setComment(getFormula());
		draw();
	}

	/**
	 * @return true if the item is an argument of the
	 * script that can be generated from the graph.
	 */
	@Override
	public boolean isArgument() {
		return _isArgument;
	}

	/**
	 * Sets whether the item is a script argument or not.
	 * 
	 * @param value the new status value.
	 */
	@Override
	public void setIsArgument( boolean value ) {
		_isArgument = value;
	}

	/**
	 * @return the list of specific properties (other than common properties).
	 */
	@Override
	public ArrayList<PropertyViewer.PropertyCouple> getSpecificProperties( ) {
		ArrayList<PropertyViewer.PropertyCouple> properties = new ArrayList<PropertyViewer.PropertyCouple>();

		_checkBox = new JCheckBox();
		_checkBox.setSelected(_isArgument);
		_checkBox.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed( ActionEvent event ) {
				_isArgument = ((JCheckBox)event.getSource()).isSelected();
				// Add to the history
				getGraph().setModified(true);
				getGraph().startActionEdition();
				getGraph().validateAction(new ChangeArgumentValueEdit(ConstantItem.this, _checkBox, _isArgument));
				getGraph().endActionEdition();
			}
		});
		properties.add(new PropertyViewer.PropertyCouple(Ariane.getString("PropertyViewer.argument"), _checkBox));
		return properties;
	}

	/**
	 * Builds or changes the prototype of the control structure
	 * from the given values.
	 * Note that the number of inputs, outputs, and parameters are always 0.
	 */
	@Override
	protected void setPrototype( String name, int inputImageCount, int inputParameterCount, int outputImageCount ) {
		// Empty
	}

	/**
	 * Changes the current constant statement with
	 * the specified one.
	 * 
	 * @param statement a constant statement.
	 * @return true if the statement is correct, false otherwise.
	 */
	@Override
	public boolean changeFormula( String statement ) {
		if (getFormula() == null || !getFormula().equals(statement)) {
			setComment(statement);
			setFormula(statement);
		}
		return true;
	}

	///// 
	///// A C C E S S O R S
	/////

	/**
	 * @return the tag name.
	 */
	@Override
	public String getTagName( ) {
		return TAGNAME;
	}

	///// 
	///// E X E C U T I O N
	/////

	/**
	 * Executes the <code>Constant</code>, namely
	 * returns the expression.
	 * 
	 * @param isFailureFatalError is FAILURE a fatal error?
	 * @return true if the operator has been executed without errors.
	 */
	@Override
	protected boolean primitiveExecute( boolean isFailureFatalError ) {
		try {
			final String constantExpression = getFormula();
			if (constantExpression == null || constantExpression.isEmpty()) {
				throw new ArianeArithmeticException(Ariane.getString("arithmetic.error19"));
			}
			getOutputResultSlot().setValue(constantExpression);
			changeStatus(Status.EXECUTED);
			return true;
		} catch (ArianeArithmeticException e) {
			if (isFailureFatalError && !Graph.isExecutionStopped()) {
				Ariane.getWorkspace().printErrorMessage(e.getErrorCode(), getAbsoluteName(), getId(), e.getLocalizedMessage());
				getGraph().stopExecution();
				changeStatus(Status.ERROR);
			}
			return false;
		}
	}

	/**
	 * Builds the contextual menu as a Popup Menu.
	 * 
	 * @return a convenient popup menu.
	 */
	@Override
	protected JPopupMenu initPopup( ) {
		JPopupMenu popupMenu = new JPopupMenu();

		JMenuItem item;
		item = new JMenuItem(Ariane.getString("Graph.Menu.copy"));
		item.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed( ActionEvent event ) {
				Ariane.getWorkspace().copy();
			}
		});
		item.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_C, KeyEvent.CTRL_MASK));
		popupMenu.add(item);

		item = new JMenuItem(Ariane.getString("Graph.Menu.cut"));
		item.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed( ActionEvent event ) {
				Ariane.getWorkspace().cut();
			}
		});
		item.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_X, KeyEvent.CTRL_MASK));
		popupMenu.add(item);

		popupMenu.addSeparator();

		// Action: executer
		item = new JMenuItem(Ariane.getString("PopupMenu.execute"));
		item.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed( ActionEvent event ) {
				getGraph().execute();
			}
		});
		item.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_R, KeyEvent.CTRL_MASK));
		popupMenu.add(item);

		item = new JMenuItem(Ariane.getString("PopupMenu.reset"));
		item.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed( ActionEvent event ) {
				resetAll();
			}
		});
		item.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_U, KeyEvent.CTRL_MASK));
		popupMenu.add(item);

		// Edition: delete
		item = new JMenuItem(Ariane.getString("PopupMenu.remove"));
		item.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed( ActionEvent event ) {
				delete();
			}
		});
		item.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_DELETE, 0));
		popupMenu.add(item);

		return popupMenu;
	}
}