/* -*- c-basic-offset: 3 -*-
 *
 * Copyright (c) 2013, GREYC.
 * All rights reserved
 *
 * You may use this file under the terms of the BSD license as follows:
 *
 * "Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of the GREYC, nor the name of its
 *     contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
 * 
 *   
 * For more information, refer to:
 * https://clouard.users.greyc.fr/Pandore
 */

package fr.greyc.pantheon.ariane.graph.item;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.FontMetrics;
import java.awt.Graphics;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyEvent;
import java.io.IOException;

import javax.swing.AbstractAction;
import javax.swing.BorderFactory;
import javax.swing.JButton;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JMenuItem;
import javax.swing.JPanel;
import javax.swing.JPopupMenu;
import javax.swing.JSpinner;
import javax.swing.JTextField;
import javax.swing.KeyStroke;
import javax.swing.SpinnerModel;
import javax.swing.SpinnerNumberModel;

import fr.greyc.pantheon.ariane.Ariane;
import fr.greyc.pantheon.ariane.graph.Graph;
import fr.greyc.pantheon.ariane.graph.slot.InputParameterSlot;
import fr.greyc.pantheon.ariane.ui.PreferenceChooser;
import fr.greyc.pantheon.ariane.utils.FadingEffects;
import fr.greyc.pantheon.ariane.utils.GDialog;
import fr.greyc.pantheon.ariane.utils.TreeItem;

/**
 * This is the base class of all control items.
 * It provides facilities to set the number of parameters, inputs and outputs.
 *
 * @author Sebastien LEFÈVRE, Richard LOBRÉAU - Feb 21, 1999
 * @version Adrien SCHRÈQUE, Hicham AMADI - Nov 05, 2000
 * @version Mathieu JOETS - Jul 09, 2001
 * @version Alban DOUBLET, Guillaume CHATELET - Jun 10, 2002
 * @version Régis CLOUARD - Oct 16, 2001
 * @version Régis CLOUARD - Jan 07, 2006
 * @version Régis CLOUARD - Jun 11, 2010 (add parameters to for and while)
 */
public abstract class ControlItem extends GraphItem {
	/** The color of the item. */
	private static final Color DEFAULT_COLOR = new Color(255, 153, 51);

	/** The color of the item. */
	private static Color _color = DEFAULT_COLOR;

	private boolean _validatedPrototype;

	/**
	 * Creates a new control structure on the given graph
	 * with the given identifier number.
	 * If the identifier =-1 then creates a unique identifier number.
	 * 
	 * @param treeItem the reference tree item.
	 * @param graph the parent graph.
	 * @param name the name of the item.
	 */
	public ControlItem( TreeItem treeItem, Graph graph, String name ) {
		super(treeItem, graph, name, null);
		createSlots(0, 0, 0, 0, 1);
	}

	/**
	 * Creates a new control structure on the given graph
	 * with the given identifier number.
	 * 
	 * @param graph the parent graph.
	 */
	public ControlItem( Graph graph ) {
		super(graph);
		createSlots(0, 0, 0, 0, 1);
	}

	/**
	 * Builds or changes the prototype of the macro-operator
	 * from the given values.
	 */
	protected abstract void setPrototype( String name, int inputImageNumber, int inputParameterNumber, int outputImageNumber );

	///// 
	///// P R E F E R E N C E S
	/////

	// Notice: theses methods cannot be generalized in the GraphItem class
	// because they use static attributes.
	/**
	 * Changes the current color.
	 * 
	 * @param col the new color.
	 */
	public static void setColor( Color col ) {
		_color = col;
	}

	/**
	 * Returns the current color.
	 * 
	 * @return the color.
	 */
	public static Color getColor( ) {
		return _color;
	}

	/**
	 * Returns the default color.
	 * 
	 * @return the color.
	 */
	public static Color getDefaultColor( ) {
		return DEFAULT_COLOR;
	}

	/**
	 * Returns the background color (same as geColor()).
	 * 
	 * @return the color.
	 */
	@Override
	public Color getBackgroundColor( ) {
		return _color;
	}

	/**
	 * Returns the description of the control structure: name+parameters value.
	 * (Used among others by tooltips).
	 * 
	 * @return a string.
	 */
	@Override
	public final String toString( ) {
		String commandLine = "(" + getId() + "): " + getName();

		for (InputParameterSlot isp : getInputParameterSlots()) {
			if (isp.getValue() != null) {
				commandLine += " " + isp.getValue();
			}
		}
		return commandLine;
	}

	///// 
	///// M E N U
	/////

	/**
	 * Builds the contextual menu as a Popup Menu.
	 *
	 * @return a convenient popup menu.
	 */
	@Override
	protected JPopupMenu initPopup( ) {
		JPopupMenu popupMenu = new JPopupMenu();

		JMenuItem item;
		item = new JMenuItem(Ariane.getString("Graph.Menu.copy"));
		item.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed( ActionEvent event ) {
				Ariane.getWorkspace().copy();
			}
		});
		item.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_C, KeyEvent.CTRL_MASK));
		popupMenu.add(item);

		item = new JMenuItem(Ariane.getString("Graph.Menu.cut"));
		item.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed( ActionEvent event ) {
				Ariane.getWorkspace().cut();
			}
		});
		item.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_X, KeyEvent.CTRL_MASK));
		popupMenu.add(item);

		popupMenu.addSeparator();

		// Action: executer
		item = new JMenuItem(Ariane.getString("PopupMenu.execute"));
		item.addActionListener(new ActionListener() {
			@Override
			public final void actionPerformed( ActionEvent event ) {
				getGraph().execute();
			}
		});
		item.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_R, KeyEvent.CTRL_MASK));
		popupMenu.add(item);

		item = new JMenuItem(Ariane.getString("PopupMenu.reset"));
		item.addActionListener(new ActionListener() {
			@Override
			public final void actionPerformed( ActionEvent event ) {
				resetAll();
			}
		});
		item.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_U, KeyEvent.CTRL_MASK));
		popupMenu.add(item);

		// Edition: delete
		item = new JMenuItem(Ariane.getString("PopupMenu.remove"));
		item.addActionListener(new ActionListener() {
			@Override
			public final void actionPerformed( ActionEvent event ) {
				delete();
			}
		});
		item.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_DELETE, 0));
		popupMenu.add(item);

		return popupMenu;
	}

	///// 
	///// P R O P E R T I E S
	/////
	/**
	 * Asks interactively the prototype of the control structure.
	 * 
	 * @param name the default name.
	 * @param iIm the default number of inputs.
	 * @param iP the default number of outputs.
	 * @param oIm the default number of parameters.
	 */
	protected final void askProperties( String name, int defaultInputImageNumber, int defaultParameterNumber, int defaultOutputImageNumber ) throws InterruptedException {
 		final GDialog dialog = new GDialog((JFrame)null, getTagName(), true);
		dialog.setAlwaysOnTop(true);
		dialog.setResizable(false);
		_validatedPrototype = false;

		String nickName;
		nickName = (name == null) ? getName() : name;
		final JTextField textFieldname = new JTextField(nickName, 10);
		SpinnerModel spModel;
		spModel = new SpinnerNumberModel(defaultParameterNumber, 0, 10, 1);
		final JSpinner textFieldiP = new JSpinner(spModel);
		spModel = new SpinnerNumberModel(defaultInputImageNumber, 0, 10, 1);
		final JSpinner textFieldiIm = new JSpinner(spModel);
		spModel = new SpinnerNumberModel(defaultOutputImageNumber, 0, 10, 1);
		final JSpinner textFieldoIm = new JSpinner(spModel);

		JPanel jPanel = new JPanel(new GridBagLayout());
		jPanel.setBorder(BorderFactory.createTitledBorder(BorderFactory.createEtchedBorder(), Ariane.getString("MacroOperator.borderTitle")));

		GridBagConstraints constraint = new GridBagConstraints();
		constraint.fill = GridBagConstraints.NONE;
		constraint.anchor = GridBagConstraints.WEST;
		constraint.ipadx = 8; constraint.ipady = 8;
		constraint.gridy = 0;

		if (name != null) { //Macro-operator
			constraint.gridy++; constraint.gridx = 0;
			jPanel.add(new JLabel(Ariane.getString("MacroOperator.name")), constraint);
			constraint.gridx = 1;
			jPanel.add(textFieldname, constraint);
		}
		if (defaultParameterNumber > 0) {
			constraint.gridy++; constraint.gridx = 0;
			jPanel.add(new JLabel(Ariane.getString("MacroOperator.parameternumber")), constraint);
			constraint.gridx = 1;
			jPanel.add(textFieldiP, constraint);
		}
		if (defaultInputImageNumber > 0) {
			constraint.gridy++; constraint.gridx = 0;
			jPanel.add(new JLabel(Ariane.getString("MacroOperator.inputnumber")), constraint);
			constraint.gridx = 1;
			jPanel.add(textFieldiIm, constraint);
		}
		if (defaultOutputImageNumber > 0) {
			constraint.gridy++; constraint.gridx = 0;
			jPanel.add(new JLabel(Ariane.getString("MacroOperator.outputnumber")), constraint);
			constraint.gridx = 1;
			jPanel.add(textFieldoIm, constraint);
		}

		JButton applyButton = new JButton(Ariane.getString("PropertyViewer.ok"), Ariane.getIcon("gtk-yes.png"));

		applyButton.addActionListener(new ActionListener() {
			@Override
			public final void actionPerformed( ActionEvent event ) {
				try {
					int ii = ((Integer)textFieldiIm.getValue());
					int ip = ((Integer)textFieldiP.getValue());
					int io = ((Integer)textFieldoIm.getValue());

					if (ii < 0 || ip < 0 || io < 0 || ii > 10 || ip > 10 || io > 10) {
						throw new IOException(Ariane.getString("controlstructure.error1"));
					}
					String newName = textFieldname.getText();
					setPrototype(newName, ii, ip, io);
					_validatedPrototype = true;
					dialog.setVisible(false);
				} catch (IOException ex) {
					Ariane.getWorkspace().printErrorMessage(ex.getLocalizedMessage());
				}
			}
		});
		
		JButton cancelButton = new JButton();
		cancelButton.setAction(new AbstractAction(Ariane.getString("SubroutineEditor.cancel"),
				Ariane.getIcon("gtk-cancel.png")) {
			@Override
			public final void actionPerformed( ActionEvent event ) {
				_validatedPrototype = false;
				dialog.setVisible(false);
			}
		});

		dialog.getContentPane().setLayout(new BorderLayout(5, 5));
		dialog.getContentPane().add(jPanel, BorderLayout.CENTER);
		JPanel panel = new JPanel();
		panel.add(applyButton);
		panel.add(cancelButton);
		dialog.setDefaultButton(applyButton);
		dialog.setCancelButton(cancelButton);
		dialog.getContentPane().add(panel, BorderLayout.SOUTH);
		dialog.pack();
		dialog.setLocationRelativeTo(Ariane.getFrame());
		FadingEffects.fadeIn(Ariane.getFrame());
		dialog.setVisible(true);
		FadingEffects.fadeOut(Ariane.getFrame());

		if (!_validatedPrototype) {
			throw new InterruptedException();
		}
	}

	///// 
	///// P R I N T I N G
	/////

	/**
	 * Overrides the default method in order to print components
	 * of the operator in a different way than they are displayed.
	 * 
	 * @param gc the graphic context.
	 */
	@Override
	public final void printComponent( Graphics gc ) {
		super.printComponent(gc);
		gc.setColor(Color.black);
		gc.setFont(PreferenceChooser.getDefaultFont());
		FontMetrics fm = gc.getFontMetrics();
		int row = fm.getHeight();
		// Name
		super.centerText(getNickName(), gc, row, false);

		// Print each parameters.
		// Print the value only of if the parameter is not connected.
		for (int i = 0; i < getInputParameterSlots().size(); i++) {
			InputParameterSlot param = getInputParameterSlots().get(i);
			row += fm.getHeight() + 1;
			if (param.isConnected()) {
				super.centerText("$" + i + ":<linked>", gc, row, true);
			} else {
				super.centerText("$" + i + ":" + param.getValue(), gc, row, true);
			}
		}
		row += fm.getHeight() + 2;
		if (getFormula() != null) {
			super.centerText(getFormula(), gc, row, true);
		}
	}
}