/* -*- c-basic-offset: 3 -*-
 *
 * Copyright (c) 2013, GREYC.
 * All rights reserved
 *
 * You may use this file under the terms of the BSD license as follows:
 *
 * "Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of the GREYC, nor the name of its
 *     contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
 * 
 *   
 * For more information, refer to:
 * https://clouard.users.greyc.fr/Pandore
 */

package fr.greyc.pantheon.ariane.graph.item;

import java.io.IOException;
import java.util.Iterator;
import java.util.List;

import fr.greyc.pantheon.ariane.Ariane;
import fr.greyc.pantheon.ariane.graph.Graph;
import fr.greyc.pantheon.ariane.graph.SubGraph;
import fr.greyc.pantheon.ariane.graph.exception.ArianeException;
import fr.greyc.pantheon.ariane.graph.exception.ArianeMissingInputException;
import fr.greyc.pantheon.ariane.graph.exception.ArianeMissingParameterException;
import fr.greyc.pantheon.ariane.graph.slot.InputDataSlot;
import fr.greyc.pantheon.ariane.graph.slot.InputMaskSlot;
import fr.greyc.pantheon.ariane.graph.slot.InputParameterSlot;
import fr.greyc.pantheon.ariane.graph.slot.OutputDataSlot;
import fr.greyc.pantheon.ariane.graph.slot.OutputResultSlot;
import fr.greyc.pantheon.ariane.ui.PreferenceChooser;
import fr.greyc.pantheon.ariane.utils.TreeItem;
import fr.greyc.pantheon.ariane.xml.XMLElement;

/**
 * A <code>ControlStructureForItem</code> implements 
 * the iterative control structure by a graph item.
 * The sequence of Items to has be iterated is inclosed
 * inside a subgraph.
 * The loop is made from the value of the parameter $0 until
 * the value of the parameter $1 by step of the value of the parameter $2.
 *
 * @author Sébastien LEFÈVRE, Richard LOBRÉAU - Feb 21, 1999
 * @version Adrien SCHRÈQUE, Hicham AMADI - Nov 05, 2000
 * @version Mathieu JOETS - Jul 09, 2001
 * @version Alban DOUBLET, Guillaume CHATELET - Jun 10, 2002
 * @version Régis CLOUARD - Oct 16, 2001
 * @version Régis CLOUARD -Jul 13, 2010 (display loop variable during execution)
 */
public final class ControlStructureForItem extends CompositeItem {
	/** The tag name of the item. */
	public static final String TAGNAME = "For";

	/**
	 * Creates a new <code>ControlStructureFor</code> on the specified
	 * graph with the given attributes values.
	 * Initializes the mouse listeners, the popup menu,
	 * and displays it in the graph.
	 * 
	 * @param treeItem the reference tree item.
	 * @param graph the parent graph.
	 * @param name the name of the item.
	 */
	public ControlStructureForItem( TreeItem treeItem, Graph graph, String name ) throws InterruptedException {
		super(treeItem, graph, name );
		askProperties(null, 1, 1, 1);
		// used to display the current loop value
		setSignature(_treeItem.getSignature());
		setComment(PreferenceChooser.EMPTY_STRING);
	}

	/**
	 * Creates a <code>ControlStructureForItem</code> on the specified
	 * graph with the given XML data.
	 * 
	 * @param graph the parent graph.
	 * @param xmlTree the XML tree of data.
	 * @throws IOException XML errors.
	 */
	public ControlStructureForItem( Graph graph, List<XMLElement> xmlTree ) throws IOException {
		super(graph);
		load(xmlTree);
		// used to display the current loop value
		setComment(PreferenceChooser.EMPTY_STRING);
		if (_treeItem == null) {
			_treeItem = restoreDefaultTreeItem();
		}
		setSignature(_treeItem.getSignature());
		draw();
	}

	/**
	 * Creates a <code>ControlStructureForItem</code>
	 * as a subroutine (undeditable item) on the specified
	 * graph with the given XML data.
	 * 
	 * @param graph the parent graph.
	 * @param xmlTree the XML tree of data.
	 * @throws IOException XML errors.
	 */
	public ControlStructureForItem( Graph graph, List<XMLElement> xmlTree, TreeItem treeItem ) throws IOException {
		super(graph, treeItem);
		load(xmlTree);
		// used to display the current loop value
		setComment(PreferenceChooser.EMPTY_STRING);
		setSignature(_treeItem.getSignature());
		setAbstractName(treeItem.getName());
		setNickName(treeItem.getName());
		draw();
	}

	@Override
	protected TreeItem restoreDefaultTreeItem() {
		setNickName(getNickName());
		setAbstractName("for");
		return Ariane.getWorkspace().getTree().getTreeItem("for", null);
	}

	/**
	 * Builds the prototype of the item from the given values.
	 * 
	 * @param name the name of the item.
	 * @param iIm the number of input images.
	 * @param iP the number of parameters.
	 * @param oIm the number of output images.
	 */
	@Override
	protected void setPrototype( String name, int iIm, int iP, int oIm ) {
		setPrototype(name, iIm, 3 + iP, oIm, iIm, 1);
	}

	/**
	 * Creates the subgraph with the convenient prototype.
	 * 
	 * @param name the name of the item.
	 * @param iIm the number of input images.
	 * @param iP the number of parameters.
	 * @param oIm the number of output images.
	 * @return the subgraph.
	 */
	@Override
	protected SubGraph createSubGraph( String name, int iIm, int iP, int oIm ) {
		return createSubGraph(getNickName(), iIm, iP - 2, oIm, iIm, 0);
	}

	/**
	 * Returns the tag name.
	 * 
	 * @return a string.
	 */
	@Override
	public String getTagName( ) {
		return TAGNAME;
	}

	///// 
	///// E X E C U T I O N
	/////

	/**
	 * Resets the item and erases the comment inside the item glyph.
	 */
	@Override
	protected void reset() {
		super.reset();
		setComment(PreferenceChooser.EMPTY_STRING);
	}

	@Override
	protected void linkItemWithVirtualItem( VirtualItem virtualOperator, boolean isFailureFatalError ) {
		// Get and check the 3 parameter values used in the conditional expression
		int[] loop = new int[3];
		for (int i = 0; i < 3; i++) {
			InputParameterSlot parameter = getInputParameterSlots().get(i);
			if (parameter.hasNumericalValue(isFailureFatalError)) {
				loop[i] = (int)parameter.getNumericalValue();
			} else {
				throw new ArianeMissingParameterException(Ariane.getString("Operator.error4") + parameter.getRank());
			}
		}
		// Set the first loop value to the virtual operator
		OutputResultSlot result1 = virtualOperator.getVirtualInputParameterSlots().get(0);
		result1.setValue(Integer.toString(loop[0]));

		// get the rest of the parameters values
		for (int i = 3; i < getInputParameterSlots().size(); i++) {
			InputParameterSlot parameter = getInputParameterSlots().get(i);
			OutputResultSlot result = virtualOperator.getVirtualInputParameterSlots().get(i - 2);
			if (parameter.hasValue(isFailureFatalError)) {
				result.setValue(parameter.getValue());
			} else {
				throw new ArianeMissingParameterException(Ariane.getString("Operator.error1") + parameter.getRank());
			}
		}

		// Get and check mask image (-> execute previous operators if needed).
		// Note: Mask is optional!
		InputMaskSlot inMasks = getInputMaskSlot();
		OutputDataSlot inMaskd = (OutputDataSlot)virtualOperator.getVirtualInputMaskSlot();
		if (inMasks.isConnected() && inMasks.hasValue(false)) {
			inMaskd.setValue(inMasks.getValue());
		}

		// Gets and checks input images.
		// (Sets the input image of the virtual operator).
		Iterator<InputDataSlot> its1 = getInputDataSlots().iterator();
		Iterator<OutputDataSlot> itd1 = virtualOperator.getVirtualInputDataSlots().iterator();
		while (its1.hasNext()) {
			InputDataSlot inImgs = its1.next();
			OutputDataSlot inImgd = itd1.next();
			if (inImgs.hasValue(isFailureFatalError)) {
				inImgd.setValue(inImgs.getValue());
			} else {
				throw new ArianeMissingInputException(Ariane.getString("Operator.error3") + inImgs.getRank());
			}
		}
	}

	/**
	 * Executes the iteration.
	 * The loop is made from the parameter value $0
	 * to the parameter value $1
	 * by step of the parameter value $2.
	 * 
	 * @param isFailureFatalError is FAILURE a fatal error?
	 * @return true if the for has been executed without errors.
	 */
	@Override
	protected boolean primitiveExecute( boolean isFailureFatalError ) {
		try {
			VirtualItem virtualOperator = getVirtualOperator();

			linkItemWithVirtualItem(virtualOperator, isFailureFatalError);
			
			int[] loop = new int[3];
			for (int i = 0; i < 3; i++) {
				InputParameterSlot parameter = getInputParameterSlots().get(i);
				if (parameter.hasNumericalValue(isFailureFatalError)) {
					loop[i] = (int)parameter.getNumericalValue();
				} else {
					throw new ArianeMissingParameterException(Ariane.getString("Operator.error4") + parameter.getRank());
				}
			}

			Ariane.getWorkspace().printSystemMessage("> For loop from " + loop[0] + " to " + loop[1] + " step " + loop[2] + " do");
			if (loop[0] > loop[1] && loop[2] > 0 
					|| loop[0] < loop[1] && loop[2] < 0) {
				Ariane.getWorkspace().printSystemMessage("> end loop for");
				return false;
			}
			// Iteration
			// Get the first value of the inner parameter
			OutputResultSlot paramd = virtualOperator.getVirtualInputParameterSlots().get(0);

			changeStatus(Status.RUNNING);
			for (int loopv = loop[0]; loopv <= loop[1] && !Graph.isExecutionStopped(); loopv += loop[2]) {
				reset();
				virtualOperator.changeStatus(Status.EXECUTED);
				
				changeStatus(Status.RUNNING);
				Ariane.getWorkspace().printSystemMessage("> loop indice " + loopv);

				// Set the value of the inner parameter with the current loop value.
				paramd.setValue(Integer.toString(loopv));
				setComment(Integer.toString(loopv));

				// The execution of the while body is split in 2 steps
				// to avoid problems when two operators use the same entry.
				
				// 1-step: Execute the graph by trying to get the intermediate images
				for (int k = 0; k < getInputDataSlots().size(); k++) {
					InputDataSlot outImgi = virtualOperator.getVirtualIntermediateDataSlots(k);
					outImgi.hasValue(isFailureFatalError);
				}

				// 2- set the result to each the intermediate images for a new loop
				for (int k = 0; k < getInputDataSlots().size(); k++) {
					InputDataSlot outImgi = virtualOperator.getVirtualIntermediateDataSlots(k);
					OutputDataSlot outImgd = virtualOperator.getVirtualInputDataSlot(k);
					if (outImgi.hasValue(isFailureFatalError)) {
						DataFile.copy(outImgi.getValue(), outImgd.getTempName());
						outImgd.setValue();
					}
				}
			}
			Ariane.getWorkspace().printSystemMessage("> end loop for");

			// Sets the new output images.
			Iterator<OutputDataSlot> its2 = getOutputDataSlots().iterator();
			Iterator<InputDataSlot>itd2 = virtualOperator.getVirtualOutputDataSlots().iterator();
			while (its2.hasNext()) {
				OutputDataSlot outImgs = its2.next();
				InputDataSlot outImgd = itd2.next();
				if (outImgd.hasValue(isFailureFatalError)) {
					outImgs.setValue(outImgd.getValue());
				}
			}

			InputParameterSlot resultd = (InputParameterSlot)virtualOperator.getVirtualOutputResultSlot();
			if (!resultd.hasValue(isFailureFatalError)) {
				throw new ArianeMissingParameterException(Ariane.getString("MacroOperator.error2"));
			}
			OutputResultSlot results = getOutputResultSlot();
			results.setValue(resultd.getValue());
			changeStatus(Status.EXECUTED);
			return true;
		} catch (ArianeException e) {
			if (isFailureFatalError && !Graph.isExecutionStopped()) {
				Ariane.getWorkspace().printErrorMessage(e.getErrorCode(), getAbsoluteName(), getId(), e.getLocalizedMessage());
				getGraph().stopExecution();
				changeStatus(Status.ERROR);
			} else {
				changeStatus(Status.UNEXECUTED);
			}
			return false;
		} catch (IOException e) {
			if (isFailureFatalError && !Graph.isExecutionStopped()) {
				Ariane.getWorkspace().printErrorMessage(Ariane.ErrorCode.INPUTOUTPUT, getAbsoluteName(), getId(), e.getLocalizedMessage());
				getGraph().stopExecution();
				changeStatus(Status.ERROR);
			} else {
				changeStatus(Status.UNEXECUTED);
			}
			return false;
		}
	}

	///// 
	///// P R O P E R T I E S
	/////

	/**
	 * Asks interactively the prototype of the structure.
	 */
	@Override
	protected void askProperties( ) throws InterruptedException {
		askProperties(null, getInputDataSlots().size(), 0, getOutputDataSlots().size());
	}
}
