/* -*- c-basic-offset: 3 -*-
 *
 * Copyright (c) 2013, GREYC.
 * All rights reserved
 *
 * You may use this file under the terms of the BSD license as follows:
 *
 * "Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of the GREYC, nor the name of its
 *     contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
 * 
 *   
 * For more information, refer to:
 * https://clouard.users.greyc.fr/Pandore
 */

package fr.greyc.pantheon.ariane.graph.item;

import fr.greyc.pantheon.ariane.Ariane;
import fr.greyc.pantheon.ariane.graph.Graph;
import fr.greyc.pantheon.ariane.graph.exception.ArianeArithmeticException;
import fr.greyc.pantheon.ariane.graph.exception.ArianeException;
import fr.greyc.pantheon.ariane.graph.exception.ArianeMissingInputException;
import fr.greyc.pantheon.ariane.graph.exception.ArianeMissingParameterException;
import fr.greyc.pantheon.ariane.graph.item.etc.LogicalExpression;
import fr.greyc.pantheon.ariane.graph.slot.InputDataSlot;
import fr.greyc.pantheon.ariane.graph.slot.InputParameterSlot;
import fr.greyc.pantheon.ariane.utils.TreeItem;
import fr.greyc.pantheon.ariane.xml.XMLElement;

import java.io.IOException;
import java.util.List;

/**
 * A <code>ControlStructureIfItem</code> implements the conditional
 * <code>ControlStructure</code>, as a tradition if (test) then ... else
 * sequence.
 * 
 * @version Régis CLOUARD - Nov 19, 2004
 * @version Sébastien LEFÈVRE, Richard LOBRÉAU - Feb 21, 1999
 * @version Adrien SCHRÈQUE, Hicham AMADI - Nov 05, 2000
 * @version Mathieu JOETS - Jul 09, 2001
 * @version Régis CLOUARD - Oct 16, 2001
 */
public class ControlStructureIfItem extends ControlStructureItem {
	/** The tag name of the item. */
	public static final String TAGNAME = "If";

	/** The boolean expression of the conditional test. */
	private LogicalExpression _condition;

	/**
	 * Creates a <code>ControlStructureIfItem</code>
	 * with the given attributes.
	 * Initializes the mouse listeners, the popup menu,
	 * and displays it in the graph.
	 * 
	 * @param treeItem the source tree item.
	 * @param graph the parent graph.
	 * @param name the item name.
	 */
	public ControlStructureIfItem( TreeItem treeItem, Graph graph, String name ) {
		super(treeItem, graph, name);
		createSlots(1, 2, 0, 2, 0);
		setAbstractName(name);
		setFormula(""); // IMPERATIVE to signify that it uses formula.
		draw();
	}

	/**
	 * Creates a <code>ControlStructureIfItem</code> on the given
	 * graph with the given XML data.
	 * 
	 * @param graph the parent graph.
	 * @param xmlTree the XML tree of data.
	 * @throws IOException XML errors.
	 */
	public ControlStructureIfItem( Graph graph, List<XMLElement> xmlTree ) throws IOException {
		super(graph);
		load(xmlTree);
		// Compile the condition
		if (getFormula() == null) {
			if (getBinary() == null) {
				setFormula(""); // IMPERATIVE to signify that it uses formula.
			} else { // Keep this for sake of compatibility.
				_condition = new LogicalExpression(getBinary());
				setFormula(getBinary());
				setBinary(null);
			}
		} else {
			_condition = new LogicalExpression(getFormula());
		}
		draw();
	}

	/**
	 * Builds or changes the prototype o
	 * f the control structure
	 * from the given values.
	 * 
	 * @param name the name of the item.
	 * @param iIm the number of input images.
	 * @param iP the number of parameters.
	 * @param oIm the number of output images.
	 */
	@Override
	protected final void setPrototype( String name, int iIm, int iP, int oIm ) {
		// Empty
	}

	/**
	 * Returns the tag name.
	 * 
	 * @return a string.
	 */
	@Override
	public final String getTagName() {
		return TAGNAME;
	}

	///// 
	/////  E X E C U T I O N 
	/////

	/**
	 * Executes the conditional control, it means:
	 * <ul>
	 * <li>check if the two input are validated.
	 * <li>check if the two parameters are available.
	 * <li>check if the boolean expression gives a result.
	 * </ul>
	 * 
	 * @param isFailureFatalError is FAILURE be noticed to user?
	 * @return true if the operator has been executed without errors.
	 */
	@Override
	protected final boolean primitiveExecute( boolean isFailureFatalError ) {
		try {
			if (getFormula().isEmpty()) {
				throw new ArianeArithmeticException(Ariane.getString("arithmetic.error0"));
			}
	
			// gets input parameters. 
			double[] val = new double[2];
			for (int i = 0; i < 2; i++) {
				InputParameterSlot parameter = getInputParameterSlots().get(i);
				if (parameter.hasNumericalValue(isFailureFatalError)) {
					val[i] = parameter.getNumericalValue();
				} else {
					throw new ArianeMissingParameterException(Ariane.getString("Operator.error4") + parameter.getRank());
				}
			}

			// gets input images.
			InputDataSlot input = getInputDataSlots().get(0);
			if (!input.hasValue(isFailureFatalError)) {
				throw new ArianeMissingInputException(Ariane.getString("Operator.error3") + input.getRank());
			}

			// use the conditional expression result to select the right output.
			int outputRank;
			outputRank = _condition.evaluate(val[0], val[1]) ? 0 : 1;

			// validate the selected output.
			getOutputDataSlots().get(outputRank).setValue(input.getValue());
			// unvalidate the other one.
			getOutputDataSlots().get(1 - outputRank).reset();

			getOutputResultSlot().setValue(Integer.toString(outputRank));
			changeStatus(Status.EXECUTED);
			return true;
		} catch (ArianeException e) {
			if (isFailureFatalError && !Graph.isExecutionStopped()) {
				Ariane.getWorkspace().printErrorMessage(e.getErrorCode(), getAbsoluteName(), getId(), e.getLocalizedMessage());
				getGraph().stopExecution();
				changeStatus(Status.ERROR);
			}
			return false;
		}
	}

	///// 
	/////  P R O P E R T I E S 
	/////

	/**
	 * Changes the current boolean expression with the 
	 * given one.
	 * 
	 * @param condition the conditional expression.
	 * @return true if the formula is valid, false otherwise.
	 */
	@Override
	public final boolean changeFormula( String condition ) {
		if (getFormula() != null && getFormula().equals(condition)) {
			return true;
		}
		try {
			_condition = new LogicalExpression(condition);
			setFormula(condition);
			return true;
		} catch (ArianeArithmeticException e) {
			Ariane.getWorkspace().printErrorMessage(e.getLocalizedMessage());
		}
		return false;
	}
}