/* -*- c-basic-offset: 3 -*-
 *
 * Copyright (c) 2013, GREYC.
 * All rights reserved
 *
 * You may use this file under the terms of the BSD license as follows:
 *
 * "Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of the GREYC, nor the name of its
 *     contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
 * 
 *   
 * For more information, refer to:
 * https://clouard.users.greyc.fr/Pandore
 */

package fr.greyc.pantheon.ariane.graph.item;
import java.io.IOException;
import java.util.List;

import fr.greyc.pantheon.ariane.Ariane;
import fr.greyc.pantheon.ariane.graph.Graph;
import fr.greyc.pantheon.ariane.graph.exception.ArianeException;
import fr.greyc.pantheon.ariane.graph.exception.ArianeMissingInputException;
import fr.greyc.pantheon.ariane.graph.exception.ArianeMissingParameterException;
import fr.greyc.pantheon.ariane.graph.slot.InputDataSlot;
import fr.greyc.pantheon.ariane.graph.slot.InputParameterSlot;
import fr.greyc.pantheon.ariane.utils.TreeItem;
import fr.greyc.pantheon.ariane.xml.XMLElement;

/**
 * A <code>ControlStructureSelectItem</code> implements the a
 * selector of multiple inputs. The selected input
 * is given by the parameter value.
 *
 * @author Régis CLOUARD - Oct 01, 1998
 * @version Sébastien LEFÈVRE, Richard LOBRÉAU - Feb 21, 1999
 * @version Adrien SCHRÈQUE, Hicham AMADI - Nov 05, 2000
 * @version Mathieu JOETS - Jul 09, 2001
 * @version Régis CLOUARD - Oct 16, 2001
 */
public class ControlStructureSelectItem extends ControlStructureItem {
	/** The tag name of the item. */
	public static final String TAGNAME = "Select";

	/**
	 * Creates a <code>ControlStructureSelectItem</code>
	 * with the given attributes.
	 * Initializes the mouse listeners, the popup menu,
	 * and displays it in the graph.
	 * 
	 * @param treeItem the reference tree item.
	 * @param graph the parent graph.
	 * @param name the item name.
	 */
	public ControlStructureSelectItem( TreeItem treeItem, Graph graph, String name ) throws InterruptedException {
		super(treeItem, graph, name);
		askProperties(null, 1, 0, 0);
		setSignature(treeItem.getSignature());
	}

	/**
	 * Creates a <code>ControlStructureSelectItem</code> on the given
	 * graph with the given XML data.
	 * 
	 * @param graph the parent graph.
	 * @param xmlTree the XML tree of data.
	 * @throws IOException XML errors.
	 */
	public ControlStructureSelectItem( Graph graph, List<XMLElement> xmlTree ) throws IOException {
		super(graph);
		load(xmlTree);
		draw();
	}

	/**   
	 * Builds or changes the prototype of the control structure
	 * from the given values.
	 * 
	 * @param name the name of the item.
	 * @param iIm the number of input images.
	 * @param iP the number of parameters.
	 * @param oIm the number of output images.
	 */
	@Override
	protected final void setPrototype( String name, int iIm, int iP, int oIm ) {
		setAbstractName(name);
		createSlots(iIm, 1, 0, 1, 0);
		draw();
	}

	/**
	 * Returns the tag name.
	 * 
	 * @return a string.
	 */
	@Override
	public final String getTagName( ) {
		return TAGNAME;
	}

	///// 
	/////  E X E C U T I O N 
	/////

	/**
	 * Executes a <i>Select</i>, it means select
	 * the input rank indicated hy
	 * the parameter value.
	 * 
	 * @param isFailureFatalError is FAILURE a fatal error?
	 * @return true if the operator has been executed without errors.
	 */
	@Override
	protected final boolean primitiveExecute( boolean isFailureFatalError ) {
		try {
			// Gets and checks the parameter (-> executes forward operators if needed).
			InputParameterSlot parameter = getInputParameterSlots().get(0);
			int inputRank;
			if (parameter.hasNumericalValue(isFailureFatalError)) {
				inputRank = (int)parameter.getNumericalValue();
			} else {
				throw new ArianeMissingParameterException(Ariane.getString("Operator.error1") + parameter.getRank());
			}

			// Checks consistency of the parameter.
			if (inputRank < 0 || inputRank >= getInputDataSlots().size()) {
				throw new ArithmeticException(Ariane.getString("controlstructure.error1")); 
			}

			// Gets and checks the input image (-> executes forward operators if needed).
			InputDataSlot isi = getInputDataSlots().get(inputRank);
			if (isi != null && isi.hasValue(isFailureFatalError)) {
				// Validate the output.
				(getOutputDataSlots().get(0)).setValue(isi.getValue());
				getOutputResultSlot().setValue(Float.toString(inputRank));
			} else {
				int rank = (isi == null) ? 0 : isi.getRank();
				throw new ArianeMissingInputException(Ariane.getString("Operator.error3") + rank);
			}
			changeStatus(Status.EXECUTED);
			return true;
		} catch (ArianeException e) {
			if (isFailureFatalError && !Graph.isExecutionStopped()) {
				Ariane.getWorkspace().printErrorMessage(e.getErrorCode(), getAbsoluteName(), getId(), e.getLocalizedMessage());
				getGraph().stopExecution();
				changeStatus(Status.ERROR);
			}
			return false;
		}
	}

	///// 
	/////  P R O P E R T I E S
	/////

	/**   
	 * Asks interactively the prototype of the structure.
	 */
	protected final void askProperties() throws InterruptedException {
		askProperties(null, 1, 0, 0);
	}
}