/* -*- c-basic-offset: 3 -*-
 *
 * Copyright (c) 2013, GREYC.
 * All rights reserved
 *
 * You may use this file under the terms of the BSD license as follows:
 *
 * "Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of the GREYC, nor the name of its
 *     contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
 * 
 *   
 * For more information, refer to:
 * https://clouard.users.greyc.fr/Pandore
 */

package fr.greyc.pantheon.ariane.graph.item;

import fr.greyc.pantheon.ariane.Ariane;
import fr.greyc.pantheon.ariane.graph.Graph;
import fr.greyc.pantheon.ariane.graph.SubGraph;
import fr.greyc.pantheon.ariane.graph.exception.ArianeArithmeticException;
import fr.greyc.pantheon.ariane.graph.exception.ArianeException;
import fr.greyc.pantheon.ariane.graph.exception.ArianeMissingInputException;
import fr.greyc.pantheon.ariane.graph.exception.ArianeMissingParameterException;
import fr.greyc.pantheon.ariane.graph.item.etc.LogicalExpression;
import fr.greyc.pantheon.ariane.graph.slot.InputDataSlot;
import fr.greyc.pantheon.ariane.graph.slot.InputMaskSlot;
import fr.greyc.pantheon.ariane.graph.slot.InputParameterSlot;
import fr.greyc.pantheon.ariane.graph.slot.OutputDataSlot;
import fr.greyc.pantheon.ariane.graph.slot.OutputResultSlot;
import fr.greyc.pantheon.ariane.ui.PreferenceChooser;
import fr.greyc.pantheon.ariane.utils.TreeItem;
import fr.greyc.pantheon.ariane.xml.XMLElement;

import java.io.IOException;
import java.util.Iterator;
import java.util.List;

/**
 * A <code>ControlStructureWhileItem</code> implements the
 * repetitive structure control by a graph item.
 * The sequence of Items to has be repeated is inclosed
 * inside a subgraph.
 * 
 * @author Sébastien LEFÈVRE, Richard LOBRÉAU - Feb 21, 1999
 * @version Adrien SCHRÈQUE, Hicham AMADI - Nov 05, 2000
 * @version Mathieu JOETS - Jul 09, 2001
 * @version Alban DOUBLET, Guillaume CHATELET - Jun 10, 2002
 * @version Régis CLOUARD - Oct 16, 2001 
 * @version Régis CLOUARD - Jun 11, 2010 (Add several optional parameters)
 * @version Régis CLOUARD - Mar 24, 2011 (fix bug on execution)
 */
public final class ControlStructureWhileItem extends CompositeItem {
	/** The tag name of the item. */
	public static final String TAGNAME = "While";

	/** The boolean expression of the conditional test. */
	private LogicalExpression _condition;

	/**
	 * Creates a new <code>ControlStructureWhileItem</code> on the specified
	 * graph with the given attributes.
	 * 
	 * @param treeItem the reference tree item.
	 * @param graph the parent graph.
	 * @param name the default name of the component.
	 */
	public ControlStructureWhileItem( TreeItem treeItem, Graph graph, String name ) throws InterruptedException {
		super(treeItem, graph, name );
		askProperties(null, 1, 1, 1);
		// used to display the current loop value
		setSignature(_treeItem.getSignature());
		setComment(PreferenceChooser.EMPTY_STRING);
	}

	/**
	 * Creates a <code>ControlStructureWhileItem</code> on the specified
	 * graph from the given XML data.
	 * 
	 * @param graph the parent graph.
	 * @param xmlTree the XML tree of data.
	 * @throws IOException XML errors.
 */
	public ControlStructureWhileItem( Graph graph, List<XMLElement> xmlTree ) throws IOException {
		super(graph);
		load(xmlTree);
		if (getFormula() == null) {
			if (getBinary() != null) { // Keep for sake of compatibility
				_condition = new LogicalExpression(getBinary());
				setFormula(getBinary());
				setBinary(null);
			}
			setFormula(""); // IMPERATIVE to signify that it uses formula.
		} else {
			_condition = new LogicalExpression(getFormula());
		}
		// used to display the current loop value
		setComment(PreferenceChooser.EMPTY_STRING);
		if (_treeItem == null) {
			_treeItem = restoreDefaultTreeItem();
		}
		setSignature(_treeItem.getSignature());
		draw();
	}

	/**
	 * Creates a <code>ControlStructureWhileItem</code> 
	 * as a subroutine (undeditable item) on the specified
	 * graph from the given XML data.
	 * 
	 * @param graph the parent graph.
	 * @param xmlTree the XML tree of data.
	 * @throws IOException XML errors.
 */
	public ControlStructureWhileItem( Graph graph, List<XMLElement> xmlTree, TreeItem treeItem ) throws IOException {
		super(graph, treeItem);
		load(xmlTree);
		if (getFormula() == null) {
			if (getBinary() != null) { // Keep for sake of compatibility
				_condition = new LogicalExpression(getBinary());
				setFormula(getBinary());
				setBinary(null);
			}
			setFormula(""); // IMPERATIVE to signify that it uses formula.
		} else {
			_condition = new LogicalExpression(getFormula());
		}
		// used to display the current loop value
		setComment(PreferenceChooser.EMPTY_STRING);
		setSignature(_treeItem.getSignature());
		setAbstractName(treeItem.getName());
		setNickName(treeItem.getName());
		draw();
	}

	@Override
	protected TreeItem restoreDefaultTreeItem() {
		setNickName(getNickName());
		setAbstractName("while");
		return Ariane.getWorkspace().getTree().getTreeItem("while", null);
	}

	/**
	 * Builds the prototype of the item from the given values.
	 * 
	 * @param name the name of the item.
	 * @param iIm the number of input images.
	 * @param iP the number of parameters.
	 * @param oIm the number of output images.
	 */
	@Override
	protected void setPrototype( String name, int iIm, int iP, int oIm ) {
		super.setPrototype(name, iIm, 2 + iP, oIm, iIm, 2);
		setFormula(""); // IMPERATIVE to signify that it uses formula.
		draw();
	}

	/**
	 * Creates the subgraph with the convenient prototype.
	 * 
	 * @param name the name of the item.
	 * @param iIm the number of input images.
	 * @param iP the number of parameters.
	 * @param oIm the number of output images.
	 * @return the subgraph.
	 */
	@Override
	protected SubGraph createSubGraph( String name, int iIm, int iP, int oIm ) {
		return createSubGraph(getNickName(), iIm, iP, oIm, iIm, 2);
	}

	/**
	 * Returns the tag name.
	 * 
	 * @return a string.
	 */
	@Override
	public String getTagName( ) {
		return TAGNAME;
	}

	///// 
	///// E X E C U T I O N
	/////

	@Override
	protected void linkItemWithVirtualItem( VirtualItem virtualOperator, boolean isFailureFatalError ) {
		// Get the rest of the parameters
		for (int i = 0; i < getInputParameterSlots().size(); i++) {
			InputParameterSlot parameter = getInputParameterSlots().get(i);
			OutputResultSlot result = virtualOperator.getVirtualInputParameterSlots().get(i);
			if (parameter.hasValue(isFailureFatalError)) {
				result.setValue(parameter.getValue());
			} else {
				throw new ArianeMissingParameterException(Ariane.getString("Operator.error1") + parameter.getRank());
			}
		}

		// Get and check the mask image (-> execute previous operators if needed)
		// Note: Mask is optional
		InputMaskSlot inMasks = getInputMaskSlot();
		OutputDataSlot inMaskd = (OutputDataSlot)virtualOperator.getVirtualInputMaskSlot();
		if (inMasks.isConnected() && inMasks.hasValue(false)) {
			inMaskd.setValue(inMasks.getValue());
		}

		// Gets and check the input images
		Iterator<InputDataSlot> its = getInputDataSlots().iterator();
		Iterator<OutputDataSlot> itd = virtualOperator.getVirtualInputDataSlots().iterator();
		while (its.hasNext()) {
			InputDataSlot inImgs = its.next();
			OutputDataSlot inImgd = itd.next();
			if (inImgs.hasValue(isFailureFatalError)) {
				inImgd.setValue(inImgs.getValue());
			} else {
				throw new ArianeMissingInputException(Ariane.getString("Operator.error3") + inImgs.getRank());
			}
		}
	}

	/**
	 * Executes the repetition while the related condition is true.
	 * 
	 * @param isFailureFatalError is FAILURE a fatal error?
	 * @return true if the operator has been executed without errors.
	 */
	@Override
	protected boolean primitiveExecute( boolean isFailureFatalError ) {
		try {
			if (getFormula().isEmpty()) {
				throw new ArianeArithmeticException(Ariane.getString("arithmetic.error0"));
			}

			VirtualItem virtualOperator = getVirtualOperator();
			linkItemWithVirtualItem(virtualOperator, isFailureFatalError);

			double[] whilevariables = new double[2];
			for (int i = 0; i < 2; i++) {
				InputParameterSlot parameter = getInputParameterSlots().get(i);
				if (parameter.hasNumericalValue(isFailureFatalError)) {
					whilevariables[i] = (int)parameter.getNumericalValue();
				} else {
					throw new ArianeMissingParameterException(Ariane.getString("Operator.error4") + parameter.getRank());
				}
			}

			changeStatus(Status.RUNNING);

			// Repetition
			int iteration = 0;
			Ariane.getWorkspace().printSystemMessage("> While loop (" + _condition.toHtml() + ") do");
			while (_condition.evaluate(whilevariables[0], whilevariables[1]) && !Graph.isExecutionStopped()) {
				iteration++;

				Ariane.getWorkspace().printSystemMessage(">cycle: " + iteration);
				setComment(Integer.toString(iteration));

				// Set the value of the 2 inner parameters.
				for (int i = 0; i < 2; i++) {
					virtualOperator.getVirtualInputParameterSlot(i).setValue(Float.toString((float)whilevariables[i]));
				}
				reset();
				virtualOperator.changeStatus(Status.EXECUTED); // Prevent the rexecution of the virtual operator.

				changeStatus(Status.RUNNING);

				// The execution of the while body is split in 2 steps
				// to avoid problems when two operators use the same entry.

				// 1-step: Execute the graph by trying to get the intermediate images
				for (int k = 0; k < getInputDataSlots().size(); k++) {
					InputDataSlot outImgi = virtualOperator.getVirtualIntermediateDataSlots(k);
					outImgi.hasValue(isFailureFatalError);
				}

				// 2- set the result to each the intermediate images for a new loop
				for (int k = 0; k < getInputDataSlots().size(); k++) {
					InputDataSlot outImgi = virtualOperator.getVirtualIntermediateDataSlots(k);
					OutputDataSlot outImgd = virtualOperator.getVirtualInputDataSlot(k);
					if (outImgi.hasValue(isFailureFatalError)) {
						DataFile.copy(outImgi.getValue(), outImgd.getTempName());
						outImgd.setValue();
					}
				}

				// Get the new parameters value.
				for (int i = 0; i < 2; i++) {
					InputParameterSlot parameter = virtualOperator.getVirtualIntermediateResultSlot(i);
					if (parameter.isConnected() && parameter.hasNumericalValue(isFailureFatalError)) {
						whilevariables[i] = (int)parameter.getNumericalValue();
					}
				}
			}
			Ariane.getWorkspace().printSystemMessage("end while loop");

			// Set the new output images.
			Iterator<OutputDataSlot> its1 = getOutputDataSlots().iterator();
			Iterator<InputDataSlot> itd1 = virtualOperator.getVirtualOutputDataSlots().iterator();
			while (its1.hasNext()) {
				OutputDataSlot outImgs = its1.next();
				InputDataSlot outImgd = itd1.next();
				if (outImgd.hasValue(isFailureFatalError)) {
					outImgs.setValue(outImgd.getValue());
				}
			}

			InputParameterSlot resultd = (InputParameterSlot)virtualOperator.getVirtualOutputResultSlot();
			if (!resultd.hasValue(isFailureFatalError)) {
				throw new ArianeMissingParameterException(Ariane.getString("MacroOperator.error2"));
			}
			OutputResultSlot results = getOutputResultSlot();
			results.setValue(resultd.getValue());
			changeStatus(Status.EXECUTED);
			return true;
		} catch (ArianeException e) {
			if (isFailureFatalError && !Graph.isExecutionStopped()) {
				Ariane.getWorkspace().printErrorMessage(e.getErrorCode(), getAbsoluteName(), getId(), e.getLocalizedMessage());
				getGraph().stopExecution();
				changeStatus(Status.ERROR);
			} else {
				changeStatus(Status.UNEXECUTED);
			}
			return false;
		} catch (IOException e) {
			if (isFailureFatalError && !Graph.isExecutionStopped()) {
				Ariane.getWorkspace().printErrorMessage(Ariane.ErrorCode.INPUTOUTPUT, getAbsoluteName(), getId(), e.getLocalizedMessage());
				getGraph().stopExecution();
				changeStatus(Status.ERROR);
			} else {
				changeStatus(Status.UNEXECUTED);
			}
			return false;
		}
	}

	///// 
	///// P R O P E R T I E S
	/////

	/**
	 * Changes the current boolean expression with the given one.
	 * 
	 * @param condition the conditional expression.
	 * @return true if the formula is valid, false otherwise.
	 */
	@Override
	public boolean changeFormula( String condition ) {
		if (getFormula() != null && getFormula().equals(condition)) {
			return true;
		}
		try {
			_condition = new LogicalExpression(condition);
			setFormula(condition);
			return true;
		} catch (ArianeArithmeticException e) {
			Ariane.getWorkspace().printErrorMessage(e.getLocalizedMessage());
		}
		return false;
	}

	/**
	 * Asks interactively the prototype of the item.
	 */
	@Override
	protected void askProperties( ) throws InterruptedException {
		askProperties(null, 1, 0, 0);
	}
}