/* -*- c-basic-offset: 3 -*-
 *
 * Copyright (c) 2013, GREYC.
 * All rights reserved
 *
 * You may use this file under the terms of the BSD license as follows:
 *
 * "Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of the GREYC, nor the name of its
 *     contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
 * 
 *   
 * For more information, refer to:
 * https://clouard.users.greyc.fr/Pandore
 */

package fr.greyc.pantheon.ariane.graph.item;

import fr.greyc.pantheon.ariane.Ariane;
import fr.greyc.pantheon.ariane.graph.Graph;
import fr.greyc.pantheon.ariane.graph.command.ChangeDatafileEdit;
import fr.greyc.pantheon.ariane.graph.exception.ArianeRuntimeException;
import fr.greyc.pantheon.ariane.graph.item.etc.OperatorProcessBuilder;
import fr.greyc.pantheon.ariane.ui.PreferenceChooser;
import fr.greyc.pantheon.ariane.utils.TreeItem;
import fr.greyc.pantheon.ariane.xml.XMLElement;

import java.awt.Color;
import java.awt.FontMetrics;
import java.awt.Graphics;
import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.LinkedList;
import java.util.List;

import javax.swing.JButton;
import javax.swing.JTextField;

/**
 * A <code>DataFile</code> is either a loader, a saver or a viewer
 * of a data files.
 * 
 * @author Régis CLOUARD - Oct 01, 1998
 * @version Sébastien LEFÈVRE, Richard LOBRÉAU - Feb 21, 1999
 * @version Adrien SCHRÈQUE, Hicham AMADI - Nov 05, 2000
 * @version Mathieu JOETS - Jul 09, 2001 
 * @version Régis CLOUARD - Oct 16, 2001
 * @version Régis CLOUARD - Dec 03, 2004
 * @version Régis CLOUARD - Sep 12, 2007 (split into Loader, Saver)
 * @version Régis CLOUARD - Jun 20, 2010 (add Viewer)
 */
public abstract class DataFile extends GraphItem {
	/** The title of related windows. */
	private final String _windowTitle;

	/** The last loading directory. One value for all instances. */
	protected static String currentLoadingDir;

	/** The last saving directory. One value for all instances. */
	protected static String currentWorkingDir;

	/** The complete name of the image filename. */
	private String _currentFilename;

	/** The color of the item. */
	private static final Color DEFAULT_COLOR = new Color(102, 255, 102); 

	/** The color of the item. */
	private static Color _color = getDefaultColor();

	/**
	 * Creates an <code>ImageFile</code> item on the given graph with
	 * the specified properties.
	 * 
	 * @param treeItem the reference tree item.
	 * @param graph the parent graph.
	 * @param name the name of the item.
	 * @param binary the name of the related file.
	 * @param category the name of the category of the operator.
	 * @param windowTitle the title of the related windows.
	 */
	protected DataFile( TreeItem treeItem, Graph graph, String name, String binary, String category, String windowTitle ) {
		super(treeItem, graph, name, category);
		_windowTitle = windowTitle;
		setBinary(binary);
	}

	/**
	 * Creates an <code>ImageFile</code> item on the given graph with the
	 * specified XML data.
	 * 
	 * @param graph the parent graph.
	 * @param xmlTree the XML tree of data.
	 * @param windowTitle the title of the related windows.
	 * @param hasResult Do it add result slot or not.
	 * @throws IOException XML errors.
	 */
	protected DataFile( Graph graph, List<XMLElement> xmlTree, String windowTitle, boolean hasResult ) throws IOException {
		super(graph);
		_windowTitle = windowTitle;
		if (hasResult) {
			createSlots(0, 0, 0, 0, 1);
		}
		load(xmlTree);
	}

	/**
	 * Creates a <code>Loader</code> or a <code>Saver</code>
	 * on the given graph with the given XML data.
	 * Notice: Loader has 1 input image and 0 output image and
	 * saver has 0 input image and 1 output image.
	 * !!!! This method is kept for sake of compatibility.
	 * 
	 * @param graph the parent graph.
	 * @param xmlTree the XML tree of data.
	 * @return the item.
	 * @throws IOException XML errors.
	 */
	public static final DataFile getInstance(Graph graph, List<XMLElement> xmlTree ) throws IOException {
		int inputDataSlots = 0;
		for (XMLElement element : xmlTree) {
			if (element.getName().equalsIgnoreCase("ImagesEntree")) {
				inputDataSlots = Integer.parseInt(element.getValue());
			}
		}
		if (inputDataSlots > 0) {
			return new SaverItem(graph, xmlTree);
		}
		return new LoaderItem(graph, xmlTree);
	}

	public static String getFileNameWithoutExtension( String binary ) {
		return binary.substring(binary.lastIndexOf(File.separator) + 1);
	}

	protected final void setCurrentFilename( String filename ) {
		_currentFilename = filename;
	}

	protected abstract void resetButton( JButton button );

	public final void changeBinaryName( JTextField textfield1, JTextField textfield2, JButton button, String binary, boolean historical) {
		resetAll();
		String oldBinary = getBinary();
		setBinary(binary);

		if (getName().equals(getNickName()) || oldBinary != null && DataFile.getFileNameWithoutExtension(oldBinary).equals(getNickName())) {
			if (binary == null) {
				setNickName(null);
			} else {
				setNickName(DataFile.getFileNameWithoutExtension(binary));
			}
		}
		textfield1.setText(getNickName());
		textfield2.setText(getBinary());
		if (button != null) {
			if (binary == null) {
				resetButton(button);
			} else {
				button.setText(DataFile.getFileNameWithoutExtension(binary));
			}
		}

		updateName();
		getGraph().setModified(true); // The graph is modified too.
		if (historical) {
			getGraph().startActionEdition();
			getGraph().validateAction(new ChangeDatafileEdit(this, textfield1, textfield2, button, binary, oldBinary));
			getGraph().endActionEdition();
		}
	}

	/**
	 * Sets the name of the related physical file.
	 * If the name of the file is specified as a relative path 
	 * than the absolute name is built with the directory of the project.
	 * 
	 * @param filename a file name (absolute or relative).
	 * @param current_directory the name of the current directory or null if the related graph file has no name.
	 */
	public static final String setPathToFilename( String filename, String currentDirectory ) {
		if (filename == null || filename.length() == 1) {
			return "";
		}
		String binary = filename;
		// relative name
		if (binary.charAt(0) == '.') {
			binary = binary.substring(2);
			File file = new File(currentDirectory, binary.replaceAll("\\\\", "/"));
			return file.getAbsolutePath();
		} else if (binary.startsWith("$FILE_DIRECTORY")) {
			String fileDirectory = PreferenceChooser.get(PreferenceChooser.FILE_DIRECTORY);
			binary = binary.substring("$FILE_DIRECTORY".length() + 1);
			File file = new File(fileDirectory, binary.replaceAll("\\\\", "/"));
			return file.getAbsolutePath();
		}
		// absolute name
		File binaryFilename = new File(binary);
		if (binaryFilename.exists()) {
			try {
				return binaryFilename.getCanonicalPath();
			} catch (IOException e) {
			}
		}
		// Keep for the sake of compatibility
		File file = new File(currentDirectory, binary.replaceAll("\\\\", "/"));
		if (file.exists()) {
			return file.getAbsolutePath();
		}
		// default
		return binaryFilename.getAbsolutePath();
	}

	/**
	 * Returns the name of the file either
	 * with relative path if the file is in the filePath directory
	 * or with an absolute path otherwise.
	 * 
	 * @param fileName the absolute path the file.
	 * @param directory the base directory.
	 * @return the name.
	 */
	public static final String getRelativePath( String filename, String directory ) {
		// Case of simple cut-copy/paste or no filename.
		if (filename == null || directory == null || filename.trim().isEmpty() || filename.trim().charAt(0) != File.separatorChar) {
			return filename;
		}
		File file = new File(filename);
		String filePath = file.getParent();
		String fileDirectory = PreferenceChooser.get(PreferenceChooser.FILE_DIRECTORY);
		if (fileDirectory != null && filePath != null && filePath.startsWith(fileDirectory)) {
			return "$FILE_DIRECTORY" + filePath.substring(fileDirectory.length()) + File.separator + file.getName();
		} else {
			File path = new File(directory);
			if (path != null && filePath != null && filePath.startsWith(directory)) {
				return "." + filePath.substring(directory.length()) + File.separator + file.getName();
			}
		}
		return file.getAbsolutePath();
	}

	/**
	 * Returns the description of the data file : id + data file name.
	 * 
	 * @return a string.
	 */
	@Override
	public final String toString( ) {
		if (getBinary() == null) {
			return "(" + getId() + "):" + getName();
		} else {
			return "(" + getId() + "):" + getBinary();
		}
	}

	/**
	 * Copies a file into an another file.
	 * Overrides the system command 'copy' in order to be independent
	 * from operative system.
	 * 
	 * @param from the source file.
	 * @param to the destination file.
	 * @throws IOException in case of IO problems.
	 * @author Guillaume CHATELET - 03/14/2002
	 */
	public static final void copy( String from, String to ) throws IOException {
		// Try to copy a file into itself: nothing to do!
		if (from.equals(to)) {
			return;
		}
		InputStream in = null;
		OutputStream out = null;
		try {
			InputStream inFile = new FileInputStream(from);
			in = new BufferedInputStream(inFile);
			OutputStream outFile = new FileOutputStream(to);
			out = new BufferedOutputStream(outFile);
			while (true) {
				int data = in.read();
				if (data == -1) {
					break;
				}
				out.write(data);
			}
		} finally {
			if (in != null) {
				in.close();
			}
			if (out != null) {
				out.close();
			}
		}
	}

	///// 
	///// P R E F E R E N C E S
	/////

	/**
	 * Changes the current color.
	 * 
	 * @param col the new color.
	 */
	public static void setColor( Color col ) {
		_color = col;
	}

	/**
	 * Returns the current color.
	 * 
	 * @return the color.
	 */
	public static Color getColor( ) {
		return _color;
	}

	/**
	 * Returns the default color.
	 * 
	 * @return the color.
	 */
	public static Color getDefaultColor( ) {
		return DEFAULT_COLOR;
	}

	/**
	 * Returns the background color.
	 * 
	 * @return the color.
	 */
	@Override
	public Color getBackgroundColor( ) {
		return _color;
	}

	///// 
	///// E X E C U T I O N 
	/////

	/**
	 * * @return the absolute name of the file content viewer.
	 */
	public String getViewerCommand() {
		return GraphItem.getAbsoluteBinary(getCommandVisu());
	}

	/**
	 * Displays the file with the specified software (in preferences).
	 * The data file needs to be executed before.
	 * @throws InterruptedException thread problem.
	 * @throws IOException bad input file format.
	 * @throws ArianeRuntimeException problems with command execution.
	 */
	protected final void displayImage( ) throws InterruptedException, IOException {
		if (getCommandVisu().trim().isEmpty()) {
			throw new ArianeRuntimeException(Ariane.getString("Operator.error5"));
		}

		if (_currentFilename == null) {
			_currentFilename = getBinary();
		}
		if (_currentFilename == null) {
			return;
		}
		File file = new File(_currentFilename);
		if (getExecutionStatus() == Status.EXECUTED || file.exists()) {
			LinkedList<String> commandLine = new LinkedList<String>();
			commandLine.add(getViewerCommand());
			commandLine.add(_currentFilename);
			OperatorProcessBuilder processBuilder = new OperatorProcessBuilder(commandLine);
			Process proc1 = processBuilder.start();

			// Check execution consistency
			InputStream inputStream = proc1.getInputStream();
			int exitValue = proc1.waitFor();

			String resultString = "";

			if (exitValue != 0 || resultString.startsWith("FAILURE")) {
				// FAILURE with the operator
				byte[] errorByte = new byte[inputStream.available()];
				inputStream.read(errorByte);
				String errorString = new String(errorByte);
				// Problem: cannot get the system error message
				// such as Segmentation Fault, Arithmetic Exception
				if (errorString.length() == 0) {
					errorString = Ariane.getString("Operator.error2");
				}
				throw new ArianeRuntimeException(errorString);
			} else {
				// OK
				byte[] output = new byte[inputStream.available()];
				inputStream.read(output);
				String outputString = new String(output);
				if (outputString.length() > 0) {
					Ariane.getWorkspace().printOutputMessage(outputString);
				}
			}
		}
	}

	///// 
	///// P R I N T I N G 
	/////

	/**
	 * Overrides the default method in order to print components
	 * of the imagefile in different way then they are displayed.
	 * 
	 * @param gc the graphic context.
	 */
	@Override
	public final void printComponent( Graphics gc ) {
		super.printComponent(gc);
		int row;

		gc.setColor(new Color(0, 0, 0));
		gc.setFont(PreferenceChooser.getDefaultFont());

		FontMetrics fontMetrics = gc.getFontMetrics();
		row = fontMetrics.getHeight();

		super.centerText(Ariane.getString(_windowTitle), gc, row, false);
	}
}