/* -*- c-basic-offset: 3 -*-
 *
 * Copyright (c) 2013, GREYC.
 * All rights reserved
 *
 * You may use this file under the terms of the BSD license as follows:
 *
 * "Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of the GREYC, nor the name of its
 *     contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
 * 
 *   
 * For more information, refer to:
 * https://clouard.users.greyc.fr/Pandore
 */

package fr.greyc.pantheon.ariane.graph.item;

import fr.greyc.pantheon.ariane.Ariane;
import fr.greyc.pantheon.ariane.graph.Graph;
import fr.greyc.pantheon.ariane.graph.exception.ArianeArithmeticException;
import fr.greyc.pantheon.ariane.graph.exception.ArianeException;
import fr.greyc.pantheon.ariane.graph.exception.ArianeMissingFormulaException;
import fr.greyc.pantheon.ariane.graph.exception.ArianeMissingParameterException;
import fr.greyc.pantheon.ariane.graph.item.etc.StringExpression;
import fr.greyc.pantheon.ariane.graph.slot.InputParameterSlot;
import fr.greyc.pantheon.ariane.utils.TreeItem;
import fr.greyc.pantheon.ariane.xml.XMLElement;

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyEvent;
import java.io.IOException;
import java.util.List;

import javax.swing.JMenuItem;
import javax.swing.JPopupMenu;
import javax.swing.KeyStroke;

/**
 * A <code>FormatItem</code> implements a calculator item.
 * 
 * @author Sébastien LEFÈVRE, Richard LOBRÉAU - Feb 21, 1999
 * @version Adrien SCHRÈQUE, Hicham AMADI - Nov 05, 2000
 * @version Mathieu JOETS - Jul 09, 2001
 * @version Régis CLOUARD - Oct 16, 2001
 * @version Régis CLOUARD - Nov 19, 2004
 * @version Frédéric LAIR - Jun 28, 2007 (add more than two parameters)
 * @version Régis CLOUARD- Mar 30, 2011 (refine)
 */
public final class FormatItem extends ControlItem {
	/** The tag name of the item. */
	public static final String TAGNAME = "Format";

	/** The arithmetic expression. */
	private StringExpression _expression;

	/**
	 * Creates a new <code>FormatItem</code> on the specified
	 * graph with the given attribute values.
	 * 
	 * @param treeItem the source tree item.
	 * @param graph the graph that holds the item.
	 * @param name the default name of the item.
	 */
	public FormatItem( TreeItem treeItem, Graph graph, String name ) throws InterruptedException {
		super(treeItem, graph, name);
		askProperties(null, 0, 2, 0);
		draw();
	}

	/**
	 * Creates a <code>FormatItem</code> on the specified
	 * graph with the given XML data.
	 * 
	 * @param graph the parent graph.
	 * @param elements the XML tree of data.
	 * @throws IOException XML errors.
	 */
	public FormatItem( Graph graph, List<XMLElement> elements ) throws IOException {
		super(graph);
		load(elements);
		// Compile formula if it exists.
		if (getFormula() != null) {
			_expression = new StringExpression(getFormula(), getInputParameterSlots().size()); 
		}
		setComment(getFormula());
		draw();
	}

	/**
	 * Builds the prototype of the control structure from the given values.
	 * 
	 * @param name the name of the item.
	 * @param iIm the number of input images.
	 * @param iP the number of parameters.
	 * @param oIm the number of output images.
	 */
	@Override
	protected void setPrototype( String name, int iIm, int iP, int oIm ) {
		setAbstractName(name);
		setFormula(""); // IMPERATIVE to signify that it uses formula.
		createSlots(0, iP, 0, 0, 0);
	}

	///// 
	///// A C C E S S O R S
	/////

	/**
	 * @return the tag name.
	 */
	@Override
	public String getTagName( ) {
		return TAGNAME;
	}

	///// 
	///// E X E C U T I O N
	/////

	/**
	 * Executes the <code>FormatItem</code>, means
	 * compute a result from the arithmetic formula.
	 * 
	 * @param isFailureFatalError is FAILURE a fatal error?
	 * @return true if the operator has been executed without errors.
	 */
	@Override
	protected boolean primitiveExecute( boolean isFailureFatalError ) {
		try {
			if (getFormula().isEmpty()) {
				throw new ArianeMissingFormulaException(Ariane.getString("arithmetic.error0"));
			}
			int numberOfParameters = getInputParameterSlots().size();
			String[] val = new String[numberOfParameters];

			for (int i = 0; i < numberOfParameters; i++) {
				InputParameterSlot parameter = getInputParameterSlots().get(i);
				if (parameter.hasValue(isFailureFatalError)) {
					val[i] = parameter.getValue();
				} else {
					throw new ArianeMissingParameterException(Ariane.getString("Operator.error1") + parameter.getRank());
				}
			}

			final String result = _expression.evaluate(val);
			getOutputResultSlot().setValue(result);
			Ariane.getWorkspace().printSystemMessage("format(" + getFormula() + ")=" + result);
			changeStatus(Status.EXECUTED);
			return true;
		} catch (ArianeException e) {
			if (isFailureFatalError && !Graph.isExecutionStopped()) {
				Ariane.getWorkspace().printErrorMessage(e.getErrorCode(), getAbsoluteName(), getId(), e.getLocalizedMessage());
				getGraph().stopExecution();
				changeStatus(Status.ERROR);
			}
		}
		return false;
	}

	///// 
	///// P R O P E R T I E S
	/////

	/**
	 * Builds the contextual menu as a Popup Menu.
	 * 
	 * @return a convenient popup menu.
	 */
	@Override
	protected JPopupMenu initPopup( ) {
		JPopupMenu popupMenu = new JPopupMenu();

		JMenuItem item;
		item = new JMenuItem(Ariane.getString("Graph.Menu.copy"));
		item.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed( ActionEvent event ) {
				Ariane.getWorkspace().copy();
			}
		});
		item.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_C, KeyEvent.CTRL_MASK));
		popupMenu.add(item);

		item = new JMenuItem(Ariane.getString("Graph.Menu.cut"));
		item.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed( ActionEvent event ) {
				Ariane.getWorkspace().cut();
			}
		});
		item.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_X, KeyEvent.CTRL_MASK));
		popupMenu.add(item);

		popupMenu.addSeparator();

		// Action: executer
		item = new JMenuItem(Ariane.getString("PopupMenu.execute"));
		item.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed( ActionEvent event ) {
				getGraph().execute();
			}
		});
		item.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_R, KeyEvent.CTRL_MASK));
		popupMenu.add(item);

		item = new JMenuItem(Ariane.getString("PopupMenu.reset"));
		item.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed( ActionEvent event ) {
				resetAll();
			}
		});
		item.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_U, KeyEvent.CTRL_MASK));
		popupMenu.add(item);

		// Edition: delete
		item = new JMenuItem(Ariane.getString("PopupMenu.remove"));
		item.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed( ActionEvent event ) {
				delete();
			}
		});
		item.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_DELETE, 0));
		popupMenu.add(item);

		return popupMenu;
	}

	/**
	 * Changes the arithmetic formula.
	 * 
	 * @param textFormula a formula.
	 */
	@Override
	public boolean changeFormula( String textFormula ) {
		if (getFormula() == null || !getFormula().equals(textFormula)) {
			try {
				_expression = new StringExpression(textFormula, getInputParameterSlots().size());
				setFormula(textFormula);
				setComment(textFormula);
			} catch (ArianeArithmeticException e) {
				Ariane.getWorkspace().printErrorMessage(e.getLocalizedMessage());
				return false;
			}
		}
		return true;
	}
}