/* -*- c-basic-offset: 3 -*-
 *
 * Copyright (c) 2013, GREYC.
 * All rights reserved
 *
 * You may use this file under the terms of the BSD license as follows:
 *
 * "Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of the GREYC, nor the name of its
 *     contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
 * 
 *   
 * For more information, refer to:
 * https://clouard.users.greyc.fr/Pandore
 */

package fr.greyc.pantheon.ariane.graph.item;

import fr.greyc.pantheon.ariane.Ariane;
import fr.greyc.pantheon.ariane.graph.Graph;
import fr.greyc.pantheon.ariane.graph.SubGraph;
import fr.greyc.pantheon.ariane.graph.command.ChangeBinaryEdit;
import fr.greyc.pantheon.ariane.graph.command.ChangeFormulaEdit;
import fr.greyc.pantheon.ariane.graph.command.RenameItemEdit;
import fr.greyc.pantheon.ariane.graph.item.etc.ItemSignatureParser;
import fr.greyc.pantheon.ariane.graph.item.etc.PropertyViewer;
import fr.greyc.pantheon.ariane.graph.slot.InputSlot;
import fr.greyc.pantheon.ariane.graph.slot.InputDataSlot;
import fr.greyc.pantheon.ariane.graph.slot.InputMaskSlot;
import fr.greyc.pantheon.ariane.graph.slot.InputParameterSlot;
import fr.greyc.pantheon.ariane.graph.slot.OutputSlot;
import fr.greyc.pantheon.ariane.graph.slot.OutputDataSlot;
import fr.greyc.pantheon.ariane.graph.slot.OutputResultSlot;
import fr.greyc.pantheon.ariane.graph.slot.Slot;
import fr.greyc.pantheon.ariane.ui.PreferenceChooser;
import fr.greyc.pantheon.ariane.ui.Workspace;
import fr.greyc.pantheon.ariane.utils.RoundedBorder;
import fr.greyc.pantheon.ariane.utils.RoundedPanel;
import fr.greyc.pantheon.ariane.utils.TreeItem;
import fr.greyc.pantheon.ariane.xml.XMLElement;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.FontMetrics;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Insets;
import java.awt.Point;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.FocusEvent;
import java.awt.event.FocusListener;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.io.File;
import java.io.IOException;
import java.lang.reflect.MalformedParameterizedTypeException;
import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;

import javax.swing.BorderFactory;
import javax.swing.ImageIcon;
import javax.swing.JButton;
import javax.swing.JComponent;
import javax.swing.JLabel;
import javax.swing.JPopupMenu;
import javax.swing.JTextArea;
import javax.swing.JTextField;
import javax.swing.border.Border;
import javax.swing.border.EtchedBorder;

/**
 * A <code>GraphItem</code> defines an item that can be displayed in a graph.
 * It is graphically represented by a box with input and output slots.
 * @see OperatorItem
 * @see MacroOperatorItem
 * @see ControlStructure
 * 
 * @author Sébastien LEFÈVRE, Richard LOBRÉAU - Feb 21, 1999
 * @version Adrien SCHRÈQUE, Hicham AMADI - Nov 05, 2000
 * @version Mathieu JOETS - Jul 09, 2001
 * @version Alban DOUBLET, Guillaume CHATELET - Jun 10, 2002
 * @version Régis CLOUARD - Oct 16, 2001
 * @version Régis CLOUARD - Dec 06, 2004
 * @version Frédéric LAIR - Jun 11, 2007
 * @version Régis CLOUARD - Jun 10, 2010 (add nickname)
 */
public abstract class GraphItem extends RoundedPanel implements MouseListener, MouseMotionListener, FocusListener, KeyListener {
	/** The tree item that contains the properties of this kind of graph item. */
	protected TreeItem _treeItem;

	/** The related graph that supports this item. */
	private Graph _graph;
	/** The identifier number; supposed to be unique. */
	private int _identifier;
	/** The identifier number read during a reading or a pasting. */
	private int _readIdentifier = -1;
	/** The current read or paste session number. */
	private int _currentReadPasteSession;

	/** The icon for unexecuted status. */
	private static ImageIcon _unexecutedStatusIcon;
	/** The icon for running status. */
	private static ImageIcon _runningStatusIcon;
	/** The icon for executed status. */
	private static ImageIcon _executedStatusIcon;
	/** The icon for error status. */
	private static ImageIcon _errorStatusIcon;

	//	private static final Color SELECTED_SHADOW_COLOR = new Color(255, 150, 150);
	/** The color for selected shadow. */
	private static final Color SELECTED_HIGHLIGHT_COLOR = new Color(255, 0, 00);

	private static final Border UNSELECTED_BORDER = new RoundedBorder(EtchedBorder.RAISED);
	private static final Border SELECTED_BORDER = new RoundedBorder(EtchedBorder.LOWERED, SELECTED_HIGHLIGHT_COLOR);

	/** The name of the item. */
	private String _name;

	/** The nickname of the item. */
	private String _nickName;

	/** The label that displays the name in the property viewer. */
	private JTextField _textFieldName;

	/** The label that displays the name on the item representation. */
	private JLabel _labelName;

	/** An optional comment that can be displayed inside the graph item. */
	private JComponent _commentPanel;

	/** The name of the executable command. Ex: "pbmp2pan". */
	private String _binary;

	/** The name of executable category. */
	private String _category;

	/** The arithmetic or logical expression: ex: $0<=$1. */
	private String _formula;

	private final List<InputDataSlot> _inputDataSlots = new ArrayList<InputDataSlot>();
	private final List<OutputDataSlot> _outputDataSlots = new ArrayList<OutputDataSlot>();
	private final List<InputParameterSlot> _inputParameterSlots = new ArrayList<InputParameterSlot>();
	private final List<OutputResultSlot> _outputResultSlots = new ArrayList<OutputResultSlot>();
	private InputMaskSlot _inputMaskSlot;

	/** The label that visualizes the status of the item. */
	private JLabel _labelStatus;

	/** The popup menu.*/
	private static JPopupMenu _popupMenu;

	/** The command for getting execution result. Ex: "pstatus". */
	private static String _commandStatus = "";

	/** The command for displaying output images. Ex: "pvisu". */
	private static String _commandVisu = "";

	/** The window of properties */
	private PropertyViewer _properties;

	/** The current status of the item. */
	private Status _executionStatus = Status.UNEXECUTED;

	/** The status of the item between selected or not. */
	private boolean _selected;

	/**
	 * The various status of a graph item
	 */
	public enum Status {
		/** Unexecuted state. */
		UNEXECUTED,
		/** Running state. */
		RUNNING,
		/** Executed state. */
		EXECUTED,
		/** Error state.*/
		ERROR
	}

	static {
		_unexecutedStatusIcon = Ariane.getIcon("StatusIcon0.png");
		_runningStatusIcon = Ariane.getIcon("StatusIcon1.png");
		_executedStatusIcon = Ariane.getIcon("StatusIcon2.png");
		_errorStatusIcon = Ariane.getIcon("StatusIcon3.png");
	}

	/**
	 * Creates a new graph item from a file.
	 * @param graph the graph that will support this item.
	 */
	public GraphItem( Graph graph ) {
		_graph = graph;
		setId();
		initialize();
		_commentPanel = getCommentEditor();
	}

	/**
	 * Creates a new graph item from the explorer.
	 * If the identifier = -1 then creates
	 * a new unique identifier number.
	 * @param treeItem the reference tree item.
	 * @param graph the graph that will support this item.
	 * @param name the name of the item.
	 * @param category the category of the item
	 */
	public GraphItem( TreeItem treeItem, Graph graph, String name, String category ) {
		_graph = graph;
		_treeItem = treeItem;
		_name = name;
		_category = category;
		setId();
		initialize();
		_commentPanel = getCommentEditor();
	}

	/**
	 * Initializes the graph item.
	 */
	private void initialize( ) {
		_labelName = new JLabel();
		_labelName.setFont(PreferenceChooser.getDefaultFont());
		_textFieldName = new JTextField();
		_textFieldName.addFocusListener(new FocusListener() {
			@Override
			public final void focusLost( FocusEvent event ) {
				setNickName(((JTextField)(event.getSource())).getText(), true);
			}
			@Override
			public final void focusGained( FocusEvent event ) {
				/* empty */
			}
		});
		_textFieldName.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed( ActionEvent event ) {
				setNickName(((JTextField)(event.getSource())).getText(), true);
			}
		});

		_labelStatus = new JLabel();
		setLayout(new GridBagLayout());
		addMouseListener(this);
		addMouseMotionListener(this);
		addFocusListener(this);
		addKeyListener(this);
	}

	protected void removeHook() {
		// empty
	}

	/**
	 * Removes the item from the graph, but it is still in the history.
	 * Disconnects it from all other items.
	 */
	public final void remove() {
		disconnectAllSlots();
		if (_properties != null) {
			_properties.setVisible(false);
		}
		removeHook();
	}

	protected void disposeHook() {
		// Empty
	}

	/**
	 * Disposes definitively the item from the graph.
	 * Tries to avoid memory leak.
	 */
	public final void dispose() {
		disposeHook();
		disconnectAllSlots();
		if (_properties != null) {
			_properties.setVisible(false);
		}
		for (InputDataSlot inputDataSlot : _inputDataSlots) {
			inputDataSlot .dispose();
		}
		_inputDataSlots.clear();
		for (InputParameterSlot inputParameterSlot: _inputParameterSlots) {
			inputParameterSlot.dispose();
		}
		_inputParameterSlots.clear();

		if (_inputMaskSlot != null) {
			_inputMaskSlot.dispose();
		}
		_inputMaskSlot = null;

		for (OutputDataSlot outputDataSlot : _outputDataSlots) {
			outputDataSlot.dispose();
		}
		_outputDataSlots.clear();

		for (OutputResultSlot outputResultSlot : _outputResultSlots) {
			outputResultSlot.dispose();
		}
		_outputResultSlots.clear();
		_graph = null;
	}

	/**
	 * Deletes the items from the graph.
	 * Pushes it into the undo list.
	 */
	protected final void delete( ) {
		_graph.deleteItem(this);
	}

	/**
	 * Sets the default value the parameters.
	 * @param signature the formatted string with the values (e.g. $0:toto=0; $1:titi=10; ..) 
	 */
	protected final void setSignature( String signature ) {
		if (signature != null && !signature.trim().isEmpty()) {
			// Set the default value for the parameters
			try {
				String[][] values = ItemSignatureParser.parse(signature, _inputParameterSlots.size());
				for ( int i = 0; i < _inputParameterSlots.size(); i++ ) {
					_inputParameterSlots.get(i).setName(values[i][0]);
					_inputParameterSlots.get(i).setValue(values[i][1]);
				}
			} catch ( MalformedParameterizedTypeException e) {
				Ariane.getWorkspace().printErrorMessage(getName() + Ariane.getString("Operator.error6"));
			}
		}
	}

	///// 
	///// E X E C U T I O N
	/////

	/**
	 * Executes the item.
	 */
	public final void execute( ) {
		execute(true);
	}

	/**
	 * Executes the item.
	 * @param imperative is FAILURE a fatal error?
	 * @return true if the item has been executed successfully.
	 */
	public final boolean execute( boolean imperative ) {
		if (Graph.isExecutionStopped()) {
			return false;
		}
		if (getExecutionStatus() == Status.EXECUTED) {
			return true;
		}
		if (primitiveExecute(imperative)) {
			updatePropertyViewer();
			return true;
		}
		return false;
	}

	/**
	 * This is the internal execution function.
	 * @param isFailureFatalError is FAILURE a fatal error?
	 * @return true if the item has been executed successfully.
	 */
	protected abstract boolean primitiveExecute( boolean isFailureFatalError );

	/**
	 * Returns the related subgraph of the item.
	 * @return a graph or null.
	 */
	public SubGraph getSubGraph() {
		return null;
	}

	public final Status getExecutionStatus( ) {
		return _executionStatus;
	}

	/**
	 * Changes the current status of the item from
	 * the specified value. Changes also its parent's status.
	 * @param value the new status (UNEXECUTED, EXECUTED, ERROR).
	 */
	protected final void changeStatus( Status value ) {
		if (!_executionStatus.equals(value)) {
			_executionStatus = value;
			if (_executionStatus == Status.ERROR && _graph.getSuperGraphItem() != null) {
				_graph.getSuperGraphItem().changeStatus(value);
			}
			displayStatus();
		}
	}

	/**
	 * A hook for customizing the comment editor.
	 * @return a text component.
	 */
	protected JComponent getCommentEditor() {
		JLabel label = new JLabel();
		label.setMaximumSize(new Dimension(150, 10000));

		int size = PreferenceChooser.getDefaultFont().getSize();
		label.setFont(PreferenceChooser.getDefaultFont().deriveFont(size - 1.0f));
		label.setBorder(BorderFactory.createEmptyBorder(2, 2, 2, 2));
		return label;
	}

	protected final void setComment( String comment ) {
		if (_commentPanel instanceof JLabel) {
			((JLabel)_commentPanel).setText(comment);
		} else if (_commentPanel instanceof JTextArea) {
			((JTextArea)_commentPanel).setText(comment);
		}
	}

	/**
	 * Displays the current status of the item
	 * in the related item widget.
	 */
	private void displayStatus( ) {
		switch(_executionStatus) {
		case UNEXECUTED:
			_labelStatus.setIcon(_unexecutedStatusIcon);
			break;
		case RUNNING:
			_labelStatus.setIcon(_runningStatusIcon);
			break;
		case EXECUTED:
			_labelStatus.setIcon(_executedStatusIcon);
			break;
		case ERROR:
			_labelStatus.setIcon(_errorStatusIcon);
			break;
		default:
			break;
		}
	}

	/**
	 * Resets the item and all its forwarded connected items
	 * and its parent.
	 */
	public void resetAll( ) {
		if (_executionStatus != Status.UNEXECUTED) {
			reset();
			if (_outputDataSlots != null) {
				for (OutputDataSlot slot: _outputDataSlots) {
					slot.reset();
					// Reset all connected items.
					if (slot.isConnected()) {
						for (InputSlot inputSlot: slot.getConnectedSlots()) {
							inputSlot.getGraphItem().resetAll();
						}
					}
				}
			}

			if (!_outputResultSlots.isEmpty()) {
				getOutputResultSlot().reset();
				if (getOutputResultSlot().isConnected()) {
					for (InputSlot inputSlot : getOutputResultSlot().getConnectedSlots()) {
						inputSlot.getGraphItem().resetAll();
					}
				}
			}
			// Its parent
			if (_graph.getSuperGraphItem() != null) {
				_graph.getSuperGraphItem().resetAll();
			}
			updatePropertyViewer();
		}
	}

	/**
	 * Resets the item without all its forwarded connected items
	 * and without its parent.
	 */
	protected void reset() {
		changeStatus(Status.UNEXECUTED);
	}

	/**
	 * Disconnects all input and output slots.
	 */
	public final void disconnectAllSlots( ) {
		for (OutputSlot slot : _outputDataSlots) {
			slot.disconnectAll();
		}
		if (_inputParameterSlots != null) {
			for (InputSlot slot : _inputParameterSlots) {
				slot.disconnect();
			}
		}
		for (InputSlot slot : _inputDataSlots) {
			slot.disconnect();
		}
		if (_inputMaskSlot != null) {
			_inputMaskSlot.disconnect();
		}
		if (!_outputResultSlots.isEmpty()) {
			getOutputResultSlot().disconnectAll();
		}
	}

	// A hook function.
	public void fireConnection( InputSlot slot, boolean connection ) { // A regler
		if (_properties != null) {
			_properties.validate();
			_properties.repaint();
		}
	}

	///// 
	///// A C C E S S O R S
	/////

	/**
	 * Returns the identifier number of the item.
	 * 
	 * @return an integer.
	 */
	public int getId( ) {
		return _identifier;
	}

	/**
	 * Sets the identifier as it was stored in the file.
	 */
	private void setId( ) {
		// Set a unique identifier number!
		_identifier = _graph.getNextItemIdentifier();
		// Keeps the current read session.
		_currentReadPasteSession = Workspace.getCurrentReadPasteSession();
	}

	/**
	 * Sets the identifier of the item. This identifier is
	 * guaranty to be unique in a same graph.
	 * If the parameter identifier (i.e., the previous identifier
	 * in a previous session) is > -1, then the identifier is kept
	 * in _readIdentifier.
	 * 
	 * @param identifier the previous identifier number.
	 */
	protected final void setInternalId( int identifier ) {
		_readIdentifier = identifier;
	}

	/**
	 * Returns the current identifier of the read session
	 * (use for loading and pasting items).
	 * @return the number of the current loading session.
	 */
	public final int getCurrentReadPasteSession( ) {
		return _currentReadPasteSession;
	}

	/**
	 * Returns the identifier number of the item
	 * in a previous session (saved or pasted).
	 * @return the identifier number.
	 */
	public final int getReadIdentifier( ) {
		return _readIdentifier;
	}

	/**
	 * Returns the category of the item.
	 * @return the name of the category.
	 */
	public final String getCategory( ) {
		if (_treeItem == null) {
			return _category;
		}
		return _treeItem.getCategory();
	}

	/**
	 * Returns the address of the related documentation HTML page.
	 * @return the complete or relative address.
	 */
	public final String getDocURL( ) {
		if (_treeItem == null) {
			return "";
		}
		return _treeItem.getDocURL();
	}

	/**
	 * Returns the absolute path of the executable if possible.
	 * Possible cases are:
	 * <ul>
	 * <li>If the binary already contains the absolute path,
	 * <li>if the binary is in the associated library.
	 * </ul>
	 * In all other cases, the binary is supposed to be accessible by
	 * the environment variable PATH, and the function returns the binary itself.
	 * 
	 * @param binary the abstract name of the executable.
	 * @return the absolute path to a known executable or the binary itself.
	 */
	public static final String getAbsoluteBinary( String binary ) {
		// !!!
		// !!! Note: even if I modify the environment variable PATH, the processBuilder cannot find
		// !!! the required Shell command. It seems that it needs the absolute path !
		// !!!
		if (binary == null) { return null; }
		// User operator
		File unixFile;
		File windowsFile;
		String binaryDirectory = PreferenceChooser.get(PreferenceChooser.BINARY_DIRECTORY);

		if (binaryDirectory != null) {
			windowsFile = new File(binaryDirectory + File.separator + File.separator + binary + ".exe");
			if (windowsFile.exists() && windowsFile.isFile()) {
				// An operator of the library
				return windowsFile.getAbsolutePath();
			}
			unixFile = new File(binaryDirectory + File.separator + File.separator + binary);
			if (unixFile.exists() && unixFile.isFile()) {
				// An operator of the library
				return unixFile.getAbsolutePath();
			}
		}

		// Already an absolute path
		if (binary.indexOf(File.separator) >= 0 ) {
			return binary;
		}

		// Unix case
		windowsFile = new File(Ariane.getLibraryPath() + File.separator + "bin" + File.separator + binary + ".exe");
		if (windowsFile.exists()) {
			// An operator of the library
			return windowsFile.getAbsolutePath();
		}
		unixFile = new File(Ariane.getLibraryPath() + File.separator + "bin" + File.separator + binary);
		if (unixFile.exists()) {
			// An operator of the library
			return unixFile.getAbsolutePath();
		}
		// Default case: return the binary itself
		return binary;
	}

	/**
	 * Returns the name of the executable.
	 * 
	 * @return the name.
	 */
	public String getBinary( ) {
		return _binary;
	}

	/**
	 * Returns the name of the executable using a relative
	 * path name if filepath is specified.
	 * 
	 * @param filePath the file path or null.
	 * @return the name.
	 */
	public String getBinary( String filePath ) {
		return _binary;
	}

	/**
	 * Sets the path to the executable.
	 * final 
	 * @param binary the complete path.
	 */
	protected void setBinary( String binary ) {
		_binary = binary;
	}

	/**
	 * Returns the name of the formula using a relative
	 * path name if it is a file and the filepath is specified.
	 * @param filePath the file path or null.
	 * @return the name.
	 */
	public String getFormula( String filepath ) {
		return _formula;
	}

	public boolean displayFormula() {
		return true;
	}

	/**
	 * Returns the formula of the logical or arithmetic expression.
	 * 
	 * @return the formula.
	 */
	public final String getFormula( ) {
		return _formula;
	}

	/**
	 * Returns the formula of the logical or arithmetic expression.
	 * 
	 * @param textFormula the formula.
	 * @return true if the formula is valid, false otherwise.
	 */
	public boolean changeFormula( String textFormula ) {
		return true;
	}

	/**
	 * Sets the new formula.
	 * @param formula the formula.
	 */
	public void setFormula( String formula ) {
		_formula = formula;
		resetAll();
		getGraph().setModified(true);
		_labelName.setText(getNickName());
		getGraph().repaint();
	}

	/**
	 * Returns the name of the item.
	 * @return the name.
	 */
	@Override
	public final String getName( ) {
		return _name;
	}

	/**
	 * Returns the nickname of the item if any
	 * otherwise returns the name.
	 * @return the nickname.
	 */
	public String getNickName( ) {
		return (_nickName == null) ? getName() : _nickName;
	}

	/**
	 * Returns the absolute name of the item, including
	 * the graph and subgraphs names.
	 * @return the absolute name.
	 */
	protected final String getAbsoluteName( ) {
		return getGraph().getName() + "/" + getNickName();
	}

	/**
	 * Sets the item name.
	 * @param name the new name.
	 */
	public final void setAbstractName( String name ) {
		if (name.length() > 0) {
			_name = name;
		}
	}

	public final void setTemporaryNickName( String name ) {
		_nickName = name;
		updateName();
	}

	/**
	 * Sets the item nickname.
	 * 
	 * @param name the new nickname.
	 */
	public void setNickName( String name ) {
		if (name != null && name.length() > 0) {
			_nickName = name;
			_textFieldName.setText(name);
		} else {
			_nickName = null;
			_textFieldName.setText("");
		}
		updatePropertyViewer();
		getGraph().setModified(true);
	}

	/**
	 * Updates the related property viewer if any.
	 */
	public final void updatePropertyViewer() {
		if (_properties != null) {
			_properties.update(this);
		}
	}

	protected final void removePropertyViewer() {
		if (_properties != null) {
			_properties.setVisible(false);
		}
		_properties = null;
	}

	/**
	 * Sets the new nickname to the item.
	 * 
	 * @param newName the new nickname.
	 * @param historical true if the action must be stored in the action
	 *        history (for use with undo/redo).
	 */
	public final void setNickName( String newName, boolean historical ) {
		String oldName = _nickName;
		setNickName(newName);
		if (historical) {
			getGraph().startActionEdition();
			getGraph().validateAction(new RenameItemEdit(this, newName, oldName));
			getGraph().endActionEdition();
		}
		updateName();
		revalidate();
		getGraph().repaint();
	}

	/**
	 * Returns the graph that holds this item.
	 * 
	 * @return the Graph.
	 */
	public final Graph getGraph( ) {
		return _graph;
	}

	/**
	 * @return the related tag name.
	 */
	public abstract String getTagName( );

	protected final void createSlots( int inputDataSlotNumber, int inputParameterSlotNumber, int inputMaskSlotNumber, int outputDataSlotNumber, int outputResultSlotNumber ) {
		for (int i = 0; i < inputDataSlotNumber; i++) {
			_inputDataSlots.add(new InputDataSlot(this, i));
		}
		if (inputMaskSlotNumber > 0) {
			_inputMaskSlot = new InputMaskSlot(this);
		}
		for (int i = 0; i < inputParameterSlotNumber; i++) {
			_inputParameterSlots.add(new InputParameterSlot(this, i));
		}
		for (int i = 0; i < outputDataSlotNumber; i++) {
			_outputDataSlots.add(new OutputDataSlot(this, i));
		}
		for (int i = 0; i < outputResultSlotNumber; i++) {
			_outputResultSlots.add(new OutputResultSlot(this, i));
		}
	}

	public final List<InputDataSlot> getInputDataSlots( ) {
		return _inputDataSlots;
	}

	public final List<OutputDataSlot> getOutputDataSlots( ) {
		return _outputDataSlots;
	}

	public final InputMaskSlot getInputMaskSlot( ) {
		return _inputMaskSlot;
	}

	public final List<InputParameterSlot> getInputParameterSlots( ) {
		return _inputParameterSlots;
	}

	public final List<OutputResultSlot> getOutputResultSlots( ) {
		return _outputResultSlots;
	}

	public final OutputResultSlot getOutputResultSlot( ) {
		if (_outputResultSlots.isEmpty()) {
			return null;
		} else {
			return _outputResultSlots.get(0);
		}
	}

	/**
	 * Returns the slot with the specified type
	 * and with the specified rank <i>rank</i>
	 * .
	 * @param typeSlot the name of the type of the slot.
	 * @param rank the rank of the slot in the item.
	 * @return a slot or null.
	 */
	public final Slot getSlot( String typeSlot, int rank ) {
		if (typeSlot.equalsIgnoreCase("InSlotImage")) {
			return getInputDataSlots().get(rank);
		} else if (typeSlot.equalsIgnoreCase("InSlotMask")) {
			return getInputMaskSlot();
		} else if (typeSlot.equalsIgnoreCase("InSlotParameter")) {
			return getInputParameterSlots().get(rank);
		} else if (typeSlot.equalsIgnoreCase("OutSlotImage")) {
			if (rank >= getOutputDataSlots().size()) { return null; }
			return getOutputDataSlots().get(rank);
		} else if (typeSlot.equalsIgnoreCase("OutSlotResult")) {
			// For sake of ascendent compatibility
			if (rank < getOutputResultSlots().size()) {
				return getOutputResultSlots().get(rank);
			} else {
				return null;
			}
		}
		return null;
	}

	///// 
	///// P R E F E R E N C E S
	///// 

	/**
	 * @return the name of status binary command.
	 */
	public static final String getCommandStatus( ) {
		return _commandStatus;
	}

	/**
	 * The name of status binary command.
	 * 
	 * @param shellCommand the command name.
	 */
	public static final void setCommandStatus( String shellCommand ) {
		_commandStatus = shellCommand;
	}

	/**
	 * @return the name of the visualization binary software.
	 */
	public static final String getCommandVisu( ) {
		return _commandVisu;
	}

	/**
	 * Sets the name of the visualization binary software.
	 * 
	 * @param shellCommand the command name.
	 */
	public static final void setCommandVisu( String shellCommand ) {
		_commandVisu = shellCommand;
	}

	///// 
	///// D I S P L A Y I N G  &  P R I N T I N G
	///// 

	/**
	 * Displays the item content.
	 */
	public void draw( ) {
		removeAll();
		setBackground(getBackgroundColor());
		setBorder(UNSELECTED_BORDER);

		_labelName.setText(getNickName());
		setToolTipText(toString());

		// layout
		GridBagConstraints constraints = new GridBagConstraints();
		constraints.gridheight = 1;

		// 1. parameters slots
		constraints.anchor = GridBagConstraints.NORTHWEST;
		constraints.gridy = 0;
		constraints.insets = new Insets(1, 5, 1, 0);

		final Insets insets = new Insets(1, 0, 1, 0);
		for (int i = 0; i < _inputParameterSlots.size(); i++) {
			Slot slot = _inputParameterSlots.get(i);
			constraints.gridx = i;
			add(slot, constraints);
			constraints.insets = insets;
		}

		// 2. The mask slot
		int numberOfParamaters = _inputParameterSlots.size();
		if (numberOfParamaters == 0) {
			numberOfParamaters = 1;
		}

		constraints.insets = new Insets(0, 0, 0, 5);
		constraints.anchor = GridBagConstraints.NORTHEAST;
		constraints.gridx = numberOfParamaters;
		if (_inputMaskSlot != null) {
			add(_inputMaskSlot, constraints);
		}
		constraints.insets = new Insets(0, 0, 0, 0);

		// 3. The name
		constraints.gridwidth = numberOfParamaters + 1;
		constraints.gridy++;
		constraints.gridx = 0;
		constraints.anchor = GridBagConstraints.CENTER;
		add(_labelName, constraints);

		// 4. The lights ramp that shows the current status.
		constraints.gridwidth = GridBagConstraints.REMAINDER;
		constraints.gridy++;
		constraints.gridx = 0;
		displayStatus();
		add(_labelStatus, constraints);
	
		// 5. Optional Comments
		constraints.gridwidth = GridBagConstraints.REMAINDER;
		constraints.gridy++;
		constraints.gridx = 0;
		if (_commentPanel.getPreferredSize().width > _commentPanel.getMaximumSize().width) {
			Dimension dim = _commentPanel.getPreferredSize();
			dim.width = _commentPanel.getMaximumSize().width;
			_commentPanel.setPreferredSize(dim);
		}
		add(_commentPanel, constraints);

		// 6. The input + output slots
		constraints.insets = new Insets(0, 0, 1, 1);
		constraints.gridy++;
		constraints.gridx = 0;
		int maxSlot = _inputDataSlots.size() > _outputDataSlots.size() ? _inputDataSlots.size() : _outputDataSlots.size();
		for (int i = 0; i < maxSlot; i++) {
			if (i < _inputDataSlots.size()) {
				constraints.gridx = 0;
				constraints.anchor = GridBagConstraints.WEST;
				Slot slot = _inputDataSlots.get(i);
//				slot.updateBackgroundColor();
				add(slot, constraints);
			}
			if (i < _outputDataSlots.size()) {
				constraints.gridx = numberOfParamaters;
				constraints.anchor = GridBagConstraints.EAST;
				Slot slot = _outputDataSlots.get(i);
				add(slot, constraints);
			}
			constraints.gridy++;
		}

		// 8. The result
		int max = _inputDataSlots.size() > _outputDataSlots.size() ? _inputDataSlots.size() : _outputDataSlots.size();
		constraints.gridy += max;
		constraints.insets = new Insets(0, 0, 2, 10);
		if (!_outputResultSlots.isEmpty()) {
			constraints.gridx = numberOfParamaters;
			constraints.gridwidth = 1;
			constraints.gridheight = 1;
			constraints.anchor = GridBagConstraints.SOUTHEAST;
			add(getOutputResultSlot(), constraints);
		}
	}

	/**
	 * Centers the specified text in the component.
	 * If the length of the text is larger than the component
	 * width, then use several lines.
	 * 
	 * @param text the text to be centered.
	 * @param row the line where to display the text.
	 * @param multiline can print the text with several lines?
	 * @param gc the graphics context.
	 */
	protected final void centerText( String text, Graphics graphics, int row, boolean multiline ) {
		if (text == null) {
			return;
		}
		int line = row;
		FontMetrics fontMetrics;
		fontMetrics = graphics.getFontMetrics();
		Dimension size = getSize();
		if (fontMetrics.stringWidth(text) > size.width) {
			if (multiline) {
				String subtext = text;
				// Use several lines and do not center text.
				String tmp;
				while (subtext.length() > 0) {
					int index = subtext.length();
					tmp = subtext;
					while (fontMetrics.stringWidth(tmp) > size.width - 4) {
						tmp = subtext.substring(0, --index);
					}
					graphics.drawString(tmp, 2, line);
					line += fontMetrics.getHeight();
					subtext = subtext.substring(index);
				}
			}
			graphics.drawString(text, 2, line);
		} else { // Centers the text
			int column = (size.width - fontMetrics.stringWidth(text)) / 2;
			graphics.drawString(text, column, line);
		}
	}

	/**
	 * Cancels printing of item contents.
	 * @param gc the graphics context.
	 */
	@Override
	public final void printChildren( Graphics gc ) { /* empty */ }

	/**
	 * Prints the border of the item;
	 * 
	 * @param gc the graphics context.
	 */
	@Override
	public final void printBorder( Graphics graphics ) {
		Dimension size;
		size = new Dimension(getSize());
		Graphics2D g2d = (Graphics2D)graphics;
		g2d.setColor(Color.black);
		BasicStroke lineStroke = new BasicStroke(2.0F, BasicStroke.CAP_ROUND, BasicStroke.JOIN_ROUND);
		g2d.setStroke(lineStroke);
		g2d.drawRect(0, 0, size.width, size.height);
	}

	/**
	 * Prints the items. Just draw a line to split the
	 * name of the item and the content.
	 * 
	 * @param gc the graphics context.
	 */
	@Override
	public void printComponent( Graphics gc ) {
		Dimension size;
		int row;
		size = new Dimension(getSize());
		gc.setColor(Color.white);
		gc.fillRect(0, 0, size.width, size.height);

		gc.setColor(Color.black);
		gc.setFont(PreferenceChooser.getDefaultFont());
		FontMetrics fontMetrics;
		fontMetrics = gc.getFontMetrics(); 
		row = fontMetrics.getHeight();
		// For aestheticism purpose, split title and parameters.
		gc.drawLine(0, row + 4, size.width, row + 4);
	}

	/** 
	 * @return the current background color.
	 */
	public abstract Color getBackgroundColor( );

	/**
	 * Shows the related documentation HTML page on a 
	 * separated window.
	 */
	public final void showMan( ) {
		if (getDocURL() != null) {
			Ariane.getWorkspace().displayMan(Ariane.getString("HTMLViewer.manual") + " " + _name, getDocURL(), null);
		}
	}

	///// 
	///// S E L E C T I O N
	///// 

	public final boolean isSelected( ) {
		return _selected;
	}

	public final void setSelected( boolean selected ) {
		if (_selected != selected) {
			for ( InputSlot slot : _inputDataSlots ) {
				slot.setSelected(selected);
			}
			for ( InputSlot slot : _inputParameterSlots ) {
				slot.setSelected(selected);
			}
			for ( OutputSlot slot : _outputDataSlots ) {
				slot.setSelected(selected);
			}
			for ( OutputSlot slot : _outputResultSlots ) {
				slot.setSelected(selected);
			}
			InputMaskSlot slot = _inputMaskSlot;
			if (slot != null) {
				slot.setSelected(selected);
			}
			_selected = selected;
			setBorder(selected ? SELECTED_BORDER : UNSELECTED_BORDER);
		}
	}

	///// 
	///// M O U S E
	///// 

	@Override
	public final void mouseClicked( MouseEvent event ) {
		if (event.getClickCount() > 1) {
			onMouseDoubleClicked(event);
		}
	}

	protected void onMouseDoubleClicked( MouseEvent event ) {
		/* Hook function. */
	}

	@Override
	public final void mousePressed( MouseEvent event ) {
		if (!event.isShiftDown() && !event.isControlDown()) {
			requestFocusInWindow();
		}
		getParent().setComponentZOrder(this, 0); // Put it in front
		if (event.isPopupTrigger()) { // Unix version -> open the popup menu
			getGraph().addItemInSelection(this, false);
			drawPopup(event);
		}
	}

	@Override
	public final void mouseReleased( MouseEvent event ) {
		if (event.isShiftDown() || event.isControlDown()) {
			hidePropertyPanel();
		} else {
			showPropertyPanel();
		}
		onMouseReleased(event);
	}

	protected final void onMouseReleased( MouseEvent event ) {
		if (event.isPopupTrigger()) { // Window version -> open the popup menu
			getGraph().addItemInSelection(this, false);
			drawPopup(event);
		} else {
			if ((_popupMenu == null || !_popupMenu.isVisible())
					&& !getGraph().dropSelection()
					&& (event.isShiftDown() || event.isControlDown())) {
				getGraph().addItemInSelection(this, true);
			}
		}
	}

	@Override
	public final void mouseDragged( MouseEvent event ) {
		if (_popupMenu == null || !_popupMenu.isVisible()) {
			if (!_selected) {
				getGraph().addItemInSelection(this, event.isShiftDown() || event.isControlDown());
			}
			event.translatePoint(getLocation().x, getLocation().y);
			getGraph().dragSelection(event);
		}
	}

	@Override
	public final void mouseMoved( MouseEvent event ) {
	}

	@Override
	public final void mouseEntered( MouseEvent event ) {
	}

	@Override
	public final void mouseExited( MouseEvent event ) {
	}

	@Override
	public void focusLost( FocusEvent event ) {
	}

	protected void focusIn() {
		// Hook function.
	}

	@Override
	public final void focusGained( FocusEvent event ) {
		showPropertyPanel();
		if (isSelected()) {
			focusIn();
		} else {
			GraphItem item = GraphItem.this;
			item.getGraph().addItemInSelection(item, false);
		}
	}

	@Override
	public final void keyPressed( KeyEvent event ) {
		// redirect events to the related graph.
		getGraph().onKeyPressed(event);
	}

	@Override
	public final void keyReleased( KeyEvent event ) {
	}

	@Override
	public final void keyTyped( KeyEvent event ) {
	}

	///// 
	///// P R O P E R T Y
	///// 

	private void showPropertyPanel() {
		Ariane.getWorkspace().displayPropertyPanel(getPropertyViewer());
	}

	private void hidePropertyPanel() {
		Ariane.getWorkspace().displayPropertyPanel(null);
	}

	/**
	 * Creates a related property viewer.
	 * 
	 * @return a dialog box.
	 */
	protected PropertyViewer createPropertyViewer() {
		return new PropertyViewer(Ariane.getLanguage(), this);
	}

	/**
	 * @return the current property window.
	 */
	private PropertyViewer getPropertyViewer() {
		if (_properties == null || _properties.getLanguage() != Ariane.getLanguage()) {
			_properties = createPropertyViewer();
		} 
		_properties.setVisible(true);
		return _properties;
	}

	/**
	 * @return the list of specific properties (other than common properties).
	 */
	public List<PropertyViewer.PropertyCouple> getSpecificProperties() {
		return null;
	}

	/**
	 * Changes the formula expression and store
	 * it in undo -redo manager.
	 * @param component the component that displays the formula.
	 * @param newValue the new formula expression.
	 * @param historical true if the action must be stored in the action
	 *        history (for use with undo/redo).
	 */
	public final void changeFormula( final JComponent component, String newValue, boolean historical ) {
		changeFormula(newValue);

		// Very ugly but YAGNI...
		if (component instanceof JTextField) {
			((JTextField)component).setText(getFormulaText());
		} else if (component instanceof JButton) {
			((JButton)component).setText(getFormulaText());
		}
		updatePropertyViewer();
		if (historical) {
			String oldFormula = getFormula();
			getGraph().startActionEdition();
			getGraph().validateAction(new ChangeFormulaEdit(this, component, getFormula(), oldFormula));
			getGraph().endActionEdition();
		}
	}

	/**
	 * Changes the binary expression and store
	 * it in undo -redo manager.
	 * @param component the component that displays the formula.
	 * @param newValue the new formula expression.
	 * @param historical true if the action must be stored in the action
	 *        history (for use with undo/redo).
	 */
	public final void changeBinary( final JComponent component, String newValue, boolean historical ) {
		setBinary(newValue);
		// Very ugly but YAGNI...
		if (component instanceof JTextField) {
			((JTextField)component).setText(getBinary());
		} else if (component instanceof JButton) {
			((JButton)component).setText(getBinary());
		}
		updatePropertyViewer();
		if (historical) {
			String oldBinary = getBinary();
			getGraph().startActionEdition();
			getGraph().validateAction(new ChangeBinaryEdit(this, component, getBinary(), oldBinary));
			getGraph().endActionEdition();
		}
	}

	/**
	 * @return the user's representation of the formula.
	 */
	public String getFormulaText() {
		return getFormula();
	}

	/**
	 * @return the text associated to the formula editor.
	 */
	public String getFormulaLabel( ) {
		return Ariane.getString("PropertyViewer.formula");
	}

	/**
	 * Builds a component appropriate to display and change the formula.
	 * 
	 * @return a new component.
	 */
	public JComponent getFormulaEditor( ) {
		JTextField formulaText = new JTextField(getFormulaText());
		formulaText.addFocusListener(new FocusListener() {
			@Override
			public final void focusLost( FocusEvent event ) {
				changeFormula((JTextField)event.getSource(), ((JTextField)event.getSource()).getText(), true);
			}
			@Override
			public final void focusGained( FocusEvent event ) {
				/* empty */
			}
		});
		formulaText.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed( ActionEvent event ) {
				changeFormula((JTextField)event.getSource(), ((JTextField)event.getSource()).getText(), true);
			}
		});
		return formulaText;
	}

	/**
	 * @return a component appropriate to edit and possibly change the name of the item.
	 */
	public final JComponent getNameEditor() {
		return _textFieldName;
	}

	/**
	 * @return a component appropriate to edit and possibly change the name of the item.
	 */
	public JComponent getBinaryEditor() {
		if (getBinary() != null) {
			JTextField textField = new JTextField(getBinary());
			textField.setEnabled(false);
			textField.setEditable(false);
			return textField;
		}
		return null;
	}

	/**
	 * Updates the item content after modification
	 * through the properties dialog box.
	 */
	public final void updateName() {
		_labelName.setText(getNickName());
		setToolTipText(toString());
	}

	protected abstract JPopupMenu initPopup( );

	protected final void drawPopup( MouseEvent event ) {
		_popupMenu = initPopup();
		_popupMenu.show(this, event.getX(), event.getY());
	}

	///// 
	/////  I N P U T S   &   O U T P U T S
	///// 

	/**
	 * Saves the item as a XML tree.
	 * 
	 * @param file the filename of the project.
	 * @return the xml Tree.
	 * @throws IOException in case of output error.
	 */
	public XMLElement save( File file ) throws IOException {
		XMLElement element = new XMLElement(getTagName());
		element.addChildren(saveProperties(file));
		return element;
	}

	/**
	 * Saves properties of the graph item in an XML tree.
	 * Saves only filled properties.
	 * 
	 * @param file the filename of the project.
	 * @return the XML tree.
	 * @throws IOException in case of output error.
	 */
	public final List<XMLElement> saveProperties( File file ) throws IOException {
		LinkedList<XMLElement> xmlTree = new LinkedList<XMLElement>();
		xmlTree.add(new XMLElement("NickName", _nickName));
		xmlTree.add(new XMLElement("Name", getName()));
		xmlTree.add(new XMLElement("Identifier", Integer.toString(getId())));
		xmlTree.add(new XMLElement("X-coordinate", Integer.toString(getLocation().x)));
		xmlTree.add(new XMLElement("Y-coordinate", Integer.toString(getLocation().y)));

		String currentPath = null;
		if (file != null) {
			currentPath = file.getParentFile().getAbsolutePath();
		}
		String tmp = getBinary(currentPath);
		if (tmp != null) {
			xmlTree.add(new XMLElement("Binary", tmp));
		}
		if (getCategory() != null) {
			xmlTree.add(new XMLElement("Category", getCategory()));
		}
		if (getFormula() != null) {
			xmlTree.add(new XMLElement("Formula", getFormula(currentPath)));
		}
		if (!_inputDataSlots.isEmpty()) {
			xmlTree.add(new XMLElement("InputImages", Integer.toString(_inputDataSlots.size())));
		}
		if (!_inputParameterSlots.isEmpty()) {
			xmlTree.add(new XMLElement("Parameters", Integer.toString(_inputParameterSlots.size())));
		}
		if (!_outputDataSlots.isEmpty()) {
			xmlTree.add(new XMLElement("OutputImages", Integer.toString(_outputDataSlots.size())));
		}
		if (this instanceof IScriptArgument) {
			IScriptArgument item = (IScriptArgument)this;
			xmlTree.add(new XMLElement("Is-Argument", Boolean.toString(item.isArgument())));
		}
		return xmlTree;
	}

	/**
	 * Saves each of its slots and its related connection
	 * in an XML tree.
	 * 
	 * @return the xml Tree.
	 * @throws IOException in case of output error.
	 */
	public final List<XMLElement> saveSlots( ) throws IOException {
		LinkedList<XMLElement> domTree = new LinkedList<XMLElement>();
		for (InputSlot slot : _inputDataSlots) {
			XMLElement element = slot.save();
			if (element != null) {
				domTree.add(element);
			}
		}

		for (InputSlot slot : _inputParameterSlots) {
			XMLElement element = slot.save();
			if (element != null) {
				domTree.add(element);
			}
		}
		if (_inputMaskSlot != null) {
			XMLElement element = _inputMaskSlot.save();
			if (element != null) {
				domTree.add(element);
			}
		}
		return domTree;
	}

	/**
	 * @return the related tree item in the explorer if any.
	 */
	protected final TreeItem getRelatedTreeItem( ) {
		return Ariane.getWorkspace().getTree().getTreeItem(getName(), getCategory());
	}


	/**
	 * Loads properties of the item from the given XML tree.
	 * 
	 * @param xmlTree the tree of XML elements.
	 * @throws IOException in case of input error.
	 */
	public final void load( List<XMLElement> xmlTree ) throws IOException {
		loadProperties(xmlTree);
		loadAdditionalProperties(xmlTree);
	}

	protected void loadAdditionalProperties( List<XMLElement> xmlTree ) throws IOException {
		// Empty
	}

	/**
	 * Loads graph item properties from the given XML tree.
	 * 
	 * @param xmlTree the tree of properties.
	 * @throws IOException in case of input error.
	 */
	protected final void loadProperties( List<XMLElement> xmlTree ) throws IOException {
		int inputParameterNumber = 0;
		int inputDataNumber = 0;
		int outputDataNumber = 0;
		Point coordinates = new Point(0, 0);
		for ( XMLElement element : xmlTree ) {
			// French tags are kept for sake of compatibility.
			if (element.getName().equalsIgnoreCase("NickName")) {
				setNickName(element.getValue());
			} else if (element.getName().equalsIgnoreCase("Name") || element.getName().equalsIgnoreCase("Nom")) {
				setAbstractName(element.getValue());
			} else if (element.getName().equalsIgnoreCase("Identifier") || element.getName().equalsIgnoreCase("Identifieur")) {
				setInternalId(Integer.parseInt(element.getValue()));
			} else if (element.getName().equalsIgnoreCase("X-coordinate") || element.getName().equalsIgnoreCase("Abscisse")) {
				coordinates.x = Integer.parseInt(element.getValue());
			} else if (element.getName().equalsIgnoreCase("Y-coordinate") || element.getName().equalsIgnoreCase("Ordonnee")) {
				coordinates.y = Integer.parseInt(element.getValue());
			} else if (element.getName().equalsIgnoreCase("Binary") || element.getName().equalsIgnoreCase("Binaire")) {
				setBinary(element.getValue()); 
			} else if (element.getName().equalsIgnoreCase("Category") || element.getName().equalsIgnoreCase("Categorie")) {
				_category = element.getValue();
			} else if (element.getName().equalsIgnoreCase("Formula")) {
				setFormula(element.getValue());
			} else if (element.getName().equalsIgnoreCase("Parameters") || element.getName().equalsIgnoreCase("ParametresEntree")) {
				inputParameterNumber = Integer.parseInt(element.getValue());
			} else if (element.getName().equalsIgnoreCase("InputImages") || element.getName().equalsIgnoreCase("ImagesEntree")) {
				inputDataNumber = Integer.parseInt(element.getValue());
			} else if (element.getName().equalsIgnoreCase("OutputImages") || element.getName().equalsIgnoreCase("ImagesSortie")) {
				outputDataNumber = Integer.parseInt(element.getValue());
			} else if (element.getName().equalsIgnoreCase("Is-Argument") && this instanceof IScriptArgument) {
				IScriptArgument item = (IScriptArgument)this;
				item.setIsArgument(Boolean.parseBoolean(element.getValue()));
			}
		}
		for (int i = 0; i < inputDataNumber; i++) {
			_inputDataSlots.add(new InputDataSlot(this, i));
		}
		for (int i = 0; i < inputParameterNumber; i++) {
			_inputParameterSlots.add(new InputParameterSlot(this, i));
		}
		for (int i = 0; i < outputDataNumber; i++) {
			_outputDataSlots.add(new OutputDataSlot(this, i));
		}
		if (_treeItem == null) {
			_treeItem = getRelatedTreeItem();
		}
		setLocation(coordinates);
	}

	/**
	 * Loads a graph item from the given XML tree.
	 * 
	 * @param graph the graph.
	 * @param xmlTree the XML tree.
	 * @param translation translate or not the item location (+5 pixels).
	 * @throws IOException in case of input error.
	 */
	public static final List<GraphItem> load( Graph graph, List<XMLElement> xmlTree, boolean translation ) throws IOException {
		List<GraphItem> items = new ArrayList<GraphItem>();
		if (xmlTree == null) {
			return items;
		}

		for ( XMLElement element : xmlTree ) {
			GraphItem item = null;
			if (element.getName().equalsIgnoreCase(OperatorItem.TAGNAME) || element.getName().equalsIgnoreCase("Operateur")) {
				item = new OperatorItem(graph, element.getChildren());
			} else if (element.getName().equalsIgnoreCase(NoteItem.TAGNAME)) {
				item = new NoteItem(graph, element.getChildren());
			} else if (element.getName().equalsIgnoreCase("Fichier")) {
				// Kept for the sake of compatibility
				item = DataFile.getInstance(graph, element.getChildren());
			} else if (element.getName().equalsIgnoreCase(LoaderItem.TAGNAME)) {
				item = new LoaderItem(graph, element.getChildren());
			} else if (element.getName().equalsIgnoreCase(SaverItem.TAGNAME)) {
				item = new SaverItem(graph, element.getChildren());
			} else if (element.getName().equalsIgnoreCase(ViewerItem.TAGNAME)) {
				item = new ViewerItem(graph, element.getChildren());
			} else if (element.getName().equalsIgnoreCase(FileNameItem.TAGNAME)
					|| element.getName().equalsIgnoreCase("Datafile")) {
				item = new FileNameItem(graph, element.getChildren());
			} else if (element.getName().equalsIgnoreCase(ConstantItem.TAGNAME)) {
				item = new ConstantItem(graph, element.getChildren());
			} else if (element.getName().equalsIgnoreCase(SystemItem.TAGNAME)) {
				item = new SystemItem(graph, element.getChildren());
			} else if (element.getName().equalsIgnoreCase(ScriptItem.TAGNAME)) {
				item = new ScriptItem(graph, element.getChildren());
			} else if (element.getName().equalsIgnoreCase(MacroOperatorItem.TAGNAME)
					|| element.getName().equalsIgnoreCase("MacroOperateur")) {
				item = new MacroOperatorItem(graph, element.getChildren());
			} else if (element.getName().equalsIgnoreCase(ControlStructureMergeItem.TAGNAME)) {
				item = new ControlStructureMergeItem(graph, element.getChildren());
			} else if (element.getName().equalsIgnoreCase(EvalItem.TAGNAME)) {
				item = new EvalItem(graph, element.getChildren());
			} else if (element.getName().equalsIgnoreCase(FormatItem.TAGNAME)) {
				item = new FormatItem(graph, element.getChildren());
			} else if (element.getName().equalsIgnoreCase(ControlStructureIfItem.TAGNAME)) {
				item = new ControlStructureIfItem(graph, element.getChildren());
			} else if (element.getName().equalsIgnoreCase(ControlStructureSelectItem.TAGNAME)) {
				item = new ControlStructureSelectItem(graph, element.getChildren());
			} else if (element.getName().equalsIgnoreCase(ControlStructureWhileItem.TAGNAME)) {
				item = new ControlStructureWhileItem(graph, element.getChildren());
			} else if (element.getName().equalsIgnoreCase("Until")) { // Keep for compatibility
				item = new ControlStructureWhileItem(graph, element.getChildren());
			} else if (element.getName().equalsIgnoreCase(ControlStructureForItem.TAGNAME)) {
				item = new ControlStructureForItem(graph, element.getChildren());
			}
			if (item != null) {
				Point coordinates = item.getLocation();
				graph.addItem(item, coordinates, true);
				items.add(item);
			}
		}
		return items;
	}
}
