/* -*- c-basic-offset: 3 -*-
 *
 * Copyright (c) 2013, GREYC.
 * All rights reserved
 *
 * You may use this file under the terms of the BSD license as follows:
 *
 * "Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of the GREYC, nor the name of its
 *     contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
 * 
 *   
 * For more information, refer to:
 * https://clouard.users.greyc.fr/Pandore
 */

package fr.greyc.pantheon.ariane.graph.item;

import fr.greyc.pantheon.ariane.Ariane;
import fr.greyc.pantheon.ariane.graph.Graph;
import fr.greyc.pantheon.ariane.graph.exception.ArianeDataFileIOException;
import fr.greyc.pantheon.ariane.graph.item.etc.FileImagePreviewer;
import fr.greyc.pantheon.ariane.graph.item.etc.FileTypeViewer;
import fr.greyc.pantheon.ariane.graph.item.etc.PropertyViewer;
import fr.greyc.pantheon.ariane.graph.slot.InputSlot;
import fr.greyc.pantheon.ariane.graph.slot.InputParameterSlot;
import fr.greyc.pantheon.ariane.ui.PreferenceChooser;
import fr.greyc.pantheon.ariane.ui.Workspace;
import fr.greyc.pantheon.ariane.utils.TreeItem;
import fr.greyc.pantheon.ariane.xml.XMLElement;

import java.awt.Component;
import java.awt.Font;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyEvent;
import java.awt.event.MouseEvent;
import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

import javax.swing.JButton;
import javax.swing.JComponent;
import javax.swing.JDialog;
import javax.swing.JFileChooser;
import javax.swing.JMenuItem;
import javax.swing.JPopupMenu;
import javax.swing.JTextField;
import javax.swing.KeyStroke;

/**
 * A <code>LoaderItem</code> manages data files loading
 * within a graph.
 * 
 * @version Régis CLOUARD - Dec 09, 2007 (created from ImageFile)
 * @version Régis CLOUARD - Jun 20, 2010 (enhancement)
 * @version Régis CLOUARD - Jan 19, 2012 (add execution of a folder of images)
 */
public final class LoaderItem extends DataFile {
	/** The tag name of the item. */
	public static final String TAGNAME = "Loader";

	/** The button for loading a new file.  */
	private final JButton _jbuttonSelect = new JButton();

	/** The component that holds the current name. */
	private final JTextField _binaryEditor = new JTextField(getBinary());
	/** The date of the last modification of the related file. */
	private long _lastModified;

	private static StaticOpenFileChooser _fileChooser = new StaticOpenFileChooser();

	/**
	 * Creates a <code>LoaderItem</code> on the given graph
	 * with the specified properties.
	 * 
	 * @param treeItem the source tree item.
	 * @param graph the graph that holds the item.
	 * @param name the name of the item.
	 * @param category the category name of the item.
	 */
	public LoaderItem( TreeItem treeItem, Graph graph, String name, String category ) {
		super(treeItem, graph, name, null, category, "loader");
		createSlots(0, 1, 0, 1, 1);
		draw();
	}

	/**
	 * Creates a <code>LoaderItem</code> on the given graph
	 * with the given XML data.
	 * 
	 * @param graph the graph that holds the item.
	 * @param xmlTree an XML tree of data.
	 * @throws IOException XML errors.
	 */
	public LoaderItem( Graph graph, List<XMLElement> xmlTree ) throws IOException {
		super(graph, xmlTree, "loader", true);
		draw();
	}

	/**
	 * Returns the tag name.
	 * 
	 * @return a string.
	 */
	@Override
	public String getTagName( ) {
		return TAGNAME;
	}

	/**
	 * Sets the name of the related physical file.
	 * If the name of the file is specified as a relative path 
	 * than the absolute name is built with the directory name
	 * of the project.
	 * 
	 * @param binary a file name (absolute or relative)
	 * @param directory a home directory for the binaries.
	 */
	@Override
	public void setBinary( String binary ) {
		if (binary != null && !binary.trim().isEmpty()) {
			String currentDirectory;
			if (getGraph().getRoot().getFileName() == null) {
				currentDirectory = null;
			} else {
				currentDirectory = getGraph().getRoot().getFileName().getParent();
			}
			super.setBinary(DataFile.setPathToFilename(binary, currentDirectory));
		}
	}

	/**
	 * Returns the name of the file either with relative path if the file is
	 * in the filePath directory or with an absolute path otherwise.
	 * 
	 * @param filePath a home directory for the files.
	 * @return the file name.
	 */
	@Override
	public String getBinary( String directory ) {
		return getRelativePath(super.getBinary(), directory);
	}

	/**
	 * Resynchronizes the item with the related file.
	 * if the related file has been changed, resets the item.
	 */
	public void synchronizeWithFileSystem() {
		if (getBinary() != null) {
			File file = new File(getBinary());
			if (_lastModified < file.lastModified()) {
				resetAll();
				_lastModified = file.lastModified();
			}
		}
	}

	///// 
	///// E X E C U T I O N
	/////

	/**
	 * Executes a Loader. It means: loads the file given
	 * by the user in the graph.
	 * If it exists, sets the current state to EXECUTED,
	 * otherwise sets the current state to ERROR.
	 * 
	 * @param isFailureFatalError is FAILURE a fatal error?
	 * @return true if the operator has been executed without errors.
	 */
	@Override
	protected boolean primitiveExecute( boolean isFailureFatalError ) {
		try {
			setCurrentFilename(null);
			final InputParameterSlot parameter = getInputParameterSlots().isEmpty() ? null : getInputParameterSlots().get(0);
			File file;
			if (parameter != null && parameter.hasValue(isFailureFatalError)) {
				changeBinaryName((JTextField)getNameEditor(), _binaryEditor, _jbuttonSelect, null, false);
				file = new File(parameter.getValue());
			} else if (getBinary() == null || getBinary().isEmpty()) {
				throw new ArianeDataFileIOException(Ariane.ErrorCode.MISSINGFILE, Ariane.getString("ImageFile.error3"));
			} else {
				file = new File(getBinary());
				_lastModified = file.lastModified();
			}
			if (!file.exists()) {
				throw new ArianeDataFileIOException(Ariane.ErrorCode.UNKNOWNFILE, Ariane.getString("ImageFile.error4")
						+ " \""
						+ file.getAbsolutePath()
						+ "\"");
			}
			final String filename = file.getAbsolutePath();
			getOutputDataSlots().get(0).setValue(filename);
			setCurrentFilename(filename);
			if (getOutputResultSlot() != null) { // Keep this test for sake of compatibility
				getOutputResultSlot().setValue(filename);
			}
		} catch (ArianeDataFileIOException e) {
			if (isFailureFatalError && !Graph.isExecutionStopped()) {
				Ariane.getWorkspace().printErrorMessage(e.getErrorCode(), getAbsoluteName(), getId(), e.getLocalizedMessage());
				getGraph().stopExecution();
				changeStatus(Status.ERROR);
			}
			return false;
		}

		// Validate the result.
		changeStatus(Status.EXECUTED);
		return true;
	}

	///// 
	///// M E N U 
	/////

	/**
	 * Builds the contextual menu as a Popup Menu.
	 * 
	 * @return a convenient popup menu.
	 */
	@Override
	protected JPopupMenu initPopup( ) {
		JPopupMenu popupMenu = new JPopupMenu();
		popupMenu.setFont(PreferenceChooser.getFont(8).deriveFont(Font.BOLD));

		JMenuItem item;
		item = new JMenuItem(Ariane.getString("Graph.Menu.copy"));
		item.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed( ActionEvent event ) {
				Ariane.getWorkspace().copy();
			}
		});
		item.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_C, KeyEvent.CTRL_MASK));
		popupMenu.add(item);

		item = new JMenuItem(Ariane.getString("Graph.Menu.cut"));
		item.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed( ActionEvent event ) {
				Ariane.getWorkspace().cut();
			}
		});
		item.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_X, KeyEvent.CTRL_MASK));
		popupMenu.add(item);

		popupMenu.addSeparator();
		
		item = new JMenuItem(Ariane.getString("PopupMenu.execute"));
		item.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed( ActionEvent event ) {
				getGraph().execute();
			}
		});
		item.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_R, KeyEvent.CTRL_MASK));
		popupMenu.add(item);

		item = new JMenuItem(Ariane.getString("PopupMenu.executeDirectory"));
		item.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed( ActionEvent event ) {
				getGraph().executeFolder(LoaderItem.this);
			}
		});
		popupMenu.add(item);

		item = new JMenuItem(Ariane.getString("PopupMenu.reset"));
		item.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed( ActionEvent event ) {
				resetAll();
			}
		});
		item.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_U, KeyEvent.CTRL_MASK));
		popupMenu.add(item);

		item = new JMenuItem(Ariane.getString("PopupMenu.remove"));
		item.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed( ActionEvent event ) {
				delete();
			}
		});
		item.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_DELETE, 0));
		popupMenu.add(item);

		popupMenu.addSeparator();

		item = new JMenuItem(Ariane.getString("PopupMenu.selectFile"));
		item.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed( ActionEvent event ) {
				selectFile();
			}
		});
		popupMenu.add(item);

		return popupMenu;
	}

	/**
	 * Opens a windows from which the user can select a file.
	 * Keeps the selected directory as the current directory.
	 */
	private void selectFile( ) {
		final String currentFile = getBinary();
		if (currentFile == null || currentFile.trim().isEmpty() || !(new File(currentFile).exists())) {
			if (currentLoadingDir == null) {
				currentLoadingDir = Workspace.getCurrentDir();
			}
			_fileChooser.setCurrentDirectory(new File(currentLoadingDir));
		} else {
			File file = new File(currentFile);
			_fileChooser.setSelectedFile(file);
		}
		if (_fileChooser.showOpenDialog(this) == JFileChooser.APPROVE_OPTION) {
			String path = _fileChooser.getSelectedFile().getPath();
			setCurrentFilename(null);
			currentLoadingDir = _fileChooser.getSelectedFile().getParent();
			changeBinaryName((JTextField)getNameEditor(), _binaryEditor, _jbuttonSelect, path, true);
		}
	}

	@Override
	protected void resetButton( JButton button ) {
		button.setText(Ariane.getString("PropertyViewer.load") + "...");
	}

	/**
	 * @return the list of specific properties (other than common properties).
	 */
	@Override
	public List<PropertyViewer.PropertyCouple> getSpecificProperties( ) {
		ArrayList<PropertyViewer.PropertyCouple> properties = new ArrayList<PropertyViewer.PropertyCouple>();

		String name;
		if (getBinary() == null) {
			name = "";
		} else {
			File file = new File(getBinary());
			name = file.getName();
		}

		if (name.isEmpty()) {
			resetButton(_jbuttonSelect);
		} else {
			_jbuttonSelect.setText(name);
		}
		_jbuttonSelect.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed( ActionEvent event ) {
				selectFile();
			}
		});
		properties.add(new PropertyViewer.PropertyCouple(Ariane.getString("PropertyViewer.fileName"), _jbuttonSelect));
		return properties;
	}

	@Override
	public void fireConnection( InputSlot slot, boolean connection ) {
		if (slot == getInputParameterSlots().get(0)) {
			_jbuttonSelect.setEnabled(!connection);
		}
		super.fireConnection(slot, connection);
	}

	/**
	 * Invoked when the mouse button has been
	 * double clicked (pressed and released) on a graph item.
	 * Shows the dialog box for setting the item's properties.
	 * 
	 * @param e the event attributes.
	 */
	@Override
	protected void onMouseDoubleClicked( MouseEvent event ) {
		_jbuttonSelect.doClick();
	}

	/**
	 * @return a component appropriate to edit and possibly change the name of the item.
	 */
	@Override
	public JComponent getBinaryEditor() {
		return null;
	}

	/**
	 * Defined to keep the size and location of the 
	 * previous opening of the JFileChooser;
	 */
	private static class StaticOpenFileChooser extends JFileChooser {
		private JDialog _dialog;
		public StaticOpenFileChooser() {
			setFileView(new FileTypeViewer());
			setDialogTitle(Ariane.getString("FileChooser.titleOpenFile"));
			setFileSelectionMode(JFileChooser.FILES_ONLY);
			setAccessory(new FileImagePreviewer(this));
		}

		@Override
		protected JDialog createDialog( Component parent ) {
			JDialog dialog = super.createDialog(parent);
			if (_dialog != null) {
				dialog.setLocation(_dialog.getLocation());
				dialog.setSize(_dialog.getSize());
			}
			_dialog = dialog;
			return dialog;
		}
	}
}