/* -*- c-basic-offset: 3 -*-
 *
 * Copyright (c) 2013, GREYC.
 * All rights reserved
 *
 * You may use this file under the terms of the BSD license as follows:
 *
 * "Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of the GREYC, nor the name of its
 *     contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
 * 
 *   
 * For more information, refer to:
 * https://clouard.users.greyc.fr/Pandore
 */

package fr.greyc.pantheon.ariane.graph.item;

import java.io.IOException;
import java.util.Iterator;
import java.util.List;

import fr.greyc.pantheon.ariane.Ariane;
import fr.greyc.pantheon.ariane.graph.Graph;
import fr.greyc.pantheon.ariane.graph.SubGraph;
import fr.greyc.pantheon.ariane.graph.exception.ArianeException;
import fr.greyc.pantheon.ariane.graph.exception.ArianeMissingInputException;
import fr.greyc.pantheon.ariane.graph.exception.ArianeMissingParameterException;
import fr.greyc.pantheon.ariane.graph.slot.InputDataSlot;
import fr.greyc.pantheon.ariane.graph.slot.InputMaskSlot;
import fr.greyc.pantheon.ariane.graph.slot.InputParameterSlot;
import fr.greyc.pantheon.ariane.graph.slot.OutputDataSlot;
import fr.greyc.pantheon.ariane.graph.slot.OutputResultSlot;
import fr.greyc.pantheon.ariane.utils.TreeItem;
import fr.greyc.pantheon.ariane.xml.XMLElement;

/**
 * A <code>MacroOperatorItem</code> is a graph item that is composed by
 * a subgraph of items.
 *
 * @author Sébastien LEFÈVRE, Richard LOBRÉAU - Feb 21, 1999
 * @version Adrien SCHRÈQUE, Hicham AMADI - Nov 05, 2000
 * @version Mathieu JOETS - Jul 09, 2001
 * @version Alban DOUBLET, Guillaume CHATELET - Jun 10, 2002
 * @version Régis CLOUARD - Oct 16, 2001
 * @version Régis CLOUARD - Jan 16, 2007
 */
public final class MacroOperatorItem extends CompositeItem {
	/** The tag name of the item. */
	public static final String TAGNAME = "MacroOperator";
	private static final String ABSTRACT_NAME = "macro-operator";

	/**
	 * Creates a <code>MacroOperatorItem</code> on the given graph with 
	 * properties interactively asked to the user.
	 * Initializes the mouse listeners, the popup menu,
	 * and displays it in the graph.
	 * 
	 * @param treeItem the reference tree item.
	 * @param graph the parent graph.
	 */
	public MacroOperatorItem( TreeItem treeItem, Graph graph ) throws InterruptedException {
		super(treeItem, graph, "");
		setAbstractName(ABSTRACT_NAME);
		askProperties(ABSTRACT_NAME + getId(), 1, 1, 1);
	}

	/**
	 * Creates a <code>MacroOperatorItem</code> on the given graph with
	 * the given XML data.
	 * 
	 * @param graph the parent graph.
	 * @param xmlTree the XML tree of data.
	 * @throws IOException XML errors.
	 */
	public MacroOperatorItem( Graph graph, List<XMLElement> xmlTree ) throws IOException {
		super(graph);
		load(xmlTree);
		// Keep the 4 following lines for sake of compatibility
		if (_treeItem == null) {
			_treeItem = restoreDefaultTreeItem();
		}
		if (_treeItem != null) {
			setSignature(_treeItem.getSignature());
		}
		createSlots(0, 0, 1, 0, 0);
		draw();
	}

	/**
	 * Creates a <code>MacroOperatorItem</code> 
	 * as a subroutine (undeditable item) on the given graph with
	 * the given XML data.
	 * 
	 * @param graph the parent graph.
	 * @param xmlTree the XML tree of data.
	 * @param treetem the related TreeItem.
	 * @throws IOException XML errors.
	 */
	public MacroOperatorItem( Graph graph, List<XMLElement> xmlTree, TreeItem treeItem ) throws IOException {
		super(graph, treeItem);
		load(xmlTree);
		setSignature(_treeItem.getSignature());
		createSlots(0, 0, 1, 0, 0);
		setAbstractName(treeItem.getName());
		setNickName(treeItem.getName());
		draw();
	}

	@Override
	protected TreeItem restoreDefaultTreeItem() {
		setNickName(getNickName());
		setAbstractName(ABSTRACT_NAME);
		return Ariane.getWorkspace().getTree().getTreeItem(ABSTRACT_NAME, null);
	}

	/**
	 * Builds or changes the prototype of the macro-operator
	 * from the given values.
	 * Creates and changes also the related subgraph.
	 */
	@Override
	protected void setPrototype( String name, int inputDataCount, int parameterCount, int outputDataCount ) {
		setPrototype(name, inputDataCount, parameterCount, outputDataCount, 0, 0);
	}

	/**
	 * Creates the subgraph with the convenient prototype.
	 */
	@Override
	protected SubGraph createSubGraph( String name, int inputDataCount, int parameterCount, int outputDataCount ) {
		return createSubGraph(getNickName(), inputDataCount, parameterCount, outputDataCount, 0, 0);
	}

	/**
	 * Returns the tag name.
	 * 
	 * @return a string.
	 */
	@Override
	public String getTagName( ) {
		return TAGNAME;
	}

	///// 
	///// E X E C U T I O N
	/////

	@Override
	protected void linkItemWithVirtualItem( VirtualItem virtualOperator, boolean isFailureFatalError ) {
		// Gets and checks parameters (-> executes forward operators if needed).
		Iterator<InputParameterSlot> its = getInputParameterSlots().iterator();
		Iterator<OutputResultSlot> itd   = virtualOperator.getVirtualInputParameterSlots().iterator();
		while (its.hasNext()) {
			InputParameterSlot parameter = its.next();
			OutputResultSlot result = itd.next();
			if (parameter.hasValue(isFailureFatalError)) {
				result.setValue(parameter.getValue());
			} else {
				throw new ArianeMissingParameterException(Ariane.getString("Operator.error1") + parameter.getRank());
			}
		}

		// Gets and checks mask (-> executes forward operators if needed).
		// Note: Mask is optional.
		InputMaskSlot inMasks = getInputMaskSlot();
		OutputDataSlot inMaskd = (OutputDataSlot)virtualOperator.getVirtualInputMaskSlot();
		if (inMasks.isConnected() && inMasks.hasValue(false)) {
			inMaskd.setValue(inMasks.getValue());
		}

		// Gets and checks input data (-> executes forward operators if needed).
		Iterator<InputDataSlot> its1  = getInputDataSlots().iterator();
		Iterator<OutputDataSlot> itd1 = virtualOperator.getVirtualInputDataSlots().iterator();
		while (its1.hasNext()) {
			InputDataSlot input = its1.next();
			OutputDataSlot output = itd1.next();
			if (input.hasValue(isFailureFatalError)) {
				output.setValue(input.getValue());
			} else {
				throw new ArianeMissingInputException(Ariane.getString("Operator.error3") + input.getRank());
			}
		}

	}

	/**
	 * Executes the content of the macro-operator.
	 * Execution consists in:
	 * <ul>
	 * <li>Gets input data.
	 * <li>Executes the subgraph.
	 * <li>Gets the results.
	 * </ul>
	 * Considers that a fatal error is in case of :
	 * <ul>
	 * <li>imperative: any errors.
	 * <li>non imperative: only systems or slot errors (unset).
	 * </ul>
	 * 
	 * @param isFailureFatalError is FAILURE a fatal error?
	 * @return true if the operator has been executed without errors.
	 */
	@Override
	protected boolean primitiveExecute( boolean isFailureFatalError ) {
		try {
			VirtualItem virtualOperator = getVirtualOperator();

			linkItemWithVirtualItem(virtualOperator, isFailureFatalError);

			changeStatus(Status.RUNNING);

			// Sets the output (executes the subgraph if needed).
			Iterator<OutputDataSlot> its2 = getOutputDataSlots().iterator();
			Iterator<InputDataSlot> itd2 = virtualOperator.getVirtualOutputDataSlots().iterator();
			while (its2.hasNext()) {
				OutputDataSlot output = its2.next();
				InputDataSlot input = itd2.next();
				if (input.hasValue(isFailureFatalError)) {
					output.setValue(input.getValue());
				}
			}

			// Sets the result.
			InputParameterSlot resultd = (InputParameterSlot)virtualOperator.getVirtualOutputResultSlot();
			if (!resultd.hasValue(isFailureFatalError)) {
				throw new ArianeMissingParameterException(Ariane.getString("MacroOperator.error2"));
			}

			OutputResultSlot results = getOutputResultSlot();
			results.setValue(resultd.getValue());
			changeStatus(Status.EXECUTED);
			return true;
		} catch (ArianeException e) {
			if (isFailureFatalError && !Graph.isExecutionStopped()) {
				Ariane.getWorkspace().printErrorMessage(e.getErrorCode(), getAbsoluteName(), getId(), e.getLocalizedMessage());
				getGraph().stopExecution();
				changeStatus(Status.ERROR);
			} else {
				changeStatus(Status.UNEXECUTED);
			}
			return false;
		}
	}

	///// 
	///// P R O P E R T I E S
	/////

	/**
	 * Asks interactively the prototype of the Macro.
	 */
	@Override
	protected void askProperties() throws InterruptedException {
		askProperties(getName(), getInputDataSlots().size(), getInputParameterSlots().size(), getOutputDataSlots().size());
	}
}