/* -*- c-basic-offset: 3 -*-
 *
 * Copyright (c) 2013, GREYC.
 * All rights reserved
 *
 * You may use this file under the terms of the BSD license as follows:
 *
 * "Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of the GREYC, nor the name of its
 *     contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
 * 
 *   
 * For more information, refer to:
 * https://clouard.users.greyc.fr/Pandore
 */

package fr.greyc.pantheon.ariane.graph.item;

import java.awt.Color;
import java.awt.FontMetrics;
import java.awt.Graphics;
import java.awt.GridBagConstraints;
import java.awt.Insets;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.FocusEvent;
import java.awt.event.KeyEvent;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.io.IOException;
import java.util.List;
import java.util.StringTokenizer;

import javax.swing.JLabel;
import javax.swing.JMenuItem;
import javax.swing.JPopupMenu;
import javax.swing.JTextArea;
import javax.swing.KeyStroke;

import fr.greyc.pantheon.ariane.Ariane;
import fr.greyc.pantheon.ariane.graph.Graph;
import fr.greyc.pantheon.ariane.graph.command.ChangeTextEdit;
import fr.greyc.pantheon.ariane.graph.item.etc.PropertyViewer;
import fr.greyc.pantheon.ariane.ui.PreferenceChooser;
import fr.greyc.pantheon.ariane.utils.TreeItem;
import fr.greyc.pantheon.ariane.xml.XMLElement;

/**
 * A <code>NoteItem</code> is a comment item that can be added to a graph.
 * It contains only text and is visualized as a <code>JTextArea</code>.
 *
 * @author Sébastien LEFÈVRE, Richard LOBRÉAU - Feb 21, 1999
 * @version Adrien SCHRÈQUE, Hicham AMADI - Nov 05, 2000
 * @version Mathieu JOETS - Jul 09, 2001
 * @version Alban DOUBLET, Guillaume CHATELET - Jun 10, 2002
 * @version Régis CLOUARD - Oct 16, 2001
 * @version Régis CLOUARD - Nov 26, 2004 (Prefer formula rather than binary for the text).
 * @version Régis CLOUARD - Nov 22, 2007 (fix bug on print)
 * @version Régis CLOUARD - Mar 30, 2011 (min size)
 */
public final class NoteItem extends GraphItem {
	/** The tag name of the item. */
	public static final String TAGNAME = "Note";

	/** The text to be displayed. */
	private JTextArea _labelNote;

	/** The color of the item. */
	private static final Color DEFAULT_COLOR = new Color(255, 255, 51);
	
	/** The color of the item. */
	private static Color _color = DEFAULT_COLOR;

	/** An empty constant JPanel for item which hasn't properties (eg. notes). */
	private static final transient PropertyViewer BLANK_PROPERTY_PANEL = new PropertyViewer();

	/**
	 * Creates a <code>NoteItem</code> on the given graph with the given properties.
	 * Initializes the mouse listeners, the popup menu,
	 * and displays it in the graph.
	 * 
	 * @param treeItem the reference tree item.
	 * @param graph the parent graph.
	 * @param name the name of the item.
	 */
	public NoteItem( TreeItem treeItem, Graph graph, String name ) {
		super(treeItem, graph, name, null);
		draw();
	}

	/**
	 * Creates a <code>NoteItem</code> on the given graph with the given XML data.
	 * 
	 * @param graph the parent graph.
	 * @param xmlTree the XML tree of data.
	 * @throws IOException XML errors.
	 */
	public NoteItem( Graph graph, List<XMLElement> xmlTree ) throws IOException {
		super(graph);
		load(xmlTree);
		if (getBinary() != null) { // KEEP THIS FOR COMPATIBILITY REASONS
			setFormula(getBinary());
			setBinary(null);
		}
		draw();
	}

	@Override
	public String toString( ) {
		return "(" + getId() + "):" + TAGNAME;
	}

	@Override
	public String getTagName( ) {
		return TAGNAME;
	}

	/**
	 * Invoked when the mouse button has been
	 * double clicked (pressed and released) on a component.
	 * Overrides the <code>GraphItem.onMouseDoubleClicked()</code>.
	 * Shows the man page of this item.
	 * 
	 * @param e the event attributes.
	 */
	@Override
	protected void onMouseDoubleClicked( MouseEvent event ) {
		showMan();
	}

	/**
	 * Invoked when a component gains the keyboard focus.
	 */
	@Override
	protected void focusIn( ) {
		_labelNote.moveCaretPosition(_labelNote.getText().length());
	}

	/**
	 * Invoked when a JtextArea looses the keyboard focus.
	 * Change the text of the related element in the model.
	 * 
	 * @param e the FocusEvent.
	 */
	@Override
	public void focusLost( FocusEvent event ) {
		if (event.getSource() == _labelNote) {
			String newFormula = _labelNote.getText();
			if (!newFormula.equals(getFormula()) && isShowing()) {
				getGraph().startActionEdition();
				getGraph().validateAction(new ChangeTextEdit(this, _labelNote, _labelNote.getText(), getFormula()));
				getGraph().endActionEdition();
				setFormula(_labelNote.getText());
				getGraph().setModified(true);
			}
		}
		_labelNote.getCaret().setVisible(false);
		_labelNote.setCaretPosition(0);
	}

	public static void setColor( Color colOR ) {
		_color = colOR;
	}

	 public static Color getColor( ) {
		return _color;
	}

	public static Color getDefaultColor( ) {
		return DEFAULT_COLOR;
	}

	@Override
	public Color getBackgroundColor( ) {
		return _color;
	}

	///// 
	///// D I S P L A Y I N G   /   P R I N T I N G
	/////

	/**
	 * Overrides the paint method of GraphItem, because a note
	 * has a different look.
	 */
	@Override
	public void draw( ) {
		removeAll();
		setBackground(getBackgroundColor());
		JLabel labelName = new JLabel();
		labelName.setFont(PreferenceChooser.getDefaultFont());
		labelName.setForeground(new Color(0, 0, 0));
		labelName.setText(getName());
		setToolTipText(toString());
		// layout
		GridBagConstraints constraint = new GridBagConstraints();
		constraint.insets = new Insets(2, 2, 3, 2);
		constraint.gridheight = 1;
		constraint.gridwidth = 1;
		constraint.gridy++; constraint.gridx = 0;
		constraint.anchor = GridBagConstraints.CENTER;
		add(labelName, constraint);
		if (_labelNote == null) {
			_labelNote = new JTextArea(getFormula(), 1, 2);
			_labelNote.setFont(PreferenceChooser.getDefaultFont());
			_labelNote.setBackground(Color.WHITE);
			_labelNote.setForeground(Color.BLACK);
			_labelNote.setTabSize(2);
		}
		// Mouse events.
		_labelNote.addMouseListener(new MouseAdapter( ) {
			@Override
			public void mousePressed( MouseEvent event ) {
				if (isSelected()) {
					_labelNote.getCaret().setVisible(true);
				} else {
					NoteItem.this.onMouseReleased(event);
				}
			}
		});

		_labelNote.addFocusListener(this);
		_labelNote.setEditable(true);
		_labelNote.getCaret().setVisible(false);

		// The lights ramp that shows the current state.
		constraint.gridwidth = GridBagConstraints.REMAINDER;
		constraint.gridy++; constraint.gridx = 0;
		add(_labelNote, constraint);
	}

	/**
	 * Overrides the default method in order to print components
	 * of the operator in a different way than they are displayed.
	 * 
	 * @param gc the graphic context.
	 */
	@Override
	public void printComponent( Graphics graphicContext ) {
		super.printComponent(graphicContext);
		FontMetrics fontMetric;
		fontMetric = graphicContext.getFontMetrics();
		int row = fontMetric.getHeight();
		super.centerText(getName(), graphicContext, row, false);
		row += fontMetric.getHeight() + 4;
		graphicContext.setFont(PreferenceChooser.getDefaultFont());
		if (getFormula() != null) {
			StringTokenizer stringTokenizer = new StringTokenizer(getFormula(), "\n");
			while (stringTokenizer.hasMoreTokens()) {
				graphicContext.drawString(stringTokenizer.nextToken(), 5, row);
				row += fontMetric.getHeight();
			}
		}
	}

	/**
	 * Executes the note : actually, does nothing.
	 * 
	 * @param isFailureFatalError is FAILURE a fatal error?
	 * @return true if the operator has been executed without errors.
	 */
	@Override
	protected boolean primitiveExecute( boolean isFailureFatalError ) {
		return false;
	}

	/**
	 * Builds the contextual menu as a Popup Menu.
	 * 
	 * @return a convenient popup menu.
	 */
	@Override
	protected JPopupMenu initPopup( ) {
		JPopupMenu popupMenu = new JPopupMenu();

		JMenuItem item;
		item = new JMenuItem(Ariane.getString("PopupMenu.remove"));
		item.addActionListener(new ActionListener( ) {
			@Override
			public void actionPerformed( ActionEvent event ) {
				delete();
			}
		});
		item.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_DELETE, 0));
		popupMenu.add(item);
		return popupMenu;
	}

	@Override
	protected PropertyViewer createPropertyViewer() {
		return BLANK_PROPERTY_PANEL;
	}
}