/* -*- c-basic-offset: 3 -*-
 *
 * Copyright (c) 2013, GREYC.
 * All rights reserved
 *
 * You may use this file under the terms of the BSD license as follows:
 *
 * "Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of the GREYC, nor the name of its
 *     contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
 * 
 *   
 * For more information, refer to:
 * https://clouard.users.greyc.fr/Pandore
 */

package fr.greyc.pantheon.ariane.graph.item;

import fr.greyc.pantheon.ariane.Ariane;
import fr.greyc.pantheon.ariane.graph.Graph;
import fr.greyc.pantheon.ariane.graph.exception.ArianeException;
import fr.greyc.pantheon.ariane.graph.exception.ArianeMissingInputException;
import fr.greyc.pantheon.ariane.graph.exception.ArianeMissingParameterException;
import fr.greyc.pantheon.ariane.graph.exception.ArianeRuntimeException;
import fr.greyc.pantheon.ariane.graph.item.etc.ArgumentLineParser;
import fr.greyc.pantheon.ariane.graph.item.etc.OperatorProcessBuilder;
import fr.greyc.pantheon.ariane.graph.slot.InputDataSlot;
import fr.greyc.pantheon.ariane.graph.slot.InputParameterSlot;
import fr.greyc.pantheon.ariane.graph.slot.OutputDataSlot;
import fr.greyc.pantheon.ariane.ui.Workspace;
import fr.greyc.pantheon.ariane.utils.TreeItem;
import fr.greyc.pantheon.ariane.xml.XMLElement;

import java.awt.Color;
import java.awt.FontMetrics;
import java.awt.Graphics;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyEvent;
import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.util.LinkedList;
import java.util.List;

import javax.swing.JMenuItem;
import javax.swing.JPopupMenu;
import javax.swing.KeyStroke;

/**
 * An <code>OperatorItem</code> is a graph item that encapsulates
 * a Shell command. It uses inputs to produce outputs with
 * specific parameter values.
 * @see GraphItem
 * 
 * @author Régis CLOUARD - Oct 01, 1998
 * @version Sébastien LEFÈVRE, Richard LOBRÉAU - Feb 21, 1999
 * @version Adrien SCHRÈQUE, Hicham AMADI - Nov 05, 2000
 * @version Mathieu JOETS - Jul 09, 2001
 * @version Alban DOUBLET, Guillaume CHATELET - Jun 10, 2002
 * @version Régis CLOUARD - Oct 16, 2001
 * @version Régis CLOUARD - Dec 06, 2004
 * @version Régis CLOUARD - Nov 24, 2007 (add Pandore path to binary operator name)
 * @version Régis CLOUARD - Jun 29, 2009 (add operator output messages)
 */
public final class OperatorItem extends GraphItem {
	/** The tag name of the item. */
	public static final String TAGNAME = "Operator";

	public static final String SEPARATOR = "&";

	private static final Color DEFAULT_COLOR = new Color(102, 102, 255);

	private static Color _color = DEFAULT_COLOR;

	/**
	 * Creates an <code>OperatorItem</code> on the given graph with
	 * the given properties. Initializes the mouse listeners,
	 * the popup menu, and display it in the graph.
	 * If the identifier =-1 then creates a unique identifier number.
	 * 
	 * @param treeItem the reference tree item.
	 * @param graph the parent graph.
	 * @param name the name of the item.
	 * @param category the name of the category of the operator.
	 * @param inputDataNumber the number of input data.
	 * @param inputParameterNumberthe number of parameters.
	 * @param outputDataNumber the number of output data.
	 * @param signature a string of the form : $0:toto=1; $1:titi=10;
	 */
	public OperatorItem( TreeItem treeItem, Graph graph, String name, String category,
			int inputDataNumber, int inputParameterNumber, int outputDataNumber, String signature ) {
		super(treeItem, graph, name, category);
		createSlots(inputDataNumber, inputParameterNumber, 1, outputDataNumber, 1);
		setSignature(signature);
		draw();
	}

	/**
	 * Creates an <code>OperatorItem</code> on the given graph
	 * with the given XML data.
	 * 
	 * @param graph the parent graph.
	 * @param xmlTree the XML tree of data.
	 * @throws IOException XML errors.
	 */
	public OperatorItem( Graph graph, List<XMLElement> xmlTree ) throws IOException {
		super(graph);
		// Creates mask and result because they are not saved in the xmlTree.
		createSlots(0, 0, 1, 0, 1);
		load(xmlTree);
		if (_treeItem != null) {
			setSignature(_treeItem.getSignature());
		} else {
			Ariane.getWorkspace().printErrorMessage(Ariane.ErrorCode.RUNTIME, getAbsoluteName(), getId(), Ariane.getString("Operator.error7"));
		}
		draw();
	}

	/**
	 * Returns the description of the operator: name + parameters value.
	 * 
	 * @return a string description of the operator.
	 */
	@Override
	public String toString( ) {
		String commandLine = "(" + getId() + "): " + getName();

		for (InputParameterSlot isp : getInputParameterSlots()) {
			if (isp.getValue() != null) {
				commandLine += " " + isp.getValue();
			}
		}
		return commandLine;
	}

	@Override
	public String getTagName( ) {
		return TAGNAME;
	}

	///// 
	///// P R E F E R E N C E S
	///// 

	// Note: the following methods cannot be generalized in the GraphItem class
	// because they use static attributes.

	/**
	 * Changes the current color.
	 * @param color the new color.
	 */
	public static void setColor( Color color ) {
		_color = color;
	}

	/**
	 * Returns the current color.
	 * 
	 * @return the color.
	 */
	public static Color getColor( ) {
		return _color;
	}

	/**
	 * Returns the default color.
	 * 
	 * @return the color.
	 */
	public static Color getDefaultColor( ) {
		return DEFAULT_COLOR;
	}

	/**
	 * Returns the background color (same as geColor()).
	 * 
	 * @return the color.
	 */
	@Override
	public Color getBackgroundColor( ) {
		return _color;
	}

	///// 
	///// M E N U
	///// 

	/**
	 * Builds the contextual menu as a Popup Menu.
	 */
	@Override
	protected JPopupMenu initPopup( ) {
		JPopupMenu popupMenu = new JPopupMenu();

		JMenuItem item;
		item = new JMenuItem(Ariane.getString("Graph.Menu.copy"));
		item.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed( ActionEvent event ) {
				Ariane.getWorkspace().copy();
			}
		});
		item.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_C, KeyEvent.CTRL_MASK));
		popupMenu.add(item);

		item = new JMenuItem(Ariane.getString("Graph.Menu.cut"));
		item.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed( ActionEvent event ) {
				Ariane.getWorkspace().cut();
			}
		});
		item.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_X, KeyEvent.CTRL_MASK));
		popupMenu.add(item);

		popupMenu.addSeparator();
		
		item = new JMenuItem(Ariane.getString("PopupMenu.execute"));
		item.addActionListener(new ActionListener( ) {
			@Override
			public void actionPerformed( ActionEvent event ) {
				getGraph().execute();
			}
		});
		item.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_R, KeyEvent.CTRL_MASK));
		popupMenu.add(item);

		item = new JMenuItem(Ariane.getString("PopupMenu.reset"));
		item.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed( ActionEvent event ) {
				resetAll();
			}
		});
		item.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_U, KeyEvent.CTRL_MASK));
		popupMenu.add(item);

		item = new JMenuItem(Ariane.getString("PopupMenu.remove"));
		item.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed( ActionEvent event ) {
				delete();
			}
		});
		item.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_DELETE, 0));
		popupMenu.add(item);

		return popupMenu;
	}

	///// 
	///// P R I N T I N G
	///// 

	/**
	 * Overrides the default method in order to print
	 * the operator in different way that is is normally displayed.
	 * Prints the name of the operator and each parameter value.
	 * 
	 * @param gc the graphics context.
	 */
	@Override
	public void printComponent( Graphics gc ) {
		super.printComponent(gc);
		FontMetrics fontMetrics = gc.getFontMetrics();
		int row = fontMetrics.getHeight();

		// The name of the operator.
		super.centerText(getNickName(), gc, row, false);
		// Print each parameters.
		// Print the value only of if the parameter is not connected.
		for (int i = 0; i < getInputParameterSlots().size(); i++) {
			InputParameterSlot param = getInputParameterSlots().get(i);
			row += fontMetrics.getHeight() + 1;
			if (param.isConnected()) {
				super.centerText("$" + i + ":<linked>", gc, row, true);
			} else {
				super.centerText("$" + i + ":" + param.getValue(), gc, row, true);
			}
		}
	}

	///// 
	///// E X E C U T I O N
	///// 

	/**
	 * Returns the relative name of the specified file
	 * if it is a temporary file otherwise returns
	 * the same file name.
	 * 
	 * @param filename a file name;
	 * @return a relative file name.
	 */
	private String getRelativeFilePathName( String filename ) {
		String dir = Workspace.getTempDir();
		if (filename.startsWith(dir)) {
			return filename.substring(dir.length());
		}
		return filename;
	}

	/**
	 * Returns the name of the executable.
	 * 
	 * @return the binary name.
	 */
	public static String getCommandPart( String binary ) {
		if (binary == null ) {
			return null;
		}
		String[] format =  binary.split(OperatorItem.SEPARATOR);
		if (format.length > 0) {
			return format[0];
		}
		return null;
	}

	public static String getFormatPart( String binary ) {
		if (binary != null) {
			int index = binary.indexOf(OperatorItem.SEPARATOR);
			if (index > 0 && index + 1 < binary.length()) {
				return binary.substring(index + 1);
			} else {
				return "[$p] [-m $m] [$i] [$o]";
			}
		}
		return null;
	}

	/**
	 * Returns the command line of the operator.
	 * 
	 * @return the binary name.
	 */
	@Override
	public String getBinary( ) {
		return (_treeItem == null) ? super.getBinary() : _treeItem.getBinary();
	}

	/**
	 * Builds the command line as specified in the command format
	 * and with the arguments. The binary name of the program
	 * is returns with the absolute pathname.
	 * Syntax: format string uses:
	 *  $p for parameter; $p0 for the first parameter
	 *  $m for the mask
	 *  $i for input; $i0 for the first input data
	 *  $o for output; $o0 for the first output data.
	 *  
	 * @param maskLine the mask value.
	 * @param parameterLine the parameter values.
	 * @param outputLine the output values.
	 * @return the command line.
	 */
	public List<String> buildAbstractCommandLine( List<String> maskLine, List<String> parameterLine, List<String> inputLine, List<String> outputLine ) {
		return ArgumentLineParser.parse(getCommandPart(getBinary()), getFormatPart(getBinary()), maskLine, parameterLine, inputLine, outputLine);
	}

	/**
	 * Builds the command line as specified in the command format
	 * and with the arguments. The binary name of the program
	 * is returns with the absolute pathname.
	 *  %p for parameter; %p0 for the first parameter
	 *  %m for the mask
	 *  %i for input; %i0 for the first input data file.
	 *  %o for output; %o0 for the first output data file.
	 *  
	 * @param maskLine the mask value.
	 * @param parameterLine the parameter values.
	 * @param outputLine the output values.
	 * @return the command line.
	 */
	public List<String> buildCommandLine( List<String> maskLine, List<String> parameterLine, List<String> inputLine, List<String> outputLine ) {
		return ArgumentLineParser.parse(GraphItem.getAbsoluteBinary(getCommandPart(getBinary())), getFormatPart(getBinary()), maskLine, parameterLine, inputLine, outputLine);
	}

	/**
	 * Executes the related command. The execution is done
	 * only if the current state is "unexecuted" or "error".
	 * Returns true if the execution is done without fatal error,
	 * where fatal error is in case of:
	 * <ul>
	 * <li>imperative: any errors.
	 * <li>non imperative: only system or parameter unset errors.
	 * </ul>
	 * 
	 * @param isFailureFatalError is FAILURE a fatal error?
	 */
	@Override
	protected boolean primitiveExecute( boolean isFailureFatalError ) {
		try {
			if (getBinary() == null) {
				throw new ArianeRuntimeException(Ariane.getString("Operator.error7"));
			}

			List<String> parameterLine = new LinkedList<String>();
			// Get the parameter values
			for (InputParameterSlot slot : getInputParameterSlots()) {
				if (slot.hasValue(isFailureFatalError)) {
					String temp = slot.getValue();
					parameterLine.add(temp);
				} else {
					throw new ArianeMissingParameterException(Ariane.getString("Operator.error1") + slot.getRank());
				}
			}
			// Get the optional mask
			List<String> maskLine = new LinkedList<String>();
			if (getInputMaskSlot().isConnected() && getInputMaskSlot().hasValue(false)) {
				String temp = getInputMaskSlot().getValue();
				maskLine.add(temp);
			}

			// Get the input data
			List<String> inputLine = new LinkedList<String>();
			for (InputDataSlot slot: getInputDataSlots()) {
				if (slot.hasValue(isFailureFatalError)) {
					inputLine.add(slot.getValue());
				} else {
					throw new ArianeMissingInputException(Ariane.getString("Operator.error3") + slot.getRank());
				}
			}

			// Create outputs data as temporary files
			List<String> outputLine = new LinkedList<String>();
			for (OutputDataSlot slot : getOutputDataSlots()) {
				outputLine.add(slot.getTempName());
			}

			List<String> commandLine = buildCommandLine(maskLine, parameterLine, inputLine, outputLine);
			String consoleMessage = getName();
			for (String part: commandLine ) {
				consoleMessage += " " + getRelativeFilePathName(part);
			}
			Ariane.getWorkspace().printSystemMessage(consoleMessage);

			// Execution of the operator binary
			changeStatus(Status.RUNNING);
			OperatorProcessBuilder processBuilder = new OperatorProcessBuilder(commandLine);
			long start = System.currentTimeMillis();
			Process proc1 = processBuilder.start();
			getGraph().setCurrentProcess(proc1);

			BufferedReader reader = new BufferedReader(new InputStreamReader(proc1.getInputStream()));
			String line;
			while ((line = reader.readLine()) != null) {
				Ariane.getWorkspace().printOutputMessage(line);
			}
			int exitValue = proc1.waitFor();
			getGraph().setCurrentProcess(null);

			if (exitValue == 0) {
				// Get the status value
				if (getCommandStatus().trim().length() > 0) {
					processBuilder = new OperatorProcessBuilder(GraphItem.getAbsoluteBinary(getCommandStatus()));
					Process proc2 = processBuilder.start();
					InputStream resultStream = proc2.getInputStream();
					proc2.waitFor();

					// Exploit the results
					byte[] resultByte = new byte[resultStream.available() + 1];
					resultStream.read(resultByte);
					String resultString = new String(resultByte);
					getOutputResultSlot().setValue(resultString.trim());
				} else {
					getOutputResultSlot().setValue("SUCCESS");
				}
			}

			// I mean if an execution fails than the error code !=0.
			if (exitValue != 0) { // '|| resultString.startsWith("FAILURE")) {
				throw new ArianeRuntimeException(Ariane.getString("Operator.error2"));
			}
			// Everything is OK -> validate the result of all outputs.
			for (OutputDataSlot slot : getOutputDataSlots()) {
				slot.setValue();
			}
			long end = System.currentTimeMillis();
			Ariane.getWorkspace().printSystemMessage("(" + (end - start) / 1000.0f + "s)");
			changeStatus(Status.EXECUTED);
			return true;
		} catch (ArianeRuntimeException e) {
			if (isFailureFatalError && !Graph.isExecutionStopped()) {
				Ariane.getWorkspace().printErrorMessage(Ariane.ErrorCode.RUNTIME, getAbsoluteName(), getId(), e.getLocalizedMessage());
				getGraph().stopExecution();
				changeStatus(Status.ERROR);
			} else {
				changeStatus(Status.UNEXECUTED);
			}
			return false;
		} catch (ArianeException e) {
			if (isFailureFatalError && !Graph.isExecutionStopped()) {
				Ariane.getWorkspace().printErrorMessage(e.getErrorCode(), getAbsoluteName(), getId(), e.getLocalizedMessage());
				getGraph().stopExecution();
				changeStatus(Status.ERROR);
			} else {
				changeStatus(Status.UNEXECUTED);
			}
			return false;
		} catch (IOException e) {
			if (isFailureFatalError && !Graph.isExecutionStopped()) {
				Ariane.getWorkspace().printErrorMessage(Ariane.ErrorCode.INPUTOUTPUT, getAbsoluteName(), getId(), e.getLocalizedMessage());
				getGraph().stopExecution();
				changeStatus(Status.ERROR);
			} else {
				changeStatus(Status.UNEXECUTED);
			}
			return false;
		} catch (InterruptedException e) {
			if (isFailureFatalError && !Graph.isExecutionStopped()) {
				Ariane.getWorkspace().printErrorMessage(Ariane.ErrorCode.INTERRUPTEDEXECUTION, getAbsoluteName(), getId(), e.getLocalizedMessage());
				getGraph().stopExecution();
				changeStatus(Status.ERROR);
			} else {
				changeStatus(Status.UNEXECUTED);
			}
			return false;
		}
	}
}