/* -*- c-basic-offset: 3 -*-
 *
 * Copyright (c) 2013, GREYC.
 * All rights reserved
 *
 * You may use this file under the terms of the BSD license as follows:
 *
 * "Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of the GREYC, nor the name of its
 *     contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
 * 
 *   
 * For more information, refer to:
 * https://clouard.users.greyc.fr/Pandore
 */

package fr.greyc.pantheon.ariane.graph.item;

import fr.greyc.pantheon.ariane.Ariane;
import fr.greyc.pantheon.ariane.graph.Graph;
import fr.greyc.pantheon.ariane.graph.exception.ArianeMissingInputException;
import fr.greyc.pantheon.ariane.graph.item.etc.PropertyViewer;
import fr.greyc.pantheon.ariane.graph.slot.InputSlot;
import fr.greyc.pantheon.ariane.graph.slot.InputDataSlot;
import fr.greyc.pantheon.ariane.graph.slot.InputParameterSlot;
import fr.greyc.pantheon.ariane.ui.Workspace;
import fr.greyc.pantheon.ariane.utils.TreeItem;
import fr.greyc.pantheon.ariane.xml.XMLElement;

import java.awt.Component;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyEvent;
import java.awt.event.MouseEvent;
import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

import javax.swing.JButton;
import javax.swing.JComponent;
import javax.swing.JDialog;
import javax.swing.JFileChooser;
import javax.swing.JMenuItem;
import javax.swing.JPopupMenu;
import javax.swing.JTextField;
import javax.swing.KeyStroke;

/**
 * A <code>saverItem</code> manages saving
 * of a data file in a graph.
 * 
 * @version Régis CLOUARD - Dec 09, 2007 (created from ImageFile)
 */
public final class SaverItem extends DataFile {
	/** The tag name of the item. */
	public static final String TAGNAME = "Saver";
	/** The button for loading a new file.  */
	private final JButton _jbuttonSelect = new JButton();
	/** The component that holds the current name. */
	private final JTextField _binaryEditor = new JTextField(getBinary());

	private static StaticSaveFileChooser _fileChooser = new StaticSaveFileChooser();

	/**
	 * Creates a <code>SaverItem</code> on the specified graph
	 * with the specified properties.
	 * Initializes the mouse listeners, the popup menu, and
	 * display it in the graph.
	 * 
	 * @param treeItem the source tree item.
	 * @param graph the graph that hold the item.
	 * @param name the name of the item.
	 * @param category the category name of the operator.
	 */
	public SaverItem( TreeItem treeItem, Graph graph, String name, String category ) {
		super(treeItem, graph, name, null, category, "saver");
		createSlots(1, 1, 0, 1, 1);
		draw();
	}

	/**
	 * Creates a <code>SaverItem</code> on the specified graph
	 * with the given XML data.
	 * 
	 * @param graph the graph that holds the item.
	 * @param xmlTree an XML tree of data.
	 * @throws IOException XML errors.
	 */
	public SaverItem( Graph graph, List<XMLElement> xmlTree ) throws IOException {
		super(graph, xmlTree, "saver", true);
		// For compatibility reasons
		if (getOutputDataSlots().isEmpty()) {
			createSlots(0, 0, 0, 1, 0);
		}
		draw();
	}

	/**
	 * Returns the tag name.
	 * 
	 * @return a string.
	 */
	@Override
	public String getTagName( ) {
		return TAGNAME;
	}

	/**
	 * Sets the name of the related physical file.
	 * If the name of the file is specified as a relative path 
	 * than the absolute name is built with the directory of the project.
	 * 
	 * @param binary a file name (absolute or relative)
	 */
	@Override
	public void setBinary( String binary ) {
		if (binary != null && !binary.trim().isEmpty()) {
			String currentDirectory;
			if (getGraph().getRoot().getFileName() == null) {
				currentDirectory = null;
			} else {
				currentDirectory = getGraph().getRoot().getFileName().getParent();
			}
			super.setBinary(DataFile.setPathToFilename(binary, currentDirectory));
		}
	}

	/**
	 * Returns the name of the file either with relative path if the file is
	 * in the filePath directory or with an absolute path otherwise.
	 * 
	 * @param directory a home directory.
	 * @return the file name.
	 */
	@Override
	public String getBinary( String directory ) {
		return getRelativePath(super.getBinary(), directory);
	}

	///// 
	///// E X E C U T I O N
	/////

	/**
	 * Executes a Saver. It means: sets the input file into
	 * the related physical file.
	 * 
	 * @param isFailureFatalError is FAILURE a fatal error?
	 * @return true if the operator has been executed without errors.
	 */
	@Override
	protected boolean primitiveExecute( boolean isFailureFatalError ) {
		try {
			setCurrentFilename(null);
			final InputDataSlot isi = getInputDataSlots().get(0);
			if (isi.hasValue(isFailureFatalError)) {
				File file;
				InputParameterSlot parameter = getInputParameterSlots().isEmpty() ? null : getInputParameterSlots().get(0);
				if (parameter != null && parameter.hasValue(isFailureFatalError)) {
					changeBinaryName((JTextField)getNameEditor(), _binaryEditor, _jbuttonSelect, null, false);
					file = new File(parameter.getValue());
				} else if (getBinary() == null || getBinary().isEmpty()) {
					throw new ArianeMissingInputException(Ariane.getString("ImageFile.error7"));
				} else {
					file = new File(getBinary());
				}
				copy(isi.getValue(), file.getAbsolutePath());
				String exec = Ariane.getString("ImageFile.save") + ": " + file.getAbsolutePath();
				Ariane.getWorkspace().printSystemMessage(exec);
				final String filename = file.getAbsolutePath();
				setCurrentFilename(filename);
				getOutputDataSlots().get(0).setValue(filename);
				if (getOutputResultSlot() != null) { // Keep this test for sake of compatibility
					getOutputResultSlot().setValue(filename);
				}
			} else {
				throw new ArianeMissingInputException(Ariane.getString("ImageFile.error5"));
			}
			// Validate the result.
			changeStatus(Status.EXECUTED);
			return true;
		} catch (ArianeMissingInputException e) {
			if (isFailureFatalError && !Graph.isExecutionStopped()) {
				Ariane.getWorkspace().printErrorMessage(e.getErrorCode(), getAbsoluteName(), getId(), e.getLocalizedMessage());
				getGraph().stopExecution();
				changeStatus(Status.ERROR);
			}
			return false;
		} catch (IOException e) {
			if (isFailureFatalError && !Graph.isExecutionStopped()) {
				Ariane.getWorkspace().printErrorMessage(Ariane.ErrorCode.INPUTOUTPUT, getAbsoluteName(), getId(), Ariane.getString("ImageFile.error6") + e.getLocalizedMessage());
				getGraph().stopExecution();
				changeStatus(Status.ERROR);
			}
			return false;
		}
	}

	@Override
	public void fireConnection( InputSlot slot, boolean connection ) {
		if (!getInputParameterSlots().isEmpty() && slot == getInputParameterSlots().get(0)) {
			_jbuttonSelect.setEnabled(!connection);
		}
		super.fireConnection(slot, connection);
	}

	///// 
	///// M E N U 
	/////

	/**
	 * Builds the contextual menu as a Popup Menu.
	 * 
	 * @return a convenient popup menu.
	 */
	@Override
	protected JPopupMenu initPopup( ) {
		JPopupMenu popupMenu = new JPopupMenu();
		
		JMenuItem item;
		item = new JMenuItem(Ariane.getString("Graph.Menu.copy"));
		item.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed( ActionEvent event ) {
				Ariane.getWorkspace().copy();
			}
		});
		item.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_C, KeyEvent.CTRL_MASK));
		popupMenu.add(item);

		item = new JMenuItem(Ariane.getString("Graph.Menu.cut"));
		item.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed( ActionEvent event ) {
				Ariane.getWorkspace().cut();
			}
		});
		item.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_X, KeyEvent.CTRL_MASK));
		popupMenu.add(item);

		popupMenu.addSeparator();

		item = new JMenuItem(Ariane.getString("PopupMenu.execute"));
		item.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed( ActionEvent event ) { getGraph().execute(); }
		});
		item.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_R, KeyEvent.CTRL_MASK));
		popupMenu.add(item);

		item = new JMenuItem(Ariane.getString("PopupMenu.reset"));
		item.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed( ActionEvent event ) { resetAll(); }
		});
		item.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_U, KeyEvent.CTRL_MASK));
		popupMenu.add(item);

		item = new JMenuItem(Ariane.getString("PopupMenu.remove"));
		item.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed( ActionEvent event ) { delete(); }
		});
		item.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_DELETE, 0));
		popupMenu.add(item);

		popupMenu.addSeparator();
		item = new JMenuItem(Ariane.getString("PopupMenu.selectFile"));
		item.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed( ActionEvent event ) { selectFile(); }
		});
		popupMenu.add(item);

		return popupMenu;
	}

	/**
	 * Opens a windows from which user can type a file name.
	 * Keeps the directory as the current directory.
	 */
	private void selectFile( ) {
		if (currentWorkingDir == null) {
			currentWorkingDir = Workspace.getCurrentDir();
		}
		_fileChooser.setCurrentDirectory(new File(currentWorkingDir));
		final String currentFile = getBinary();
		if (currentFile != null && !currentFile.trim().isEmpty()) {
			_fileChooser .setSelectedFile(new File(currentFile));
		}
		if (_fileChooser.showSaveDialog(this) == JFileChooser.APPROVE_OPTION) {
			final String path = _fileChooser.getSelectedFile().getPath();
			setCurrentFilename(null);
			currentWorkingDir = _fileChooser.getSelectedFile().getParent();
			changeBinaryName((JTextField)getNameEditor(), _binaryEditor, _jbuttonSelect, path, true);
		}
	}

	@Override
	protected void resetButton( JButton button ) {
		button.setText(Ariane.getString("PropertyViewer.saveas") + "...");
	}

	/**
	 * @return the list of specific properties (other than common properties).
	 */
	@Override
	public List<PropertyViewer.PropertyCouple> getSpecificProperties( ) {
		ArrayList<PropertyViewer.PropertyCouple> properties = new ArrayList<PropertyViewer.PropertyCouple>();

		String name;
		if (getBinary() == null) {
			name = "";
		} else {
			File file = new File(getBinary());
			name = file.getName();
		}

		if (name.length() == 0) {
			resetButton(_jbuttonSelect);
		} else { 
			_jbuttonSelect.setText(name);
		}	
		_jbuttonSelect.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed( ActionEvent event ) {
				selectFile();
			}
		});
		properties.add(new PropertyViewer.PropertyCouple(Ariane.getString("PropertyViewer.path"), _jbuttonSelect));
		return properties;
	}

	/**
	 * Invoked when the mouse button has been
	 * double clicked (pressed and released) on a graph item.
	 * Shows the dialog box for setting the item's properties.
	 * 
	 * @param e the event attributes.
	 */
	@Override
	protected void onMouseDoubleClicked( MouseEvent event ) {
		_jbuttonSelect.doClick();
	}

	/**
	 * @return a component appropriate to edit
	 *               and possibly change the name of the item.
	 */
	@Override
	public JComponent getBinaryEditor() {
		return null;
	}

	/**
	 * Defined to keep the size and location of the
	 * previous opening of the JFileChooser;
	 */
	private static class StaticSaveFileChooser extends JFileChooser {
		private JDialog _dialog;
		public StaticSaveFileChooser() {
			setDialogTitle(Ariane.getString("FileChooser.titleSaveFile"));
			setDialogType(JFileChooser.SAVE_DIALOG);
		}
		@Override
		protected JDialog createDialog( Component parent ) {
			JDialog dialog = super.createDialog(parent);
			if (_dialog != null) {
				dialog.setLocation(_dialog.getLocation());
				dialog.setSize(_dialog.getSize());
			}
			_dialog = dialog;
			return dialog;
		}
	}
}