/* -*- c-basic-offset: 3 -*-
 *
 * Copyright (c) 2013, GREYC.
 * All rights reserved
 *
 * You may use this file under the terms of the BSD license as follows:
 *
 * "Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of the GREYC, nor the name of its
 *     contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
 * 
 *   
 * For more information, refer to:
 * https://clouard.users.greyc.fr/Pandore
 */

package fr.greyc.pantheon.ariane.graph.item;

import fr.greyc.pantheon.ariane.Ariane;
import fr.greyc.pantheon.ariane.graph.Graph;
import fr.greyc.pantheon.ariane.graph.command.ChangeTextEdit;
import fr.greyc.pantheon.ariane.graph.exception.ArianeArithmeticException;
import fr.greyc.pantheon.ariane.ui.PreferenceChooser;
import fr.greyc.pantheon.ariane.utils.TreeItem;
import fr.greyc.pantheon.ariane.xml.XMLElement;

import java.awt.Color;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.FocusEvent;
import java.awt.event.KeyEvent;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.io.IOException;
import java.util.List;

import javax.swing.JComponent;
import javax.swing.JMenuItem;
import javax.swing.JPopupMenu;
import javax.swing.JTextArea;
import javax.swing.KeyStroke;
import javax.swing.text.AttributeSet;
import javax.swing.text.BadLocationException;
import javax.swing.text.PlainDocument;

/**
 * A <code>ScriptItem</code> implements a long text value.
 * 
 * @author Régis CLOUARD - Aug 24, 2012
 */
public final class ScriptItem extends ControlItem {
	/** The tag name of the item. */
	public static final String TAGNAME = "Script";

	/** The text to be displayed. */
	private JTextArea _scriptText;

	private boolean _visible;

	/**
	 * Creates a <code>ConstantItem</code> with the given
	 * attributes.
	 * Initializes the mouse listeners, the popup menu,
	 * and displays it in the graph.
	 * 
	 * @param treeItem the reference tree item.
	 * @param graph the parent graph.
	 * @param name the name of the item.
	 */
	public ScriptItem( TreeItem treeItem, Graph graph, String name ) {
		super(treeItem, graph, name);
		_visible = true;
		_scriptText.setVisible(true);
		setFormula(""); // IMPERATIVE to signify that it uses formula.
		setComment(getFormula());
		draw();
	}

	/**
	 * Creates a <code>ConstantItem</code> on the given
	 * graph with the given XML data.
	 * 
	 * @param graph the parent graph.
	 * @param elements the XML tree of data.
	 * @throws IOException XML errors.
	 */
	public ScriptItem( Graph graph, List<XMLElement> elements ) throws IOException {
		super(graph);
		setFormula(""); // IMPERATIVE to signify that it uses formula.
		load(elements);
		setComment(getFormula());
		draw();
	}

	/**
	 * Builds or changes the prototype of the control structure
	 * from the given values.
	 * Note that the number of inputs, outputs, and parameters are always 0.
	 * @param name the name of the item.
	 * @param iIm the number of input images.
	 * @param iP the number of parameters.
	 * @param oIm the number of output images.
	 */
	@Override
	protected void setPrototype( String name, int iIm, int iP, int oIm ) {
		setAbstractName(name);
	}

	@Override
	public boolean displayFormula() {
		return false;
	}

	///// 
	///// A C C E S S O R S
	/////

	/**
	 * @return the tag name.
	 */
	@Override
	public String getTagName( ) {
		return TAGNAME;
	}

	///// 
	///// E X E C U T I O N
	/////

	/**
	 * Executes the <code>Constant</code>, namely
	 * returns the expression.
	 * 
	 * @param isFailureFatalError is FAILURE a fatal error?
	 * @return true if the operator has been executed without errors.
	 */
	@Override
	protected boolean primitiveExecute( boolean isFailureFatalError ) {
		try {
			final String formula = getFormula();
			if (formula == null || formula.isEmpty()) {
				throw new ArianeArithmeticException(Ariane.getString("arithmetic.error19"));
			}
			getOutputResultSlot().setValue(formula);
			changeStatus(Status.EXECUTED);
			return true;
		} catch (ArianeArithmeticException e) {
			if (isFailureFatalError && !Graph.isExecutionStopped()) {
				Ariane.getWorkspace().printErrorMessage(e.getErrorCode(), getAbsoluteName(), getId(), e.getLocalizedMessage());
				getGraph().stopExecution();
				changeStatus(Status.ERROR);
			}
			return false;
		}
	}

	/**
	 * Builds the contextual menu as a Popup Menu.
	 * 
	 * @return a convenient popup menu.
	 */
	@Override
	protected JPopupMenu initPopup( ) {
		JPopupMenu popupMenu = new JPopupMenu();

		JMenuItem item;
		item = new JMenuItem(Ariane.getString("Graph.Menu.copy"));
		item.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed( ActionEvent event ) {
				Ariane.getWorkspace().copy();
			}
		});
		item.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_C, KeyEvent.CTRL_MASK));
		popupMenu.add(item);

		item = new JMenuItem(Ariane.getString("Graph.Menu.cut"));
		item.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed( ActionEvent event ) {
				Ariane.getWorkspace().cut();
			}
		});
		item.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_X, KeyEvent.CTRL_MASK));
		popupMenu.add(item);

		popupMenu.addSeparator();

		// Action: executer
		item = new JMenuItem(Ariane.getString("PopupMenu.execute"));
		item.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed( ActionEvent event ) {
				getGraph().execute();
			}
		});
		item.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_R, KeyEvent.CTRL_MASK));
		popupMenu.add(item);

		item = new JMenuItem(Ariane.getString("PopupMenu.reset"));
		item.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed( ActionEvent event ) {
				resetAll();
			}
		});
		item.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_U, KeyEvent.CTRL_MASK));
		popupMenu.add(item);

		// Edition: delete
		item = new JMenuItem(Ariane.getString("PopupMenu.remove"));
		item.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed( ActionEvent event ) {
				delete();
			}
		});
		item.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_DELETE, 0));
		popupMenu.add(item);

		return popupMenu;
	}

	/**
	 * A hook for customizing the comment editor.
	 * @return a Jlabel.
	 */
	@Override
	protected JComponent getCommentEditor() {
		// Override the class in order to refresh
		// the graph when the text area is modified.
		_scriptText = new JTextArea() {
			@Override
			public void repaint() {
				getGraph().repaint();
			}
		};
		_scriptText.setFont(PreferenceChooser.getDefaultFont());
		_scriptText.setBackground(Color.WHITE);
		_scriptText.setForeground(Color.BLACK);
		_scriptText.setDocument(new ScriptDocument());
		_scriptText.setTabSize(2);

		// Mouse events.
		_scriptText.addMouseListener(new MouseAdapter( ) {
			@Override
			public void mouseReleased( MouseEvent event ) {
				if (isSelected()) {
					_scriptText.getCaret().setVisible(true);
				} else {
					ScriptItem.this.onMouseReleased(event);
				}
			}
		});

		_scriptText.addFocusListener(this);
		_scriptText.setEditable(true);
		_scriptText.getCaret().setVisible(false);
		_scriptText.setVisible(_visible);

		return _scriptText;
	}

	/**
	 * Invoked when a component gains the keyboard focus.
	 */
	@Override
	protected void focusIn( ) {
		_scriptText.moveCaretPosition(_scriptText.getText().length());
	}

	/**
	 * Invoked when a JtextArea looses the keyboard focus.
	 * Change the text of the related element in the model.
	 * 
	 * @param e the FocusEvent.
	 */
	@Override
	public void focusLost( FocusEvent event ) {
		if (event.getSource() == _scriptText) {
			String newFormula = _scriptText.getText();
			if (!newFormula.equals(getFormula()) && isShowing()) {
				getGraph().startActionEdition();
				getGraph().validateAction(new ChangeTextEdit(this, _scriptText, _scriptText.getText(), getFormula()));
				getGraph().endActionEdition();
				setFormula(_scriptText.getText());
				getGraph().setModified(true);
			}
		}
		_scriptText.getCaret().setVisible(false);
		_scriptText.setCaretPosition(0);
	}

	/**
	 * Invoked when the mouse button has been
	 * double clicked (pressed and released) on a graph item.
	 * Shows the dialog box for setting the item's properties.
	 * 
	 * @param e the event attributes.
	 */
	@Override
	protected void onMouseDoubleClicked( MouseEvent event ) {
		_visible = !_visible;
		_scriptText.setVisible(_visible);
	}

	/**
	 * Defines the plain document that stores the string
	 * of the field.
	 */
	static class ScriptDocument extends PlainDocument {
		@Override
		public final void insertString( int offs, String str, AttributeSet attributeSet ) throws BadLocationException {
			final StringBuffer read = new StringBuffer(str);
			final StringBuffer filtered = new StringBuffer(getText(0, offs));
			String inserted = "";

			for ( int i = 0; i < read.length(); i++) {
				if (read.charAt(i) == '\n' && read.length() == 1) {
					String indentation = "";
					for ( int k = filtered.length() - 1; k >= 0; k--) {
						if (filtered.charAt(k) == '\n') {
							for ( int l = k + 1; l < filtered.length(); l++) {
								if (filtered.charAt(l) == ' ' || filtered.charAt(l) == '\t') {
									indentation += filtered.charAt(l);
								} else {
									break;
								}
							}
							break;
						}
					}
					String toBeInserted = read.charAt(i) + indentation;
					inserted += toBeInserted;
					filtered.append(toBeInserted);
				} else {
					inserted += read.charAt(i);
					filtered.append(read.charAt(i));
				}
			}
			super.insertString(offs, inserted, attributeSet);
		}
	}
}