/* -*- c-basic-offset: 3 -*-
 *
 * Copyright (c) 2013, GREYC.
 * All rights reserved
 *
 * You may use this file under the terms of the BSD license as follows:
 *
 * "Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of the GREYC, nor the name of its
 *     contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
 * 
 *   
 * For more information, refer to:
 * https://clouard.users.greyc.fr/Pandore
 */

package fr.greyc.pantheon.ariane.graph.item;

import fr.greyc.pantheon.ariane.Ariane;
import fr.greyc.pantheon.ariane.graph.Graph;
import fr.greyc.pantheon.ariane.graph.exception.ArianeException;
import fr.greyc.pantheon.ariane.graph.exception.ArianeMissingInputException;
import fr.greyc.pantheon.ariane.graph.exception.ArianeMissingParameterException;
import fr.greyc.pantheon.ariane.graph.exception.ArianeRuntimeException;
import fr.greyc.pantheon.ariane.graph.item.etc.ArgumentLineParser;
import fr.greyc.pantheon.ariane.graph.item.etc.OperatorProcessBuilder;
import fr.greyc.pantheon.ariane.graph.slot.InputDataSlot;
import fr.greyc.pantheon.ariane.graph.slot.InputParameterSlot;
import fr.greyc.pantheon.ariane.graph.slot.OutputDataSlot;
import fr.greyc.pantheon.ariane.utils.TreeItem;
import fr.greyc.pantheon.ariane.xml.XMLElement;

import java.awt.Color;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.FocusEvent;
import java.awt.event.FocusListener;
import java.awt.event.KeyEvent;
import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.util.LinkedList;
import java.util.List;

import javax.swing.JComponent;
import javax.swing.JMenuItem;
import javax.swing.JPopupMenu;
import javax.swing.JTextField;
import javax.swing.KeyStroke;

/**
 * An <code>SystemItem</code> is a graph item that
 * executes an operating system command. It stores the command line
 * and gets its parameters, inputs and outputs from slots.
 * @see GraphItem
 * 
 * @author Régis CLOUARD - Aug 25, 2012
 */
public final class SystemItem extends ControlItem {
	/** The tag name of the item. */
	public static final String TAGNAME = "System";

	/** The color of the item. */
	private static final Color DEFAULT_COLOR = new Color(0, 153, 255);

	/** The color of the item. */
	private static Color _color = DEFAULT_COLOR;

	/**
	 * Creates a new <code>CommandItem</code> on the specified
	 * graph with the given attribute values.
	 * 
	 * @param treeItem the source tree item.
	 * @param graph the graph that holds the item.
	 * @param name the default name of the item.
	 */
	public SystemItem( TreeItem treeItem, Graph graph, String name ) throws InterruptedException {
		super(treeItem, graph, name);
		askProperties(null, 1, 1, 1);
		setBinary("");
		draw();
	}

	/**
	 * Creates a <code>CommandItem</code> on the specified
	 * graph with the given XML data.
	 * 
	 * @param graph the parent graph.
	 * @param elements the XML tree of data.
	 * @throws IOException XML errors.
	 */
	public SystemItem( Graph graph, List<XMLElement> elements ) throws IOException {
		super(graph);
		load(elements);
		draw();
	}

	/**
	 * Builds the prototype of the command item from the given values.
	 */
	@Override
	protected void setPrototype( String name, int inputImageNumber, int inputParameterNumber, int outputImageNumber ) {
		setAbstractName(name);
		createSlots(inputImageNumber, inputParameterNumber, 0, outputImageNumber, 1);
		draw();
	}

	///// 
	///// P R E F E R E N C E S
	///// 
	
	// Note: theses methods cannot be generalized in the GraphItem class
	// because they use static attributes.

	/**
	 * Changes the current color.
	 * 
	 * @param color the new color.
	 */
	public static void setColor( Color color ) {
		_color = color;
	}

	/**
	 * Returns the current color.
	 * 
	 * @return the color.
	 */
	public static Color getColor( ) {
		return _color;
	}

	/**
	 * Returns the default color.
	 * 
	 * @return the color.
	 */
	public static Color getDefaultColor( ) {
		return DEFAULT_COLOR;
	}

	/**
	 * Returns the background color (same as geColor()).
	 * 
	 * @return the color.
	 */
	@Override
	public Color getBackgroundColor( ) {
		return _color;
	}

	///// 
	///// A C C E S S O R S
	/////

	/**
	 * @return the tag name.
	 */
	@Override
	public String getTagName( ) {
		return TAGNAME;
	}

	///// 
	///// E X E C U T I O N
	/////

	/**
	 * Builds the command line as specified in the command format
	 * and with the arguments. The binary name of the program
	 * is returns with the absolute pathname.
	 * Syntax: format string uses:
	 *  $p for parameter; $p0 for the first parameter
	 *  $m for the mask
	 *  $i for input; $i0 for the first input image
	 *  $o for output; $o0 for the first output image.
	 *  
	 * @param maskLine the mask value.
	 * @param parameterLine the parameter values.
	 * @param outputLine the output values.
	 * @return the command line.
	 */
	public List<String> buildAbstractCommandLine( List<String> maskLine, List<String> parameterLine, List<String> inputLine, List<String> outputLine ) {
		return ArgumentLineParser.parse("", getBinary(), maskLine, parameterLine, inputLine, outputLine);
	}

	/**
	 * Executes the <code>ScriptItem</code>, means
	 * compute a result from the arithmetic formula.
	 * 
	 * @param isFailureFatalError is FAILURE a fatal error?
	 * @return true if the operator has been executed without errors.
	 */
	@Override
	protected boolean primitiveExecute( boolean isFailureFatalError ) {
		try {
			List<String> parameterLine = new LinkedList<String>();
			// Get the parameter values
			for (InputParameterSlot slot : getInputParameterSlots()) {
				if (slot.hasValue(isFailureFatalError)) {
					parameterLine.add(slot.getValue());
				} else {
					throw new ArianeMissingParameterException(Ariane.getString("Operator.error1") + slot.getRank());
				}
			}

			// Get the input images
			List<String> inputLine = new LinkedList<String>();
			for (InputDataSlot slot: getInputDataSlots()) {
				if (slot.hasValue(isFailureFatalError)) {
					inputLine.add(slot.getValue());
				} else {
					throw new ArianeMissingInputException(Ariane.getString("Operator.error3") + slot.getRank());
				}
			}

			// Create outputs images as temporary files
			List<String> outputLine = new LinkedList<String>();
			for (OutputDataSlot slot : getOutputDataSlots()) {
				outputLine.add(slot.getTempName());
			}

			List<String> maskLine = new LinkedList<String>();
			List<String> commandLine = ArgumentLineParser.parse("", getBinary(), maskLine, parameterLine, inputLine, outputLine );

			String consoleMessage = "";
			for (int i = 0; i < commandLine.size(); i++ ) {
				consoleMessage += " " + commandLine.get(i);
			}
			Ariane.getWorkspace().printSystemMessage(consoleMessage);

			commandLine.clear();
			String osName = System.getProperty("os.name").toLowerCase();
			if (osName.startsWith("windows")) {
				commandLine.add("cmd");
				commandLine.add("/C");
				commandLine.add(consoleMessage);
			} else {
				String shell = System.getenv("SHELL");
				if (shell.length() == 0) {
					shell = "/bin/sh";
				}
				commandLine.add(shell);
				commandLine.add("-c");
				commandLine.add(consoleMessage);
			}

			// Execution of the operator binary
			changeStatus(Status.RUNNING);
			OperatorProcessBuilder pb = new OperatorProcessBuilder(commandLine);
			Process proc1 = pb.start();
			getGraph().setCurrentProcess(proc1);

			BufferedReader br = new BufferedReader(new InputStreamReader(proc1.getInputStream()));
			String line;
			while ((line = br.readLine()) != null) {
				Ariane.getWorkspace().printOutputMessage(line);
			}
			int exitValue = proc1.waitFor();
			getGraph().setCurrentProcess(null);

			String resultString = "";
			if (exitValue == 0) {
				// Get the status value
				if (getCommandStatus().trim().length() > 0) {
					pb = new OperatorProcessBuilder(GraphItem.getAbsoluteBinary(getCommandStatus()));
					Process proc2 = pb.start();
					InputStream resultStream = proc2.getInputStream();
					proc2.waitFor();

					// Exploit the results
					byte[] resultByte = new byte[resultStream.available() + 1];
					resultStream.read(resultByte);
					resultString = new String(resultByte);
					getOutputResultSlot().setValue(resultString.trim());
				} else {
					getOutputResultSlot().setValue("SUCCESS");
				}
			}
			// I mean if an execution fails than the error code !=0. 
			if (exitValue != 0 ) {
				throw new ArianeRuntimeException(Ariane.getString("Operator.error2"));
			}
			// Everything is OK -> validate the result of all outputs.
			for (OutputDataSlot slot : getOutputDataSlots()) {
				slot.setValue();
			}
			changeStatus(Status.EXECUTED);
			return true;
		} catch (ArianeRuntimeException e) {
			if (isFailureFatalError && !Graph.isExecutionStopped()) {
				Ariane.getWorkspace().printErrorMessage(Ariane.ErrorCode.RUNTIME, getAbsoluteName(), getId(), e.getLocalizedMessage());
				getGraph().stopExecution();
				changeStatus(Status.ERROR);
			} else {
				changeStatus(Status.UNEXECUTED);
			}
			return false;
		} catch (ArianeException e) {
			if (isFailureFatalError && !Graph.isExecutionStopped()) {
				Ariane.getWorkspace().printErrorMessage(e.getErrorCode(), getAbsoluteName(), getId(), e.getLocalizedMessage());
				getGraph().stopExecution();
				changeStatus(Status.ERROR);
			} else {
				changeStatus(Status.UNEXECUTED);
			}
			return false;
		} catch (IOException e) {
			if (isFailureFatalError && !Graph.isExecutionStopped()) {
				Ariane.getWorkspace().printErrorMessage(Ariane.ErrorCode.INPUTOUTPUT, getAbsoluteName(), getId(), e.getLocalizedMessage());
				getGraph().stopExecution();
				changeStatus(Status.ERROR);
			} else {
				changeStatus(Status.UNEXECUTED);
			}
			return false;
		} catch (InterruptedException e) {
			if (isFailureFatalError && !Graph.isExecutionStopped()) {
				Ariane.getWorkspace().printErrorMessage(Ariane.ErrorCode.INTERRUPTEDEXECUTION, getAbsoluteName(), getId(), e.getLocalizedMessage());
				getGraph().stopExecution();
				changeStatus(Status.ERROR);
			} else {
				changeStatus(Status.UNEXECUTED);
			}
			return false;
		}
	}

	///// 
	///// P R O P E R T I E S
	/////

	/**
	 * Builds the contextual menu as a Popup Menu.
	 * 
	 * @return a convenient popup menu.
	 */
	@Override
	protected JPopupMenu initPopup( ) {
		JPopupMenu popupMenu = new JPopupMenu();

		JMenuItem item;
		item = new JMenuItem(Ariane.getString("Graph.Menu.copy"));
		item.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed( ActionEvent event ) {
				Ariane.getWorkspace().copy();
			}
		});
		item.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_C, KeyEvent.CTRL_MASK));
		popupMenu.add(item);

		item = new JMenuItem(Ariane.getString("Graph.Menu.cut"));
		item.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed( ActionEvent event ) {
				Ariane.getWorkspace().cut();
			}
		});
		item.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_X, KeyEvent.CTRL_MASK));
		popupMenu.add(item);

		popupMenu.addSeparator();

		// Action: executer
		item = new JMenuItem(Ariane.getString("PopupMenu.execute"));
		item.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed( ActionEvent event ) {
				getGraph().execute();
			}
		});
		item.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_R, KeyEvent.CTRL_MASK));
		popupMenu.add(item);

		item = new JMenuItem(Ariane.getString("PopupMenu.reset"));
		item.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed( ActionEvent event ) {
				resetAll();
			}
		});
		item.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_U, KeyEvent.CTRL_MASK));
		popupMenu.add(item);

		// Edition: delete
		item = new JMenuItem(Ariane.getString("PopupMenu.remove"));
		item.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed( ActionEvent event ) {
				delete();
			}
		});
		item.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_DELETE, 0));
		popupMenu.add(item);

		return popupMenu;
	}

	/**
	 * @return a component appropriate to edit and possibly change the name of the item.
	 */
	@Override
	public JComponent getBinaryEditor() {
		if (getBinary() != null) {
			JTextField textField = new JTextField(getBinary());
			textField.addFocusListener(new FocusListener() {
				@Override
				public void focusLost( FocusEvent event ) {
					changeBinary((JTextField)event.getSource(), ((JTextField)event.getSource()).getText(), true);
					resetAll();
					getGraph().setModified(true);
				}
				@Override
				public void focusGained( FocusEvent event ) {
					/* Empty */
				}
			});
			textField.setEnabled(true);
			textField.setEditable(true);
			return textField;
		}
		return null;
	}
}