/* -*- c-basic-offset: 3 -*-
 *
 * Copyright (c) 2013, GREYC.
 * All rights reserved
 *
 * You may use this file under the terms of the BSD license as follows:
 *
 * "Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of the GREYC, nor the name of its
 *     contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
 * 
 *   
 * For more information, refer to:
 * https://clouard.users.greyc.fr/Pandore
 */

package fr.greyc.pantheon.ariane.graph.item;

import java.awt.Color;
import java.io.File;
import java.io.IOException;
import java.util.List;

import javax.swing.JPopupMenu;

import fr.greyc.pantheon.ariane.Ariane;
import fr.greyc.pantheon.ariane.graph.Graph;
import fr.greyc.pantheon.ariane.graph.SubGraph;
import fr.greyc.pantheon.ariane.graph.exception.ArianeException;
import fr.greyc.pantheon.ariane.graph.slot.InputDataSlot;
import fr.greyc.pantheon.ariane.graph.slot.InputParameterSlot;
import fr.greyc.pantheon.ariane.graph.slot.OutputDataSlot;
import fr.greyc.pantheon.ariane.graph.slot.OutputResultSlot;
import fr.greyc.pantheon.ariane.graph.slot.Slot;
import fr.greyc.pantheon.ariane.xml.XMLElement;

/**
 * A <code>VirtualItem</code> is used to reference
 * inputs and outputs of a <code>SubGraph</code>.
 * It is invisible as it has no graphical representation.
 *
 * @author Sébastien LEFÈVRE, Richard LOBRÉAU - Feb 21, 1999
 * @version Adrien SCHRÈQUE, Hicham AMADI - Nov 05, 2000
 * @version Mathieu JOETS - Jul 09, 2001
 * @version Régis CLOUARD - Oct 16, 2001
 * @version Regis CLOUARD - Jan 07, 2007
 */
public class VirtualItem extends GraphItem {
	/** Number of output images. */
	private final int _outputDataSlotNumber;

	/**
	 * Creates a VirtualOperator with the given attributes.
	 * Initializes the mouse listeners, the popup menu
	 * and displays the content in the graph.
	 * 
	 * @param graph the parent graph.
	 * @param name the name of the macro-operator.
	 * @param iIm the number of input images.
	 * @param iP the number of parameters.
	 * @param oIm the number of output images.
	 * @param iiIm the number of intermediate input images.
	 * @param iiP the number of intermediate input parameters.
	 */
	public VirtualItem( SubGraph graph, String name, int inputDataNumber, int inputParameterNumber, int outputDataNumber, int intermediateInputDataNumber, int intermediateInputParameterNumber ) {
		super(null, graph, name, null);
		setInternalId(0);       // IMPERATIVE = 0.
		_outputDataSlotNumber = outputDataNumber;
		createSlots(_outputDataSlotNumber + intermediateInputDataNumber,
				1 + intermediateInputParameterNumber,
				0,
				inputDataNumber + 1,
				inputParameterNumber);

		List<OutputDataSlot> inputDataSlots = getOutputDataSlots();
		for (int i = 0; i < inputDataNumber; i++) {
			OutputDataSlot slot = inputDataSlots.get(i);
			slot.setToolTipText(graph.getSuperGraphItem().getNickName() + Ariane.getString("VirtualOperator.input") + i);
			inputDataSlots.add(slot);
		}
		{
			OutputDataSlot slot = inputDataSlots.get(inputDataNumber);
			slot.setToolTipText(graph.getSuperGraphItem().getNickName() + Ariane.getString("VirtualOperator.mask"));
			inputDataSlots.add(slot);
		}

		List<InputDataSlot> outputDataSlots = getInputDataSlots();
		for (int i = 0; i < _outputDataSlotNumber; i++) {
			InputDataSlot slot = outputDataSlots.get(i);
			slot.setToolTipText(graph.getSuperGraphItem().getNickName() + Ariane.getString("VirtualOperator.output") + i);
			outputDataSlots.add(slot);
		}
		for (int i = _outputDataSlotNumber; i < _outputDataSlotNumber + intermediateInputDataNumber; i++) {
			InputDataSlot slot = outputDataSlots.get(i);
			int number = i - _outputDataSlotNumber;
			slot.setToolTipText(graph.getSuperGraphItem().getNickName() + Ariane.getString("VirtualOperator.reinput") + number);
			outputDataSlots.add(slot);
		}

		List<InputParameterSlot> outResultSlots = getInputParameterSlots();
		// OutResult
		{
			InputParameterSlot slot = outResultSlots.get(0);
			slot.setToolTipText(graph.getSuperGraphItem().getNickName() + Ariane.getString("VirtualOperator.result"));
			outResultSlots.add(slot);
		}

		// Intermediate parameters
		for (int i = 1; i < 1 + intermediateInputParameterNumber; i++) {
			InputParameterSlot slot = outResultSlots.get(i);
			int number = i + inputParameterNumber - intermediateInputParameterNumber - 1;
			slot.setToolTipText(graph.getSuperGraphItem().getNickName() + Ariane.getString("VirtualOperator.reparameter") + number); 
			outResultSlots.add(slot);
		}

		// Input parameters
		List<OutputResultSlot> inputParameterSlots = getOutputResultSlots();
		for (int i = 0; i < inputParameterNumber; i++) {
			OutputResultSlot slot = inputParameterSlots.get(i);
			slot.setToolTipText(graph.getSuperGraphItem().getNickName() + Ariane.getString("VirtualOperator.parameter") + i);
			inputParameterSlots.add(slot);
		}
		setFocusable(false);
	}

	/**
	 * Returns the identifier number of a VirtualOperator which is always 0.
	 * 
	 * @return the identifier number.
	 */
	@Override
	public final int getId( ) {
		return 0;
	}

	public final OutputDataSlot getVirtualInputDataSlot( int index ) {
		return getOutputDataSlots().get(index);
	}

	public final List<OutputDataSlot> getVirtualInputDataSlots( ) {
		return getOutputDataSlots();
	}

	public final List<InputDataSlot> getVirtualOutputDataSlots( ) {
		return getInputDataSlots();
	}

	public final List<OutputResultSlot> getVirtualInputParameterSlots( ) {
		return getOutputResultSlots();
	}

	public final InputDataSlot getVirtualOutputDataSlot( int index ) {
		return getInputDataSlots().get(index);
	}

	public final OutputResultSlot getVirtualInputParameterSlot( int index ) {
		return getOutputResultSlots().get(index);
	}

	public final InputParameterSlot getVirtualIntermediateResultSlot( int index ) {
		return getInputParameterSlots().get(1 + index);
	}

	public final Slot getVirtualInputMaskSlot( ) {
		return getOutputDataSlots().get(getOutputDataSlots().size() - 1);
	}

	public final Slot getVirtualOutputResultSlot( ) {
		return getInputParameterSlots().get(0);
	}

	public final InputDataSlot getVirtualIntermediateDataSlots( int index ) {
		return getInputDataSlots().get(index + _outputDataSlotNumber);
	}

	/**
	 * Resets the item and all its forwarded connected items
	 * and its parent.
	 */
	@Override
	public final void resetAll( ) {
		if (getExecutionStatus() != Status.UNEXECUTED) {
			// Overrides the graphitem reset in order to
			// avoid resetting the inter images and inter parameters
			// during a loop since a loop need to reset the
			// contents of the subgraph for each iterations.
			reset();
			if (getOutputDataSlots() != null) {
				for (OutputDataSlot slot: getOutputDataSlots()) {
					slot.reset();
				}
			}
			if (!getOutputResultSlots().isEmpty()) {
				getOutputResultSlot().reset();
			}
			// Its parent
			if (getGraph().getSuperGraphItem() != null) {
				getGraph().getSuperGraphItem().resetAll();
			}
		}
	}

	/**
	 * Executes the related subgraph and returns the
	 * status of the execution (always true).
	 * 
	 * @param isFailureFatalError is FAILURE is considered as an error?
	 * @return true if the virtual operator has been executed without errors.
	 */
	@Override
	protected final boolean primitiveExecute( boolean isFailureFatalError ) {
		CompositeItem item = (CompositeItem)getGraph().getSuperGraphItem();
		try {
			item.linkItemWithVirtualItem(this, isFailureFatalError);
			changeStatus(Status.EXECUTED);
			return true;
		} catch (ArianeException e) {
			changeStatus(Status.UNEXECUTED);
			return false;
		}
	}

	/**
	 * Saves a VirtualOperator as an XML tree.
	 * Does nothing!
	 * 
	 * @param file the current filename of the project.
	 * @return the XML tree containing the subgraph.
	 * @throws IOException XML errors.
	 */
	@Override
	public final XMLElement save( File file ) throws IOException {
		return null;
	}

	/**
	 * Never paint this item.
	 */
	@Override
	public final void draw( ) {
		// Do nothing
	}

	/**
	 * Returns the tag name.
	 * 
	 * @return a string.
	 */
	@Override
	public final String getTagName( ) {
		return null;
	}

	/**
	 * Returns the background color.
	 */
	@Override
	public final Color getBackgroundColor( ) {
		return Graph.getBackgroundColor();
	}

	@Override
	protected final JPopupMenu initPopup( ) {
		return null;
	}
}