/* -*- c-basic-offset: 3 -*-
 *
 * Copyright (c) 2013, GREYC.
 * All rights reserved
 *
 * You may use this file under the terms of the BSD license as follows:
 *
 * "Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of the GREYC, nor the name of its
 *     contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
 * 
 *   
 * For more information, refer to:
 * https://clouard.users.greyc.fr/Pandore
 */

package fr.greyc.pantheon.ariane.graph.item.etc;

import java.util.LinkedList;
import java.util.List;

/**
 * This class parses the template string for
 * the argument line of an operator.
 * Argument line uses : $m, $p, $i, $o $p0, $i1 ...[] wildcards.
 *
 * @author Régis CLOUARD - Dec 1, 2011
 */
public final class ArgumentLineParser {

	private ArgumentLineParser() { }

	public static List<String> parse( String command, String customFormat, List<String> maskLine, List<String> parameterLine, List<String> inputLine, List<String> outputLine ) {
		String format = customFormat.replaceAll("\\[", "\\[ ").replaceAll("\\]", " \\]");
		// for sake of compatibility: the Pandore default command line
		if (format.trim().isEmpty()) {
			format = " [$p] [-m $m] [$i] [$o]";
		}

		List<String> commandLine = new LinkedList<String>();

		// add the program name.
		commandLine.add(command);

		List<String> optionalCommandLine = null;
		List<String> temp = commandLine;
		String parsedText = "";
		boolean isSet = false;
		boolean isString = false;

		int index = 0;
		while (index < format.length()) {
			char character = format.charAt(index++);
			switch(character) {

			case '"': {
				isString = !isString;
				parsedText += "\"";
			} break;

			case ' ': { // space is a separator except if it in a string.
				if (isString) {
					parsedText += " ";
					break;
				} else {
					boolean end = false;
					while (!end && index < format.length()) {
						character = format.charAt(index++);
						if (character != ' ') {
							end = true;
							index--;
						}
					}
					temp.add(parsedText);
					parsedText = "";
				}
			} break;

			case '[': {
				optionalCommandLine = new LinkedList<String>();
				temp.add(parsedText);
				parsedText = "";
				temp = optionalCommandLine;
				isSet = false;
			} break;

			case ']': {
				if (isSet) {
					temp.add(parsedText);
					commandLine.addAll(temp);
				}
				parsedText = "";
				temp = commandLine;
			} break;

			case '$': {
				boolean end = false;
				while (!end && index < format.length()) {
					character = format.charAt(index++);
					switch(character) {
					case '$':
						parsedText += "$";
						end = true;
						break;

					case '{':
						break;
					case '}':
						break;

					case 'm': {
						isSet = !maskLine.isEmpty();
						temp.add(parsedText);
						parsedText = "";
						temp.addAll(maskLine);
					} break;

					case 'p': {
						if (index < format.length()) {
							character = format.charAt(index++);
							if (Character.isDigit(character)) {
								int position = character - '0';
								if (position < parameterLine.size()) {
									parsedText += parameterLine.get(position);
									isSet = true;
								}
							} else if (character == '}') {
								isSet = !parameterLine.isEmpty();
								temp.add(parsedText);
								parsedText = "";
								temp.addAll(parameterLine);
							} else {
								isSet = !parameterLine.isEmpty();
								temp.add(parsedText);
								parsedText = "";
								temp.addAll(parameterLine);
								index--;
							}
						} else {
							isSet = !parameterLine.isEmpty();
							temp.add(parsedText);
							parsedText = "";
							temp.addAll(parameterLine);
						}
					} break;
					case 'i': {
						if (index < format.length()) {
							character = format.charAt(index++);
							if (Character.isDigit(character)) {
								int position = character - '0';
								if (position < inputLine.size()) {
									parsedText += inputLine.get(position);
									isSet = true;
								}
							} else if (character == '}') {
								isSet = !inputLine.isEmpty();
								temp.add(parsedText);
								parsedText = "";
								temp.addAll(inputLine);
							} else {
								isSet = !inputLine.isEmpty();
								temp.add(parsedText);
								parsedText = "";
								temp.addAll(inputLine);
								index--;
							}
						} else {
							isSet = !inputLine.isEmpty();
							temp.add(parsedText);
							parsedText = "";
							temp.addAll(inputLine);
						}
					} break;
					case 'o': {
						if (index < format.length()) {
							character = format.charAt(index++);
							if (Character.isDigit(character)) {
								int position = character - '0';
								if (position < outputLine.size()) {
									parsedText += outputLine.get(position);
									isSet = true;
								}
							} else if (character == '}') {
								isSet = !outputLine.isEmpty();
								temp.add(parsedText);
								parsedText = "";
								temp.addAll(outputLine);
							} else {
								isSet = !outputLine.isEmpty();
								temp.add(parsedText);
								parsedText = "";
								temp.addAll(outputLine);
								index--;
							}
						} else {
							isSet = !outputLine.isEmpty();
							temp.add(parsedText);
							parsedText = "";
							temp.addAll(outputLine);
						}
					} break;
					default:
						index--;
						end = true;
					}
				}
			} break;
			default:
				parsedText += character;
			}
		}
		if (parsedText.length() > 0) {
			commandLine.add(parsedText);
		}

		// Remove extra spaces.
		LinkedList<String> newCommandLine = new LinkedList<String>();
		for (String commandPart : commandLine) { 
			if (!commandPart.trim().isEmpty()) {
				newCommandLine.add(commandPart.trim());
			}
		}
		return newCommandLine;
	}
}
