/* -*- c-basic-offset: 3 -*-
 *
 * Copyright (c) 2013, GREYC.
 * All rights reserved
 *
 * You may use this file under the terms of the BSD license as follows:
 *
 * "Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of the GREYC, nor the name of its
 *     contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
 * 
 *   
 * For more information, refer to:
 * https://clouard.users.greyc.fr/Pandore
 */

package fr.greyc.pantheon.ariane.graph.item.etc;

import fr.greyc.pantheon.ariane.Ariane;
import fr.greyc.pantheon.ariane.graph.exception.ArianeArithmeticException;

/**
 * A <code>ArithmeticExpression</code> is a calculus formula that includes
 * arithmetic operators, constant values and parameters values ($0, $1, etc.)
 *
 * @author Régis CLOUARD - Oct 16, 2001
 * @version Régis CLOUARD - Jan 23, 2007 (add: modulo operator)
 * @version Frédéric LAIR - Jun 28, 2007 (add: more than two parameters)
 */
public final class ArithmeticExpression {
	/** The rough formula. */
	private final String _initialExpression;

	/** The coded formula. */
	private byte[] _codedExpression;

	/** The stack of operations. */
	private double[] _stack;

	/** The table of constants used in the formula. */
	private double[] _constantStack;

	/** The codes of known arithmetic operations. */
	private static final byte PLUS = -1;
	/** The codes of known arithmetic operations. */
	private static final byte MINUS = -2;
	/** The codes of known arithmetic operations. */
	private static final byte TIMES = -3;
	/** The codes of known arithmetic operations. */
	private static final byte DIVIDE = -4;
	/** The codes of known arithmetic operations. */
	private static final byte MODULO = -5;
	/** The codes of known arithmetic operations. */
	private static final byte POWER = -6;
	/** The codes of known arithmetic operations. */
	private static final byte MIN = -11;
	/** The codes of known arithmetic operations. */
	private static final byte MAX = -12;
	/** The codes of known arithmetic operations. */
	private static final byte SIN = -21;
	/** The codes of known arithmetic operations. */
	private static final byte COS = -22;
	/** The codes of known arithmetic operations. */
	private static final byte TAN = -23;
	/** The codes of known arithmetic operations. */
	private static final byte ASIN = -24;
	/** The codes of known arithmetic operations. */
	private static final byte ACOS = -25;
	/** The codes of known arithmetic operations. */
	private static final byte ATAN = -26;
	/** The codes of known arithmetic operations. */
	private static final byte EXP = -27;
	/** The codes of known arithmetic operations. */
	private static final byte LN = -28;
	/** The codes of known arithmetic operations. */
	private static final byte LOG10 = -29;
	/** The codes of known arithmetic operations. */
	private static final byte LOG2 = -30;
	/** The codes of known arithmetic operations. */
	private static final byte ABS = -31;
	/** The codes of known arithmetic operations. */
	private static final byte SQRT = -32;
	/** The codes of known arithmetic operations. */
	private static final byte UNARYMINUS = -33;
	// Warning: -34 -> infinite is used for variables. 
	/** The code for all variables: first -34, then -35, etc. */
	private final byte[] _variables;

	/** The number of used parameter. */
	private final int _nbParam;

	/** The codes of known unary functions. */
	private final String[] _functionNamesUnary = {
			"sin", "cos", "tan", "asin",
			"acos", "atan", "exp",
			"ln", "log10", "log2", "abs", "sqrt" };

	/** The codes of known binary functions. */
	private final String[] _functionNamesBinary = { "min", "max" };

	private int _pos;

	private int _constantStackIndex;

	private int _codeSize;

	/**
	 * Creates an arithmetic expression from a formula.
	 * In case of error, generates an exception.
	 * 
	 * @param formula the rough formula.
	 * @param nbParam the number of parameters.
	 */
	public ArithmeticExpression( String formula, int nbParam ) {
		// Empty formula
		_nbParam = nbParam;
		_variables = new byte[_nbParam];
		_initialExpression = formula;
		if (formula == null || formula.trim().isEmpty()) {
			return;
		}
		_codedExpression = new byte[formula.length()];
		_constantStack = new double[formula.length()];
		for (int i = 0; i < _nbParam; i++) {
			_variables[i] = (byte) (-34 - i);
		}

		// Convert formula to computable code
		_pos = 0;
		_constantStackIndex = 0;
		_codeSize = 0;
		parseFormula();
		skip();
		if (next() != 0) {
			error(Ariane.getString("arithmetic.error1"));
		}
		int stackSize = computeStackUsage();
		_stack = new double[stackSize];
		byte[] code = new byte[_codeSize];
		System.arraycopy(_codedExpression, 0, code, 0, _codeSize);
		_codedExpression = code;
		double[] copy = new double[_constantStackIndex];
		System.arraycopy(_constantStack, 0, copy, 0, _constantStackIndex);
		_constantStack = copy;
	}

	/**
	 * Evaluates the current expression from the specified
	 * values of the two parameters.
	 * 
	 * @param val the list of value.
	 * @return the result.
	 * @throws ArianeArithmeticException in case of malformed expression.
	 */
	public double evaluate( double[] val ) {
		int top = 0;
		for (int i = 0; i < _codeSize; i++) {
			int var = 0;
			boolean flag = false;
			if (_codedExpression[i] >= 0) { // constants
				_stack[top++] = _constantStack[_codedExpression[i]];
			} else if (_codedExpression[i] >= POWER) { // binary operators
				double operande2 = _stack[--top];
				double operande1 = _stack[--top];
				double ans = Double.NaN;
				switch (_codedExpression[i]) {
				case PLUS: ans = operande1 + operande2;  break;
				case MINUS: ans = operande1 - operande2;  break;
				case TIMES: ans = operande1 * operande2;  break;
				case DIVIDE: ans = operande1 / operande2;  break;
				case MODULO: ans = operande1 % operande2;  break;
				case POWER: ans = Math.pow(operande1, operande2); break;
				default:
					break;
				}
				if (Double.isNaN(ans)) {
					return ans;
				}
				_stack[top++] = ans;
			} else {
				// parameters
				while (!flag && var < _nbParam) {
					if (_codedExpression[i] == _variables[var]) {
						flag = true;
						_stack[top++] = val[var];
					}
					var++;
				}
				// otherwise
				if (!flag) {
					double operande2;
					double operande1 = _stack[--top];
					double ans = Double.NaN;
					switch (_codedExpression[i]) {
					case MIN: operande2 = _stack[--top];
					ans = operande1 < operande2 ? operande1 : operande2;
					break;
					case MAX: operande2 = _stack[--top];
					ans = operande1 > operande2 ? operande1 : operande2;
					break;
					case SIN: ans = Math.sin(operande1);
					break;
					case COS: ans = Math.cos(operande1);
					break;
					case TAN: ans = Math.tan(operande1);
					break;
					case ASIN: if (Math.abs(operande1) <= 1.0) {
						ans = Math.asin(operande1);
					}
					break;
					case ACOS: if (Math.abs(operande1) <= 1.0) {
						ans = Math.acos(operande1); 
					}
					break;
					case ATAN: ans = Math.atan(operande1);
					break;
					case EXP: ans = Math.exp(operande1);
					break;
					case LN: if (operande1 > 0.0) {
						ans = Math.log(operande1);
					}
					break;
					case LOG2: if (operande1 > 0.0) {
						ans = Math.log(operande1) / Math.log(2);
					}
					break;
					case LOG10: if (operande1 > 0.0) { 
						ans = Math.log(operande1) / Math.log(10);
					}
					break;
					case ABS: ans = Math.abs(operande1);  break;
					case SQRT: if (operande1 >= 0.0) { ans = Math.sqrt(operande1); }
					break;
					case UNARYMINUS: ans = -operande1;
					break;
					default:
						break;
					}
					if (Double.isNaN(ans)) {
						return ans;
					}
					_stack[top++] = ans;
				}
			}
		}
		if (Double.isInfinite(_stack[0])) {
			throw new ArianeArithmeticException(Ariane.getString("arithmetic.error17"));
		} else {
			return _stack[0];
		}
	}

	/**
	 * Generates a exception error with the specified message.
	 * 
	 * @param message the message to be displayed.
	 */
	private void error( String message ) {
		throw new ArianeArithmeticException(Ariane.getString("arithmetic.error0.1") + " " + message + " " + Ariane.getString("arithmetic.error0.2") + " " + _pos + ")");
	}

	/**
	 * Calculates the maximum size of the stack.
	 * 
	 * @return the size.
	 */
	private int computeStackUsage( ) {
		int size = 0;
		int max = 0;
		for (int i = 0; i < _codeSize; i++) {
			if (_codedExpression[i] >= 0 || isVariable(i)) {
				size++;
				if (size > max) {
					max = size;
				}
			} else if (_codedExpression[i] >= POWER) {
				size--;
			}
		}
		return max;
	}

	/**
	 * tells if the component i from the coded expression is a parameter or not.
	 * 
	 * @param i the rank of the item we want to check.
	 * @return true if it is a parameter, false if not.
	 */
	private boolean isVariable( int rank ) {
		return _codedExpression[rank] <= -34 && _codedExpression[rank] > -34 - _nbParam;
	}

	/**
	 * @return the next current character in the formula.
	 */
	private char next() {
		return _pos >= _initialExpression.length() ? 0 : _initialExpression.charAt(_pos);
	}

	/**
	 * Skips white spaces.
	 */
	private void skip() {
		while (Character.isWhitespace(next())) {
			_pos++;
		}
	}

	/**
	 * Parses a mathematical expression such as:
	 * [+|-]Term[+|-]Term
	 */
	private void parseFormula( ) {
		boolean negative = false;
		skip();
		// Signe plus ou moins au debut de l'expression
		if (next() == '+' || next() == '-') {
			negative = next() == '-';
			_pos++;
			skip();
		}
		parseTerm();
		if (negative) {
			_codedExpression[_codeSize++] = UNARYMINUS;
		}
		skip();
		while (next() == '+' || next() == '-') {
			char op = next();
			_pos++;
			parseTerm();
			_codedExpression[_codeSize++] = (op == '+') ? PLUS : MINUS;
			skip();
		}
	}

	/**
	 * Parses an expression such as :
	 * Primary [*|/|^] Primary.
	 */
	private void parseTerm( ) {
		parsePrimary();
		skip();
		while (next() == '*' || next() == '/' || next() == '^' || next() == '%') {
			char op = next();
			_pos++;
			parsePrimary();
			if (op == '*') {
				_codedExpression[_codeSize++] = TIMES;
			} else if (op == '/') {
				_codedExpression[_codeSize++] = DIVIDE;
			} else if (op == '^') {
				_codedExpression[_codeSize++] = POWER;
			} else if (op == '%') {
				_codedExpression[_codeSize++] = MODULO;
			}
			skip();
		}
	}

	/**
	 * Parses an expression such as:
	 * [VARIABLE1 | VARIABLE2| FUNCTION | NUMBER | (FORMULA) ]
	 */
	private void parsePrimary( ) {
		skip();
		char ch = next();
		if (ch == '$') {
			_pos++;
			if ( Integer.valueOf(next()) < (48 + _nbParam) && Integer.valueOf(next()) >= 48) {
				_codedExpression[_codeSize++] = _variables[-48 + Integer.valueOf(next())];
			} else { //rajouter des conditions. Du genre while
				error(Ariane.getString("arithmetic.error2") + " " + ch + next());
			}
			_pos++;
		} else if (Character.isLetter(ch)) {
			parseFunction();
		} else if (Character.isDigit(ch) || ch == '.') {
			parseNumber();
		} else if (ch == '(') {
			_pos++;
			parseFormula();
			skip();
			if (next() != ')') {
				error(Ariane.getString("arithmetic.error3"));
			}
			_pos++;
		} else if (ch == ')') {
			error(Ariane.getString("arithmetic.error4"));
		} else if (ch == '+' || ch == '-' || ch == '*' || ch == '/' || ch == '^') {
			error(Ariane.getString("arithmetic.error5.1") + ch + Ariane.getString("arithmetic.error5.2"));
		} else if (ch == 0) {
			error(Ariane.getString("arithmetic.error6"));
		} else {
			error(Ariane.getString("arithmetic.error7.1") + ch + Ariane.getString("arithmetic.error7.2"));
		}
	}

	/**
	 * Parses a word that must be a function name.
	 */
	private void parseFunction( ) {
		String fonctionName = "";
		while (Character.isLetterOrDigit(next())) {
			fonctionName += next();
			_pos++;
		}
		fonctionName = fonctionName.toLowerCase();
		for (int i = 0; i < _functionNamesUnary.length; i++) {
			if (fonctionName.equals(_functionNamesUnary[i])) {
				skip();
				if (next() != '(') {
					error(Ariane.getString("arithmetic.error8.1`") + fonctionName + Ariane.getString("arithmetic.error8.2"));
				}
				_pos++;
				parseFormula();
				skip();
				if (next() != ')') {
					error(Ariane.getString("arithmetic.error9") + fonctionName + "'.");
				}
				_pos++;
				_codedExpression[_codeSize++] = (byte)(SIN - i);
				return;
			}
		}
		for (int i = 0; i < _functionNamesBinary.length; i++) {
			if (fonctionName.equals(_functionNamesBinary[i])) {
				skip();
				if (next() != '(') {
					error(Ariane.getString("arithmetic.error10.1") + fonctionName + Ariane.getString("arithmetic.error10.2"));
				}
				_pos++;
				parseFormula();
				skip();
				if (next() != ',') {
					error(Ariane.getString("arithmetic.error11") + fonctionName + "'.");
				}
				_pos++;
				parseFormula();
				skip();
				if (next() != ')') {
					error(Ariane.getString("arithmetic.error12") + fonctionName + "'.");
				}
				_pos++;
				_codedExpression[_codeSize++] = (byte)(MIN - i);
				return;
			}
		}
		error(Ariane.getString("arithmetic.error13") + fonctionName );
	}

	/**
	 * Parses a numerical value which must be a real value.
	 */
	private void parseNumber() {
		String numerical = "";
		while (Character.isDigit(next())) {
			numerical += next();
			_pos++;
		}
		if (next() == '.') {
			numerical += next();
			_pos++;
			while (Character.isDigit(next())) {
				numerical += next();
				_pos++;
			}
		}
		if (".".equals(numerical)) {
			error(Ariane.getString("arithmetic.error14"));
		}
		if (next() == 'E' || next() == 'e') {
			numerical += next();
			_pos++;
			if (next() == '+' || next() == '-') {
				numerical += next();
				_pos++;
			}
			if (!Character.isDigit(next())) {
				error(Ariane.getString("arithmetic.error15"));
			}
			while (Character.isDigit(next())) {
				numerical += next();
				_pos++;
			}
		}
		double value;
		try {
			value = Double.parseDouble(numerical);
		} catch (NumberFormatException e) {
			value = Double.NaN;
		}
		if (Double.isNaN(value)) {
			error(Ariane.getString("arithmetic.error16.1") + numerical + Ariane.getString("arithmetic.error16.2"));
		}
		_codedExpression[_codeSize++] = (byte)_constantStackIndex;
		_constantStack[_constantStackIndex++] = value;
	}

	/**
	 * @return the string representation of the expression.
	 */
	@Override
	public String toString() {
		return _initialExpression;
	}
}