/* -*- c-basic-offset: 3 -*-
 *
 * Copyright (c) 2013, GREYC.
 * All rights reserved
 *
 * You may use this file under the terms of the BSD license as follows:
 *
 * "Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of the GREYC, nor the name of its
 *     contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
 * 
 *   
 * For more information, refer to:
 * https://clouard.users.greyc.fr/Pandore
 */

package fr.greyc.pantheon.ariane.graph.item.etc;

import fr.greyc.pantheon.ariane.Ariane;
import fr.greyc.pantheon.ariane.graph.exception.ArianeArithmeticException;

/**
 * A <code>LogicalExpression</code> is defined from a
 * boolean formula that includes logical operators and
 * and the two parameters values ($0 and $1).
 *
 * @author Régis CLOUARD - Oct 16, 2001
 * @version Régis CLOUARD - Dec 10, 2004
 */
public final class LogicalExpression {
	/** The rough formula. */
	private final String _initialExpression;

	/** The coded formula. */
	private byte[] _codedExpression;

	/** The stack of operations. */
	private final boolean[] _stack;

	/** The table of constants used in the formula. */
	private double[] _constantStack;

	/** The codes of known boolean operations. */
	private static final byte GT = -1;
	/** The codes of known boolean operations. */
	private static final byte LT = -2;
	/** The codes of known boolean operations. */
	private static final byte GE = -3;
	/** The codes of known boolean operations. */
	private static final byte LE = -4;
	/** The codes of known boolean operations. */
	private static final byte EQ = -5;
	/** The codes of known boolean operations. */
	private static final byte NEQ = -6;
	/** The codes of known boolean operations. */
	private static final byte AND = -7;
	/** The codes of known boolean operations. */
	private static final byte OR = -8;
	/** The codes of known boolean operations. */
	private static final byte XOR = -9;
	/** The codes of known boolean operations. */
	private static final byte NOT = -10;
	/** The codes of known boolean operations. */
	private static final byte VARIABLE1 = -22;
	/** The codes of known boolean operations. */
	private static final byte VARIABLE2 = -23;

	/** Internal variables. */
	private int _pos;
	/** Internal variables. */
	private int _constantStackIndex;
	/** Internal variables. */
	private int _codeSize;

	/**
	 * Creates a boolean expression from a formula.
	 * In case of error, generates an exception.
	 * 
	 * @param formula the rough formula.
	 */
	public LogicalExpression( String formula ) {
		// No formula...
		_initialExpression = formula;
		if (formula == null || formula.trim().equals("")) {
			_stack = null;
			return;
		}

		_codedExpression = new byte[formula.length()];
		_constantStack = new double[formula.length()];
		// Transforme la formule en codes evaluables.
		_pos = 0;
		_constantStackIndex = 0;
		_codeSize = 0;

		parseFormula();
		skip();
		if (next() != 0) {
			error(Ariane.getString("boolean.error1"));
		}
		// Minimization de l'espace occupe.
		int stackSize = computeStackUsage();
		_stack = new boolean[stackSize];
		byte[] code = new byte[_codeSize];
		System.arraycopy(_codedExpression, 0, code, 0, _codeSize);
		_codedExpression = code;
		double[] tmp = new double[_constantStackIndex];
		System.arraycopy(_constantStack, 0, tmp, 0, _constantStackIndex);
		_constantStack = tmp;
	}

	/**
	 * Evaluates the current expression from the specified
	 * values of the two parameters.
	 * 
	 * @param val1 the value of the 1st parameter,
	 * @param val2 the value of the 2nd parameter.
	 * @return the result.
	 * @throws ArianeArithmeticException in case of arithmetic errors.
	 */
	public boolean evaluate( double val1, double val2 ) {
		// La pile des valeurs numeriques.
		double[] doubleStack = new double[3];
		int doubleTop = 0;
		try {
			int top = 0;
			for (int i = 0; i < _codeSize; i++) {
				if (_codedExpression[i] >= 0) {
					doubleStack[doubleTop++] = _constantStack[_codedExpression[i]];
				} else if (_codedExpression[i] == NOT) {
					boolean operand1 = _stack[--top];
					_stack[top++] = !operand1;
				} else if (_codedExpression[i] > XOR && _codedExpression[i] <= AND) {
					boolean operand2 = _stack[--top];
					boolean operand1 = _stack[--top];
					boolean ans = false;
					switch (_codedExpression[i]) {
					case AND:   ans = operand1 && operand2;
					break;
					case OR:    ans = operand1 || operand2;
					break;
					case XOR:   ans = (operand1 || operand2) && !(operand1 && operand2);
					break;
					default:
						break;
					}
					_stack[top++] = ans;
				} else if (_codedExpression[i] >= NEQ && _codedExpression[i] <= GT) {
					double operand2 = doubleStack[--doubleTop];
					double operand1 = doubleStack[--doubleTop];
					boolean ans = false;
					switch (_codedExpression[i]) {
					case GT:    ans = operand1 > operand2;  break;
					case GE:    ans = operand1 >= operand2;  break;
					case LT:    ans = operand1 < operand2;  break;
					case LE:    ans = operand1 <= operand2;  break;
					case EQ: ans = operand1 == operand2;  break;
					case NEQ:  ans = operand1 != operand2;  break;
					default:
						break;
					}
					_stack[top++] = ans;
				} else if (_codedExpression[i] == VARIABLE1) {
					doubleStack[doubleTop++] = val1;
				} else if (_codedExpression[i] == VARIABLE2) {
					doubleStack[doubleTop++] = val2;
				}
			}
		} catch (Exception e) {
			throw new ArianeArithmeticException(e.getLocalizedMessage());
		}
		return _stack[0];
	}   

	/**
	 * Generates a exception error with the specified message.
	 * 
	 * @param message the message to be displayed.
	 */
	private void error( String message ) { 
		throw new ArianeArithmeticException(Ariane.getString("boolean.error0.1") + " " + message + Ariane.getString("boolean.error0.2") + " " + _pos + ")");
	}

	/**
	 * Calculates the maximum size of the stack.
	 * 
	 * @return the size.
	 */
	private int computeStackUsage( ) {
		int size = 0;   // Pile pour chaque operation.
		int max = 0; // Taille maximum observee.
		for (int i = 0; i < _codeSize; i++) {
			if (_codedExpression[i] >= 0 || _codedExpression[i] == VARIABLE1 || _codedExpression[i] == VARIABLE2) {
				size++;
				if (size > max) {
					max = size;
				}
			}
		}
		return max;
	}

	/**
	 * @return the next current character in the expression.
	 */
	private char next( ) { 
		return _pos >= _initialExpression.length() ? 0 : _initialExpression.charAt(_pos);
	}

	/**
	 * Skips white spaces.
	 */
	private void skip( ) {
		while (Character.isWhitespace(next())) {
			_pos++;
		}
	}

	/**
	 * Parses a boolean expression such as:
	 * [!]Term [ '&' | '|' | '^' ] [!]Term
	 */
	private void parseFormula( ) {
		boolean negative = false;
		skip();
		// Signe not au debut de l'expression
		if (next() == '!' ) {
			negative = true;
			_pos++;
			skip();
		}
		parseTerm();
		if (negative) {
			_codedExpression[_codeSize++] = NOT;
		}
		skip();
		while (next() == '&' || next() == '|' || next() == '^') {
			char code = next();
			_pos++;
			negative = false;
			if (next() == '!')  {
				negative = true;
				_pos++;
				skip();
			}
			parseTerm();
			if (negative) {
				_codedExpression[_codeSize++] = NOT;
			}
			if (code == '&') {
				_codedExpression[_codeSize++] = AND;
			} else if (code == '|') {
				_codedExpression[_codeSize++] = OR;
			} else if (code == '^') {
				_codedExpression[_codeSize++] = XOR;
			}
			skip();
		}
	}

	/**
	 * Parses a boolean expression such as:
	 * <pre>
	 * TERM :: NUMERIC | (FORMULA)
	 *</pre>
	 */
	private void parseTerm( ) {
		skip();
		char ch = next();
		// VARIABLE
		if (ch == '$') {
			parsePrimary();
		// NUMBER
		} else if (Character.isDigit(ch) || ch == '.') {
			parsePrimary();
		// (FORMULA)
		} else if (ch == '(') {
			_pos++;
			parseFormula();
			skip();
			if (next() != ')') {
				error(Ariane.getString("boolean.error2"));
			}
			_pos++;
		} else if (ch == ')') {
			error(Ariane.getString("boolean.error3"));
		} else if (ch == '&' || ch == '|' || ch == '^' || ch == '=' || ch == '!') {
			error(Ariane.getString("boolean.error4.1") + ch + Ariane.getString("boolean.error4.2"));
		} else if (ch == 0) {
			error(Ariane.getString("boolean.error5"));
		} else {
			error(Ariane.getString("boolean.error6.1") + ch + Ariane.getString("boolean.error6.2"));
		}
	}

	/**
	 * Parses a boolean expression such as:
	 * <pre>
	 * PRIMARY :: NUMERIC '>' | '<' | '<=' | '>=' | '==' | '!='  NUMERIC
	 *</pre>
	 */
	private void parsePrimary( ) {
		parseNumeric();
		skip();
		if (next() == '<' || next() == '>' || next() == '=' || next() == '!') {
			char op = next();
			byte code = LE;
			_pos++;
			if (next() == '=') {
				_pos++;
				if (op == '<') {
					code = LE;
				} else if (op == '>') {
					code = GE;
				} else if (op == '=') {
					code = EQ;
				} else { // if (op == '!')
					code = NEQ;
				}
			} else {
				if (op == '<') {
					code = LT;
				} else if (op == '>') {
					code = GT;
				} else {
					error(Ariane.getString("boolean.error7"));
				}
			}
			parseNumeric();
			_codedExpression[_codeSize++] = code;
		} else {
			error(Ariane.getString("boolean.error8"));
		}
	}

	/**
	 * Parses a boolean expression such as:
	 * <pre>
	 * NUMERIC :: VARIABLE1 | VARIABLE2| NUMBER
	 * ou VARIABLE :: $0 | $1
	 *</pre>
	 */
	private void parseNumeric( ) {
		skip();
		char ch = next();
		// VARIABLE
		if (ch == '$') {
			_pos++;
			if (next() == '0') {
				_codedExpression[_codeSize++] = VARIABLE1;
			} else if (next() == '1') {
				_codedExpression[_codeSize++] = VARIABLE2;
			} else {
				error(Ariane.getString("boolean.error9") + " " + ch + _pos);
			}
			_pos++;
		} else if (Character.isDigit(ch) || ch == '.') { // NUMBER
			parseNumber();
		} else {
			error(Ariane.getString("boolean.error10"));
		}
	}

	/**
	 * Parses a numerical value which must be a real value.
	 */
	private void parseNumber( ) {
		String number = "";
		while (Character.isDigit(next())) {
			number += next();
			_pos++;
		}
		if (next() == '.') {
			number += next();
			_pos++;
			while (Character.isDigit(next())) {
				number += next();
				_pos++;
			}
		}
		if (".".equals(number)) {
			error(Ariane.getString("boolean.error11"));
		}
		if (next() == 'E' || next() == 'e') {
			number += next();
			_pos++;
			if (next() == '+' || next() == '-') {
				number += next();
				_pos++;
			}
			if (!Character.isDigit(next())) {
				error(Ariane.getString("boolean.error12"));
			}
			while (Character.isDigit(next())) {
				number += next();
				_pos++;
			}
		}
		double value;
		try {
			value = Double.parseDouble(number);
		} catch (NumberFormatException e) {
			value = Double.NaN;
		}
		if (Double.isNaN(value)) {
			error(Ariane.getString("boolean.error13.1")
					+ number
					+ Ariane.getString("boolean.error13.2"));
		}
		_codedExpression[_codeSize++] = (byte)_constantStackIndex;
		_constantStack[_constantStackIndex++] = value;
	}

	/**
	 * Transforms text string to html string.
	 * For example, converts < to &lt;
	 * 
	 * @return the modified string.
	 */
	public String toHtml() {
		return _initialExpression.replaceAll("<", "&lt;");
	}

	/**
	 * @return the string representation of the logical expression.
	 */
	@Override
	public String toString() {
		return _initialExpression;
	}
}