/* -*- c-basic-offset: 3 -*-
 *
 * Copyright (c) 2013, GREYC.
 * All rights reserved
 *
 * You may use this file under the terms of the BSD license as follows:
 *
 * "Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of the GREYC, nor the name of its
 *     contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
 * 
 *   
 * For more information, refer to:
 * https://clouard.users.greyc.fr/Pandore
 */

package fr.greyc.pantheon.ariane.graph.item.etc;

import fr.greyc.pantheon.ariane.Ariane;
import fr.greyc.pantheon.ariane.graph.exception.ArianeArithmeticException;

/**
 * A <code>StringExpression</code> allows string manipulation.
 *
 * @author Régis CLOUARD - Apr 26, 2011
 */
public final class StringExpression {
	/** The rough formula. */
	private final String _initialExpression;

	/** The coded formula. */
	private byte[] _codedExpression;

	/** The stack of string operands. */
	private String[] _stack;

	/** The table of constants used in the formula. */
	private String[] _constantStack;

	/** The codes of known arithmetic operations. */
	private static final byte PLUS = -1;
	/** The codes of known arithmetic operations. */
	private static final byte MINUS = -2;
	/** The codes of known arithmetic operations. */
	private static final byte POWER = -6;
	/** The codes of known arithmetic operations. */
	private static final byte MIN = -11;
	/** The codes of known arithmetic operations. */
	private static final byte QUOTE = -21;
	// Warning: -34 -> infinite is used for variables. 
	/** The code for all variables: first -34, then -35, etc. */
	private final byte[] _variables;

	/** The number of used parameter. */
	private final int _nbParam;

	/** The codes of known unary functions. */
	private final String[] _functionNamesUnary = { "quote" };

	/** The codes of known binary functions. */
	private final String[] _functionNamesBinary = { "concatenate", "substring" };

	private int _pos;

	private int _constantStackIndex;

	private int _codeSize;

	/**
	 * Creates a string from string.
	 * In case of error, generates an exception.
	 * 
	 * @param formula the rough formula.
	 * @param nbParam the number of parameters.
	 */
	public StringExpression( String formula, int nbParam ) {
		// Empty formula
		_nbParam = nbParam;
		_variables = new byte[_nbParam];
		_initialExpression = formula;
		if (formula == null || formula.trim().isEmpty()) {
			return;
		}
		for (int i = 0; i < _nbParam; i++) {
			_variables[i] = (byte) (-34 - i);
		}
		_codedExpression = new byte[formula.length()];
		_constantStack = new String[formula.length()];

		// Convert formula to computable code
		_pos = 0;
		_constantStackIndex = 0;
		_codeSize = 0;
		parseFormula();
		skip();
		if (next() != 0) {
			error(Ariane.getString("arithmetic.error1"));
		}
		int stackSize = computeStackUsage();
		_stack = new String[stackSize];
		byte[] code = new byte[_codeSize];
		System.arraycopy(_codedExpression, 0, code, 0, _codeSize);
		_codedExpression = code;
		String[] tmp = new String[_constantStackIndex];
		System.arraycopy(_constantStack, 0, tmp, 0, _constantStackIndex);
		_constantStack = tmp;
	}

	/**
	 * Evaluates the current expression from the specified
	 * values of the two parameters.
	 * 
	 * @param val the list of value.
	 * @return the result.
	 * @throws ArianeArithmeticException in case of malformed expression.
	 */
	public String evaluate( String[] val ) {
		int top = 0;
		boolean flag = false; // used for while condition
		int var = 0;
		for (int i = 0; i < _codeSize; i++) {
			var = 0;
			flag = false;
			if (_codedExpression[i] >= 0) { // constants
				_stack[top++] = _constantStack[_codedExpression[i]];
			} else if (_codedExpression[i] >= POWER) { // binary operators
				String op2 = _stack[--top];
				String op1 = _stack[--top];
				String ans = null;
				if (_codedExpression[i] == PLUS) {
					ans = op1 + op2;
				}
				if (ans == null) {
					return ans;
				}
				_stack[top++] = ans;
			} else {
				// parameters
				while (!flag && var < _nbParam) {
					if (_codedExpression[i] == _variables[var]) {
						flag = true;
						_stack[top++] = val[var];
					}
					var++;
				}
				// otherwise
				if (!flag) {
					//						String op2;
					String op1 = _stack[--top];
					String ans = null;
					if (_codedExpression[i] == QUOTE) {
						ans = "'" + op1 + "'"; 
					} else {
						return null;
					}
					_stack[top++] = ans;
				}
			}
		}
		return _stack[0];
	}

	/**
	 * Generates a exception error with the specified message.
	 * 
	 * @param message the message to be displayed.
	 */
	private void error( String message ) {
		throw new ArianeArithmeticException(Ariane.getString("arithmetic.error0.1") + " " + message + " " + Ariane.getString("arithmetic.error0.2") + " " + _pos + ")");
	}

	/**
	 * Calculates the maximum size of the stack.
	 * 
	 * @return the size.
	 */
	private int computeStackUsage( ) {
		int size = 0;
		int max = 0;
		for (int i = 0; i < _codeSize; i++) {
			if (_codedExpression[i] >= 0 || isVariable(i)) {
				size++;
				if (size > max) {
					max = size;
				}
			} else if (_codedExpression[i] >= POWER) {
				size--;
			}
		}
		return max;
	}

	/**
	 * Tells if the specified component from the coded expression is a parameter or not.
	 * 
	 * @param rank the rank of the item we want to check.
	 * @return true if it is a parameter, false if not.
	 */
	private boolean isVariable( int rank ) {
		return _codedExpression[rank] <= -34 && _codedExpression[rank] > -34 - _nbParam;
	}

	/**
	 * @return the next current character in the formula.
	 */
	private char next( ) {
		return _pos >= _initialExpression.length() ? 0 : _initialExpression.charAt(_pos);
	}

	/**
	 * Skips white spaces.
	 */
	private void skip( ) {
		while (Character.isWhitespace(next())) {
			_pos++;
		}
	}

	/**
	 * Parses a mathematical expression such as:
	 * [+|-]Term[+|-]Term
	 */
	private void parseFormula( ) {
		skip();
		parseTerm();
		skip();
		while (next() == '+' || next() == '-') {
			char op = next();
			_pos++;
			parseTerm();
			_codedExpression[_codeSize++] = (op == '+') ? PLUS : MINUS;
			skip();
		}
	}

	/**
	 * Parses an expression such as:
	 * [VARIABLE1 | VARIABLE2| FUNCTION | NUMBER | (FORMULA) ]
	 */
	private void parseTerm( ) {
		skip();
		char ch = next();
		if (ch == '$') {
			_pos++;
			if (Integer.valueOf(next()) < (48 + _nbParam) && Integer.valueOf(next()) >= 48) {
				_codedExpression[_codeSize++] = _variables[-48 + Integer.valueOf(next())];
			} else { //rajouter des conditions. Du genre while
				error(Ariane.getString("arithmetic.error2") + " " + ch + next());
			}
			_pos++;
		} else if (Character.isLetter(ch)) {
			parseFunction();
		} else if (ch == '\"') {
			parseString();
		} else if (ch == ')') {
			error(Ariane.getString("arithmetic.error4"));
		} else {
			error(Ariane.getString("arithmetic.error7.1") + ch + Ariane.getString("arithmetic.error7.2"));
		}
	}

	/**
	 * Parses a word that must be a function name.
	 */
	private void parseFunction( ) {
		String functionName = "";
		while (Character.isLetterOrDigit(next())) {
			functionName += next();
			_pos++;
		}
		functionName = functionName.toLowerCase();
		for (int i = 0; i < _functionNamesUnary.length; i++) {
			if (functionName.equals(_functionNamesUnary[i])) {
				skip();
				if (next() != '(') {
					error(Ariane.getString("arithmetic.error8.1`") + functionName + Ariane.getString("arithmetic.error8.2"));
				}
				_pos++;
				parseFormula();
				skip();
				if (next() != ')') {
					error(Ariane.getString("arithmetic.error9") + functionName + "'.");
				}
				_pos++;
				_codedExpression[_codeSize++] = (byte)(QUOTE - i);
				return;
			}
		}
		for (int i = 0; i < _functionNamesBinary.length; i++) {
			if (functionName.equals(_functionNamesBinary[i])) {
				skip();
				if (next() != '(') {
					error(Ariane.getString("arithmetic.error10.1") + functionName + Ariane.getString("arithmetic.error10.2"));
				}
				_pos++;
				parseFormula();
				skip();
				if (next() != ',') {
					error(Ariane.getString("arithmetic.error11") + functionName + "'.");
				}
				_pos++;
				parseFormula();
				skip();
				if (next() != ')') {
					error(Ariane.getString("arithmetic.error12") + functionName + "'.");
				}
				_pos++;
				_codedExpression[_codeSize++] = (byte)(MIN - i);
				return;
			}
		}
		error(Ariane.getString("arithmetic.error13") + functionName );
	}

	/**
	 * Parses a numerical value which must be a real value.
	 */
	private void parseString( ) {
		String value = "";
		_pos++;
		while (next() != '\"') {
			value += next();
			_pos++;
		}
		_pos++;
		_codedExpression[_codeSize++] = (byte)_constantStackIndex;
		_constantStack[_constantStackIndex++] = value;
	}

	/**
	 * @return the string representation of the expression.
	 */
	@Override
	public String toString() {
		return _initialExpression;
	}
}