/* -*- c-basic-offset: 3 -*-
 *
 * Copyright (c) 2013, GREYC.
 * All rights reserved
 *
 * You may use this file under the terms of the BSD license as follows:
 *
 * "Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of the GREYC, nor the name of its
 *     contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
 * 
 *   
 * For more information, refer to:
 * https://clouard.users.greyc.fr/Pandore
 */

package fr.greyc.pantheon.ariane.graph.slot;

import fr.greyc.pantheon.ariane.Ariane;
import fr.greyc.pantheon.ariane.graph.command.ConnectItemEdit;
import fr.greyc.pantheon.ariane.graph.command.DisconnectItemEdit;
import fr.greyc.pantheon.ariane.graph.exception.ArianeSlotException;
import fr.greyc.pantheon.ariane.graph.item.GraphItem;
import fr.greyc.pantheon.ariane.xml.XMLElement;

import java.awt.Color;
import java.awt.Graphics;
import java.io.IOException;

/**
 * The class <code>InputSlot</code> defines input connection slot.
 * Only one <code>OutputSlot</code> can be connected to one
 * <code>InputSlot</code>. It is a drop site for an <code>OutputSlot</code>.
 * Three types of input slots are available:
 * <ul>
 * <li>@see InputdataSlot
 * <li>@see InputMaskSlot
 * <li>@see InputParameterSlot
 * </ul>
 * 
 * @author Régis CLOUARD - Oct 01, 1998
 * @version Sébastien LEFÈVRE, Richard LOBRÉAU - Feb 21, 1999
 * @version Adrien SCHRÈQUE, Hicham AMADI - Nov 05, 2000
 * @version Mathieu JOETS - Jul 09, 2001
 * @version Régis CLOUARD - Oct 16, 2001
 * @version Frédéric LAIR - May 23, 2007
 * @version Régis CLOUARD - 2007-11-24 (restructuration)
 */
public abstract class InputSlot extends Slot {
	/** The connected output slot. */
	private transient OutputSlot _connectedSlot;

	/** The link between this and the connected slot. */
	private transient Link _link;

	/**
	 * Creates an input slot on the given graph item.
	 * 
	 * @param item the graph item that will hold the slot.
	 * @param rank the rank in the list of input slot of the graph item.
	 */
	public InputSlot( GraphItem item, int rank ) {
		super(item, rank);
	}

	/**
	 * @return the link of the slot.
	 */
	public final Link getLink( ) {
		return _link;
	}

	protected final void setLink( Link link ) {
		_link = link;
	}

	@Override
	public final void drop( Slot from ) {
		setNormalBorder();
		try {
			OutputSlot outputSlot = (OutputSlot)from;
			// check if the slot are in the same graph!
			if (getGraph() != outputSlot.getGraph()) {
				throw new ArianeSlotException(Ariane.getString("Slot.error3"));
			}
			// check slots consistency.
			boolean connect = this instanceof InputParameterSlot && outputSlot instanceof OutputResultSlot
					|| this instanceof InputDataSlot && outputSlot instanceof OutputDataSlot
					|| this instanceof InputMaskSlot && outputSlot instanceof OutputDataSlot;
			if (connect) {
				getGraph().startActionEdition();
				outputSlot.connectTo(this, null, true);
				getGraph().endActionEdition();
				getGraph().requestFocusInWindow();
			} else {
				throw new ArianeSlotException(Ariane.getString("Slot.error2"));
			}
		} catch (ArianeSlotException e) {
			Ariane.getWorkspace().printErrorMessage(Ariane.getString("Slot.error1") + e.getMessage());
		} catch (ClassCastException e) {
			Ariane.getWorkspace().printErrorMessage(Ariane.getString("Slot.error1") + Ariane.getString("Slot.error2"));
		}
	}

	///// 
	///// C O N N E C T I O N   /   D I S C O N N E C T I O N 
	/////

	/**
	 * Connects the slot with the given slot.
	 *<br>
	 * 'Passive' method. First of all, disconnects the slot if needed.
	 * Then calls the connect method of the given slot.
	 * 
	 * @param outputSlot the slot to be connected.
	 */
	protected final void connectFrom( OutputSlot outputSlot ) {
		if (!outputSlot.equals(_connectedSlot)) {
			disconnect();
			outputSlot.connectTo(this, null, true);
		}
	}

	public final OutputSlot getConnectedSlot( ) {
		return _connectedSlot;
	}

	@Override
	public final boolean isConnected( ) {
		return _connectedSlot != null;
	}

	/**
	 * Connects the slot with the given slot,
	 * or disconnects the slot if the given slot is null.
	 * Resets the related graph item.
	 * 
	 * @param slot the slot to be connected or null.
	 * @param link a regular link or null.
	 * @param historical true if the action must be stored in the action
	 *        history (for use with undo/redo).
	 */
	protected final void setConnectedSlot( OutputSlot slot, Link link, boolean historical ) {
		getGraphItem().resetAll();
		getGraphItem().getGraph().setModified(true);
		if (slot == null) { // case of disconnection.
			setDisconnectedIcon();
			if (historical) {
				getGraph().validateAction(new DisconnectItemEdit(getGraph(), this, _connectedSlot));
			}
			_link.dispose();
			getGraphItem().fireConnection(this, false);
		} else { // case of connection
			setConnectedIcon();
			if (link != null ) {
				_link = link;
			} else {
				_link = createLink();
			}
			if (historical) {
				getGraph().validateAction(new ConnectItemEdit(getGraph(), this, slot));
			}
			getGraphItem().fireConnection(this, true);
		}
		_connectedSlot = slot;
	}

	@Override
	public final void disconnectAll() {
		disconnect();
	}

	/**
	 * Disconnects the slot.
	 * 'passive' method : calls the disconnection method
	 * of the connected slot.
	 */
	public final void disconnect( ) {
		if (_connectedSlot != null) {
			_connectedSlot.disconnectFrom(this, true);
			disposeLink();
		}
	}

	/**
	 * Selects the slot and the connected slot.
	 * @param isSelected true if it is selected, false otherwise.
	 */
	@Override
	public final void setSelected( boolean isSelected ) {
		if (isConnected()) {
			highlightSlot(isSelected);
			_connectedSlot.highlightSlot(isSelected);
		}
	}

	protected abstract Link createLink();

	/**
	 * Disposes the link.
	 */
	public void disposeLink( ) { 
		/* empty */
	}

	/**
	 * Returns the current color of that link type.
	 * 
	 * @return the user defined color.
	 */
	protected abstract Color getLinkColor( );

	/**
	 * Draws the link between the slot and its connected slot.
	 * (Try a sophisticated algorithm).
	 * 
	 * @param gc the graphic context.
	 */
	public final void drawLink( Graphics gc ) {
		if (isConnected()) {
			_link.draw(gc, getLinkColor(), isSelected());
		}
	}

	/**
	 * Saves the slot and its values as an XML tree.
	 * 
	 * @return the root of the XML tree.
	 * @throws IOException in case of output errors.
	 */
	public abstract XMLElement save( ) throws IOException;

}