/* -*- c-basic-offset: 3 -*-
 *
 * Copyright (c) 2013, GREYC.
 * All rights reserved
 *
 * You may use this file under the terms of the BSD license as follows:
 *
 * "Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of the GREYC, nor the name of its
 *     contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
 * 
 *   
 * For more information, refer to:
 * https://clouard.users.greyc.fr/Pandore
 */

package fr.greyc.pantheon.ariane.graph.slot;


import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Cursor;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Point;
import java.awt.RenderingHints;
import java.awt.Stroke;
import java.awt.geom.Line2D;
import java.awt.geom.Point2D;
import java.awt.geom.QuadCurve2D;
import java.util.ArrayList;
import java.util.List;

/**
 * A <code>Link</code> is a connection between two slots.
 * It is represented as a polyline, namely a series of straight lines.
 *
 * @author Régis CLOUARD - Oct 16, 2001
 * @version Frédéric LAIR - May 16, 2007
 * @version Régis CLOUARD - Nov 2007-11-24 (restructuration)
 * @version Régis CLOUARD - Nov 28, 2010 (highlight links of selected items)
 */
public abstract class Link {
	/** The predefined stroke for regular link. */
	private static final BasicStroke REGULAR_LINE_STROKE = new BasicStroke(2.0F, BasicStroke.CAP_ROUND, BasicStroke.JOIN_ROUND);
	/** The predefined stroke for selected link. */
	private static final BasicStroke SELECTED_LINE_STRKE = new BasicStroke(4.0F, BasicStroke.CAP_ROUND, BasicStroke.JOIN_ROUND);
	/** Minimum gap between the link and the slot coordinate when moving link. */
	protected static final int GAP = 10;
	/** The target of the link (either an input image, a mask, or a parameter). */
	private final InputSlot _targetSlot;
	/** True if the user a point to the link. */
	private boolean _userDefined;
	/** The polyline that represents a link. */
	private Line2D.Float[] _polyline;
	/** The current selected part of the link. */
	private int _selectedLine = -1;
	/** True if the link is selected. */
	private boolean _selected;

	/**
	 * Creates a new link with the given input slot
	 * as the source.
	 * 
	 * @param slot the input slot that supports this slot.
	 */
	public Link( InputSlot slot ) {
		_targetSlot = slot;
	}

	/**
	 * Deletes permanently the link.
	 */
	protected final void dispose( ) {
		_targetSlot.setSelected(false);
	}

	/**
	 * Defines the link from a series of intermediate points.
	 * 
	 * @param path a series of intermediate points.
	 * @param userDefined the state of the path (defined by user or not).
	 *        this parameter is kept for compatibility reasons;
	 */
	protected final void setPath( List<Point> path, boolean userDefined ) {
		if (path.isEmpty()) {
			_userDefined = false;
		} else {
			_userDefined = userDefined;
			buildPolyline(path);
		}
	}

	protected final Line2D.Float[] getPolyline() {
		return _polyline;
	}

	/**
	 * Returns the current link as a series of intermediate points
	 * if this line is reshaped by the user.
	 * If the link is the default link, namely not reshaped by the user,
	 * the method returns an empty array of size 0.
	 * 
	 * @return the current link as a series of intermediate points.
	 */
	protected final List<Point> getPath( ) {
		ArrayList<Point> path = new ArrayList<Point>();
		if (_userDefined) {
			for (Line2D.Float line : _polyline) {
				path.add(new Point((int)line.x1, (int)line.y1));
			}
			path.add(new Point((int)_polyline[_polyline.length - 1].x2, (int)_polyline[_polyline.length - 1].y2));
		}
		return path;
	}

	/**
	 * @return true if the user added a point to the link.
	 */
	protected final boolean isUserDefined( ) {
		return _userDefined;
	}

	/**
	 * Sets the value of user defined.
	 * 
	 * @param value the new value.
	 */
	protected final void setUserDefined( boolean value ) {
		_userDefined = value;
	}

	/**
	 * @return true if the link has only be moved and not reshaped.
	 */
	private boolean hasMoved() {
		int dx1 = (int)(_targetSlot.getConnectedSlot().getAnchor().getX() - _targetSlot.getAnchor().getX());
		int dx2 = (int)(_polyline[0].x1 - _polyline[_polyline.length - 1].x2);
		int dy1 = (int)(_targetSlot.getConnectedSlot().getAnchor().getY() - _targetSlot.getAnchor().getY());
		int dy2 = (int)(_polyline[0].y1 - _polyline[_polyline.length - 1].y2);
		return dx1 == dx2 && dy1 == dy2;
	}

	private boolean isModified() {
		return _polyline[0].x1 != _targetSlot.getConnectedSlot().getAnchor().getX()
				|| _polyline[0].y1 != _targetSlot.getConnectedSlot().getAnchor().getY()
				|| _polyline[_polyline.length - 1].x2 != _targetSlot.getAnchor().getX()
				|| _polyline[_polyline.length - 1].y2 != _targetSlot.getAnchor().getY();
	}
	/**
	 * Moves the link keeping the same relative position
	 * with regards to the  slots.
	 */
	private void movePolyline() {
		final double dx = _targetSlot.getConnectedSlot().getAnchor().getX() - _polyline[0].x1;
		final double dy = _targetSlot.getConnectedSlot().getAnchor().getY() - _polyline[0].y1;
		for (Line2D.Float line : _polyline) {
			line.x1 += dx;
			line.x2 += dx;
			line.y1 += dy;
			line.y2 += dy;
		}
	}

	/**
	 * Modifies the link because one the source or target
	 * items has been moved.
	 */
	protected abstract boolean modifyPolyline();

	/**
	 * Returns the current polyline that represents
	 * the link or creates a fresh new one.
	 * 
	 * @return the different line parts.
	 */
	private void buildPolyline( ) {
		if (_polyline == null) {
			// create a fresh polyline.
			buildPolyline(createPolyline());
		} else if (isModified()) {
			if (hasMoved()) {
				movePolyline();
			} else {
				if (!_userDefined || !modifyPolyline()) {
					// -> create a fresh new polyline.
					buildPolyline(createPolyline());
					_userDefined = false;
				}
			}
		}
	}

	/**
	 * Creates the list of lines from the specified
	 * list of points.
	 * 
	 * @param path the list of intermediate points.
	 */
	private void buildPolyline( List<Point> path ) {
		_polyline = new Line2D.Float[path.size() - 1];
		// create the lines from the intermediate points
		for (int i = 0; i < path.size() - 1; i++) {
			final Point point1 = path.get(i);
			final Point point2 = path.get(i + 1);
			_polyline[i] = new Line2D.Float((float)point1.getX(), (float)point1.getY(), (float)point2.getX(), (float)point2.getY());
		}
	}

	/**
	 * Builds the polyline that corresponds to the link.
	 *
	 * @return the series of straight lines.
	 */
	protected abstract List<Point> createPolyline( );

	/**
	 * Returns the rank of the closest line of the polyline
	 * to the specified coordinate.
	 *
	 * @param coordinate a point.
	 * @return the rank of the line within the polyline.
	 */
	private int getRank( Point coordinate ) {
		int rank = 0;
		double distance = Double.MAX_VALUE;
		for (int i = 0; i < _polyline.length; i++) {
			if (_polyline[i].ptSegDist(coordinate.x, coordinate.y) < distance) {
				distance = _polyline[i].ptSegDist(coordinate.x, coordinate.y);
				rank = i;
			}
		}
		return rank;
	}

	/**
	 * Returns the closest distance to the link. Checks each
	 * part of the link and returns the closed distance.
	 * 
	 * @param x the x coordinate of the mouse pointer.
	 * @param y the y coordinate of the mouse pointer.
	 * @return the distance of the link.
	 */
	public final double distance( double x, double y ) {
		if (_polyline == null) {
			return 1e30;
		}
		double distance = Double.MAX_VALUE;
		for (Line2D.Float line : _polyline) {
			if (line.ptSegDist(x, y) < distance) {
				distance = line.ptSegDist(x, y);
			}
		}
		return distance;
	}

	/**
	 * Draws the link between the target slot and its connected
	 * source slot with the specified color.
	 * 
	 * @param gc the graphic context.
	 * @param color the color.
	 * @param isSelected true if the related graph item is selected.
	 */
	protected final void draw( Graphics gc, Color color, boolean isSelected ) {
		Graphics2D gc2D = (Graphics2D)gc;

		// store old stroke.
		final Stroke oldStroke = gc2D.getStroke();
		gc2D.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
		gc2D.setRenderingHint(RenderingHints.KEY_RENDERING, RenderingHints.VALUE_RENDER_QUALITY);
		gc2D.setStroke(isSelected || _selected ? SELECTED_LINE_STRKE : REGULAR_LINE_STROKE);
		gc2D.setPaint(color);


		buildPolyline();
		double gap = 2.0;
		Point2D p0;
		Point2D p1;
		Point2D pi1 = new Point2D.Double();
		Line2D line = new Line2D.Double();
		QuadCurve2D curve = new QuadCurve2D.Double();

		if (_polyline[0].getP1().getY() == _polyline[_polyline.length - 1].getP2().getY()) {
			line.setLine(_polyline[0].getP1(), _polyline[_polyline.length - 1].getP2());
			gc2D.draw(line);
		} else if (_polyline[0].getP1().getX() == _polyline[_polyline.length - 1].getP2().getX()) {
			line.setLine(_polyline[0].getP1(), _polyline[_polyline.length - 1].getP2());
			gc2D.draw(line);
		} else {
			for (int i = 0; i < _polyline.length; i++) {
				p0 = _polyline[i].getP1();
				p1 = _polyline[i].getP2();

				// start point
				if (i != 0) {
					p1 = _polyline[i].getP2();
					if (p0.getY() == p1.getY()) {
						if (p0.getX() < p1.getX()) {
							p0.setLocation(Math.min(p1.getX(), p0.getX() + gap), p0.getY());
						} else {
							p0.setLocation(Math.max(p1.getX(), p0.getX() - gap), p0.getY());
						}
					} else {
						if (p0.getY() > p1.getY()) {
							p0.setLocation(p0.getX(), Math.max(p1.getY(), p0.getY() - gap));
						} else {
							p0.setLocation(p0.getX(), Math.min(p1.getY(), p0.getY() + gap));
						}
					}
				}

				// end point
				if (i < _polyline.length - 1) {
					if (p0.getY() == p1.getY()) {
						if (p0.getX() < p1.getX()) {
							p1.setLocation(Math.max(p0.getX(), p1.getX() - gap), p1.getY());
						} else {
							p1.setLocation(Math.min(p0.getX(), p1.getX() + gap), p1.getY());
						}
					} else {
						if (p0.getY() > p1.getY()) {
							p1.setLocation(p1.getX(), Math.min(p0.getY(), p1.getY() + gap));
						} else {
							p1.setLocation(p1.getX(), Math.max(p0.getY(), p1.getY() - gap));
						}
					}
				}
				if (i > 0) {
					curve.setCurve(pi1, _polyline[i].getP1(), p0);
					gc2D.draw(curve);
				}
				line.setLine(p0, p1);
				gc2D.draw(line);
				pi1.setLocation(p1);
			}
		}
		gc2D.setStroke(oldStroke);
	}

	/**
	 * Returns the cursor that corresponds to that the
	 * only allowed direction of link reshape
	 * (vertically or horizontally) depending of the current
	 * selected line in the polyline.
	 * 
	 * @return the convenient cursor.
	 */
	public final Cursor getCursor( ) {
		if (_selectedLine > 0) {
			if (_polyline[_selectedLine].x1 == _polyline[_selectedLine].x2) {
				return Cursor.getPredefinedCursor(Cursor.W_RESIZE_CURSOR);
			} else {
				return Cursor.getPredefinedCursor(Cursor.N_RESIZE_CURSOR);
			}
		}
		return null;
	}

	/**
	 * Sets the new state of the link. If the new state is true
	 * then selects the link according to the specified coordinate
	 * of the mouse pointer, otherwise it is unselected.
	 * 
	 * @param state the new state.
	 * @param xycoord the coordinate of the mouse pointer.
	 */
	public final void setSelected( boolean state, Point xycoord ) {
		if (state) {
			_selectedLine = getRank(xycoord);
			if (_selectedLine == 0 || _selectedLine == _polyline.length - 1) {
				_selectedLine = -1; // We only consider intermediate lines.
			}
			_selected = true;
			_targetSlot.setSelected(true);
		} else {
			_selectedLine = -1;
			_selected = false;
			_targetSlot.setSelected(false);
		}
	}

	/**
	 * Moves the current selected intermediate line
	 * of the link according to the new coordinates.
	 * 
	 * @param xycoord the new coordinates.
	 */
	public abstract void move( Point xycoord);

	/**
	 * @return the current selected line rank.
	 */
	public final int getSelectedLine() {
		return _selectedLine;
	}

	/**
	 * Moves the specified intermediate line
	 * of the link according to the new coordinates.
	 * 
	 * @param lineRank the rank of the line within the link.
	 * @param xycoord the new coordinates.
	 */
	public final void move( int lineRank, Point xycoord) {
		_selectedLine = lineRank;
		move(xycoord);
		_selectedLine = -1;
	}

	protected final Point getSourceCoordinate() {
		return _targetSlot.getConnectedSlot().getAnchor();
	}

	protected final Point getTargetCoordinate() {
		return _targetSlot.getAnchor();
	}
}