/* -*- c-basic-offset: 3 -*-
 *
 * Copyright (c) 2013, GREYC.
 * All rights reserved
 *
 * You may use this file under the terms of the BSD license as follows:
 *
 * "Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of the GREYC, nor the name of its
 *     contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
 * 
 *   
 * For more information, refer to:
 * https://clouard.users.greyc.fr/Pandore
 */

package fr.greyc.pantheon.ariane.graph.slot;

import java.awt.Point;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.File;
import java.io.IOException;
import java.util.ArrayList;

import javax.swing.ImageIcon;
import javax.swing.JButton;

import fr.greyc.pantheon.ariane.Ariane;
import fr.greyc.pantheon.ariane.graph.item.DataFile;
import fr.greyc.pantheon.ariane.graph.item.GraphItem;
import fr.greyc.pantheon.ariane.graph.item.etc.ButtonImage;
import fr.greyc.pantheon.ariane.graph.item.etc.FileImagePreviewer;
import fr.greyc.pantheon.ariane.graph.item.etc.OperatorProcessBuilder;
import fr.greyc.pantheon.ariane.ui.PreferenceChooser;
import fr.greyc.pantheon.ariane.ui.Workspace;
import fr.greyc.pantheon.ariane.xml.XMLElement;

/**
 * The class <code>OutputDataSlot</code> defines output image slot.
 * It can be connected to several <code>InputDataSlot</code>.
 * An <code>OutputDataSlot</code> can be set with a data file.
 *
 * @author Régis CLOUARD - Oct 01, 1998
 * @version Sébastien LEFÈVRE, Richard LOBRÉAU - Feb 21, 1999
 * @version Adrien SCHRÈQUE, Hicham AMADI - Nov 05, 2000
 * @version Mathieu JOETS - Jul 09, 2001
 * @version Régis CLOUARD - Oct 16, 2001
 */
public class OutputDataSlot extends OutputSlot {
	/** The connected rendering of the slot. */
	private static final transient ImageIcon CONNECTED_ICON;

	/** The disconnected rendering  of the slot. */
	private static final transient ImageIcon DISCONNECTED_ICON;

	/** The name of the related image or file. */
	private String _value;

	static {
		CONNECTED_ICON = Ariane.getIcon("DataSlotOn.png");
		DISCONNECTED_ICON = Ariane.getIcon("DataSlotOff.png");
	}

	/**
	 * Creates an output image slot on the given graph item.
	 * 
	 * @param item the graph item that will hold the slot.
	 * @param rank the rank in the list of input slot of the graph item.
	 */
	public OutputDataSlot( GraphItem item, int rank ) {
		super(item, rank);
		setIcon(DISCONNECTED_ICON);
	}

	@Override
	public final String toString( ) {
		if (getGraphItem() instanceof DataFile) {
			return "OutputDataSlot: " + getGraphItem().getBinary();
		}
		if (_value != null) {
			return "OutputDataSlot: " + _value;
		}
		return "OutputDataSlot: " + getTempName();
	}

	@Override
	protected final void setConnectedIcon( ) {
		setIcon(CONNECTED_ICON);
	}

	@Override
	protected final void setDisconnectedIcon( ) {
		setIcon(DISCONNECTED_ICON);
	}

	/**
	 * Returns the anchor coordinates of the slot.
	 * 
	 * @return anchor point.
	 */
	@Override
	public final Point getAnchor( ) {
		return new Point((int)(getLocation().getX() + getWidth() + getGraphItem().getLocation().getX() + 3),
				(int)( getLocation().getY() + getHeight() / 2 + getGraphItem().getLocation().getY()));
	}

	private String createTempFilename() {
		return PreferenceChooser.createTemporaryName(getGraph().getIdentifier(), getGraphItem().getId(), getRank());
	}

	/**
	 * Creates the complete filename for a temporary file.
	 * <br>Uses the value of temporary directory defined
	 * in the preferences.
	 * <br><i>Does not use the java method <tt>createTempFile</tt>
	 * because it creates the filename in the /temp directory.
	 * 
	 * @return the complete filename.
	 */
	public final String getTempName( ) {
		return Workspace.getTempDir() + createTempFilename();
	}

	/**
	 * Returns the name of the image or file where
	 * output data are stored.
	 * 
	 * @return the filename.
	 */
	public final String getValue( ) {
		return _value;
	}

	/**
	 * Sets a new temporary filename
	 * as the new result value.
	 */
	public final void setValue( ) {
		_value = getTempName();
		setValidated();
	}

	/**
	 * Set the given filename as the new result value.
	 * 
	 * @param fileName the new result.
	 */
	public final void setValue( String fileName ) {
		_value = fileName;
		setValidated();
	}

	/**
	 * Creates the button that allows to display the result.
	 * 
	 * @return the JButton.
	 */
	public final JButton getButton( ) {
		JButton button = new ButtonImage(this, Ariane.getString("OutSlotImage.display"));
		button.addActionListener(new ActionListener( ) {
			@Override
			public void actionPerformed( ActionEvent event ) {
				showDataFileContents();
			}
		});
		return button;
	}

	/**
	 * Displays the result by using the visualization software
	 * given in the preferences.
	 */
	public final void showDataFileContents( ) {
		if (isValidate()) {
			try {
				ArrayList<String> commandLine = new ArrayList<String>();
				File file = new File(getValue());
				if (file.exists() && FileImagePreviewer.isText(file, false)) {
					String command = PreferenceChooser.get(PreferenceChooser.TEXT_EDITOR);
					if (command != null && !command.isEmpty()) {
						commandLine.add(command);
					} else {
						return;
					}
				} else {
					commandLine.add(GraphItem.getAbsoluteBinary(GraphItem.getCommandVisu()));
				}
				commandLine.add(getValue());
				OperatorProcessBuilder pb = new OperatorProcessBuilder(commandLine);
				pb.start();
			} catch (IOException e) {
				Ariane.getWorkspace().printErrorMessage(Ariane.getString("Slot.error5") + getValue());
			}
		}
	}

	@Override
	protected final boolean isAcceptableDropSite( Slot slot ) {
		return slot.getGraphItem() != this.getGraphItem()
				&& (slot instanceof InputDataSlot || slot instanceof InputMaskSlot);
	}

	///// 
	///// I N P U T   /   O U T P U T
	///// 

	/**
	 * Saves the slot with its values as an XML tree.
	 * 
	 * @param e the XML parent in the tree. 
	 * @return the root of the XML tree.
	 */
	@Override
	public final XMLElement save( XMLElement element ) throws IOException {
		element.addChildren(new XMLElement("StartItemId", Integer.toString(getGraphItem().getId())));
		element.addChildren(new XMLElement("StartSlotType", "OutSlotImage"));
		element.addChildren(new XMLElement("StartSlotIndex", Integer.toString(getRank())));
		return element;
	}
}