/* -*- c-basic-offset: 3 -*-
 *
 * Copyright (c) 2013, GREYC.
 * All rights reserved
 *
 * You may use this file under the terms of the BSD license as follows:
 *
 * "Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of the GREYC, nor the name of its
 *     contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
 * 
 *   
 * For more information, refer to:
 * https://clouard.users.greyc.fr/Pandore
 */

package fr.greyc.pantheon.ariane.graph.slot;

import java.awt.Point;
import java.io.IOException;

import javax.swing.ImageIcon;

import fr.greyc.pantheon.ariane.Ariane;
import fr.greyc.pantheon.ariane.graph.item.GraphItem;
import fr.greyc.pantheon.ariane.xml.XMLElement;

/**
 * The class <code>OutputResultSlot</code> defines result slot.
 * <br>Result value can be :
 * <ul>
 * <li>SUCCESS
 * <li>FAILURE
 * <li>a float
 * <li>a string.
 * </ul>
 * <br>It can be connected to several <code>InputParameterSlot</code>.
 *
 * @author Régis CLOUARD - Oct 01, 1998
 * @version Sébastien LEFÈVRE, Richard LOBRÉAU - Feb 21, 1999
 * @version Adrien SCHRÈQUE, Hicham AMADI - Nov 05, 2000
 * @version Mathieu JOETS - Jul 09, 2001
 * @version Régis CLOUARD - Oct 16, 2001
 * @version Régis CLOUARD - Nov 24, 2004
 */
public final class OutputResultSlot extends OutputSlot {
	/** The result value  */
	private String _result;

	/** The connected rendering of the slot. */
	private static final transient ImageIcon CONNECTED_ICON;

	/** The disconnected rendering  of the slot. */
	private static final transient ImageIcon DISCONNECTED_ICON;

	static {
		CONNECTED_ICON = Ariane.getIcon("ValueSlotOn.png");
		DISCONNECTED_ICON = Ariane.getIcon("ValueSlotOff.png");
	}

	/**
	 * Creates an output result slot on the given graph item.
	 * 
	 * @param item the graph item that will hold the slot.
	 * @param rank the rank in the input slots list of the graph item.
	 */
	public OutputResultSlot( GraphItem item, int rank ) {
		super(item, rank);
		setIcon(DISCONNECTED_ICON);
	}

	@Override
	public String toString( ) {
		return "OutputResultSlot: " + getValue();
	}

	@Override
	protected void setConnectedIcon( ) {
		setIcon(CONNECTED_ICON);
	}

	@Override
	protected void setDisconnectedIcon( ) {
		setIcon(DISCONNECTED_ICON);
	}

	/**
	 * Returns the anchor coordinates of the slot.
	 * 
	 * @return anchor point.
	 */
	@Override
	public Point getAnchor( ) {
		return new Point((int)(getLocation().getX() + getWidth() / 2 + getGraphItem().getLocation().getX()),
				(int)(getLocation().getY() + getHeight() + getGraphItem().getLocation().getY()));
	}

	///// 
	/////  V A L U E 
	/////

	/**
	 * Sets the given result as the new result.
	 * For boolean value (isNUmeric = FALSE),
	 * value 0 corresponds to FAILURE and 1 to SUCCESS.
	 * 
	 * @param result the string that contains the result.
	 */
	public void setValue( String result ) {
		_result = result;
		setValidated();
	}

	/**
	 * Returns the current value.
	 * 
	 * @return a String.
	 */
	public String getValue() {
		return _result;
	}

	@Override
	protected boolean isAcceptableDropSite( Slot slot ) {
		return slot.getGraphItem() != this.getGraphItem() && slot instanceof InputParameterSlot;
	}

	/**
	 * Saves the slot with its values as an XML tree.
	 * 
	 * @param e the XML parent in the tree. 
	 * @return the root of the XML tree.
	 */
	@Override
	public XMLElement save( XMLElement element ) throws IOException {
		element.addChildren(new XMLElement("StartItemId", Integer.toString(getGraphItem().getId())));
		element.addChildren(new XMLElement("StartSlotType", "OutSlotResult"));
		element.addChildren(new XMLElement("StartSlotIndex", Integer.toString(getRank())));
		return element;
	}
}