/* -*- c-basic-offset: 3 -*-
 *
 * Copyright (c) 2013, GREYC.
 * All rights reserved
 *
 * You may use this file under the terms of the BSD license as follows:
 *
 * "Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of the GREYC, nor the name of its
 *     contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
 * 
 *   
 * For more information, refer to:
 * https://clouard.users.greyc.fr/Pandore
 */

package fr.greyc.pantheon.ariane.graph.slot;

import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

import fr.greyc.pantheon.ariane.Ariane;
import fr.greyc.pantheon.ariane.graph.exception.ArianeSlotException;
import fr.greyc.pantheon.ariane.graph.item.GraphItem;
import fr.greyc.pantheon.ariane.xml.XMLElement;

/**
 * The class <code>OutputSlot</code> defines an output connection slot.
 * Several <code>InputSlot</code> can be connected to an output slot.
 * Two types of output slots are available:
 * <ul>
 * <li>@see OutputDataSlot
 * <li>@see outputResultSlot
 * </ul>
 *
 * @author Régis CLOUARD - Oct 01, 1998
 * @version Sébastien LEFÈVRE, Richard LOBRÉAU - Feb 21, 1999
 * @version Adrien SCHRÈQUE, Hicham AMADI - Nov 05, 2000
 * @version Mathieu JOETS - Jul 09, 2001
 * @version Régis CLOUARD - Oct 16, 2001
 */
public abstract class OutputSlot extends Slot {
	/** Is the result of the slot validate? */
	private boolean _validate;

	/** The list of connected input slots. */
	private final transient List<InputSlot> _connectedSlots = new ArrayList<InputSlot>();

	public OutputSlot( GraphItem item, int rank ) {
		super(item, rank);
	}

	@Override
	public final void drop( Slot from ) {
		setNormalBorder();
		try {
			InputSlot inputSlot = (InputSlot)from;
			// check if the slot are in the same graph!
			if (getGraph() != inputSlot.getGraph()) {
				throw new ArianeSlotException(Ariane.getString("Slot.error3"));
			}
			// check slots consistency.
			boolean connect = this instanceof OutputDataSlot && inputSlot instanceof InputDataSlot
					|| this instanceof OutputDataSlot && inputSlot instanceof InputMaskSlot
					|| this instanceof OutputResultSlot && inputSlot instanceof InputParameterSlot;
			if (connect) {
				getGraph().startActionEdition();
				inputSlot.connectFrom(this);
				getGraph().endActionEdition();
				getGraph().requestFocusInWindow();
			} else {
				throw new ArianeSlotException(Ariane.getString("Slot.error2"));
			}
		} catch (ArianeSlotException e) {
			Ariane.getWorkspace().printErrorMessage(Ariane.getString("Slot.error1") + e.getMessage());
		} catch (ClassCastException e) {
			Ariane.getWorkspace().printErrorMessage(Ariane.getString("Slot.error1") + Ariane.getString("Slot.error2"));
		}
	}

	///// 
	///// C O N N E C T I O N   /   D I S C O N N E C T I O N 
	/////

	/**
	 * Connects the given slot with the slot.
	 * <br>
	 * If the given slot is already connected to the slot does nothing.
	 * Else, disconnects the given slot if needed
	 * and adds the slot to the list of connected slots.
	 * 
	 * @param toSlot the input slot to be connected.
	 * @param link an existing link or null
	 * @param historical true if the action must be stored in the action
	 *        history (for use with undo/redo).
	 */
	public final void connectTo( InputSlot toSlot, Link link, boolean historical ) {
		if (!_connectedSlots.contains(toSlot)) {
			toSlot.disconnectAll();
			_connectedSlots.add(toSlot);
			setConnectedIcon();
			toSlot.setConnectedSlot(this, link, historical);
		}
	}

	public final List<InputSlot> getConnectedSlots( ) {
		return _connectedSlots;
	}

	@Override
	public final boolean isConnected( ) {
		return !_connectedSlots.isEmpty();
	}

	/**
	 * Disconnected the slot from the specified slot.
	 * 
	 * @param inputSlot the input slot to be disconnected from.
	 * @param historical true if the action must be stored in the action
	 *        history (for use with undo/redo).
	 */
	public final void disconnectFrom( InputSlot inputSlot, boolean historical ) {
		inputSlot.setConnectedSlot(null, null, historical);
		inputSlot.disposeLink();
		if (_connectedSlots.remove(inputSlot) && _connectedSlots.isEmpty()) {
			setDisconnectedIcon();
		}
	}

	/**
	 * Disconnects all the connected slots.
	 * Implementation: calls the disconnection method of each slot
	 * since disconnection remove slot from the list.
	 */
	@Override
	public final void disconnectAll( ) {
		while (!_connectedSlots.isEmpty()) {
			disconnectFrom(_connectedSlots.get(0), true);
		}
	}

	public final boolean isValidate( ) {
		return _validate;
	}

	protected final void setValidated( ) {
		_validate = true;
	}

	@Override
	public final void setSelected( boolean isSelected ) {
		if (isConnected()) {
			highlightSlot(isSelected);
			for ( InputSlot inputSlot : getConnectedSlots() ) {
				inputSlot.highlightSlot(isSelected);
			}
		}
	}

	public final void reset( ) {
		_validate = false;
	}

	/**
	 * Determines if the slot has a validate result,
	 * if needed executes the related forward operators
	 * to create output images.
	 * 
	 * @param imperative is FAILURE a fatal error?
	 * @return true if the slot has a result.
	 */
	public final boolean hasValue( boolean imperative ) {
		return getGraphItem().execute(imperative) && isValidate();
	}

	/**
	 * Saves the slot with its values as an XML tree.
	 * 
	 * @param e the XML parent in the tree.
	 * @return the root of the XML tree.
	 * @throws IOException is case of output errors.
	 */
	public abstract XMLElement save( XMLElement element ) throws IOException;
}