/* -*- c-basic-offset: 3 -*-
 *
 * Copyright (c) 2013, GREYC.
 * All rights reserved
 *
 * You may use this file under the terms of the BSD license as follows:
 *
 * "Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of the GREYC, nor the name of its
 *     contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
 * 
 *   
 * For more information, refer to:
 * https://clouard.users.greyc.fr/Pandore
 */

package fr.greyc.pantheon.ariane.graph.slot;

import fr.greyc.pantheon.ariane.graph.Graph;
import fr.greyc.pantheon.ariane.graph.item.GraphItem;
import fr.greyc.pantheon.ariane.xml.XMLElement;

import java.awt.Color;
import java.awt.Cursor;
import java.awt.Graphics;
import java.awt.Point;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

import javax.swing.BorderFactory;
import javax.swing.JLabel;
import javax.swing.border.LineBorder;

/**
 * The class <code>Slot</code> defines connection slot between 2 graph items.
 * The two derived classes <code>InputSlot</code> and <code>OutputSlot</code>
 * distinguish the input from output slots. A slot is graphically represented
 * an icon (using JLabel). A slot is both a draggable and a droppable component.
 * 
 * @author Régis CLOUARD - Oct 01, 1998
 * @version Sébastien LEFÈVRE, Richard LOBRÉAU - Feb 21, 1999
 * @version Adrien SCHRÈQUE, Hicham AMADI - Nov 05, 2000
 * @version Mathieu JOETS - Jul 09, 2001
 * @version Régis CLOUARD - Oct 16, 2001
 * @version Régis CLOUARD - Nov 10, 2004
 * @version Frédéric LAIR - May 23, 2007
 */
public abstract class Slot extends JLabel implements MouseListener, MouseMotionListener {
	private static final int LINE_WIDTH = 1;
	/** Border for a selected looking of the component. */
	private static final transient LineBorder ACTIVE_BORDER = (LineBorder)BorderFactory.createLineBorder(Color.RED, LINE_WIDTH);
	/** Border for a normal looking of the component.*/
	private static final transient LineBorder NORMAL_BORDER = (LineBorder)BorderFactory.createLineBorder(new Color(0, 0, 0, 0), LINE_WIDTH);

	/** The graph item that supports the slot. */
	private transient GraphItem _graphItem;

	/** The rank of the slot in the slots list of the graph item. */
	private final int _rank;

	/** Flag that indicates if the slot is currently selected. */
	private boolean _selected;

	private static Slot _targetSlot;

	private static Slot _sourceSlot;

	public Slot( GraphItem item, int rank ) {
		super();
		_rank = rank;
		_graphItem = item;
		setNormalBorder();
		setFocusable(false);

		addMouseListener(this);
		addMouseMotionListener(this);
	}

	/**
	 * Disposes the slot (definitively).
	 * Tries to avoid memory leak.
	 * <b>IT IS SUPPOSED TO BE DISCONNECTED!</b>
	 */
	public final void dispose( ) {
		_graphItem = null;
	}

	private void dropIntoGraph( ) {
		if (_targetSlot == null) {
			getGraph().drop(this);
		} else {
			drop(_targetSlot);
		}
	}

	public abstract void drop( Slot slot );

	protected abstract boolean isAcceptableDropSite( Slot slot );

	/**
	 * @return true if it is selected.
	 */
	protected final boolean isSelected() {
		return _selected;
	}

	public abstract void setSelected( boolean isSelected );

	protected final void highlightSlot( boolean isSelected ) {
		_selected = isSelected;
		if (isSelected) {
			setActiveBorder();
		} else {
			setNormalBorder();
		}
	}

	/**
	 * Sets the border with active look.
	 */
	private void setActiveBorder( ) {
		setBorder(ACTIVE_BORDER);
	}

	/**
	 * Sets the border with normal look.
	 */
	protected final void setNormalBorder( ) {
		setBorder(NORMAL_BORDER);
	}

	/**
	 * @return the rank of the slot in the slot list of the graph item.
	 */
	public final int getRank( ) {
		return _rank;
	}

	/**
	 * Sets the icon with the connected look.
	 */
	protected abstract void setConnectedIcon( );

	/**
	 * Sets the icon with the disconnected look.
	 */
	protected abstract void setDisconnectedIcon( );

	/**
	 * Determines if the slot is connected.
	 * 
	 * @return true if the item is connected.
	 */
	protected abstract boolean isConnected( );

	public abstract void disconnectAll();

	/**
	 * Returns the first anchor coordinates of the slot.
	 * 
	 * @return anchor point.
	 */
	protected abstract Point getAnchor();

	/**
	 * Returns the graph item that supports the slot.
	 * 
	 * @return a reference to the graph item.
	 */
	public final GraphItem getGraphItem( ) {
		return _graphItem;
	}

	/**
	 * Returns the related graph.
	 * 
	 * @return a reference to the graph.
	 */
	public final Graph getGraph( ) {
		return _graphItem.getGraph();
	}

	/**
	 * overrides the default print method to print
	 * a filled rectangle.
	 * 
	 * @param gc the graph context.
	 */
	@Override
	public final void print( Graphics gc ) {
		gc.fillRect(0, 0, getWidth(), getHeight());
	}

	private static void connectCompatibleSlots( GraphItem gFrom, GraphItem gTo, SlotAttributes attributes ) {
		if (gFrom != null && gTo != null) {
			OutputSlot sFrom = (OutputSlot)gFrom.getSlot(attributes.typeFrom, attributes.rankFrom);
			InputSlot sTo = (InputSlot)gTo.getSlot(attributes.typeTo, attributes.rankTo);
			if (sFrom != null && sTo != null) {
				sFrom.connectTo(sTo, null, true);
				Link link = sTo.createLink();
				link.setPath(attributes.path, attributes.userDefined);
				sTo.setLink(link);
			}
		}
	}

	/**
	 * Loads a set of connections from an XML tree.
	 * 
	 * @param graph the graph that supports the connections.
	 * @param items the XML tree.
	 * @throws ArianeSlotException for problems.
	 */
	public static final void load( Graph graph, List<XMLElement> items ) throws IOException {
		if (items == null) { // Empty set.
			return;
		}
		for (XMLElement el : items) {
			if (el.getName().equalsIgnoreCase("Connection") || el.getName().equalsIgnoreCase("Connexion") ) {
				SlotAttributes attributes = new SlotAttributes();

				for (XMLElement elc : el.getChildren()) {
					if (elc.getName().equalsIgnoreCase("EndItemId") || elc.getName().equalsIgnoreCase("IdElementArrivee")) {
						attributes.idTo = Integer.parseInt(elc.getValue());
					} else if (elc.getName().equalsIgnoreCase("StartItemId") || elc.getName().equalsIgnoreCase("IdElementDepart")) {
						attributes.idFrom = Integer.parseInt(elc.getValue());
					} else if (elc.getName().equalsIgnoreCase("EndSlotType") || elc.getName().equalsIgnoreCase("TypeSlotArrivee")) {
						attributes.typeTo = elc.getValue();
					} else if (elc.getName().equalsIgnoreCase("StartSlotType") || elc.getName().equalsIgnoreCase("TypeSlotDepart")) {
						attributes.typeFrom = elc.getValue();
					} else if (elc.getName().equalsIgnoreCase("StartSlotIndex") || elc.getName().equalsIgnoreCase("NumeroSlotDepart")) {
						attributes.rankFrom = Integer.parseInt(elc.getValue());
					} else if (elc.getName().equalsIgnoreCase("EndSlotIndex") || elc.getName().equalsIgnoreCase("NumeroSlotArrivee")) {
						attributes.rankTo = Integer.parseInt(elc.getValue());
					} else if (elc.getName().equalsIgnoreCase("ParameterValue")) {
						attributes.parameterValue = elc.getValue();
					} else if (elc.getName().equalsIgnoreCase("UserDefined")) {
						attributes.userDefined = elc.getValue().equalsIgnoreCase("true");
					} else if (elc.getName().equalsIgnoreCase("LinkPath") && elc.getChildren() != null) {
						for (XMLElement pt : elc.getChildren()) {
							final Point point = new Point();
							if (pt.getName().equalsIgnoreCase("LinkPoint")) {
								int pos = 0;
								for (XMLElement elP : pt.getChildren()) {
									if (elP.getName().equalsIgnoreCase("IdPoint")) {
										pos = Integer.parseInt(elP.getValue());
									} else if (elP.getName().equalsIgnoreCase("X-coordinate")) {
										point.x = Integer.parseInt(elP.getValue());
									} else if (elP.getName().equalsIgnoreCase("Y-coordinate")) {
										point.y = Integer.parseInt(elP.getValue());
									}
								}
								if (pos > 0) {
									attributes.path.add(pos, point);
								} else {
									attributes.path.add(point);
								}
							}
						}
					}
				}
				if (attributes.typeTo == null) {
					throw new IOException("Unknown connection type");
				}
				if (attributes.typeTo.equalsIgnoreCase("InSlotParameter")) { // Parameters
					GraphItem gTo = graph.getGraphItem(attributes.idTo);
					if (attributes.idFrom > -1) { // Connection parameter -> result
						GraphItem gFrom = graph.getGraphItem(attributes.idFrom);
						connectCompatibleSlots(gFrom, gTo, attributes);
					}
					if (gTo != null) { // At least one parameter value.
						InputParameterSlot par = gTo.getInputParameterSlots().get(attributes.rankTo);
						if (par != null) {
							par.setValue(attributes.parameterValue);
						}
					}
				} else if (attributes.typeTo.equalsIgnoreCase("InSlotMask") || attributes.typeTo.equalsIgnoreCase("InSlotImage")) {
					GraphItem gFrom = graph.getGraphItem(attributes.idFrom);
					GraphItem gTo = graph.getGraphItem(attributes.idTo);
					connectCompatibleSlots(gFrom, gTo, attributes);
				}
			}
		}
	}

	/////////
	///////// MOUSE
	/////////

	@Override
	public final void mouseReleased( MouseEvent event ) {
		setCursor(null);
		getGraph().stopElasticConnection();
		setSelected(false);
		dropIntoGraph();
	}

	@Override
	public final void mousePressed( MouseEvent event ) {
		synchronized (this) {
			_targetSlot = null;
			_sourceSlot = this;
		}
		setCursor(Cursor.getPredefinedCursor(Cursor.CROSSHAIR_CURSOR));
		getGraph().startElasticConnection(getLocationOnScreen().x + 4, getLocationOnScreen().y + 4);
	}

	@Override
	public final void mouseExited( MouseEvent event ) {
		if (_sourceSlot != null) {
			if (!_selected) {
				setNormalBorder();
			}
			_targetSlot = null;
		}
	}

	@Override
	public final void mouseEntered( MouseEvent event ) {
		if (_sourceSlot != null) {
			_targetSlot = this;
			// Accepts the drop action for a move action.
			if (_sourceSlot.isAcceptableDropSite(_targetSlot)) {
				setActiveBorder();
			}
		}
	}

	@Override
	public final void mouseClicked( MouseEvent event ) {
		/* empty */
	}

	@Override
	public final void mouseMoved( MouseEvent event ) {
		/* empty */
	}

	@Override
	public final void mouseDragged( MouseEvent event ) {
		getGraph().drawElasticConnection(event.getLocationOnScreen());
	}

	private static class SlotAttributes { // A messenger
		//CHECKSTYLE:OFF
		public int idTo = -1;
		public int idFrom = -1;
		public int rankFrom = -1;
		public int rankTo = -1;
		public String typeTo;
		public String typeFrom;
		public String parameterValue;
		public final List<Point> path = new ArrayList<Point>();
		public boolean userDefined;
		//CHECKSTYLE:ON
	}
}
