/* -*- c-basic-offset: 3 -*-
 *
 * Copyright (c) 2013, GREYC.
 * All rights reserved
 *
 * You may use this file under the terms of the BSD license as follows:
 *
 * "Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of the GREYC, nor the name of its
 *     contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
 * 
 *   
 * For more information, refer to:
 * https://clouard.users.greyc.fr/Pandore
 */

package fr.greyc.pantheon.ariane.ui;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Component;
import java.awt.Point;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.PipedInputStream;
import java.io.PipedOutputStream;
import java.io.PrintStream;
import java.io.PrintWriter;
import java.util.Locale;

import javax.swing.JButton;
import javax.swing.JFileChooser;
import javax.swing.JLabel;
import javax.swing.JMenuItem;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JPopupMenu;
import javax.swing.JTextField;
import javax.swing.JTextPane;
import javax.swing.text.BadLocationException;
import javax.swing.text.DefaultHighlighter;
import javax.swing.text.Document;
import javax.swing.text.Highlighter;
import javax.swing.text.JTextComponent;
import javax.swing.text.Style;
import javax.swing.text.StyleConstants;
import javax.swing.text.StyleContext;
import javax.swing.text.StyledDocument;

import fr.greyc.pantheon.ariane.Ariane;
import fr.greyc.pantheon.ariane.utils.GDialog;

/**
 * The <i>Console</i> is used to display styled text.
 * 
 * @author Régis CLOUARD - Dec 29, 2006
 * @author Régis CLOUARD - Dec 12, 2008 (distinction error message)
 * @author Régis CLOUARD - Jun 30, 2010 (improve rendering)
 * @author Régis CLOUARD - Apr 01, 2011 (handle \n \t, etc in the console + save)
 */
public final class Console extends JTextPane {
	private static final Highlighter HIGH_LIGHTER = new DefaultHighlighter();
	private final WordSearcher _searcher;
	/** Custom output PrintStream. */
	private PrintStream _printOutputStream;
	/** Custom system PrintStream. */
	private PrintStream _printSystemStream;
	/** Custom error PrintStream. */
	private PrintStream _printErrorStream;
	/** The StyleContext for the Console. */
	private final ConsoleStyleContext STYLE_CONTEXT = new ConsoleStyleContext();

	private ReadWriteThread _outputReadWriteThread;
	private ReadWriteThread _systemReadWriteThread;
	private ReadWriteThread _errorReadWriteThread;

	/**
	 * Creates a new console.
	 */
	public Console( ) {
		setEditable(false);
		_searcher = new WordSearcher(this);
		addMouseListener(new MouseAdapter() {
			private void showPopupMenu(final MouseEvent mouseEvent) {
				JPopupMenu popup = new JPopupMenu();
				JMenuItem erase = new JMenuItem(Ariane.getString("Workspace.console.clear"));
				erase.addActionListener(new ActionListener() {
					@Override
					public void actionPerformed( ActionEvent event ) {
						clear();
					}
				});
				JMenuItem search = new JMenuItem(Ariane.getString("Workspace.console.search"));
				search.addActionListener(new ActionListener() {
					@Override
					public void actionPerformed( ActionEvent event ) {
						search(mouseEvent.getLocationOnScreen());
					}
				});
				JMenuItem save = new JMenuItem(Ariane.getString("Workspace.console.save"));
				save.addActionListener(new ActionListener() {
					@Override
					public void actionPerformed( ActionEvent event ) {
						save();
					}
				});
				popup.add(erase);
				popup.add(search);
				popup.add(save);
				popup.show((Component)mouseEvent.getSource(), mouseEvent.getX(), mouseEvent.getY());
			}
			/** Opens the contextual menu */
			@Override
			public void mousePressed(MouseEvent event) {
				if (event.isPopupTrigger()) {
					showPopupMenu(event);
				}
			}
			/** Opens the contextual menu */
			@Override
			public void mouseReleased(MouseEvent event) {
				if (event.isPopupTrigger()) {
					showPopupMenu(event);
				}
			}
		});

		setHighlighter(HIGH_LIGHTER);
		redirect();
	}

	/**
	 * Clears the console.
	 */
	public void clear( ) {
		StyledDocument document = getStyledDocument();
		try {
			document.remove(0, document.getLength());
		} catch (BadLocationException e) {
		}
	}

	/**
	 * Searches a word in the log console.
	 */
	private void search( Point mouseCoordinates ) {
		final GDialog dialog = new GDialog(Ariane.getFrame(), "Find", false);
		dialog.setLayout(new BorderLayout(1, 1));
		dialog.setAlwaysOnTop(true);

		JPanel content = new JPanel();
		final JTextField textfield = new JTextField(10);
		content.add(new JLabel(Ariane.getString("Workspace.console.searchTitle")));
		content.add(textfield);
		dialog.add(content, BorderLayout.CENTER);

		JButton okButton = new JButton(Ariane.getString("Workspace.console.findButton"));
		okButton.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed( ActionEvent event ) {
				String word = textfield.getText().trim();
				int offset = _searcher.search(word);
				if (offset != -1) {
					try {
						Console.this.scrollRectToVisible(Console.this.modelToView(offset));
					} catch ( BadLocationException exception ) {
					}
				}
				dialog.dispose();
			}
		});
		JButton cancelButton = new JButton(Ariane.getString("Workspace.console.closeButton"));
		cancelButton.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed( ActionEvent event ) {
				dialog.dispose();
			}
		});
		JPanel buttonPanel = new JPanel();
		buttonPanel.add(okButton);
		buttonPanel.add(cancelButton);
		dialog.add(buttonPanel, BorderLayout.SOUTH);
		dialog.pack();
		dialog.setLocation(mouseCoordinates);
		dialog.setVisible(true);
		dialog.setDefaultButton(okButton);
		dialog.setCancelButton(cancelButton);
	}

	/**
	 * Saves the console contents in a file.
	 */
	private void save( ) {
		String currentDir = Workspace.getCurrentDir(); 
		File currentFileName = new File(currentDir);

		JFileChooser fileChooser = new JFileChooser(currentFileName);
		fileChooser.setDialogType(JFileChooser.SAVE_DIALOG);
		fileChooser.setDialogTitle(Ariane.getString("FileChooser.titleSaveLog"));

		if (fileChooser.showSaveDialog(this) == JFileChooser.APPROVE_OPTION) {
			File theFile = fileChooser.getSelectedFile();
			if (theFile == null || !theFile.exists()) {
				return;
			}
			if (JOptionPane.showConfirmDialog(Ariane.getFrame(),
					Ariane.getString("Workspace.dialog.message.part1")
					+ theFile.getName()
					+ Ariane.getString("Workspace.dialog.message.part2"),
					Ariane.getString("Workspace.dialog.title"),
					JOptionPane.YES_NO_OPTION) == JOptionPane.NO_OPTION) {
				return;
			}
			PrintWriter out = null;
			try {
				String text = getDocument().getText(0, getDocument().getLength());
				out = new PrintWriter(new BufferedWriter(new FileWriter(theFile)));
				out.print(text);
				Ariane.getWorkspace().printSystemMessage(theFile + Ariane.getString("Workspace.saving.success"));
			} catch (IOException | BadLocationException ex) {
				Ariane.getWorkspace().printErrorMessage(Ariane.getString("Workspace.saving.error1") + getName() + " " + ex.getLocalizedMessage());
			} finally {
				if (out != null) {
					out.close();
				}
			}
		}
	}

	/**
	 * Use to search and highlight the word.
	 */
	final class WordSearcher {
		private final JTextComponent _comp;
		private final Highlighter.HighlightPainter _painter;

		protected WordSearcher( JTextComponent comp ) {
			_comp = comp;
			_painter = new DefaultHighlighter.DefaultHighlightPainter(Color.LIGHT_GRAY);
		}

		/**
		 * Searches for the given word and returns the offset of the
		 * first occurrence. Highlight are added for all occurrence found.
		 */
		private int search( String word ) {
			// remove any existing highlights for last word.
			Highlighter highlighter = _comp.getHighlighter();
			highlighter.removeAllHighlights();
			if (word == null || word.equals("")) {
				return -1;
			}

			// look for the word - insensitive search
			String content;
			try {
				Document document = _comp.getDocument();
				content = document.getText(0, document.getLength()).toLowerCase();
			} catch ( BadLocationException e ) {
				return -1;
			}
			String lowerCaseWord = word.toLowerCase(Locale.US);
			int lastIndex = 0;
			int wordSize = lowerCaseWord.length();
			int firstOffset = -1;
			while ((lastIndex = content.indexOf(lowerCaseWord, lastIndex)) != -1) {
				int endIndex = lastIndex + wordSize;
				try {
					highlighter.addHighlight(lastIndex, endIndex, _painter);
				} catch ( BadLocationException e ) {

				}
				if (firstOffset == -1) {
					firstOffset = lastIndex;
				}
				lastIndex = endIndex;
			}
			return firstOffset;
		}
	}

	/**
	 * Displays the specified text in the console.
	 * The text is supposed to be a message from
	 * the execution of the items.
	 * 
	 * @param text the text to be displayed.
	 */
	public void displayOutputMessage( String text ) {
		int maxLength = 200;
		String printedText = text;
		while (printedText.length() > maxLength) {
			_printOutputStream.println(printedText.substring(0, maxLength));
			_printOutputStream.flush();
			synchronized (_outputReadWriteThread) {
				_outputReadWriteThread.notifyAll();
			}
			printedText = printedText.substring(maxLength);
		}
		_printOutputStream.println(printedText);
		_printOutputStream.flush();
		synchronized (_outputReadWriteThread) {
			_outputReadWriteThread.notifyAll();
		}
	}

	/**
	 * Displays the specified text in the console.
	 * The text is supposed to be an error message.
	 * 
	 * @param text the text to be displayed.
	 */
	public void displayErrorMessage( String text ) {
		_printErrorStream.println(text);
		_printErrorStream.flush();
		synchronized (_errorReadWriteThread) {
			_errorReadWriteThread.notifyAll();
		}
	}

	/**
	 * Displays the specified text in the console.
	 * The text is supposed to be an Ariane message.
	 * 
	 * @param text the text to be displayed.
	 */
	public void displaySystemMessage( String text ) {
		_printSystemStream.println(text);
		_printSystemStream.flush();
		synchronized (_systemReadWriteThread) {
			_systemReadWriteThread.notifyAll();
		}
	}

	/**
	 * Redirects the flows.
	 */
	private void redirect() {
		try {
			PipedInputStream piCustom = new PipedInputStream();
			PipedOutputStream poCustom = new PipedOutputStream(piCustom);
			_printOutputStream = new PrintStream(poCustom, false);
			_outputReadWriteThread = new ReadWriteThread(piCustom, ConsoleStyleContext.CUSTOM);
			_outputReadWriteThread.start();

			PipedInputStream piOut = new PipedInputStream();
			PipedOutputStream poOut = new PipedOutputStream(piOut);
			_printSystemStream = new PrintStream(poOut, true);
			_systemReadWriteThread = new ReadWriteThread(piOut, ConsoleStyleContext.SYSTEMOUT);
			_systemReadWriteThread.start();

			PipedInputStream piErr = new PipedInputStream();
			PipedOutputStream poErr = new PipedOutputStream(piErr);
			_printErrorStream = new PrintStream(poErr, true);
			_errorReadWriteThread = new ReadWriteThread(piErr, ConsoleStyleContext.SYSTEMERR);
			_errorReadWriteThread.start();
		} catch (IOException e) {
		}
	}

	/**
	 * The thread that writes everything to the text area.
	 */
	final class ReadWriteThread extends Thread {
		/** The InputStream of this Thread */
		private final PipedInputStream _pipedInputStream;
		/** The type (CUSTOM, SYSTEMOUT, SYSTEMERR) of this Thread */
		private final String _type;

		ReadWriteThread( PipedInputStream pi, String type ) {
			super();
			_pipedInputStream = pi;
			_type = type;
		}

		@Override
		public void run() {
			final Document doc = Console.this.getDocument();
			final BufferedReader br = new BufferedReader(new InputStreamReader(_pipedInputStream));

			while (true) {
				try {
					String line;
					while (_pipedInputStream.available() > 0 ) {
						while ((line = br.readLine()) != null) {
							try {
								synchronized (doc) {
									if (doc.getLength() > 100000) {
										doc.remove(0, line.length());
									}
									doc.insertString(doc.getLength(), line + "\n", STYLE_CONTEXT.getStyle(_type));
									Console.this.setCaretPosition(Console.this.getDocument().getLength());
								}
								sleep(50);
							} catch (BadLocationException e) {
							}
						}
					}
					synchronized (this) {
						wait(); // Do not occupy the CPU no more
					}
				} catch (IOException | InterruptedException e) {
				}
			}
		}
	}

	/**
	 * The style context defining the styles of each type of PrintStream.
	 */
	final class ConsoleStyleContext extends StyleContext {
		/** The name of the Style used for Custom messages. */
		public static final String CUSTOM = "Custom";
		/** The name of the Style used for system messages. */
		public static final String SYSTEMOUT = "SystemOut";
		/** The name of the Style used for error messages. */
		public static final String SYSTEMERR = "SystemErr";

		/** Creates the style context for the Console. */
		public ConsoleStyleContext() {
			super();
			Style root = getStyle(DEFAULT_STYLE);
			Style style = addStyle(CUSTOM, root);
			StyleConstants.setForeground(style, Color.BLUE);
			style = addStyle(SYSTEMOUT, root);
			StyleConstants.setForeground(style, Color.BLACK);
			style = addStyle(SYSTEMERR, root);
			StyleConstants.setForeground(style, Color.RED);
		}
	}
}