/* -*- c-basic-offset: 3 -*-
 *
 * Copyright (c) 2013, GREYC.
 * All rights reserved
 *
 * You may use this file under the terms of the BSD license as follows:
 *
 * "Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of the GREYC, nor the name of its
 *     contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
 * 
 *   
 * For more information, refer to:
 * https://clouard.users.greyc.fr/Pandore
 */

package fr.greyc.pantheon.ariane.ui;

import java.awt.Color;
import java.awt.Component;
import java.awt.FlowLayout;
import java.awt.Font;
import java.awt.dnd.DnDConstants;
import java.awt.dnd.DropTarget;
import java.awt.dnd.DropTargetAdapter;
import java.awt.dnd.DropTargetDragEvent;
import java.awt.dnd.DropTargetDropEvent;
import java.awt.event.KeyEvent;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.io.File;
import java.util.Hashtable;
import java.util.Map;

import javax.swing.BorderFactory;
import javax.swing.JButton;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTabbedPane;
import javax.swing.KeyStroke;
import javax.swing.UIManager;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

import fr.greyc.pantheon.ariane.Ariane;
import fr.greyc.pantheon.ariane.graph.Graph;

/**
 * The <code>GraphEditor</code> is the editor of graph.
 * Each graph is displayed in a <code>JTabbedPane</code>.
 * Beware of the fact that each graph is first for all inserted
 * in a JScrollPane.
 * 
 * @author Régis CLOUARD - Dec 13, 2004
 * @version Régis CLOUARD - Dec 01, 2007 (change JInternalFrame to JTabbedPane) 
 */
public final class GraphEditor extends JTabbedPane implements ChangeListener {
	/** The workspace that support the graph. */
	private final Workspace _workspace;
	/** The index table of current graphs. */
	private final Map<Component, Graph> _indexTable = new Hashtable<Component, Graph>(); // NOPMD by regis on 8/8/14 10:37 AM

	/**
	 * Creates the tabbed pane manager.
	 * 
	 * @param workspace the workspace that supports the graph editor.
	 */
	public GraphEditor( Workspace workspace ) {
		_workspace = workspace;
		setTabLayoutPolicy(JTabbedPane.SCROLL_TAB_LAYOUT);
		addChangeListener(this);
		Font font = getFont();
		setFont(font.deriveFont(Font.PLAIN, 11f));

		new DropTarget(this, new DropTargetAdapter() {
			@Override
			public void dragEnter( DropTargetDragEvent event ) {
				event.acceptDrag(DnDConstants.ACTION_MOVE);
			}

			@Override
			public void drop( DropTargetDropEvent event ) {
				if (getTabCount() == 0) {
					_workspace.newGraph().getDropTarget().drop(event);
				} else {
					event.rejectDrop();
				}
			}
		});
	}

	/**
	 * Returns the JScrollPane component that holds the
	 * specified graph or null if any.
	 * 
	 * @param graph the graph.
	 * @return the JScrollPane that holds the graph.
	 */
	private JScrollPane getScrollPane( Graph graph ) {
		for (int i = 0; i < getTabCount(); i++) {
			JScrollPane scrollpane = (JScrollPane)getComponentAt(i);
			if (_indexTable.get(scrollpane) == graph) {
				return scrollpane;
			}
		}
		return null;
	}

	/**
	 * Returns the graph supported by the given scrollpane.
	 * 
	 * @param scrollpane a scrollpane.
	 * @return the graph or null if any.
	 */
	private Graph getGraph( Component scrollpane ) {
		if (scrollpane == null) {
			return null;
		}
		return _indexTable.get(scrollpane);
	}

	/**
	 * Returns the graph supported by the nth scrollpane
	 * in the JTabbedPane.
	 * 
	 * @param index the index of the scrollpane.
	 * @return the graph or null if any.
	 */
	private Graph getGraphAt( int index ) {
		if (index < 0) {
			return null;
		}
		return getGraph(getComponentAt(index));
	}

	/**
	 * Checks whether the specified graph already exists.
	 * If the graph already exists, puts it in front.
	 * 
	 * @param filename a regular file name.
	 * @return true if the graph already exists.
	 */
	public boolean exists( String filename ) {
		for (int i = 0; i < getTabCount(); i++) {
			JScrollPane scrollpane = (JScrollPane)getComponentAt(i);
			Graph graph = _indexTable.get(scrollpane);
			if (graph.getFileName() != null && graph.getFileName().equals(new File(filename))) {
				super.setSelectedComponent(scrollpane);
				_workspace.setCurrentGraph(graph);
				return true;
			}
		}
		return false;
	}

	/**
	 * Adds a new frame in the editor.
	 * 
	 * @param graph the graph to be added.
	 */
	public void add( Graph graph ) {
		JScrollPane scrollpane = getScrollPane(graph);
		if ( scrollpane == null ) {
			scrollpane = new JScrollPane(graph);
			// Discard the arrows events.
			scrollpane.setInputMap(WHEN_IN_FOCUSED_WINDOW, null); 
			scrollpane.setInputMap(WHEN_FOCUSED, null); 
			scrollpane.setInputMap(WHEN_ANCESTOR_OF_FOCUSED_COMPONENT , null); 
			scrollpane.getVerticalScrollBar().setUnitIncrement(10);
			scrollpane.getHorizontalScrollBar().setUnitIncrement(10);
			scrollpane.unregisterKeyboardAction(KeyStroke.getKeyStroke(KeyEvent.VK_DOWN, 0));
			_indexTable.put(scrollpane, graph);
			super.add(scrollpane);
			setTabComponentAt(indexOfComponent(scrollpane), new TabTitle(this, scrollpane, graph.getName(), graph.isRoot()));
			super.setSelectedComponent(scrollpane);
			_workspace.setCurrentGraph(graph);
		} else {
			super.setSelectedComponent(scrollpane);
			_workspace.setCurrentGraph(graph);
		}
	}

	/**
	 * Removes the specified graph from the editor.
	 * 
	 * @param graph the graph to be removed.
	 */
	public void remove( Graph graph ) {
		JScrollPane scrollpane = getScrollPane(graph);
		if (scrollpane != null) {
			super.remove(scrollpane);
			_indexTable.remove(scrollpane);
			_workspace.setCurrentGraph(getGraph(getSelectedComponent()));
		}
	}

	/**
	 * Closes the specified graph from the editor.
	 * 
	 * @param index the index of the graph in the editor.
	 */
	protected void close( int index ) {
		Graph graph = getGraphAt(index);
		graph.close();
	}

	/**
	 * Sets the title to the given graph.
	 * 
	 * @param graph a graph.
	 * @param name the new name.
	 */
	public void setTitle( Graph graph, String name ) {
		if (graph != null) {
			int index = indexOfComponent(getScrollPane(graph));
			graph.setName(name);
			if (index >= 0) {
				TabTitle component = (TabTitle)getTabComponentAt(index);
				component.setTitle(graph.getName(), graph.isRoot());
			}
		}
	}

	/**
	 * Closes all graphs.
	 * 
	 * @return true if all graph has been closed.
	 */
	public boolean closeAllGraphs( ) {
		// Closes only the root graph
		// because root graph deletes it.
		for (int i = getTabCount() - 1; i >= 0; i--) {
			Graph graph = getGraphAt(i);
			if (graph.isRoot() && !graph.close()) {
				return false;
			}
		}
		return true;
	}

	/**
	 * Invoked when the user select a new pane.
	 * 
	 * @param event a ChangeEvent object.
	 */ 
	@Override
	public void stateChanged( ChangeEvent event ) {
		JTabbedPane source = (JTabbedPane)event.getSource();
		Graph graph = getGraphAt(source.getSelectedIndex());
		_workspace.setCurrentGraph(graph);
	}

	/**
	 * Repaints all graphs.
	 */
	public void repaintAllGraphs( ) {
		for (int i = 0; i < getTabCount(); i++) {
			Graph graph = getGraphAt(i);
			graph.repaintAll();
		}
	}

	class TabTitle extends JPanel {
		private JLabel _titleLabel;
		
		TabTitle(final JTabbedPane tabbedPane, final JScrollPane panel, String title, boolean isRoot) {
			super(new FlowLayout(FlowLayout.LEFT, 0, 0));
			setOpaque(false);
			_titleLabel = new JLabel();
			setTitle(title, isRoot);
			_titleLabel.setBorder(BorderFactory.createEmptyBorder(0, 0, 0, 5));
			add(_titleLabel);
			JButton closeButton = new JButton(Ariane.getIcon("window-close.png"));
			closeButton.addMouseListener(new MouseAdapter() {
				@Override
				public void mouseClicked( MouseEvent event ) {
					int index = ((GraphEditor)tabbedPane).indexOfComponent(panel);
					((GraphEditor)tabbedPane).close(index);
				}
			});
			closeButton.setOpaque(false);
			closeButton.setBorder(null);
			closeButton.setContentAreaFilled(false);
			add(closeButton);
		}

		public void setTitle( String title, boolean isRoot ) {
			String label;
			if (isRoot) {
				final Color color = (Color)UIManager.get("nimbusOrange");
				String code = "#" + toBrowserHexValue(color.getRed()) + toBrowserHexValue(color.getGreen()) + toBrowserHexValue(color.getBlue());
				label = "<html><b><font color=" + code + ">" + title + "</font></b></html>";
			} else {
				label = title;
			}
			_titleLabel.setText(label);
		}
	}

	private static String toBrowserHexValue( int number ) {
		StringBuilder builder = new StringBuilder(Integer.toHexString(number & 0xff));
		while (builder.length() < 2) {
			builder.append('0');
		}
		return builder.toString().toUpperCase();
	}
}