/* -*- c-basic-offset: 3 -*-
 *
 * Copyright (c) 2013, GREYC.
 * All rights reserved
 *
 * You may use this file under the terms of the BSD license as follows:
 *
 * "Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of the GREYC, nor the name of its
 *     contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
 * 
 *   
 * For more information, refer to:
 * https://clouard.users.greyc.fr/Pandore
 */

package fr.greyc.pantheon.ariane.ui;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.net.URI;
import java.net.URISyntaxException;
import java.net.URL;

import javax.swing.JEditorPane;
import javax.swing.JFrame;
import javax.swing.JScrollPane;
import javax.swing.WindowConstants;
import javax.swing.event.HyperlinkEvent;
import javax.swing.event.HyperlinkListener;

import fr.greyc.pantheon.ariane.Ariane;

/**
 * A <code>HTMLDocumentViewers</code> displays HTML pages
 * as real HTML pages.
 *
 * @author Régis CLOUARD - Oct 16, 2001
 * @version Régis CLOUARD - dec 06, 2004
 * @version Régis CLOUARD - Jul 11, 2010
 */
public final class HTMLDocumentViewer extends JFrame implements HyperlinkListener {
	/** The HTML panel. */
	private final JEditorPane _webEditorPane;
	/** The page used when no page is found. */
	private static final String ERROR_PAGE = "operator.html";

	/**
	 * Creates the HTML frame but does not display it.
	 * 
	 * @param title the title of the window.
	 * @param page the URL of the HTML page.
	 */
	public HTMLDocumentViewer( String title, URI page ) {
		_webEditorPane = new JEditorPane();
		_webEditorPane.addHyperlinkListener(this);
		getContentPane().add(new JScrollPane(_webEditorPane));
		setDefaultCloseOperation(WindowConstants.DISPOSE_ON_CLOSE);
		setBounds(200, 25, 800, 600);
		setPage(title, page);
		setLocationRelativeTo(Ariane.getFrame());
		setVisible(true);
	}

	/**
	 * Finds the absolute address of the given document either in the
	 * operators library documentation or in the Ariane manual pages.
	 * 
	 * @param document The name or path of an HTML page.
	 * @param fragment the index in the page as "#name".
	 * @return the address or null.
	 */
	public static URI getURI( String document, String fragment ) {
		if (document.isEmpty()) {
			try {
				return Ariane.getDocumentationURI(ERROR_PAGE);
			} catch ( URISyntaxException e ) {
				return null;
			}
		}
		URI page;
		// 1. direct documentation file name?
		try {
			if (fragment == null ) {
				page = new URI(document);
			} else {
				page = new URI(document + fragment);
			}
			// Check whether the file exists.
			InputStream stream = page.toURL().openStream(); // try to open it.
			stream.close();
			return page;
		} catch (IOException | URISyntaxException | IllegalArgumentException ioee) { }

		// 2. Relative path to library documentation
		File file = new File(Ariane.getLibraryPath() + Ariane.getDocPath() + File.separator + document);
		try {
			String address = file.getCanonicalPath();
			// Library documentation?
			page = new URI("file", address, fragment);
			// Check whether the file exists.
			InputStream stream = page.toURL().openStream(); // try to open it.
			stream.close();
			return page;
		} catch (IOException | URISyntaxException ioee) { }

		// 3. Kept for compatibility reason.
		file = new File(Ariane.getLibraryPath() + File.separator + document);
		try {
			// Check whether the file exists.
			String address = file.getCanonicalPath();
			page = new URI("file", address, fragment);
			InputStream stream = page.toURL().openStream(); // try to open it.
			stream.close();
			return page;
		} catch (IOException | URISyntaxException ioee) {
		}

		// 4. Kept for compatibility reason.
		try {
			// Ariane manual pages
			page = Ariane.getDocumentationURI(document);
			// The last chance: the operator page.
			if (page == null) {
				return Ariane.getDocumentationURI(ERROR_PAGE);
			} else {
				return new URI(page.toString() + ((fragment == null) ? "" : fragment));
			}
		} catch (URISyntaxException e) {
		}
		return null;
	}

	/**
	 * Opens a windows and displays the content of the page
	 * in the HTML format.
	 * First of all, finds the address of the document, either
	 * in the operators library documentation or in the Ariane manual.
	 * 
	 * @param title the title of the window.
	 * @param page the URL of the HTML page.
	 */
	private void setPage( String title, URI page ) {
		setTitle(title);
		if (page == null) {
			_webEditorPane.setContentType("text/html");
			_webEditorPane.setText("<html><body>" + Ariane.getString("HTMLViewer.unknownpage") + title + "</body></html>");
		} else {
			try {
				_webEditorPane.setPage(page.toURL());
				_webEditorPane.setEditable(false);
			} catch (IOException exec) {
				_webEditorPane.setContentType("text/html");
				_webEditorPane.setText("<html><body>" + Ariane.getString("HTMLViewer.unknownpage") + title + "</body></html>");
			}
		}
	}

	/**
	 * Follows a link, displays a new page in the editor.
	 * 
	 * @param hyperlinkevent the event.
	 */
	@Override
	public void hyperlinkUpdate( HyperlinkEvent hyperlinkevent ) {
		HyperlinkEvent.EventType eventtype = hyperlinkevent.getEventType();
		if (eventtype == HyperlinkEvent.EventType.ACTIVATED) {
			URL url = hyperlinkevent.getURL();
			try {
				_webEditorPane.setPage(url);
			} catch (IOException e) {
				_webEditorPane.setContentType("text/html");
				_webEditorPane.setText("<html><body>" + Ariane.getString("HTMLViewer.unknownpage") + url + "</body></html>");
			}
		}
	}
}
