/* -*- c-basic-offset: 3 -*-
 *
 * Copyright (c) 2013, GREYC.
 * All rights reserved
 *
 * You may use this file under the terms of the BSD license as follows:
 *
 * "Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of the GREYC, nor the name of its
 *     contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
 * 
 *   
 * For more information, refer to:
 * https://clouard.users.greyc.fr/Pandore
 */

package fr.greyc.pantheon.ariane.ui;

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.File;
import java.io.IOException;
import java.util.List;

import javax.swing.JDialog;
import javax.swing.JMenuItem;
import javax.swing.JPopupMenu;
import javax.swing.JTree;
import javax.swing.tree.DefaultTreeModel;
import javax.swing.tree.TreePath;

import fr.greyc.pantheon.ariane.Ariane;
import fr.greyc.pantheon.ariane.utils.GDialog;
import fr.greyc.pantheon.ariane.utils.MutableTreeItem;
import fr.greyc.pantheon.ariane.utils.TreeItem;
import fr.greyc.pantheon.ariane.xml.XMLElement;
import fr.greyc.pantheon.ariane.xml.XMLFileOutputStream;

/**
 * The class <code>OperatorTreeEditor</code> defines a dialog box
 * where user can modify the personal operators tree.
 * 
 * @version Régis CLOUARD - Jul 20, 2010
 */
public final class OperatorTreeEditor extends TreeEditor {

	/**
	 * Creates a dialog box that manages the sub tree
	 * of personal operators.
	 * 
	 * @param rootNode the subroutine tree.
	 */
	public OperatorTreeEditor( MutableTreeItem rootNode ) {
		super(Ariane.getString("OperatorTreeEditor.title"), rootNode);
	}

	/**
	 * Returns a new instance of the item editor.
	 * 
	 * @param dialog the parent.
	 * @param currentNode the current Node;
	 * @param treeModel the tree model.
	 * @return the dialog.
	 */
	@Override
	protected JDialog getItemEditor( GDialog dialog, MutableTreeItem currentNode, DefaultTreeModel treeModel ) {
		return new OperatorEditor(dialog, currentNode, treeModel);
	}

	@Override
	protected TreeItem createItem( String categoryName ) {
		return new TreeItem(TreeItem.Item.OPERATOR, "operator", "", null, categoryName, 0, 0, 0, "");
	}

	@Override
	protected boolean isNewEnabled( TreePath path ) {
		return path != null && path.getPathCount() < 3;
	}

	/**
	 * Creates and displays a popup menu
	 * for the current selected category in the tree.
	 * 
	 * @param tree the current tree.
	 * @param x the x coordinate position to popup the menu.
	 * @param y the y coordinate position to popup the menu.
	 */
	@Override
	protected void openPopupMenuCategory( final JTree tree, int x, int y ) {
		JPopupMenu popupMenu = new JPopupMenu();

		JMenuItem item = new JMenuItem(Ariane.getString("OperatorTreeEditor.addOperator"));
		item.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed( ActionEvent event ) {
				addNewItemBelowCurrentSelection();
			}
		});
		popupMenu.add(item);
		item = new JMenuItem(Ariane.getString("TreeEditor.editCategory"));
		item.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed( ActionEvent event ) {
				editCurrentItem();
			}
		});
		popupMenu.add(item);
		item = new JMenuItem(Ariane.getString("TreeEditor.deleteCategory"));
		item.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed( ActionEvent event ) {
				deleteSelectedItem();
			}
		});
		popupMenu.add(item);
		popupMenu.show(tree, x, y);
	}

	/**
	 * Creates and displays a popup menu
	 * for the current selected operator in the tree.
	 * 
	 * @param tree the current tree.
	 * @param x the x coordinate position to popup the menu.
	 * @param y the y coordinate position to popup the menu.
	 */
	@Override
	protected void openPopupMenuItem( final JTree tree, int x, int y ) {
		JPopupMenu popupMenu = new JPopupMenu();
		JMenuItem item;
		item = new JMenuItem(Ariane.getString("TreeEditor.modify"));
		item.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed( ActionEvent event ) {
				editCurrentItem();
			}
		});
		popupMenu.add(item);
		item = new JMenuItem(Ariane.getString("TreeEditor.delete"));
		item.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed( ActionEvent event ) {
				deleteSelectedItem();
			}
		});
		popupMenu.add(item);
		popupMenu.show(tree, x, y);
	}

	/**
	 * Stores the specified subtree in the current explorer
	 * tree and in the related file.
	 * 
	 * @param node the root node of the modified subtree.
	 * @return true if all preferences has been saved.
	 */
	@Override
	protected boolean store( MutableTreeItem node ) {
		Ariane.getWorkspace().getTree().setUserOperatorTree(node);

		File file = new File(Ariane.getPersonalOperatorFilename());
		File directory = file.getParentFile();
		try {
			directory.mkdirs();
			XMLFileOutputStream out = new XMLFileOutputStream(file, null);
			out.write(saveOperators());
			out.close();
		} catch (IOException e) {
			Ariane.getWorkspace().printErrorMessage(Ariane.getString("PreferenceChooser.error.saved") + " " + Ariane.getPreferencesFilename() + " " + e.getLocalizedMessage());
			return false;
		}
		return true;
	}

	/**
	 * Saves personal operators as a XML tree.
	 * 
	 * @return a XML tree.
	 * @throws IOException in case of IO problems.
	 */
	private static XMLElement saveOperators() throws IOException {
		List<TreeItem> operatorList = Ariane.getWorkspace().getTree().getUserOperatorList();
		if (!operatorList.isEmpty()) {
			XMLElement element = new XMLElement("OperatorLibrary");
			for (TreeItem item : operatorList) {
				element.addChildren(item.save());
			}
			return element;
		}
		return null;
	}
}