/* -*- c-basic-offset: 3 -*-
 *
 * Copyright (c) 2013, GREYC.
 * All rights reserved
 *
 * You may use this file under the terms of the BSD license as follows:
 *
 * "Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of the GREYC, nor the name of its
 *     contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
 * 
 *   
 * For more information, refer to:
 * https://clouard.users.greyc.fr/Pandore
 */

package fr.greyc.pantheon.ariane.ui;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.FlowLayout;
import java.awt.Font;
import java.awt.Frame;
import java.awt.Point;
import java.awt.Toolkit;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.FocusAdapter;
import java.awt.event.FocusEvent;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.StringTokenizer;

import javax.swing.AbstractAction;
import javax.swing.BorderFactory;
import javax.swing.Box;
import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JColorChooser;
import javax.swing.JComboBox;
import javax.swing.JFileChooser;
import javax.swing.JLabel;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JSpinner;
import javax.swing.JTabbedPane;
import javax.swing.JTextField;
import javax.swing.SpinnerModel;
import javax.swing.SpinnerNumberModel;
import javax.swing.SwingUtilities;
import javax.swing.ToolTipManager;
import javax.swing.border.EtchedBorder;

import fr.greyc.pantheon.ariane.Ariane;
import fr.greyc.pantheon.ariane.graph.Graph;
import fr.greyc.pantheon.ariane.graph.item.ControlItem;
import fr.greyc.pantheon.ariane.graph.item.ControlStructureItem;
import fr.greyc.pantheon.ariane.graph.item.DataFile;
import fr.greyc.pantheon.ariane.graph.item.GraphItem;
import fr.greyc.pantheon.ariane.graph.item.NoteItem;
import fr.greyc.pantheon.ariane.graph.item.OperatorItem;
import fr.greyc.pantheon.ariane.utils.GDialog;
import fr.greyc.pantheon.ariane.xml.XMLElement;
import fr.greyc.pantheon.ariane.xml.XMLFileOutputStream;

/**
 * The class <code>Preferences</code> stores the current values
 * for all user's preferences and defines the dialog box
 * that allows the user to change this values.
 * <br>
 * This class sounds like the @link{java.util.prefs.Preferences Preferences class},
 * but the persistence is ensured by our own methods.
 * Each preference is handled using the static get and put methods:
 * <pre>
 * Object value = PreferencesChooser.get("property_name",default_value);
 * PreferencesChooser.set("property_name", value);
 * </pre>
 *
 * @author Régis CLOUARD - Oct 01, 1998
 * @version Sébastien LEFÈVRE, Richard LOBRÉAU - Feb 21, 1999
 * @version Adrien SCHRÈQUE, Hicham AMADI - Nov 05, 2000
 * @version Mathieu JOETS - Jul 09, 2001
 * @version Régis CLOUARD - Oct 16, 2001
 * @version Régis CLOUARD - Dec 30, 2006 (revision)
 * @version Frédéric LAIR - Jun 28, 2007 (add selected Links color)
 * @version Régis CLOUARD - Jun 10, 2010 (improve rendering)
 * @version Régis CLOUARD - Jan 21, 2011 (use map to store preferences)
 * @version Régis CLOUARD - Jan 12, 2013 (add browser and text editor)
 */
public final class PreferenceChooser extends GDialog {
	public static final String UNDO_HISTORY_SIZE = "UndoHistorySize";
	public static final String USE_OPERATOR_LIBRARY = "UseOperatorLibrary";
	public static final String BINARY_DIRECTORY = "BinaryDirectory";
	private static final String LANGUAGE = "Language";
	public static final String FRAME_SIZE = "Frame.Size";
	public static final String TEXT_EDITOR = "TextEditor";
	private static final String USE_TOOLTIP = "UseToolTip";
	public static final String WEB_BROWSER = "WebBrowser";
	public static final String ARIANE_BROWSE = "Ariane.browse";
	public static final String FILE_DIRECTORY = "FileDirectory";
	private static final String FONT_SIZE = "FontSize";
	public static final String CODE_GENERATOR_PREFERENCE = "CodeGenerator.Preference";
	private static final String TEMPORARY_FILE_EXTENSION = "TemporaryFileExtension";
	public static final String FRAME_MAXIMIZED = "Frame.Maximized";
	public static final String SPLITPANE_H1_RATIO = "SplitPane.H1.Ratio";
	public static final String SPLITPANE_H2_RATIO = "SplitPane.H2.Ratio";
	public static final String SPLITPANE_V1_RATIO = "SplitPane.V1.Ratio";
	public static final String SPLITPANE_V2_RATIO = "SplitPane.V2.Ratio";
	
	/** The empty character for the item comment. */
	public static final String EMPTY_STRING = "\t";

	/** The temporary file extension. Keep this as variable for sake of efficiency. */
	private static String _temporaryFileExtension = ".tmp";

	/** The temporary file extension. Keep this as variable for sake of efficiency. */
	private static String _temporaryFilePrefix = "data";

	private static Locale _locale;

	private static Font _defaultFont;

	private static int _defaultFontSize;

	/** Stores the last opened tab index. */
	private static int _lastestSelectedTab;

	/** The table of colors of the graphical components. */
	private Color[] _colorArray;

	/** The number of colors*/
	private static final int ITEM_COUNT = 9;

	/** The table of legends of the graphical components. */
	private String[] _colorStringArray;

	/** The current workspace. */
	private final Workspace _workspace;

	/** The name of the current language. */
	private String _language;

	/** The name of the visualization program. */
	private String _visuCommand;

	/** The name of the status command. */
	private String _statusCommand;

	/** The name of the temporary directory. */
	private String _tempDir;
	/** The file directory. */
	private String _fileDirectory;
	/** The operator binary path. */
	private String _binaryDirectory;

	private static Font _fontDialogPLAIN6;
	private static Font _fontDialogPLAIN8;
	private static Font _fontDialogPLAIN9;
	private static Font _fontDialogPLAIN11;
	private static Font _fontDialogPLAIN10;
	private static Font _fontDialogPLAIN12;
	private static Font _fontDialogPLAIN16;

	/** The checkbox for tip options. */
	private final JCheckBox _tipc = new JCheckBox();

	/** The  checkbox for desktop browser. */
	private final JTextField _browser = new JTextField(10);
	private final JTextField _editor = new JTextField(10);

	/** The undo history size. */
	private JSpinner _undoSizeSpinner;

	/** Stores all the preferences. */
	private static Map<String, Object> _preferences = new HashMap<String, Object>(); // NOPMD by regis on 8/8/14 10:38 AM

	/** The tabbed pane that supports the preferences elements. */
	private final JTabbedPane _prefsTabbedPane = new JTabbedPane();

	/**
	 * Stores the specified value with the specified key
	 * in this preference node.
	 * 
	 * @param key the name of the property.
	 * @param value the property value.
	 */
	public static void putBoolean( final String key, final boolean value ) {
		_preferences.put(key, value);
	}

	/**
	 * Returns the preference value for the specified property name.
	 * 
	 * @param key the name of the property.
	 * @return the property value.
	 */
	public static boolean getBoolean( String key ) {
		if (_preferences.containsKey(key)) {
			return (Boolean)_preferences.get(key);
		} else {
			return false;
		}
	}

	/**
	 * Stores the specified value with the specified key
	 * in this preference node.
	 * 
	 * @param key the name of the property.
	 * @param value the property value.
	 */
	public static void put( String key, String value) {
		_preferences.put(key, value);
	}

	/**
	 * Returns the preference value for the specified property name.
	 * 
	 * @param key the name of the property.
	 * @return the property name.
	 */
	public static String get( String key ) {
		return (String)_preferences.get(key);
	}

	/**
	 * Stores the specified value with the specified key
	 * in this preference node.
	 * 
	 * @param key the name of the property.
	 * @param value the property value.the dimension
	 */
	public static void putDimension( String key, Dimension value ) {
		_preferences.put(key, value);
	}

	/**
	 * Returns the preference value for the specified property name.
	 * 
	 * @param key the name of the property.
	 * @return the property name.
	 */
	public static Dimension getDimension( String key ) {
		if (_preferences.containsKey(key)) {
			return (Dimension)_preferences.get(key);
		} else {
			return null;
		}
	}

	/**
	 * Stores the specified value with the specified key
	 * in this preference node.
	 * 
	 * @param key the name of the property.
	 * @param value the coordinates.
	 */
	public static void putPoint( String key, Point value ) {
		_preferences.put(key, value);
	}

	/**
	 * Returns the preference value for the specified property name.
	 * 
	 * @param key the name of the property.
	 * @return the property name.
	 */
	public static Point getPoint( String key ) {
		return (Point)_preferences.get(key);
	}

	/**
	 * Stores the specified value with the specified key
	 * in this preference node.
	 * 
	 * @param key the name of the property.
	 * @param value an integer value.
	 */
	public static void putInt( String key, int value ) {
		_preferences.put(key, value);
	}

	/**
	 * Returns the preference value for the specified property name.
	 * 
	 * @param key the name of the property.
	 * @return the property name.
	 */
	public static int getInt( String key ) {
		return (Integer)_preferences.get(key);
	}

	public static String createTemporaryName( int graphIndex, int itemIndex, int slotIndex ) {
		String name = _temporaryFilePrefix
				+ "_"
				+ graphIndex
				+ "_" + itemIndex
				+ "_" + slotIndex + PreferenceChooser._temporaryFileExtension;
		return name;
	}

	public static String getTemporaryFilePrefix() {
		return _temporaryFilePrefix;
	}

	public static String getTemporaryFileExtension() {
		return _temporaryFileExtension;
	}

	/**
	 * Creates a dialog for setting new preferences.
	 * 
	 * @param workspace the related workspace.
	 */
	public PreferenceChooser( Workspace workspace ) {
		super(Ariane.getFrame(), Ariane.getString("PreferenceChooser"), true);
		_workspace = workspace;
		setResizable(false);
		displayContent();
		pack();
		setLocationRelativeTo(getOwner());
		addWindowListener(new WindowAdapter() {
			// Redirects the window closing event to cancelAction();
			@Override
			public void windowClosing( WindowEvent event ) {
				cancelAction();
			}
		});
	}

	/**
	 * Saves all the preferences in the given filename.
	 * 
	 * @return true if all preferences has been saved.
	 */
	public static boolean save() {
		File file = new File(Ariane.getPreferencesFilename());
		File directory = file.getParentFile();
		try {
			directory.mkdirs();
			XMLFileOutputStream out = new XMLFileOutputStream(file, null);
			out.write(savePreferences());
			out.write(saveRecentFiles());
			out.close();
		} catch (IOException ex) {
			Ariane.getWorkspace().printSystemMessage(Ariane.getString("PreferenceChooser.error.saved") + " " + Ariane.getPreferencesFilename());
			return false;
		}
		return true;
	}

	/**
	 * Loads all preferences from the given XML tree.
	 * 
	 * @param xmlTree the XML tree of preferences.
	 */
	public static void load( List<XMLElement> xmlTree1 ) {
		// Discards tag <ARIANE> from the xmlTree.
		List<XMLElement> properties = xmlTree1.get(0).getChildren();
		Iterator<XMLElement> iterator = properties.iterator();

		XMLElement element;
		try {
			while (iterator.hasNext()) {
				element = iterator.next();
				if (element.getName().equalsIgnoreCase("Preferences")) {
					loadPreferences(element.getChildren());
				} else if (element.getName().equalsIgnoreCase("RecentFiles") && element.getChildren() != null) {
					loadRecentFiles(element.getChildren());
				}
			}
		} catch (IOException e) {
//			e.printStackTrace();
		}
	}

	///// 
	///// F O N T
	///// 

	/**
	 * Loads the various fonts.
	 * Sets component with the right font.
	 */
	private static void loadFonts( ) {
		int screenRes = Toolkit.getDefaultToolkit().getScreenResolution();
		int fontSize = (int)Math.round(12.0 * screenRes / 72.0);
		Font font = new Font("Arial", Font.PLAIN, fontSize);
		_fontDialogPLAIN6 = font.deriveFont(6f);
		_fontDialogPLAIN8 = font.deriveFont(8f);
		_fontDialogPLAIN9 = font.deriveFont(9f);
		_fontDialogPLAIN10 = font.deriveFont(10.0f);
		_fontDialogPLAIN11 = font.deriveFont(11.0f);
		_fontDialogPLAIN12 = font.deriveFont(12f);
		_fontDialogPLAIN16 = font.deriveFont(16f);
	}

	public static Font getFont( int size ) {
		switch(size) {
		case 6: return _fontDialogPLAIN6;
		case 8: return _fontDialogPLAIN8;
		case 9: return _fontDialogPLAIN9;
		case 10: return _fontDialogPLAIN10;
		case 11: return _fontDialogPLAIN11;
		case 12: return _fontDialogPLAIN12;
		case 16: return _fontDialogPLAIN16;
		default: return null;
		}
	}
	
	/**
	 * Sets the default preferences.
	 * If the default language is not represented into
	 * the system, then English is the default language.
	 */
	public static void initPreferences( ) {
		loadFonts();
		putDimension(FRAME_SIZE, new Dimension(1000, 660));
		putInt(SPLITPANE_V1_RATIO, 232);
		putInt(SPLITPANE_V2_RATIO, 584);
		putInt(SPLITPANE_H1_RATIO, 436);
		putInt(SPLITPANE_H2_RATIO, 433);
		putInt(CODE_GENERATOR_PREFERENCE, 0);
		putBoolean(USE_OPERATOR_LIBRARY, true);
		putBoolean(USE_TOOLTIP, true);
		putInt(UNDO_HISTORY_SIZE, 200);
		_locale = Locale.getDefault();
		if (_locale.getDisplayLanguage().equals(Locale.FRENCH.getDisplayLanguage())) {
			setLocale("Français");
		} else {
			setLocale("English");
		}
		setFontSize(10);
	}

	public static Locale getCurrentLocale() {
		return _locale;
	}
	/**
	 * Initializes contents of the dialog.
	 */
	private void initializePreferences() {
		_colorArray = new Color[ITEM_COUNT];
		_colorStringArray = new String[ITEM_COUNT];
		int index = 0;
		_colorArray[index] = NoteItem.getColor();
		_colorStringArray[index++] = Ariane.getString("PreferenceChooser.note");

		_colorArray[index] = DataFile.getColor();
		_colorStringArray[index++] = Ariane.getString("PreferenceChooser.datafile");

		_colorArray[index] = OperatorItem.getColor();
		_colorStringArray[index++] = Ariane.getString("PreferenceChooser.operator");

		_colorArray[index] = ControlItem.getColor();
		_colorStringArray[index++] = Ariane.getString("PreferenceChooser.controlItem");

		_colorArray[index] = ControlStructureItem.getColor();
		_colorStringArray[index++] = Ariane.getString("PreferenceChooser.controlStructureItem");

		_colorArray[index] = Graph.getBackgroundColor();
		_colorStringArray[index++] = Ariane.getString("PreferenceChooser.background");

		_colorArray[index] = Graph.getImageLinkColor();
		_colorStringArray[index++] = Ariane.getString("PreferenceChooser.imagelinks");

		_colorArray[index] = Graph.getValueLinkColor();
		_colorStringArray[index++] = Ariane.getString("PreferenceChooser.numericlinks");

		_colorArray[index] = Graph.getMaskLinkColor();
		_colorStringArray[index++] = Ariane.getString("PreferenceChooser.masklinks");

		_language = get(LANGUAGE);
		_defaultFontSize = getInt(FONT_SIZE);
		_visuCommand = GraphItem.getCommandVisu();
		_statusCommand = GraphItem.getCommandStatus();
		_tempDir = Workspace.getTempDirBasename();
		_fileDirectory = get(FILE_DIRECTORY);
		_binaryDirectory = get(BINARY_DIRECTORY);
		_tipc.setSelected(getBoolean(USE_TOOLTIP));
		_browser.setText(get(WEB_BROWSER));
		_editor.setText(get(TEXT_EDITOR));
	}

	/**
	 * Sets the new user's preferences values as the
	 * current preferences.
	 */
	protected void changePreferences() {
		int index = 0;
		NoteItem.setColor(_colorArray[index++]);
		DataFile.setColor(_colorArray[index++]);
		OperatorItem.setColor(_colorArray[index++]);
		ControlItem.setColor(_colorArray[index++]);
		ControlStructureItem.setColor(_colorArray[index++]);

		Graph.setBackgroundColor(_colorArray[index++]);
		Graph.setImageLinkColor(_colorArray[index++]);
		Graph.setValueLinkColor(_colorArray[index++]);
		Graph.setMaskLinkColor(_colorArray[index++]);
		changeLanguage(_language);
		setFontSize(_defaultFontSize);
		GraphItem.setCommandVisu(_visuCommand);
		GraphItem.setCommandStatus(_statusCommand);
		Workspace.setTempDir(_tempDir);
		put(FILE_DIRECTORY, _fileDirectory);
		put(BINARY_DIRECTORY, _binaryDirectory);
		putInt(UNDO_HISTORY_SIZE, (Integer)_undoSizeSpinner.getValue());
		setToolTip(_tipc.isSelected());
		put(WEB_BROWSER, _browser.getText().trim());
		put(TEXT_EDITOR, _editor.getText().trim());
	}

	/**
	 * Sets the new state value of the tool tips option.
	 * 
	 * @param state the state.
	 */
	private static void setToolTip( boolean state ) {
		putBoolean(USE_TOOLTIP, state);
		ToolTipManager.sharedInstance().setEnabled(state);
	}

	/**
	 * Changes the current language and changes all texts 
	 * in the User Interface.
	 * 
	 * @param language the new language (Français, English).
	 */
	private static void changeLanguage( String language ) {
		setLocale(language);
		Ariane.getWorkspace().changeLanguage();
	}

	/**
	 * Changes the current language and changes all texts 
	 * in the User Interface.
	 * 
	 * @param language the new language (Français, English).
	 */
	private static void setFontSize( int fontSize ) {
		putInt(FONT_SIZE, fontSize);
		_defaultFontSize = fontSize;
		switch (fontSize) {
		case 8: _defaultFont = _fontDialogPLAIN8; break;
		case 9: _defaultFont = _fontDialogPLAIN9; break;
		case 10: _defaultFont = _fontDialogPLAIN10; break;
		case 11: _defaultFont = _fontDialogPLAIN11; break;
		case 12: _defaultFont = _fontDialogPLAIN12; break;
		default: _defaultFont = _fontDialogPLAIN10; break;
		}
	}

	public static Font getDefaultFont( ) {
		return _defaultFont;
	}

	/**
	 * Sets the current language to the specified language.
	 * Loads related resources from the resource files.
	 * 
	 * @param language the name of the language.
	 */
	private static void setLocale( String language ) {
		put(LANGUAGE, language);
		if ("Français".equals(language)) {
			_locale = new Locale("fr", "FR");
		} else if ("English".equals(language)) {
			_locale = new Locale("en", "US");
		}
		Locale.setDefault(_locale);
		Ariane.setLocale(_locale);
	}

	private Box createGeneralPreferencesTab() {
		Box panel = Box.createVerticalBox();
		JPanel panel1 = new JPanel();
		panel.add(panel1);
		panel1.setLayout(new FlowLayout(FlowLayout.LEFT));
		panel1.setBorder(BorderFactory.createTitledBorder(BorderFactory.createEtchedBorder(), Ariane.getString("PreferenceChooser.general.language")));

		// List of known languages.
		panel1.add(new JLabel(Ariane.getString("PreferenceChooser.general.language.label")));
		String[] languages = {"Français", "English"};
		JComboBox<String> list = new JComboBox<String>(languages);
		list.setSelectedItem(_language);
		list.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed( ActionEvent event ) {
				@SuppressWarnings("unchecked")
				JComboBox<String> comboBox = (JComboBox<String>)event.getSource();
				_language = (String)comboBox.getSelectedItem();
			}
		});
		panel1.add(list);

		JPanel panel2 = new JPanel(new FlowLayout(FlowLayout.LEFT));
		panel.add(panel2);
		panel2.setBorder(BorderFactory.createTitledBorder(BorderFactory.createEtchedBorder(), Ariane.getString("PreferenceChooser.general.tooltip")));
		panel2.add(_tipc);
		panel2.add(new JLabel(Ariane.getString("PreferenceChooser.general.tooltip.text")));


		JPanel panel3 = new JPanel(new FlowLayout(FlowLayout.LEFT));
		panel.add(panel3);
		panel3.setBorder(BorderFactory.createTitledBorder(BorderFactory.createEtchedBorder(EtchedBorder.RAISED), Ariane.getString("PreferenceChooser.undoHistoryTitle")));
		SpinnerModel spinnerModel;
		spinnerModel = new SpinnerNumberModel(getInt(UNDO_HISTORY_SIZE), 10, 500, 1);
		_undoSizeSpinner = new JSpinner(spinnerModel);
		panel3.add(new JLabel(Ariane.getString("PreferenceChooser.undoHistoryText")));
		panel3.add(_undoSizeSpinner);

		// Browser
		JPanel panel4 = new JPanel(new FlowLayout(FlowLayout.LEFT));
		panel.add(panel4);
		panel4.setBorder(BorderFactory.createTitledBorder(BorderFactory.createEtchedBorder(), Ariane.getString("PreferenceChooser.general.browser")));
		panel4.add(new JLabel(Ariane.getString("PreferenceChooser.general.browser.text")));
		panel4.add(_browser);
		JButton browser1 = new JButton(Ariane.getString(ARIANE_BROWSE));
		panel4.add(browser1);
		browser1.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed( ActionEvent event ) {
				final JFileChooser fileChooser = new JFileChooser();
				fileChooser.setDialogType(JFileChooser.OPEN_DIALOG);
				fileChooser.setDialogTitle(Ariane.getString("PreferenceChooser.general.selectProgram"));
				fileChooser.setFileSelectionMode(JFileChooser.FILES_ONLY);
				fileChooser.setCurrentDirectory(new File("."));
				int returnVal = fileChooser.showOpenDialog(PreferenceChooser.this);
				if (returnVal == JFileChooser.APPROVE_OPTION) {
					_browser.setText(fileChooser.getSelectedFile().getAbsolutePath());
				}
			}
		});

		// Editor
		JPanel panel5 = new JPanel(new FlowLayout(FlowLayout.LEFT));
		panel.add(panel5);
		panel5.setBorder(BorderFactory.createTitledBorder(BorderFactory.createEtchedBorder(), Ariane.getString("PreferenceChooser.general.editor")));
		panel5.add(new JLabel(Ariane.getString("PreferenceChooser.general.editor.text")));
		panel5.add(_editor);
		JButton browser2 = new JButton(Ariane.getString(ARIANE_BROWSE));
		panel5.add(browser2);
		browser2.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed( ActionEvent event ) {
				final JFileChooser fileChooser = new JFileChooser();
				fileChooser.setDialogType(JFileChooser.OPEN_DIALOG);
				fileChooser.setDialogTitle(Ariane.getString("PreferenceChooser.general.selectProgram"));
				fileChooser.setFileSelectionMode(JFileChooser.FILES_ONLY);
				fileChooser.setCurrentDirectory(new File("."));
				int returnVal = fileChooser.showOpenDialog(PreferenceChooser.this);
				if (returnVal == JFileChooser.APPROVE_OPTION) {
					_editor.setText(fileChooser.getSelectedFile().getAbsolutePath());
				}
			}
		});

		panel.add(Box.createGlue());
		return panel;
	}

	private Box createRuntimePreferencesTab() {
		Box panel = Box.createVerticalBox();

		// -> VISU command
		JPanel panel1 = new JPanel(new FlowLayout(FlowLayout.LEFT));
		panel.add(panel1);
		panel1.setBorder(BorderFactory.createTitledBorder(BorderFactory.createEtchedBorder(), Ariane.getString("PreferenceChooser.runtime.visuCommand")));
		final JTextField textfield2 = new JTextField(20);
		panel1.add(new JLabel(Ariane.getString("PreferenceChooser.runtime.CommandName")));

		textfield2.setText(_visuCommand);
		textfield2.addFocusListener(new FocusAdapter() {
			@Override
			public void focusLost( FocusEvent event ) {
				_visuCommand = ((JTextField)event.getSource()).getText();
			}
		});
		panel1.add(textfield2);

		// STATUS command
		JPanel panel2 = new JPanel(new FlowLayout(FlowLayout.LEFT));
		panel.add(panel2);
		panel2.setBorder(BorderFactory.createTitledBorder(BorderFactory.createEtchedBorder(), Ariane.getString("PreferenceChooser.runtime.statusCommand")));
		panel2.add(new JLabel(Ariane.getString("PreferenceChooser.runtime.CommandName")));
		final JTextField textfield3 = new JTextField(20);
		textfield3.setText(_statusCommand);
		textfield3.addFocusListener(new FocusAdapter() {
			@Override
			public void focusLost( FocusEvent event ) {
				_statusCommand = ((JTextField)event.getSource()).getText();
			}
		});
		panel2.add(textfield3);
		return panel;
	}

	private Box createGraphPreferencesTab() {
		Box panel = Box.createVerticalBox();

		// List of known languages.
		JPanel panel1 = new JPanel();
		panel.add(panel1);
		panel1.setLayout(new FlowLayout(FlowLayout.LEFT));
		panel1.setBorder(BorderFactory.createTitledBorder(BorderFactory.createEtchedBorder(), Ariane.getString("PreferenceChooser.graph.font")));

		panel1.add(new JLabel(Ariane.getString("PreferenceChooser.graph.font.size")));
		String[] fontSize = {"8", "9", "10", "11", "12"};
		JComboBox<String> list = new JComboBox<String>(fontSize);
		list.setSelectedItem(Integer.toString(_defaultFontSize));
		list.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed( ActionEvent event ) {
				@SuppressWarnings("unchecked")
				JComboBox<String> comboBox = (JComboBox<String>)event.getSource();
				_defaultFontSize = Integer.parseInt((String)comboBox.getSelectedItem());
			}
		});
		panel1.add(list);
		panel1.add(new JLabel(Ariane.getString("PreferenceChooser.graph.font.notice")));
		panel.add(panel1);

		for (int i = 5; i < ITEM_COUNT; i++) {
			JPanel panel2 = new JPanel(new FlowLayout(FlowLayout.LEFT));
			panel.add(panel2);
			panel2.setBorder(BorderFactory.createTitledBorder(BorderFactory.createEtchedBorder(), _colorStringArray[i]));
			panel2.add(new BindingColor(Ariane.getString("PreferenceChooser.backgroundColor"), i));
		}

		return panel;
	}

	private Box createColorPreferencesTab() {
		Box panel = Box.createVerticalBox();
		for (int i = 0; i < 5; i++) {
			JPanel panel1 = new JPanel(new FlowLayout(FlowLayout.LEFT));
			panel.add(panel1);
			panel1.setBorder(BorderFactory.createTitledBorder(BorderFactory.createEtchedBorder(), _colorStringArray[i]));
			panel1.add(new BindingColor(Ariane.getString("PreferenceChooser.backgroundColor"), i));
		}
		return panel;
	}

	private Box createPathPreferencesTab() {
		Box panel = Box.createVerticalBox();
		// tab 1 : Temporary paths
		JPanel panel1 = new JPanel();
		panel1.setLayout(new FlowLayout(FlowLayout.LEFT));
		panel1.setBorder(BorderFactory.createTitledBorder(BorderFactory.createEtchedBorder(), Ariane.getString("PreferenceChooser.path.temporaryDirectory")));
		JButton button1 = new JButton(Ariane.getString(ARIANE_BROWSE));
		final JTextField textfield1 = new JTextField(20);
		button1.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed( ActionEvent event ) {
				final JFileChooser fileChooser = new JFileChooser();
				fileChooser.setDialogType(JFileChooser.OPEN_DIALOG);
				fileChooser.setDialogTitle(Ariane.getString("FileChooser.titleSelectDir"));
				fileChooser.setFileSelectionMode(JFileChooser.DIRECTORIES_ONLY);
				fileChooser.setCurrentDirectory(new File(Workspace.getTempDirBasename()));
				int returnVal = fileChooser.showOpenDialog(PreferenceChooser.this);
				if (returnVal == JFileChooser.APPROVE_OPTION) {
					_tempDir = fileChooser.getSelectedFile().getAbsolutePath();
					textfield1.setText(_tempDir);
				}
			}
		});
		textfield1.setText(_tempDir);
		textfield1.addFocusListener(new FocusAdapter() {
			@Override
			public void focusLost( FocusEvent event ) {
				_tempDir = ((JTextField)event.getSource()).getText();
			}
		});
		panel1.add(textfield1);
		panel1.add(button1);
		panel.add(panel1);

		// tab 2: Temporary file extension
		JPanel panel2 = new JPanel(new FlowLayout(FlowLayout.LEFT));
		panel2.setBorder(BorderFactory.createTitledBorder(BorderFactory.createEtchedBorder(), Ariane.getString("PreferenceChooser.runtime.TemporaryFileExtension")));
		JTextField textfield2 = new JTextField(20);
		panel2.add(new JLabel(Ariane.getString("PreferenceChooser.runtime.fileExtension")));

		textfield2.setText(_temporaryFileExtension);
		textfield2.addFocusListener(new FocusAdapter() {
			@Override
			public void focusLost( FocusEvent event ) {
				_temporaryFileExtension = ((JTextField)event.getSource()).getText();
			}
		});
		panel2.add(textfield2);
		panel.add(Box.createGlue());

		JPanel panel21 = new JPanel(new FlowLayout(FlowLayout.LEFT));
		panel.add(panel21);
		panel21.setBorder(BorderFactory.createTitledBorder(BorderFactory.createEtchedBorder(), Ariane.getString("PreferenceChooser.environment.title")));
		// -> Remove temporary files
		panel21.add(new JLabel(Ariane.getString("PreferenceChooser.environment.cleaning")));
		JButton button2 = new JButton(Ariane.getString("PreferenceChooser.environment.erasing"));
		button2.addActionListener(new ActionListener() { // Erases all temporary files.
			@Override
			public void actionPerformed( ActionEvent event ) {
				int numberOfFiles = Workspace.deleteTemporaryFiles();
				if (numberOfFiles > 0) {
					JOptionPane.showMessageDialog(PreferenceChooser.this,
							numberOfFiles + Ariane.getString("PreferenceChooser.message1"),
							Ariane.getString("notice"),
							JOptionPane.INFORMATION_MESSAGE);
				}
			}
		} );
		panel21.add(button2);
		panel.add(panel2);
		panel.add(Box.createGlue());

		// File directory
		JPanel panel3 = new JPanel();
		panel3.setLayout(new FlowLayout(FlowLayout.LEFT));
		panel3.setBorder(BorderFactory.createTitledBorder(BorderFactory.createEtchedBorder(), Ariane.getString("PreferenceChooser.path.fileDirectory")));
		JButton button3 = new JButton(Ariane.getString(ARIANE_BROWSE));
		final JTextField textfield3 = new JTextField(20);
		button3.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed( ActionEvent event ) {
				final JFileChooser fileChooser = new JFileChooser();
				fileChooser.setDialogType(JFileChooser.OPEN_DIALOG);
				fileChooser.setDialogTitle(Ariane.getString("FileChooser.titleSelectDir"));
				fileChooser.setFileSelectionMode(JFileChooser.DIRECTORIES_ONLY);
				if (_fileDirectory != null ) {
					fileChooser.setCurrentDirectory(new File(_fileDirectory));
				}
				int returnVal = fileChooser.showOpenDialog(PreferenceChooser.this);
				if (returnVal == JFileChooser.APPROVE_OPTION) {
					_fileDirectory = fileChooser.getSelectedFile().getAbsolutePath();
					textfield3.setText(_fileDirectory);
				}
			}
		});
		textfield3.setText(_fileDirectory);
		textfield3.addFocusListener(new FocusAdapter() {
			@Override
			public void focusLost( FocusEvent event ) {
				_fileDirectory = ((JTextField)event.getSource()).getText();
				if (_fileDirectory.trim().isEmpty()) {
					_fileDirectory = null;
				}
			}
		});
		panel3.add(textfield3);
		panel3.add(button3);
		panel.add(panel3);

		// Personal operator paths
		JPanel panel4 = new JPanel();
		panel4.setLayout(new FlowLayout(FlowLayout.LEFT));
		panel4.setBorder(BorderFactory.createTitledBorder(BorderFactory.createEtchedBorder(), Ariane.getString("PreferenceChooser.path.binaryDirectory")));
		JButton button = new JButton(Ariane.getString(ARIANE_BROWSE));
		final JTextField textfield4 = new JTextField(20);
		button.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed( ActionEvent event ) {
				final JFileChooser fileChooser = new JFileChooser();
				fileChooser.setDialogType(JFileChooser.OPEN_DIALOG);
				fileChooser.setDialogTitle(Ariane.getString("FileChooser.titleSelectDir"));
				fileChooser.setFileSelectionMode(JFileChooser.DIRECTORIES_ONLY);
				if (_binaryDirectory != null) {
					fileChooser.setCurrentDirectory(new File(_binaryDirectory));
				}
				int returnVal = fileChooser.showOpenDialog(PreferenceChooser.this);
				if (returnVal == JFileChooser.APPROVE_OPTION) {
					_binaryDirectory = fileChooser.getSelectedFile().getAbsolutePath();
					textfield4.setText(_binaryDirectory);
				}
			}
		});
		textfield4.setText(_binaryDirectory);
		textfield4.addFocusListener(new FocusAdapter() {
			@Override
			public void focusLost( FocusEvent event ) {
				_binaryDirectory = ((JTextField)event.getSource()).getText();
				if (_binaryDirectory.trim().isEmpty()) {
					_binaryDirectory = null;
				}
			}
		});
		panel4.add(textfield4);
		panel4.add(button);
		panel.add(panel4);

		return panel;
	}

	/**
	 * Initializes the content of the dialog.
	 */
	private void displayContent() {
		initializePreferences();

		_prefsTabbedPane.addTab(Ariane.getString("PreferenceChooser.general.title"), createGeneralPreferencesTab());
		_prefsTabbedPane.addTab(Ariane.getString("PreferenceChooser.runtime.title"), createRuntimePreferencesTab());
		_prefsTabbedPane.addTab(Ariane.getString("PreferenceChooser.graph.title"), createGraphPreferencesTab());
		_prefsTabbedPane.addTab(Ariane.getString("PreferenceChooser.item.title"), createColorPreferencesTab());
		_prefsTabbedPane.addTab(Ariane.getString("PreferenceChooser.paths"), createPathPreferencesTab());
		_prefsTabbedPane.setSelectedIndex(_lastestSelectedTab);

		// Buttons panel
		final JPanel buttonPanel = new JPanel();
		JButton applyButton = new JButton(Ariane.getString("PreferenceChooser.applyText"), Ariane.getIcon("gtk-apply.png"));
		applyButton.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed( ActionEvent event ) {
				buttonPanel.requestFocusInWindow();
				// Tip: handle event before this event.
				SwingUtilities.invokeLater(new Runnable() {
					@Override
					public void run() {
						applyAction();
					}
				});
			}
		} );

		// cancel button
		JButton cancelButton = new JButton();
		cancelButton.setAction(new AbstractAction(Ariane.getString("TreeEditor.cancel"),
				Ariane.getIcon("gtk-cancel.png")) {
			@Override
			public void actionPerformed( ActionEvent event ) {
				cancelAction();
			}
		});

		JButton helpButton = new JButton();
		helpButton.setAction(new AbstractAction(null, Ariane.getIcon("help.png")) {
			@Override
			public void actionPerformed( ActionEvent event ) {
				Ariane.getWorkspace().displayMan("Ariane", "manual.html", "#tab" + _prefsTabbedPane.getSelectedIndex());
			}
		});
		buttonPanel.add(helpButton);
		buttonPanel.add(Box.createHorizontalStrut(80));
		buttonPanel.add(applyButton);
		buttonPanel.add(cancelButton);
		setDefaultButton(applyButton);
		setCancelButton(cancelButton);

		getContentPane().setLayout(new BorderLayout());
		getContentPane().add(buttonPanel, BorderLayout.SOUTH);
		getContentPane().add(_prefsTabbedPane, BorderLayout.CENTER);
		pack();
	}

	/**
	 * Closes the preferences dialog box.
	 */
	private void cancelAction() {
		setVisible(false);
		dispose();
	}

	/**
	 * Applies and saves the chosen preferences,
	 * and closes the dialog box.
	 */
	private void applyAction() {
		changePreferences();
		_workspace.repaintAllGraphs();
		save();
		setVisible(false);
		_lastestSelectedTab = _prefsTabbedPane.getSelectedIndex();
		dispose();
	}

	/**
	 * Codes a color as a triplet: r,g,b.
	 * 
	 * @param color the color to be coded.
	 * @return the color encoded as a string.
	 */
	private static String code( Color color ) {
		return color.getRed() + "," + color.getGreen() + "," + color.getBlue();
	}

	/**
	 * Decodes a color as a string to a <code>Color</code>.
	 * Color is given as a triplet r,g,b.
	 * 
	 * @param color the string to be uncoded.
	 * @return the <code>Color</code>.
	 */
	private static Color decode( String color ) {
		StringTokenizer colorValuePart = new StringTokenizer(color, ",");

		int red = Integer.parseInt(colorValuePart.nextToken());
		int green = Integer.parseInt(colorValuePart.nextToken());
		int blue = Integer.parseInt(colorValuePart.nextToken());
		return new Color(red, green, blue);
	}

	/**
	 * Saves the preferences as an XML tree.
	 * @return the XML tree.
	 * @throws IOException in case of IO problems.
	 */
	private static XMLElement savePreferences( ) throws IOException {
		XMLElement element = new XMLElement("Preferences");
		element.addChildren(new XMLElement("CurrentDir", Workspace.getCurrentDir()));
		if (!Workspace.getDefaultTempDir().equals(Workspace.getTempDirBasename())) {
			element.addChildren(new XMLElement("TempDir", Workspace.getTempDirBasename()));
		}
		element.addChildren(new XMLElement("PandoreDir", Ariane.getLibraryPath()));
		if (GraphItem.getCommandVisu().trim().length() > 0) {
			element.addChildren(new XMLElement("Command.Visu", GraphItem.getCommandVisu()));
		}
		if (GraphItem.getCommandStatus().trim().length() > 0) {
			element.addChildren(new XMLElement("Command.Status", GraphItem.getCommandStatus()));
		}
		if (!NoteItem.getDefaultColor().equals(NoteItem.getColor())) {
			element.addChildren(new XMLElement("Note.Color", code(NoteItem.getColor())));
		}
		if (!OperatorItem.getDefaultColor().equals(OperatorItem.getColor())) {
			element.addChildren(new XMLElement("Operator.Color", code(OperatorItem.getColor())));
		}
		if (!DataFile.getDefaultColor().equals(DataFile.getColor())) {
			element.addChildren(new XMLElement("Datafile.Color", code(DataFile.getColor())));
		}
		if (!ControlItem.getDefaultColor().equals(ControlItem.getColor())) {
			element.addChildren(new XMLElement("Control.Color", code(ControlItem.getColor())));
		}
		if (!ControlStructureItem.getDefaultColor().equals(ControlStructureItem.getColor())) {
			element.addChildren(new XMLElement("Module.Color", code(ControlStructureItem.getColor())));
		}
		if (!Graph.getDefaultBackgroundColor().equals(Graph.getBackgroundColor())) {
			element.addChildren(new XMLElement("Graph.Background.Color", code(Graph.getBackgroundColor())));
		}
		if (!Graph.getDefaultImageLinkColor().equals(Graph.getImageLinkColor())) {
			element.addChildren(new XMLElement("Image.Link.Color", code(Graph.getImageLinkColor())));
		}
		if (!Graph.getDefaultValueLinkColor().equals(Graph.getValueLinkColor())) {
			element.addChildren(new XMLElement("Value.Link.Color", code(Graph.getValueLinkColor())));
		}
		if (!Graph.getDefaultMaskLinkColor().equals(Graph.getMaskLinkColor())) {
			element.addChildren(new XMLElement("Mask.Link.Color", code(Graph.getMaskLinkColor())));
		}
		element.addChildren(new XMLElement(LANGUAGE, get(LANGUAGE)));
		element.addChildren(new XMLElement(FONT_SIZE, Integer.toString(getInt(FONT_SIZE))));
		element.addChildren(new XMLElement("ToolTip", Boolean.toString(getBoolean(USE_TOOLTIP))));
		String web = get(WEB_BROWSER);
		if (web != null && !web.trim().isEmpty()) {
			element.addChildren(new XMLElement(WEB_BROWSER, web));
		}
		String editor = get(TEXT_EDITOR);
		if (editor != null && !editor.trim().isEmpty()) {
			element.addChildren(new XMLElement(TEXT_EDITOR, editor));
		}

		Dimension dimension = Ariane.getFrame().getFinalSize();
		element.addChildren(new XMLElement("Frame.Width", Integer.toString(dimension.width)));
		element.addChildren(new XMLElement("Frame.Height", Integer.toString(dimension.height)));
		element.addChildren(new XMLElement(FRAME_MAXIMIZED, Boolean.toString(Ariane.getFrame().getExtendedState() == Frame.MAXIMIZED_BOTH)));

		element.addChildren(new XMLElement("SplitH1.ratio", Integer.toString(Ariane.getWorkspace().getSplitH1())));
		element.addChildren(new XMLElement("SplitH2.ratio", Integer.toString(Ariane.getWorkspace().getSplitH2())));
		element.addChildren(new XMLElement("SplitV1.ratio", Integer.toString(Ariane.getWorkspace().getSplitV1())));
		element.addChildren(new XMLElement("SplitV2.ratio", Integer.toString(Ariane.getWorkspace().getSplitV2())));
		element.addChildren(new XMLElement(USE_OPERATOR_LIBRARY, Boolean.toString(getBoolean(USE_OPERATOR_LIBRARY))));
		element.addChildren(new XMLElement(TEMPORARY_FILE_EXTENSION, _temporaryFileExtension));
		element.addChildren(new XMLElement(UNDO_HISTORY_SIZE, Integer.toString(getInt(UNDO_HISTORY_SIZE))));
		element.addChildren(new XMLElement(CODE_GENERATOR_PREFERENCE, Integer.toString(getInt(CODE_GENERATOR_PREFERENCE))));
		if (get(FILE_DIRECTORY) != null) {
			element.addChildren(new XMLElement(FILE_DIRECTORY, get(FILE_DIRECTORY)));
		}
		if (get(BINARY_DIRECTORY) != null) {
			element.addChildren(new XMLElement(BINARY_DIRECTORY, get(BINARY_DIRECTORY)));
		}
		return element;
	}

	/**
	 * Saves the recent files as an XML tree.
	 * 
	 * @return the XML tree.
	 * @throws IOException in case of IO problems.
	 */
	private static XMLElement saveRecentFiles() throws IOException {
		XMLElement element = new XMLElement("RecentFiles");
		for (String filename : Workspace.getRecentProjectFilenames()) {
			element.addChildren(new XMLElement("File", filename));
		}
		return element;
	}

	/**
	 * Sets all preferences in the current project.
	 * 
	 * @param xmlTree the XML tree of preferences.
	 * @throws IOException in case of IO problems.
	 */
	private static void loadPreferences( List<XMLElement> xmlTree ) throws IOException {
		Iterator<XMLElement> iterator = xmlTree.iterator();
		while (iterator.hasNext()) {
			XMLElement element = iterator.next();
			// A l'interieur de <Preferences>
			if (element.getName().equalsIgnoreCase("CurrentDir")) {
				Workspace.setCurrentDir(element.getValue());
			} else if (element.getName().equalsIgnoreCase("TempDir")) {
				Workspace.setTempDir(element.getValue());
			} else if (element.getName().equalsIgnoreCase("PandoreDir")) {
				Ariane.setLibraryPath(element.getValue());
			} else if (element.getName().equalsIgnoreCase("Command.Visu")) {
				GraphItem.setCommandVisu(element.getValue());
			} else if (element.getName().equalsIgnoreCase("Command.Status")) {
				GraphItem.setCommandStatus(element.getValue());
			} else if (element.getName().equalsIgnoreCase("Note.Color")) {
				NoteItem.setColor(decode(element.getValue()));
			} else if (element.getName().equalsIgnoreCase("Operator.Color")) {
				OperatorItem.setColor(decode(element.getValue()));
			} else if (element.getName().equalsIgnoreCase("Datafile.Color")) {
				DataFile.setColor(decode(element.getValue()));
			} else if (element.getName().equalsIgnoreCase("Control.Color")) {
				ControlItem.setColor(decode(element.getValue()));
			} else if (element.getName().equalsIgnoreCase("Composite.Color")) {
				ControlStructureItem.setColor(decode(element.getValue()));
			} else if (element.getName().equalsIgnoreCase("Graph.Background.Color")) {
				Graph.setBackgroundColor(decode(element.getValue()));
			} else if (element.getName().equalsIgnoreCase("Image.Link.Color")) {
				Graph.setImageLinkColor(decode(element.getValue()));
			} else if (element.getName().equalsIgnoreCase("Value.Link.Color")) {
				Graph.setValueLinkColor(decode(element.getValue()));
			} else if (element.getName().equalsIgnoreCase("Mask.Link.Color")) {
				Graph.setMaskLinkColor(decode(element.getValue()));
			} else if (element.getName().equalsIgnoreCase(LANGUAGE)) {
				setLocale(element.getValue());
			} else if (element.getName().equalsIgnoreCase(FONT_SIZE)) {
				int size = Integer.parseInt(element.getValue());
				setFontSize(size);
			} else if (element.getName().equalsIgnoreCase("ToolTip")) {
				setToolTip(Boolean.parseBoolean(element.getValue()));
			} else if (element.getName().equalsIgnoreCase(WEB_BROWSER)) {
				put(WEB_BROWSER, element.getValue());
			} else if (element.getName().equalsIgnoreCase(TEXT_EDITOR)) {
				put(TEXT_EDITOR, element.getValue());
			} else if (element.getName().equalsIgnoreCase("Frame.Width")) {
				Dimension dimension = getDimension(FRAME_SIZE);
				dimension.width = Integer.parseInt(element.getValue());
				putDimension(FRAME_SIZE, dimension);
			} else if (element.getName().equalsIgnoreCase("Frame.Height")) {
				Dimension dimension = getDimension(FRAME_SIZE);
				dimension.height = Integer.parseInt(element.getValue());
				putDimension(FRAME_SIZE, dimension);
			} else if (element.getName().equalsIgnoreCase(FRAME_MAXIMIZED)) {
				putBoolean("Frame.Maximized", Boolean.parseBoolean(element.getValue()));
			} else if (element.getName().equalsIgnoreCase("SplitH1.ratio")) {
				putInt(SPLITPANE_H1_RATIO, Integer.parseInt(element.getValue()));
			} else if (element.getName().equalsIgnoreCase("SplitH2.ratio")) {
				putInt(SPLITPANE_H2_RATIO, Integer.parseInt(element.getValue()));
			} else if (element.getName().equalsIgnoreCase("SplitV1.ratio")) {
				putInt(SPLITPANE_V1_RATIO, Integer.parseInt(element.getValue()));
			} else if (element.getName().equalsIgnoreCase("SplitV2.ratio")) {
				putInt(SPLITPANE_V2_RATIO, Integer.parseInt(element.getValue()));
			} else if (element.getName().equalsIgnoreCase(USE_OPERATOR_LIBRARY)) {
				putBoolean(USE_OPERATOR_LIBRARY, Boolean.parseBoolean(element.getValue()));
			} else if (element.getName().equalsIgnoreCase(TEMPORARY_FILE_EXTENSION)) {
				_temporaryFileExtension = element.getValue();
			} else if (element.getName().equalsIgnoreCase(UNDO_HISTORY_SIZE)) {
				putInt(UNDO_HISTORY_SIZE, Integer.parseInt(element.getValue()));
			} else if (element.getName().equalsIgnoreCase(CODE_GENERATOR_PREFERENCE)) {
				putInt(CODE_GENERATOR_PREFERENCE, Integer.parseInt(element.getValue()));
			} else if (element.getName().equalsIgnoreCase(FILE_DIRECTORY)) {
				element.addChildren(new XMLElement(FILE_DIRECTORY, get(FILE_DIRECTORY)));
				put(FILE_DIRECTORY, element.getValue());
			} else if (element.getName().equalsIgnoreCase(BINARY_DIRECTORY)) {
				element.addChildren(new XMLElement(BINARY_DIRECTORY, get(BINARY_DIRECTORY)));
				put(BINARY_DIRECTORY, element.getValue());
			}
		}
	}

	/**
	 * Gets the list of recent files.
	 * 
	 * @param xmlTree the XML tree of preferences.
	 * @throws IOException in case of IO problems.
	 */
	private static void loadRecentFiles( List<XMLElement> xmlTree ) throws IOException {
		Iterator<XMLElement> iterator = xmlTree.iterator();
		List<String> filenames = new ArrayList<String>();
		while (iterator.hasNext()) {
			XMLElement element = iterator.next();
			if (element.getName().equalsIgnoreCase("File")) {
				filenames.add(element.getValue());
			}
		}
		Workspace.setRecentProjectFilenames(filenames);
	}

	/**
	 * The class <code>BindingColor</code> is an inner class
	 * that represents a pair (item, color).
	 * It provides a pane of controls designed to allow a user
	 * to choose a color for a given item.
	 */
	private final class BindingColor extends JPanel {
		/**  The index of the color of the item. */
		private final int _numColor;
		/**  The text that describes the item. */
		private final String _legend;
		/** Chosen button. */
		private final JPanel _colorButton;

		/**
		 * Creates the pair (item/color).
		 * 
		 * @param legend the legend of the item.
		 * @param numColor the index of the color for the given item.
		 */
		public BindingColor( final String legend, int numColor ) {
			_legend = legend;
			_numColor = numColor;
			_colorButton = new JPanel();
			_colorButton.setPreferredSize(new Dimension(30, 30));
			_colorButton.setBorder(BorderFactory.createEtchedBorder());
			_colorButton.setBackground(_colorArray[_numColor]);
			_colorButton.addMouseListener(new MouseAdapter() {
				@Override
				public void mouseClicked( final MouseEvent event ) {
					selectColor(event.getComponent()); }
			});
			setLayout(new BorderLayout());
			JLabel label = new JLabel(_legend + "   ");
			label.setFont(new Font("Dialog", Font.PLAIN, 12));
			add(label, BorderLayout.CENTER);
			add(_colorButton, BorderLayout.EAST);
		}

		private void selectColor( Component item ) {
			final Color col = JColorChooser.showDialog(item, _legend, _colorArray[_numColor]);
			if (col != null) {
				_colorArray[_numColor] = col;
				_colorButton.setBackground(col);
				_colorButton.repaint();
			}
		}
	}
}