/* -*- c-basic-offset: 3 -*-
 *
 * Copyright (c) 2013, GREYC.
 * All rights reserved
 *
 * You may use this file under the terms of the BSD license as follows:
 *
 * "Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of the GREYC, nor the name of its
 *     contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
 * 
 *   
 * For more information, refer to:
 * https://clouard.users.greyc.fr/Pandore
 */

package fr.greyc.pantheon.ariane.ui;

import fr.greyc.pantheon.ariane.Ariane;
import fr.greyc.pantheon.ariane.graph.item.CompositeItem;
import fr.greyc.pantheon.ariane.graph.slot.InputParameterSlot;
import fr.greyc.pantheon.ariane.utils.GDialog;
import fr.greyc.pantheon.ariane.utils.MutableTreeItem;
import fr.greyc.pantheon.ariane.utils.SubroutineTreeItem;
import fr.greyc.pantheon.ariane.utils.TreeItem;
import fr.greyc.pantheon.ariane.xml.XMLElement;
import fr.greyc.pantheon.ariane.xml.XMLFileOutputStream;

import java.awt.BorderLayout;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.io.File;
import java.io.IOException;
import java.util.List;

import javax.swing.AbstractAction;
import javax.swing.BorderFactory;
import javax.swing.JComboBox;
import javax.swing.JDialog;
import javax.swing.JButton;
import javax.swing.JFileChooser;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JTextField;
import javax.swing.tree.DefaultTreeModel;

/**
 * The class <code>SubroutineEditor</code> defines a dialog box
 * where the user can add new subroutine in the explorer.
 * 
 * @version Régis CLOUARD - Jul 20, 2010
 */
public final class SubroutineEditor extends GDialog {
	private SubroutineTreeItem _currentItem;
	private MutableTreeItem _currentNode;
	private DefaultTreeModel _treeModel;
	private JComboBox<String> _categoryField;
	private JTextField _nameField;
	private JTextField _signatureField;
	private JTextField _docURLField;
	private boolean _inserted;
	private boolean _validated;

	/**
	 * Creates a dialog box for new settings.
	 * @param item the item to be stored.
	 */
	public SubroutineEditor( CompositeItem item ) {
		super((JDialog)null, Ariane.getString("SubroutineEditor.title"), true);
		addWindowListener(new WindowAdapter() {
			@Override
			public void windowClosing( WindowEvent event ) {
				cancelAction();
			}
		});

		try {
			XMLElement element;
			element = new XMLElement("Subroutine");
			element.addChildren(new XMLElement("Category", ""));
			element.addChildren(new XMLElement("Name", item.getNickName()));
			element.addChildren(new XMLElement("Documentation" , ""));
			// Build the default parameter values list
			String list = "";
			for (int i = 0; i < item.getInputParameterSlots().size(); i++ ) {
				InputParameterSlot parameter = item.getInputParameterSlots().get(i);
				String name;
				if (parameter.hasName()) {
					name = parameter.getName();
				} else {
					name = "";
				}
				String defaultValue;
				if (!parameter.isConnected() && parameter.hasValue(false)) {
					defaultValue = "=" + parameter.getValue();
				} else {
					defaultValue = "";
				}
				if (name.length() > 0 || defaultValue.length() > 0) {
					list += "$" + i + ":" + name + defaultValue + ";";
				}
			}
			element.addChildren(new XMLElement("Signature" , list));
			element.addChildren(item.save(null));
			_currentItem = new SubroutineTreeItem(element.getChildren());
			_inserted = true;
		} catch (IOException e) {
		}
		displayContent();
		setLocationRelativeTo(null);
	}

	/**
	 * Creates and shows the dialog box.
	 * 
	 * @param parent the parent for this dialog box.
	 * @param currentNode the node which supports the operator.
	 * @param treeModel the tree model.
	 */
	public SubroutineEditor( JDialog parent, MutableTreeItem currentNode, DefaultTreeModel treeModel ) {
		super(parent, Ariane.getString("SubroutineEditor.title"), true);
		_currentNode = currentNode;
		_currentItem = (SubroutineTreeItem)_currentNode.getUserObject();
		_treeModel = treeModel;
		_inserted = false;
		displayContent();
		setLocationRelativeTo(parent);
	}

	/**
	 * Displays the content of the dialog box.
	 */
	private void displayContent( ) {
		setResizable(false);
		JPanel textPanel = new JPanel();
		textPanel.setLayout(new GridBagLayout());
		textPanel.setBorder(BorderFactory.createTitledBorder(BorderFactory.createEtchedBorder(), Ariane.getString("SubroutineEditor.borderTitle")));
		GridBagConstraints cst = new GridBagConstraints();

		cst.fill = GridBagConstraints.BOTH;
		textPanel.add(new JLabel(Ariane.getString("SubroutineEditor.categoryText")), cst);
		String[] categories = Ariane.getWorkspace().getTree().getSubroutineCategories();
		_categoryField = new JComboBox<String>(categories);
		_categoryField.setEditable(true);
		_categoryField.setSelectedItem(_currentItem.getCategory());
		cst.gridwidth = GridBagConstraints.REMAINDER; //end row
		textPanel.add(_categoryField, cst);

		cst.gridwidth = 1;
		textPanel.add(new JLabel(Ariane.getString("SubroutineEditor.nameText")), cst);
		_nameField = new JTextField(30);
		_nameField.setText(_currentItem.getName());
		cst.gridwidth = GridBagConstraints.REMAINDER; //end row
		textPanel.add(_nameField, cst);

		// Signature
		cst.gridwidth = 1;
		textPanel.add(new JLabel(Ariane.getString("SubroutineEditor.signature")), cst);
		_signatureField = new JTextField(_currentItem.getSignature());
		cst.gridwidth = GridBagConstraints.REMAINDER; //end row
		textPanel.add(_signatureField, cst);
		textPanel.add(new JLabel(Ariane.getString("SubroutineEditor.signatureExample")), cst);

		cst.gridwidth = 1;
		textPanel.add(new JLabel(Ariane.getString("SubroutineEditor.documentationText")), cst);
		_docURLField = new JTextField(30);
		_docURLField.setText(_currentItem.getDocURL());
		textPanel.add(_docURLField, cst);
		JButton button2 = new JButton(Ariane.getString("SubroutineEditor.load"));
		button2.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed( ActionEvent event ) {
				JFileChooser fc = new JFileChooser(".");
				fc.setDialogType(JFileChooser.OPEN_DIALOG);
				fc.setFileSelectionMode(JFileChooser.FILES_ONLY);
				if (_docURLField.getText().length() > 6) {
					fc.setSelectedFile(new File(_docURLField.getText().substring(7)));
				}
				int returnVal = fc.showOpenDialog(SubroutineEditor.this);
				if (returnVal == JFileChooser.APPROVE_OPTION) {
					_docURLField.setText("file://" + fc.getSelectedFile().getAbsolutePath());
					_docURLField.setCaretPosition(0);
				}
			}
		});
		JButton button3 = new JButton(Ariane.getString("OperatorEditor.edit"));
		button3.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed( ActionEvent event ) {
				HTMLDocumentEditor editor = new HTMLDocumentEditor(SubroutineEditor.this);
				if (_docURLField.getText().length() > 6) {
					editor.setSelectedFile(new File(_docURLField.getText().substring(7)));
				}
				if (editor.edit() == HTMLDocumentEditor.OK_OPTION && editor.getSelectedFile() != null) {
					_docURLField.setText("file://" + editor.getSelectedFile().getAbsolutePath());
					_docURLField.setCaretPosition(0);
				}
			}
		});
		textPanel.add(button2, cst);
		cst.gridwidth = GridBagConstraints.REMAINDER; //end row
		textPanel.add(button3, cst);

		// cancel button
		JButton cancelButton = new JButton();
		cancelButton.setAction(new AbstractAction(Ariane.getString("SubroutineEditor.cancel"),
				Ariane.getIcon("gtk-cancel.png")) {
			@Override
			public void actionPerformed( ActionEvent event ) {
				cancelAction();
			}
		});

		// Apply button
		JButton applyButton = new JButton(Ariane.getString("SubroutineEditor.apply"), Ariane.getIcon("gtk-apply.png"));

		applyButton.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed( ActionEvent event ) {
				applyAction();
			}
		});

		JPanel buttonPanel = new JPanel();
		buttonPanel.add(applyButton);
		buttonPanel.add(cancelButton);

		setDefaultButton(applyButton);
		setCancelButton(cancelButton);

		getContentPane().setLayout(new BorderLayout());
		getContentPane().add(textPanel, BorderLayout.NORTH);
		getContentPane().add(buttonPanel, BorderLayout.SOUTH);
		pack();
	}

	/**
	 * Validates information.
	 */
	private void cancelAction( ) {
		_validated = false;
		setVisible(false);
		dispose();
	}

	/**
	 * Applies and saves the chosen preferences,
	 * and eventually closes the dialog box.
	 */
	private void applyAction( ) {
		String category = _categoryField.getSelectedItem().toString();
		if (category.trim().isEmpty()) {
			Ariane.getWorkspace().printErrorMessage(Ariane.getString("SubroutineEditor.error1"));
			return;
		}
		_currentItem.setCategory(category);
		_currentItem.setName(_nameField.getText());
		_currentItem.setSignature(_signatureField.getText());
		_currentItem.setDocURL(_docURLField.getText());
		if (_inserted) {
			if (!Ariane.getWorkspace().getTree().addSubroutine(_currentItem)) {
				return;
			}
			save();
			_validated = true;
		} else {
			_treeModel.nodeChanged(_currentNode);
			_validated = false;
		}
		setVisible(false);
		dispose();
	}

	/**
	 * Saves the current subroutine list in the convenient file.
	 * 
	 * @return true if all of the subroutines has been saved.
	 */
	private boolean save() {
		File file = new File(Ariane.getSubroutineFilename());
		File folder = file.getParentFile();
		try {
			folder.mkdirs();
			XMLFileOutputStream out = new XMLFileOutputStream(file, null);
			out.write(saveSubroutine());
			out.close();
		} catch (IOException ex) {
			Ariane.getWorkspace().printSystemMessage(Ariane.getString("PreferenceChooser.message.saved") + " " + Ariane.getPreferencesFilename());
			return false;
		}
		return true;
	}

	/**
	 * Saves subroutines in an XML tree.
	 * 
	 * @return the XML tree.
	 * @throws IOException in case of IO problems.
	 */
	private static XMLElement saveSubroutine() throws IOException {
		List<TreeItem> subRoutineList = Ariane.getWorkspace().getTree().getSubroutineList();
		if (!subRoutineList.isEmpty()) {
			XMLElement element = new XMLElement("SubroutineLibrary");
			for (TreeItem item: subRoutineList) {
				element.addChildren(item.save());
			}
			return element;
		} else {
			return null;
		}
	}
	
	public boolean isValidated() {
		return _validated;
	}

	public SubroutineTreeItem getTreeItem() {
		return _currentItem;
	}
}