/* -*- c-basic-offset: 3 -*-
 *
 * Copyright (c) 2013, GREYC.
 * All rights reserved
 *
 * You may use this file under the terms of the BSD license as follows:
 *
 * "Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of the GREYC, nor the name of its
 *     contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
 * 
 *   
 * For more information, refer to:
 * https://clouard.users.greyc.fr/Pandore
 */

package fr.greyc.pantheon.ariane.ui;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Insets;
import java.awt.Point;
import java.awt.Toolkit;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.util.ArrayList;
import java.util.List;

import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTextArea;
import javax.swing.text.BadLocationException;
import javax.swing.text.Document;
import javax.swing.tree.TreePath;

import fr.greyc.pantheon.ariane.Ariane;
import fr.greyc.pantheon.ariane.utils.MutableTreeItem;
import fr.greyc.pantheon.ariane.utils.TreeItem;

/**
 * This dialog implements a terminal where the user can
 * type in the name of operators to be added into the graph.
 * 
 * @author Régis CLOUARD - Jan 19, 2011
 */
public final class Terminal extends JPanel implements KeyListener {
	/** The default window height. */
	private static final int WINHEIGHT = 10;
	/** The default window width. */
	private static final int WINWIDTH = 30;
	/** The list of predefined shortcuts for the menu items. */
	private static final int SHORTCUT_MASK = Toolkit.getDefaultToolkit().getMenuShortcutKeyMask();
	/** The prompt string. */
	private static final String PROMPT_STRING = "$ ";
	private static final String EOL = "\n";
	/** The coordinates of the next items to create. */
	private static Point _coordinates = new Point(50, 50);

	/** The terminal editor. */
	private JTextArea _text;

	/** The current edited line. */
	private StringBuilder _commandLine = new StringBuilder();
	private int _commandPointer;
	/** History */
	private int _historyPointer;
	private final List<String> _history = new ArrayList<String>();

	/**
	 * Creates a new terminal window with default specifications.
	 */
	public Terminal( ) {
		initialise();
	}

	/**
	 * Initializes the terminal window; creates the UI.
	 */
	private void initialise() {
		_text = new JTextArea(WINHEIGHT, WINWIDTH);
		_text.setFont(PreferenceChooser.getDefaultFont());
		_text.setEditable(true);
		_text.setLineWrap(false);
		_text.setForeground(Color.LIGHT_GRAY);
		_text.setMargin(new Insets(6, 6, 6, 6));
		_text.setBackground(Color.BLACK);
		_text.addKeyListener(this);
		_text.setCaretColor(Color.WHITE);

		JScrollPane scrollPane = new JScrollPane(_text);
		setLayout(new BorderLayout());
		add(scrollPane, BorderLayout.CENTER);

		_text.setEditable(true);
		_text.requestFocus();
		_text.append(PROMPT_STRING);
	}

	/**
	 * Clears the terminal.
	 */
	private void clear() {
		_text.setText(PROMPT_STRING);
		_commandLine = new StringBuilder();
		setCaretPosition(0);
	}

	private void moveCaretPosition( int increment ) {
		if (_commandPointer + increment >= 0 && _commandPointer + increment <= _commandLine.length()) {
			setCaretPosition(_commandPointer + increment);
		}
	}

	private void setCaretPosition( int value ) {
		_commandPointer = value;
		_text.setCaretPosition(_text.getDocument().getLength() - _commandLine.length() + _commandPointer);
	}

	private void removeCurrentLine( ) {
		if (_commandLine.length() > 0) {
			Document doc = _text.getDocument();
			int end = doc.getLength();
			int start = end - _commandLine.length();
			try {
				doc.remove(start, end - start);
			} catch (BadLocationException e) {
			}
		}
	}

	/**
	 * Starts a new command in the terminal.
	 */
	private void newCommandLine() {
		_text.append(EOL + PROMPT_STRING);
		_commandLine = new StringBuilder();
		setCaretPosition(0);
	}

	/**
	 * Adds a new line in the terminal.
	 */
	private void newLine() {
		_text.append(EOL);
	}

	/**
	 * Invoked when a key is pressed.
	 */
	@Override
	public void keyPressed( KeyEvent event ) {
		if (event.getModifiers() != SHORTCUT_MASK) {
			switch (event.getKeyCode()) {
			case KeyEvent.VK_LEFT:
			case KeyEvent.VK_KP_LEFT:
				moveCaretPosition(-1);
				break;
			case KeyEvent.VK_RIGHT:
			case KeyEvent.VK_KP_RIGHT:
				moveCaretPosition(+1);
				break;
			case KeyEvent.VK_DOWN:
			case KeyEvent.VK_KP_DOWN:
				if (_historyPointer < _history.size() - 1) {
					removeCurrentLine();
					_commandLine = new StringBuilder(_history.get(++_historyPointer));
					_text.append(_commandLine.toString());
					setCaretPosition(_commandLine.length());
				} else if (_historyPointer == _history.size() - 1) {
					removeCurrentLine();
					_commandLine = new StringBuilder();
					++_historyPointer;
					_text.append(_commandLine.toString());
					setCaretPosition(_commandLine.length());
				}
				break;
			case KeyEvent.VK_UP:
			case KeyEvent.VK_KP_UP:
				if (_historyPointer > 0) {
					removeCurrentLine();
					_commandLine = new StringBuilder(_history.get(--_historyPointer));
					_text.append(_commandLine.toString());
					setCaretPosition(_commandLine.length());
				}
				break;
			default:
					break;
			}
		}
		if (event.getModifiers() != SHORTCUT_MASK) {
			event.consume();
		}
	}

	/**
	 * Invoked when a key is released.
	 */
	@Override
	public void keyReleased( KeyEvent event ) {
	}

	private int strcmp( String name1, String name2 ) {
		int minSize;
		int n1 = name1.length();
		int n2 = name2.length();

		if (n1 > n2) {
			minSize = n2;
		} else {
			minSize = n1;
		}
		byte[] v1 = name1.getBytes();
		byte[] v2 = name2.getBytes();
		for (int i = 0; i < minSize; i++ ) {
			if (v1[i] != v2[i]) {
				return i;
			}
		}
		return minSize;
	}

	/**
	 * Invoked when a key is typed.
	 */
	@Override
	public void keyTyped(KeyEvent event) {
		char ch = event.getKeyChar();
		switch(ch) {

		case 7:     // CTRL-G cancel the current command and begin a new line.
			newCommandLine();
			break;

		case '\b':  // backspace
			if (_commandPointer > 0) {
				try {
					Document doc = _text.getDocument();
					int length = doc.getLength();
					doc.remove(length - _commandLine.length() + _commandPointer - 1, 1);
					_commandLine.deleteCharAt(_commandPointer - 1);
					moveCaretPosition(-1);
				} catch (BadLocationException e) {
				}
			}
			break;

		case '\t':     // tabulation (CTRL-I)
			List<String> names = getMatchedOperatorNames(_commandLine.toString());
			if (names.size() == 1) {
				String complement = names.get(0).substring(_commandLine.length());
				_text.append(complement);
				_commandLine.append(complement);
				setCaretPosition(_commandLine.length());
			} else if (!names.isEmpty()) {
				// Completion until difference
				String tmpName = names.get(0);
				for (String name : names) {
					int index = strcmp(name, tmpName);
					tmpName = name.substring(0, index);
				}
				tmpName = tmpName.substring(_commandLine.length());
				_text.append(tmpName);
				_commandLine.append(tmpName);
				newLine();
				int count = 0; // Only display 4 operators on a same line.
				for (String name : names) {
					count++;
					if (count > 4) {
						newLine();
						count = 0;
					}
					_text.append(name + "\t");
				}
				_text.append(EOL + PROMPT_STRING + _commandLine);
				setCaretPosition(_commandLine.length());
			}
			break;

		case '\r':     // carriage return
		case '\n':     // newline
			if (_commandLine.toString().equals("clear")) {
				_history.add(_commandLine.toString());
				clear();
				_historyPointer = _history.size();
			} else {
				String command = _commandLine.toString().trim();
				if (command.length() > 0 ) {
					TreeItem treeItem = getMatchedTreeItem(command);
					if (treeItem == null) {
						_text.append(EOL + Ariane.getString("error") + Ariane.getString("Terminal.error") + command);
					} else {
						try {
							_coordinates.translate(10, 10);
							Ariane.getWorkspace().addNewItem(treeItem, _coordinates);
						} catch (InstantiationException e ) {
							_text.append(EOL + Ariane.getString("error") + e.getLocalizedMessage());
						}
					}
					_history.add(command);
				}
				newCommandLine();
				_historyPointer = _history.size();
			}
			break;

		default: // add printable characters
			int length = _text.getDocument().getLength();
			_text.insert(String.valueOf(ch), length - _commandLine.length() + _commandPointer);
			_commandLine.insert(_commandPointer, ch);
			moveCaretPosition(1);
			break;
		}

		event.consume();        // make sure the text area doesn't handle this
	}

	/**
	 * The list of all operators that can be added to the graph.
	 */
	private MutableTreeItem[] getSubTrees() {
		return new MutableTreeItem[]{ Ariane.getWorkspace().getTree().getCommonOperatorTree(),
				Ariane.getWorkspace().getTree().getUserOperatorTree(),
				Ariane.getWorkspace().getTree().getSubroutineTree()};
	}

	/**
	 * Returns the list of operators that match the given name.
	 * 
	 * @param pattern the beginning of operator names.
	 * @return the list of match operator names.
	 */
	private List<String> getMatchedOperatorNames( String pattern ) {
		List<String> result = new ArrayList<String>();
		for (MutableTreeItem subtree : getSubTrees()) {
			for (TreePath path : Ariane.getWorkspace().getTree().search(subtree, pattern)) {
				MutableTreeItem leaf;
				leaf = (MutableTreeItem)path.getLastPathComponent();
				TreeItem item = (TreeItem)leaf.getUserObject();
				// Ensure that the operator BEGINS with the pattern.
				if (item.getName().startsWith(pattern)) {
					result.add(item.getName());
				}
			}
		}
		return result;
	}

	/**
	 * Returns the tree item that matches exactly
	 * the given pattern name.
	 * 
	 * @param pattern a complete operator name.
	 * @return the match operator item or null.
	 */
	private TreeItem getMatchedTreeItem( String pattern ) {
		for (MutableTreeItem subtree : getSubTrees()) {
			for (TreePath path : Ariane.getWorkspace().getTree().search(subtree, pattern)) {
				MutableTreeItem leaf;
				leaf = (MutableTreeItem)path.getLastPathComponent();
				TreeItem item = (TreeItem)leaf.getUserObject();
				// Ensure that the operator is exactly the pattern.
				if (item.getName().equals(pattern)) {
					return item;
				}
			}
		}
		return null;
	}
}
