/* -*- c-basic-offset: 3 -*-
 *
 * Copyright (c) 2013, GREYC.
 * All rights reserved
 *
 * You may use this file under the terms of the BSD license as follows:
 *
 * "Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of the GREYC, nor the name of its
 *     contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
 * 
 *   
 * For more information, refer to:
 * https://clouard.users.greyc.fr/Pandore
 */

package fr.greyc.pantheon.ariane.ui;

import fr.greyc.pantheon.ariane.Ariane;
import fr.greyc.pantheon.ariane.generator.CodeGenerator;
import fr.greyc.pantheon.ariane.graph.Graph;
import fr.greyc.pantheon.ariane.graph.item.DataFile;
import fr.greyc.pantheon.ariane.graph.item.GraphItem;
import fr.greyc.pantheon.ariane.graph.item.etc.PropertyViewer;
import fr.greyc.pantheon.ariane.utils.FadingEffects;
import fr.greyc.pantheon.ariane.utils.MutableTreeItem;
import fr.greyc.pantheon.ariane.utils.Tree;
import fr.greyc.pantheon.ariane.utils.TreeItem;
import fr.greyc.pantheon.ariane.xml.XMLElement;
import fr.greyc.pantheon.ariane.xml.XMLFileInputStream;

import java.awt.BorderLayout;
import java.awt.Dimension;
import java.awt.Frame;
import java.awt.Point;
import java.awt.Toolkit;
import java.awt.datatransfer.Clipboard;
import java.awt.datatransfer.ClipboardOwner;
import java.awt.datatransfer.Transferable;
import java.awt.datatransfer.UnsupportedFlavorException;
import java.awt.event.ActionEvent;
import java.awt.event.KeyEvent;
import java.awt.print.PrinterException;
import java.awt.print.PrinterJob;
import java.io.File;
import java.io.IOException;
import java.net.URI;
import java.util.ArrayList;
import java.util.List;

import javax.print.attribute.HashPrintRequestAttributeSet;
import javax.print.attribute.PrintRequestAttributeSet;
import javax.print.attribute.standard.JobName;
import javax.print.attribute.standard.MediaSizeName;
import javax.print.attribute.standard.OrientationRequested;
import javax.swing.AbstractAction;
import javax.swing.Action;
import javax.swing.ActionMap;
import javax.swing.Box;
import javax.swing.ImageIcon;
import javax.swing.InputMap;
import javax.swing.JButton;
import javax.swing.JComponent;
import javax.swing.JFileChooser;
import javax.swing.JMenu;
import javax.swing.JMenuBar;
import javax.swing.JMenuItem;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JSeparator;
import javax.swing.JSplitPane;
import javax.swing.JToolBar;
import javax.swing.KeyStroke;
import javax.swing.UIManager;
import javax.swing.border.EtchedBorder;
import javax.xml.parsers.ParserConfigurationException;

import org.xml.sax.SAXException;

/**
 * The <i>Workspace</i> is the main window of the software.</br>
 * It holds an editor of graphs, an explorer and a console.
 * 
 * @author Régis CLOUARD - Oct 01, 1998
 * @version Sébastien LEFÈVRE, Richard LOBRÉAU - Feb 21, 1999
 * @version Adrien SCHRÈQUE, Hicham AMADI - Nov 05, 2000
 * @version Mathieu JOETS - Jul 09, 2001
 * @version Régis CLOUARD - Oct 16, 2001
 * @version Régis CLOUARD - Dec 15, 2004
 * @version Régis CLOUARD - Dec 01, 2007 (UI enhancement)
 * @version Régis CLOUARD - Jan 19, 2011
 * @version Régis CLOUARD - Apr 08,-2011 (new design)
 */
public final class Workspace extends JPanel implements ClipboardOwner {
	private static final int MAXIMUM_RECENT_FILES = 10;
	/** The list of recent files. */
	private static List<String> _recentProjectFilenames = new ArrayList<String>();

	private final JPanel _property;
	private final Explorer _explorer;
	private final JScrollPane _scrollPaneConsole;

	/** The editor of graphs. */
	private final GraphEditor _editor;
	/** The console for displaying information. */
	private final Console _console;
	/** The tree of all available items for building graphs. */
	private final Tree _tree;

	//// Directories.
	/** The default loading directory for images. */
	private static final String DEFAULT_WORKING_DIRCETORY;
	/** The current directory for loading and saving graphs. */
	private static String _currentDir;
	/** The default temporary directory. */
	private static final String DEFAULT_TEMPORARY_DIRECTORY;
	/** The preferred temporary directory. */
	private static String _tempDir;

	/** The current graph among all graphs. */
	private Graph _currentGraph;

	/** The menu titles */
	private JMenu _fileM;
	/** The menu titles */
	private JMenu _recentFilesSM;
	/** The menu titles */
	private JMenu _editM;
	/** The menu titles */
	private JMenu _optionsM;
	/** The menu titles */
	private JMenu _graphM;
	/** The menu titles */
	private JMenu _helpM;
	/** The actions of the menu : File */
	private Action _newA;
	/** The actions of the menu : File */
	private Action _openA;
	/** The actions of the menu : File */
	private Action _saveAsA;
	/** The actions of the menu : File */
	private Action _saveA;
	/** The actions of the menu : File */
	private Action _printA;
	/** The actions of the menu : File */
	private Action _importSubroutineA;
	/** The actions of the menu : File */
	private Action _exportSubroutineA;
	/** The actions of the menu : File */
	private Action _quitA;

	/** The menu items: File */
	private JMenuItem _newI;
	/** The menu items: File */
	private JMenuItem _openI;
	/** The menu items: File */
	private JMenuItem _saveAsI;
	/** The menu items: File */
	private JMenuItem _saveI;
	/** The menu items: File */
	private JMenuItem _printI;
	/** The menu items: File */
	private JMenuItem _importSubroutineI;
	/** The menu items: File */
	private JMenuItem _exportSubroutineI;
	/** The menu items: File */
	private JMenuItem _quitI;

	/** The buttons : File */
	private JButton _newB;
	/** The buttons : File */
	private JButton _openB;
	/** The buttons : File */
	private JButton _saveB;
	/** The buttons : File */
	private JButton _saveAsB;
	/** The buttons : File */
	private JButton _printB;
	/** The buttons : File */

	/** The actions of the menu : File */
	private Action _closeA;
	/** The actions of the menu : File */
	private Action _closeAllA;

	/** The menu items: File */
	private JMenuItem _closeI;
	/** The menu items: File */
	private JMenuItem _closeAllI;

	/** The actions of the menu : Edit */
	private Action _undoA;
	/** The actions of the menu : Edit */
	private Action _redoA;
	/** The actions of the menu : Edit */
	private Action _copyA;
	/** The actions of the menu : Edit */
	private Action _cutA;
	/** The actions of the menu : Edit */
	private Action _pasteA;
	/** The actions of the menu : Edit */
	private Action _eraseA;
	/** The actions of the menu : Edit */
	private Action _selectA;

	/** The menu items: Edit */
	private JMenuItem _undoI;
	/** The menu items: Edit */
	private JMenuItem _redoI;
	/** The menu items: Edit */
	private JMenuItem _copyI;
	/** The menu items: Edit */
	private JMenuItem _cutI;
	/** The menu items: Edit */
	private JMenuItem _pasteI;
	/** The menu items: Edit */
	private JMenuItem _eraseI;
	/** The menu items: Edit */
	private JMenuItem _selectI;

	/** The buttons : Edit*/
	private JButton _undoB;
	/** The buttons : Edit*/
	private JButton _redoB;
	/** The buttons : Edit*/
	private JButton _cutB;
	/** The buttons : Edit*/
	private JButton _copyB;
	/** The buttons : Edit*/
	private JButton _pasteB;
	/** The buttons : Edit*/
	private JButton _eraseB;

	/** The actions of the menu : Options */
	private Action _confA;
	/** The actions of the menu : Options */
	private Action _prefA;
	/** The actions of the menu : Options */
	private Action _operatorTreeA;
	/** The actions of the menu : Options */
	private Action _subroutineTreeA;

	/** The menu items : Options */
	private JMenuItem _confI;
	/** The menu items : Options */
	private JMenuItem _prefI;
	/** The menu items : Options */
	private JMenuItem _operatorTreeI;
	/** The menu items : Options */
	private JMenuItem _subroutineTreeI;

	/** The buttons : Options */
	private JButton _prefB;
	/** The buttons : Options */
	private JButton _operatorTreeB;
	/** The buttons : Options */
	private JButton _subroutineTreeB;

	/** The actions of the menu : Graph. */
	private Action _execA;
	/** The actions of the menu : Graph. */
	private Action _resetA;
	/** The actions of the menu : Graph. */
	private Action _generateA;
	/** The actions of the menu : Graph. */
	private Action _stopA;

	/** The menu items: Graph. */
	private JMenuItem _execI;
	/** The menu items: Graph. */
	private JMenuItem _resetI;
	/** The menu items: Graph. */
	private JMenuItem _generateI;
	/** The menu items: Graph. */
	private JMenuItem _stopI;

	/** The buttons : actions */
	private JButton _execB;
	/** The buttons : actions */
	private JButton _generateB;
	/** The buttons : actions */
	private JButton _stopB;

	/** The about action. */
	private Action _aboutA;
	/** The actions of the menu : Help */
	private Action _helpA;

	/** The menu items: Help */
	private JMenuItem _aboutI;
	/** The menu items: Help */
	private JMenuItem _helpI;

	/** The buttons : Help */
	private JButton _helpB;

	/** The current session identifier, where a session is a read or a paste. */
	private static int _currentReadPasteSession;

	/** The vertical split pane ratio (splitPaneHV2 + splitPaneV2). */
	private JSplitPane _splitPaneV1;
	/** The vertical split pane ratio (splipaneH  property). */
	private JSplitPane _splitPaneV2;
	/** The current horizontal split pane ratio (explorer + terminal). */
	private JSplitPane _splitPaneH1;
	/** The current horizontal split pane ratio (graph + console). */
	private JSplitPane _splitPaneH2;

	static {
		DEFAULT_WORKING_DIRCETORY = System.getProperty("user.dir") + File.separator;
		DEFAULT_TEMPORARY_DIRECTORY  = System.getProperty("java.io.tmpdir") + File.separator;
		setCurrentDir(DEFAULT_WORKING_DIRCETORY);
		setTempDir(DEFAULT_TEMPORARY_DIRECTORY);
	}

	/**
	 * Creates the main window of the software.
	 * It holds the menu, an item explorer, a console for message
	 * and a graph editor.
	 * 
	 * @param operatorsTree the list of all available operators.
	 */
	public Workspace( Tree operatorsTree ) {
		buildActions();
		_property = new JPanel(new BorderLayout());
		_property.setMinimumSize(new Dimension(70, 60));
		_tree = operatorsTree;
		_console = new Console();
		_editor = new GraphEditor(this);
		_explorer = new Explorer(_tree);
		_scrollPaneConsole = new JScrollPane(_console);
	}

	/**
	 * @return the last value passed to setDividerLocation.
	 */
	public int getSplitH1( ) {
		if (Ariane.getFrame().getExtendedState() ==  Frame.MAXIMIZED_BOTH) {
			if (_splitPaneH1.getLastDividerLocation() < 0) {
				return PreferenceChooser.getInt(PreferenceChooser.SPLITPANE_H1_RATIO);
			}
			return _splitPaneH1.getLastDividerLocation();
		}
		return _splitPaneH1.getDividerLocation();
	}

	public int getSplitH2( ) {
		if (Ariane.getFrame().getExtendedState() ==  Frame.MAXIMIZED_BOTH) {
			if (_splitPaneH2.getLastDividerLocation() < 0) {
				return PreferenceChooser.getInt(PreferenceChooser.SPLITPANE_H2_RATIO);
			}
			return _splitPaneH2.getLastDividerLocation();
		}
		return _splitPaneH2.getDividerLocation();
	}

	/**
	 * @return the current horizontal split pane value.
	 */
	public int getSplitV1( ) {
		if (Ariane.getFrame().getExtendedState() ==  Frame.MAXIMIZED_BOTH) {
			if (_splitPaneV1.getLastDividerLocation() < 0) {
				return PreferenceChooser.getInt(PreferenceChooser.SPLITPANE_V1_RATIO);
			}
			return _splitPaneV1.getLastDividerLocation();
		}
		return _splitPaneV1.getDividerLocation();
	}

	public int getSplitV2( ) {
		if (Ariane.getFrame().getExtendedState() ==  Frame.MAXIMIZED_BOTH) {
			if (_splitPaneV2.getLastDividerLocation() < 0) {
				return PreferenceChooser.getInt(PreferenceChooser.SPLITPANE_V2_RATIO);
			}
			return _splitPaneV2.getLastDividerLocation();
		}
		return _splitPaneV2.getDividerLocation();
	}

	/**
	 * Initializes the workspace content.
	 */
	public void initContent( ) {
		// Organization of the workspace in three panes.
		// H: (editor + console)
		// V: Explorer V.
		JPanel contentPane = new JPanel();

		// Terminal + Explorer
		_splitPaneH1 = new JSplitPane(JSplitPane.VERTICAL_SPLIT, _explorer, new Terminal());
		_splitPaneH1.setResizeWeight(1.0);
		_splitPaneH1.setDividerSize(5);

		// Graph + console
		_splitPaneH2 = new JSplitPane(JSplitPane.VERTICAL_SPLIT, _editor, _scrollPaneConsole);
		_splitPaneH2.setResizeWeight(1.0);
		_splitPaneH2.setDividerSize(5);

		// V2: tree | (editor + property)
		_splitPaneV2 = new JSplitPane(JSplitPane.HORIZONTAL_SPLIT, _splitPaneH2, _property);
		_splitPaneV2.setDividerSize(5);
		_splitPaneV2.setResizeWeight(1.0);

		// V1: tree | ((terminal explorer) + (editor+console+property)
		_splitPaneV1 = new JSplitPane(JSplitPane.HORIZONTAL_SPLIT, _splitPaneH1, _splitPaneV2);
		_splitPaneV1.setDividerSize(5);

		// Install the toolbar.
		contentPane.setLayout(new BorderLayout());
		contentPane.add(buildToolBar(), BorderLayout.NORTH);
		contentPane.add(_splitPaneV1, BorderLayout.CENTER);

		setLayout(new BorderLayout());
		add(contentPane, BorderLayout.CENTER);

		changeLanguage();
		setCurrentGraph(null);

		InputMap inputMap = getInputMap(JComponent.WHEN_IN_FOCUSED_WINDOW);
		ActionMap actionMap = getActionMap();
		inputMap.put(KeyStroke.getKeyStroke(new Character('0'), 0), "Display0");
		actionMap.put("Display0", new AbstractAction(null) {
			@Override
			public void actionPerformed( ActionEvent event ) {
				if (_currentGraph != null) {
					_currentGraph.displayOutputs(0);
				}
			}
		});
		inputMap.put(KeyStroke.getKeyStroke(new Character('1'), 0), "Display1");
		actionMap.put("Display1", new AbstractAction(null) {
			@Override
			public void actionPerformed( ActionEvent event ) {
				if (_currentGraph != null) {
					_currentGraph.displayOutputs(1);
				}
			}
		});
		inputMap.put(KeyStroke.getKeyStroke(new Character('2'), 0), "Display2");
		actionMap.put("Display2", new AbstractAction(null) {
			@Override
			public void actionPerformed( ActionEvent event ) {
				if (_currentGraph != null) {
					_currentGraph.displayOutputs(2);
				}
			}
		});
		inputMap.put(KeyStroke.getKeyStroke(new Character('3'), 0), "Display3");
		actionMap.put("Display3", new AbstractAction(null) {
			@Override
			public void actionPerformed( ActionEvent event ) {
				if (_currentGraph != null) {
					_currentGraph.displayOutputs(3);
				}
			}
		});
		inputMap.put(KeyStroke.getKeyStroke(new Character('4'), 0), "Display4");
		actionMap.put("Display4", new AbstractAction(null) {
			@Override
			public void actionPerformed( ActionEvent event ) {
				if (_currentGraph != null) {
					_currentGraph.displayOutputs(4);
				}
			}
		});
		inputMap.put(KeyStroke.getKeyStroke(new Character('5'), 0), "Display5"); // 5 outputs is very much
		actionMap.put("Display5", new AbstractAction(null) {
			@Override
			public void actionPerformed( ActionEvent event ) {
				if (_currentGraph != null) {
					_currentGraph.displayOutputs(5);
				}
			}
		});
	}

	public void restoreSplitPanesLocation() {
		_splitPaneH1.setDividerLocation(PreferenceChooser.getInt(PreferenceChooser.SPLITPANE_H1_RATIO));
		_splitPaneH2.setDividerLocation(PreferenceChooser.getInt(PreferenceChooser.SPLITPANE_H2_RATIO));
		_splitPaneV1.setDividerLocation(PreferenceChooser.getInt(PreferenceChooser.SPLITPANE_V1_RATIO));
		_splitPaneV2.setDividerLocation(PreferenceChooser.getInt(PreferenceChooser.SPLITPANE_V2_RATIO));
	}

	public static void setRecentProjectFilenames( List<String> filenames ) {
		_recentProjectFilenames = filenames;
	}

	public static List<String> getRecentProjectFilenames( ) {
		return _recentProjectFilenames;
	}

	/////
	///// M E N U  &  T O O L B A R
	/////

	/**
	 * Creates all actions that are used in the menu and the toolbar.
	 * An action is a method, an icon and a title.
	 */
	private void buildActions() {
		ImageIcon icon = Ariane.getIcon("new.png");
		_newA = new AbstractAction(null, icon)
		{ @Override
			public void actionPerformed( ActionEvent event ) { newGraph(); }
		};
		icon = Ariane.getIcon("open.png");
		_openA = new AbstractAction(null, icon)
		{ @Override
			public void actionPerformed( ActionEvent event ) { openFileAction(); }
		};
		icon = Ariane.getIcon("save.png");
		_saveA = new AbstractAction(null, icon)
		{ @Override
			public void actionPerformed( ActionEvent event ) { saveFileAction(); }
		};
		icon = Ariane.getIcon("saveas.png");
		_saveAsA = new AbstractAction(null, icon)
		{ @Override
			public void actionPerformed( ActionEvent event ) { saveAsFileAction(); }
		};
		_printA = new AbstractAction(null, Ariane.getIcon("print.png"))
		{ @Override
			public void actionPerformed( ActionEvent event ) { print(); }
		};
		_importSubroutineA = new AbstractAction()
		{ @Override
			public void actionPerformed( ActionEvent event ) { importSubroutinesAction(); }
		};
		_exportSubroutineA = new AbstractAction()
		{ @Override
			public void actionPerformed( ActionEvent event ) { exportSubroutinesAction(); }
		};
		_closeA = new AbstractAction()
		{ @Override
			public void actionPerformed( ActionEvent event ) { close(); }
		};
		_closeAllA = new AbstractAction()
		{ @Override
			public void actionPerformed( ActionEvent event ) { closeAll(); }
		};
		_quitA = new AbstractAction(null)
		{ @Override
			public void actionPerformed( ActionEvent event ) { quitAction();  }
		};
		_undoA = new AbstractAction(null, Ariane.getIcon("undo.png"))
		{ @Override
			public void actionPerformed( ActionEvent event ) { undo(); }
		};
		_redoA = new AbstractAction(null, Ariane.getIcon("redo.png"))
		{ @Override
			public void actionPerformed( ActionEvent event ) { redo(); }
		};
		icon = Ariane.getIcon("copy.png");
		_copyA = new AbstractAction(null, icon)
		{ @Override
			public void actionPerformed( ActionEvent event ) { copy(); }
		};
		icon = Ariane.getIcon("cut.png");
		_cutA = new AbstractAction(null, icon)
		{ @Override
			public void actionPerformed( ActionEvent event ) { cut(); }
		};
		icon = Ariane.getIcon("paste.png");
		_pasteA = new AbstractAction(null, icon)
		{ @Override
			public void actionPerformed( ActionEvent event ) { paste(); }
		};
		icon = Ariane.getIcon("delete.png");
		_eraseA = new AbstractAction(null, icon)
		{ @Override
			public void actionPerformed( ActionEvent event ) { deleteSelectedItems(); }
		};
		_selectA = new AbstractAction(null)
		{ @Override
			public void actionPerformed( ActionEvent event ) { selectAll(); }
		};
		_confA = new AbstractAction(null)
		{ @Override
			public void actionPerformed( ActionEvent event ) { configurationAction(); }
		};
		_prefA = new AbstractAction(null, Ariane.getIcon("preferences.png"))
		{ @Override
			public void actionPerformed( ActionEvent event ) { preferencesAction(); }
		};
		_operatorTreeA = new AbstractAction(null, Ariane.getIcon("operatortree.png"))
		{ @Override
			public void actionPerformed( ActionEvent event ) { treeOperatorAction(); }
		};
		_subroutineTreeA = new AbstractAction(null, Ariane.getIcon("subroutinetree.png"))
		{ @Override
			public void actionPerformed( ActionEvent event ) { treeSubroutineAction(); }
		};
		_stopA = new AbstractAction(null, Ariane.getIcon("stop.png"))
		{ @Override
			public void actionPerformed( ActionEvent event ) { stopExecution(); }
		};
		_stopA.setEnabled(false);
		icon = Ariane.getIcon("execute.png");
		_execA = new AbstractAction(null, icon)
		{ @Override
			public void actionPerformed( ActionEvent event ) { execute(); }
		};
		_execA.setEnabled(false);
		_resetA = new AbstractAction(null, null)
		{ @Override
			public void actionPerformed( ActionEvent event ) { reset(); }
		};
		_generateA = new AbstractAction(null, Ariane.getIcon("generation.png"))
		{ @Override
			public void actionPerformed( ActionEvent event ) { generate(); }
		};
		_aboutA = new AbstractAction(null)
		{ @Override
			public void actionPerformed( ActionEvent event ) { aboutAction(); }
		};
		_helpA = new AbstractAction(null, Ariane.getIcon("help.png"))
		{ @Override
			public void actionPerformed( ActionEvent event ) { helpAction(); }
		};
	}	

	public JToolBar buildToolBar( ) {
		JToolBar toolbar = new JToolBar();
		toolbar.setFloatable(true);
		toolbar.setBorder(new EtchedBorder(EtchedBorder.LOWERED));
		toolbar.setBorderPainted(true);
		toolbar.setOpaque(false);

		_newB = new JButton(_newA);
		_newB.setBorderPainted(false);
		toolbar.add(_newB);

		_openB = new JButton(_openA);
		_openB.setBorderPainted(false);
		toolbar.add(_openB);

		_saveB = new JButton(_saveA);
		_saveB.setBorderPainted(false);
		toolbar.add(_saveB);

		_saveAsB = new JButton(_saveAsA);
		_saveAsB.setBorderPainted(false);
		toolbar.add(_saveAsB);

		toolbar.addSeparator();

		_printB = new JButton(_printA);
		_printB.setBorderPainted(false);
		toolbar.add(_printB);

		toolbar.addSeparator();

		_prefB = new JButton(_prefA);
		_prefB.setBorderPainted(false);
		toolbar.add(_prefB);

		_operatorTreeB = new JButton(_operatorTreeA);
		_operatorTreeB.setBorderPainted(false);
		toolbar.add(_operatorTreeB);

		_subroutineTreeB = new JButton(_subroutineTreeA);
		_subroutineTreeB.setBorderPainted(false);
		toolbar.add(_subroutineTreeB);

		toolbar.addSeparator();

		_undoB = new JButton(_undoA);
		_undoB.setBorderPainted(false);
		toolbar.add(_undoB);

		_redoB = new JButton(_redoA);
		_redoB.setBorderPainted(false);
		toolbar.add(_redoB);

		_copyB = new JButton(_copyA);
		_copyB.setBorderPainted(false);
		toolbar.add(_copyB);

		_cutB = new JButton(_cutA);
		_cutB.setBorderPainted(false);
		toolbar.add(_cutB);

		_pasteB = new JButton(_pasteA);
		_pasteB.setBorderPainted(false);
		toolbar.add(_pasteB);

		_eraseB = new JButton(_eraseA);
		_eraseB.setBorderPainted(false);
		toolbar.add(_eraseB);

		toolbar.addSeparator();
		_execB = new JButton(_execA);
		_execB.setBorderPainted(false);
		toolbar.add(_execB);

		_stopB = new JButton(_stopA);
		_stopB.setBorderPainted(false);
		toolbar.add(_stopB);

		toolbar.addSeparator();

		_generateB = new JButton(_generateA);
		_generateB.setBorderPainted(false);
		toolbar.add(_generateB);

		toolbar.addSeparator();

		_helpB = new JButton(_helpA);
		_helpB.setBorderPainted(false);
		toolbar.add(_helpB);

		return toolbar;
	}

	public JMenuBar buildMenuBar( ) {
		// Adds all submenus.
		JMenuBar menuBar = new JMenuBar();

		// File
		_fileM = new JMenu();
		_fileM.setMnemonic('F');
		_newI = new JMenuItem(_newA);
		_newI.setMnemonic('N');
		_newI.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_N, KeyEvent.CTRL_MASK));
		_fileM.add(_newI);

		_openI = new JMenuItem(_openA);
		_openI.setMnemonic('O');
		_openI.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_O, KeyEvent.CTRL_MASK));
		_fileM.add(_openI);

		_recentFilesSM = new JMenu();
		_recentFilesSM.setMnemonic('I');
		updateRecentFilesListMenu();
		if (_recentProjectFilenames.isEmpty()) {
			_recentFilesSM.setEnabled(false);
		}
		_fileM.add(_recentFilesSM);

		_fileM.add(new JSeparator());

		_closeI = new JMenuItem(_closeA);
		_closeI.setMnemonic('C');
		_closeI.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_W, KeyEvent.CTRL_MASK));
		_fileM.add(new JSeparator());
		_fileM.add(_closeI);
		_closeAllI = new JMenuItem(_closeAllA);
		_fileM.add(_closeAllI);

		_saveAsI = new JMenuItem(_saveAsA);
		_saveAsI.setMnemonic('A');
		_fileM.add(_saveAsI);

		_saveI = new JMenuItem(_saveA);
		_saveI.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_S, KeyEvent.CTRL_MASK));
		_fileM.add(_saveI);

		_fileM.add(new JSeparator());
		_importSubroutineI = new JMenuItem(_importSubroutineA);
		_fileM.add(_importSubroutineI);
		_exportSubroutineI = new JMenuItem(_exportSubroutineA);
		_fileM.add(_exportSubroutineI);

		_printI = new JMenuItem(_printA);
		_printI.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_P, KeyEvent.CTRL_MASK));
		_printI.setMnemonic('P');
		_fileM.add(new JSeparator());
		_fileM.add(_printI);

		_quitI = new JMenuItem(_quitA);
		_quitI.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_Q, KeyEvent.CTRL_MASK));
		_quitI.setMnemonic('Q');
		_fileM.add(new JSeparator());
		_fileM.add(_quitI);

		menuBar.add(_fileM);

		// Edit
		_editM = new JMenu();
		_editM.setMnemonic('E');

		_undoI = new JMenuItem(_undoA);
		_undoI.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_Z, KeyEvent.CTRL_MASK));
		_editM.add(_undoI);

		_redoI = new JMenuItem(_redoA);
		_redoI.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_Y, KeyEvent.CTRL_MASK));
		_editM.add(_redoI);
		_editM.add(new JSeparator());

		_copyI = new JMenuItem(_copyA);
		_copyI.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_C, KeyEvent.CTRL_MASK));
		_editM.add(_copyI);

		_cutI = new JMenuItem(_cutA);
		_cutI.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_X, KeyEvent.CTRL_MASK));
		_editM.add(_cutI);

		_pasteI = new JMenuItem(_pasteA);
		_pasteI.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_V, KeyEvent.CTRL_MASK));
		_editM.add(_pasteI);

		_eraseI = new JMenuItem(_eraseA);
		_eraseI.setMnemonic('D');
		_eraseI.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_DELETE, 0));
		_editM.add(_eraseI);

		_editM.add(new JSeparator());
		_selectI = new JMenuItem(_selectA);
		_selectI.setMnemonic('A');
		_selectI.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_A, KeyEvent.CTRL_MASK));
		_editM.add(_selectI);
		menuBar.add(_editM);

		// Options
		_optionsM = new JMenu();
		_optionsM.setMnemonic('O');

		_confI = new JMenuItem(_confA);
		_optionsM.add(_confI);
		_prefI = new JMenuItem(_prefA);
		_optionsM.add(_prefI);
		_operatorTreeI = new JMenuItem(_operatorTreeA);
		_optionsM.add(_operatorTreeI);
		_subroutineTreeI = new JMenuItem(_subroutineTreeA);
		_optionsM.add(_subroutineTreeI);
		menuBar.add(_optionsM);

		// Search
		menuBar.add(_explorer.getMenu());

		// Graph
		_graphM = new JMenu();
		_graphM.setMnemonic('G');
		_execI = new JMenuItem(_execA);
		_execI.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_R, KeyEvent.CTRL_MASK));
		_graphM.add(_execI);
		_resetI = new JMenuItem(_resetA);
		_resetI.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_U, KeyEvent.CTRL_MASK));
		_graphM.add(_resetI);
		_stopI = new JMenuItem(_stopA);
		_stopI.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_G, KeyEvent.CTRL_MASK));
		_graphM.add(_stopI);
		_generateI = new JMenuItem(_generateA);
		_graphM.add(_generateI);

		menuBar.add(_graphM);

		menuBar.add(Box.createHorizontalGlue());

		// Help
		_helpM = new JMenu();
		_helpM.setMnemonic('H');
		_helpI = new JMenuItem(_helpA);
		_helpI.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_F1, 0));
		_helpM.add(_helpI);

		_aboutI = new JMenuItem(_aboutA);
		_helpM.add(_aboutI);

		menuBar.add(_helpM);

		return menuBar;
	}

	/**
	 * Adds the recent project filenames to the menu.
	 * 
	 * @param filename the file menu.
	 */
	public void addRecentProjects( String filename ) {
		if (_recentProjectFilenames.contains(filename)) {
			_recentProjectFilenames.remove(filename);
			_recentProjectFilenames.add(0, filename);
		} else {
			if (_recentProjectFilenames.size() > MAXIMUM_RECENT_FILES) {
				_recentProjectFilenames.remove(MAXIMUM_RECENT_FILES);
			}
			_recentProjectFilenames.add(0, filename);
		}
	}

	/**
	 * Adds the list of the recent project
	 * filenames to the menu.
	 */
	public void updateRecentFilesListMenu() {
		_recentFilesSM.removeAll();
		int fileIndex = 1;
		for (String filename : _recentProjectFilenames ) {
			final String file = filename;
			Action action = new AbstractAction(fileIndex + " " + filename) {
				@Override
				public void actionPerformed( ActionEvent event ) {
					openFile(file);
				}
			};
			JMenuItem item = new JMenuItem(action);
			item.setMnemonic((char)('0' + fileIndex));
			_recentFilesSM.add(item);
			fileIndex++;
		}
		_recentFilesSM.setEnabled(true);
	}

	/**
	 * Sets the new text for all component of the workspace
	 * with the current language.
	 */
	private void setInternationalText( ) {
		//Menu file
		_fileM.setText(Ariane.getString("Menu.file.title"));
		_recentFilesSM.setText(Ariane.getString("Menu.file.recentFiles"));
		_newI.setText(Ariane.getString("Menu.file.new"));
		_openI.setText(Ariane.getString("Menu.file.open"));
		_saveAsI.setText(Ariane.getString("Menu.file.saveas"));
		_saveI.setText(Ariane.getString("Menu.file.save"));
		_importSubroutineI.setText(Ariane.getString("Menu.file.importSubroutine"));
		_exportSubroutineI.setText(Ariane.getString("Menu.file.exportSubroutine"));
		_printI.setText(Ariane.getString("Menu.file.print"));
		_closeI.setText(Ariane.getString("Menu.file.close"));
		_closeAllI.setText(Ariane.getString("Menu.file.closeAll"));
		_quitI.setText(Ariane.getString("Menu.file.quit"));

		//Menu edit 
		_editM.setText(Ariane.getString("Menu.edit.title"));
		_undoI.setText(Ariane.getString("Menu.edit.undo"));
		_redoI.setText(Ariane.getString("Menu.edit.redo"));
		_copyI.setText(Ariane.getString("Menu.edit.copy"));
		_cutI.setText(Ariane.getString("Menu.edit.cut"));
		_pasteI.setText(Ariane.getString("Menu.edit.paste"));
		_eraseI.setText(Ariane.getString("Menu.edit.delete"));
		_selectI.setText(Ariane.getString("Menu.edit.selectall"));

		//Menu preferences
		_optionsM.setText(Ariane.getString("Menu.options.title"));
		_confI.setText(Ariane.getString("Menu.options.configuration"));
		_prefI.setText(Ariane.getString("Menu.options.preferences"));
		_operatorTreeI.setText(Ariane.getString("Menu.options.operators"));
		_subroutineTreeI.setText(Ariane.getString("Menu.options.subroutines"));

		//Menu graph
		_graphM.setText(Ariane.getString("Menu.graph.title"));
		_execI.setText(Ariane.getString("Menu.graph.exec"));
		_resetI.setText(Ariane.getString("Menu.graph.reset"));
		_generateI.setText(Ariane.getString("Menu.graph.generate"));
		_stopI.setText(Ariane.getString("Menu.graph.stop"));

		// Menu help
		_helpM.setText(Ariane.getString("Menu.help.title"));
		_aboutI.setText(Ariane.getString("Menu.help.about"));
		_helpI.setText(Ariane.getString("Menu.help.help"));

		// tips
		_newB.setToolTipText(Ariane.getString("Menu.file.new"));
		_openB.setToolTipText(Ariane.getString("Menu.file.open"));
		_saveB.setToolTipText(Ariane.getString("Menu.file.save"));
		_saveAsB.setToolTipText(Ariane.getString("Menu.file.saveas"));
		_printB.setToolTipText(Ariane.getString("Menu.file.print"));
		_undoB.setToolTipText(Ariane.getString("Menu.edit.undo"));
		_redoB.setToolTipText(Ariane.getString("Menu.edit.redo"));
		_copyB.setToolTipText(Ariane.getString("Menu.edit.copy"));
		_cutB.setToolTipText(Ariane.getString("Menu.edit.cut"));
		_pasteB.setToolTipText(Ariane.getString("Menu.edit.paste"));
		_eraseB.setToolTipText(Ariane.getString("Menu.edit.delete"));
		_prefB.setToolTipText(Ariane.getString("Menu.options.preferences"));
		_operatorTreeB.setToolTipText(Ariane.getString("Menu.options.operators"));
		_subroutineTreeB.setToolTipText(Ariane.getString("Menu.options.subroutines"));
		_execB.setToolTipText(Ariane.getString("Menu.graph.exec"));
		_generateB.setToolTipText(Ariane.getString("Menu.graph.generate"));
		_helpB.setToolTipText(Ariane.getString("Menu.help.help"));
		_stopB.setToolTipText(Ariane.getString("Menu.graph.stop"));

		// Change all text of the JColorChooser.
		UIManager.put("ColorChooser.cancelText", Ariane.getString("ColorChooser.cancelText"));
		UIManager.put("ColorChooser.okText", Ariane.getString("ColorChooser.okText"));
		UIManager.put("ColorChooser.resetText", Ariane.getString("ColorChooser.resetText"));
		UIManager.put("ColorChooser.previewText", Ariane.getString("ColorChooser.previewText"));
		UIManager.put("ColorChooser.sampleText", Ariane.getString("ColorChooser.sampleText"));
		UIManager.put("ColorChooser.swatchesNameText", Ariane.getString("ColorChooser.swatchesNameText"));
		UIManager.put("ColorChooser.swatchesRecentText", Ariane.getString("ColorChooser.swatchesRecentText"));
		UIManager.put("ColorChooser.rgbNameText", Ariane.getString("ColorChooser.rgbNameText"));
		UIManager.put("ColorChooser.rgbRedText", Ariane.getString("ColorChooser.rgbRedText"));
		UIManager.put("ColorChooser.rgbGreenText", Ariane.getString("ColorChooser.rgbGreenText"));
		UIManager.put("ColorChooser.rgbBlueText", Ariane.getString("ColorChooser.rgbBlueText"));
		UIManager.put("ColorChooser.hsbNameText", Ariane.getString("ColorChooser.hsbNameText"));
		UIManager.put("ColorChooser.hsbRedText", Ariane.getString("ColorChooser.hsbRedText"));
		UIManager.put("ColorChooser.hsbGreenText", Ariane.getString("ColorChooser.hsbGreenText"));
		UIManager.put("ColorChooser.hsbBlueText", Ariane.getString("ColorChooser.hsbBlueText"));
		UIManager.put("ColorChooser.hsbHueText", Ariane.getString("ColorChooser.hsbHueText"));
		UIManager.put("ColorChooser.hsbSaturationText", Ariane.getString("ColorChooser.hsbSaturationText"));
		UIManager.put("ColorChooser.hsbBrightnessText", Ariane.getString("ColorChooser.hsbBrightnessText"));

		// Change all text of the JFileChooser.
		UIManager.put("FileChooser.fileNameLabelText", Ariane.getString("FileChooser.fileNameLabelText"));
		UIManager.put("FileChooser.homeFolderToolTipText", Ariane.getString("FileChooser.homeFolderToolTipText"));
		UIManager.put("FileChooser.newFolderToolTipText", Ariane.getString("FileChooser.newFolderToolTipText"));
		UIManager.put("FileChooser.listViewButtonToolTipText", Ariane.getString("FileChooser.listViewButtonToolTipText"));
		UIManager.put("FileChooser.detailsViewButtonToolTipText", Ariane.getString("FileChooser.detailsViewButtonToolTipText"));
		UIManager.put("FileChooser.saveButtonText" , Ariane.getString("FileChooser.saveButtonText"));
		UIManager.put("FileChooser.openButtonText", Ariane.getString("FileChooser.openButtonText"));
		UIManager.put("FileChooser.cancelButtonText", Ariane.getString("FileChooser.cancelButtonText"));
		UIManager.put("FileChooser.updateButtonText", Ariane.getString("FileChooser.updateButtonText"));
		UIManager.put("FileChoosre.saveButtonToolTipText", Ariane.getString("FileChooser.saveButtonToolTipText"));
		UIManager.put("FileChooser.openButtonToolTipText", Ariane.getString("FileChooser.openButtonToolTipText"));
		UIManager.put("FileChooser.cancelButtonToolTipText", Ariane.getString("FileChooser.cancelButtonToolTipText"));
		UIManager.put("FileChooser.updateButtonToolTipText", Ariane.getString("FileChooser.updateButtonToolTipText"));
		UIManager.put("FileChooser.lookInLabelText", Ariane.getString("FileChooser.lookInLabelText"));
		UIManager.put("FileChooser.saveInLabelText", Ariane.getString("FileChooser.saveInLabelText"));
		UIManager.put("FileChooser.filesOfTypeLabelText", Ariane.getString("FileChooser.filesOfTypeLabelText"));
		UIManager.put("FileChooser.upFolderToolTipText", Ariane.getString("FileChooser.upFolderToolTipText"));
		UIManager.put("FileChooser.acceptAllFileFilterText", Ariane.getString("FileChooser.acceptAllFileFilterText"));

		UIManager.put("OptionPane.cancelButtonText", Ariane.getString("OptionPane.cancelButtonText"));
		UIManager.put("OptionPane.noButtonText", Ariane.getString("OptionPane.noButtonText"));
		UIManager.put("OptionPane.okButtonText", Ariane.getString("OptionPane.okButtonText"));
		UIManager.put("OptionPane.yesButtonText", Ariane.getString("OptionPane.yesButtontext"));
	}

	/**
	 * Changes the current texts of the interface
	 * from the new language.
	 */
	public void changeLanguage( ) {
		setInternationalText();
		_tree.setInternationalText();
		_explorer.setInternationalText();
	}

	private static void createTemporaryFolder( String directoryName ) {
		// Check existence.
		File directory = new File(directoryName);
		if (directory.isDirectory()) {
			File file;
			Long index = 1L;
			do {
				file = new File(directoryName + "ariane" + index++);
			} while (!file.mkdirs());
			_tempDir = file.getAbsolutePath() + File.separator;
		}
	}

	/**
	 * Sets the new temporary directory with the specified directory.
	 * @param dir the new directory.
	 */
	protected static void setTempDir( String dir ) {
		if (dir.endsWith(File.separator)) {
			createTemporaryFolder(dir);
		} else {
			createTemporaryFolder(dir + File.separator);
		}
	}

	/**
	 * Returns the default temporary directory.
	 * 
	 * @return the complete directory name
	 */
	public static String getDefaultTempDir( ) {
		return DEFAULT_TEMPORARY_DIRECTORY;
	}

	public static String getTempDirBasename( ) {
		File file = new File(_tempDir);
		return file.getParent() + File.separator;
	}

	public static String getTempDir( ) {
		return _tempDir;
	}

	/**
	 * Returns the list of all defined items
	 * for building graphs.
	 * 
	 * @return the tree of items.
	 */
	public Tree getTree( ) { return _tree; }

	/**
	 * Displays the workspace and
	 * in it the specified graph if any.
	 * 
	 * @param graphs a set of graphs or null.
	 */
	public void display( Graph[] graphs ) {
		for (Graph graph : graphs) {
			if (graph != null) {
				addGraph(graph);
			}
		}
	}

	/**
	 * Displays the specified property panel in the property area.
	 * @param propertyPanel a panel.
	 */
	public void displayPropertyPanel( PropertyViewer propertyPanel ) {
		_property.removeAll();
		if (propertyPanel != null) { _property.add(propertyPanel, BorderLayout.CENTER); }
		_property.revalidate();
		_property.repaint();
	}

	///// 
	///// C O N S O L E
	///// 

	/**
	 * Displays the given text in the console as
	 * a system message.
	 * 
	 * @param mesg the string to be displayed.
	 */
	public void printOutputMessage( String mesg ) {
		_console.displayOutputMessage(mesg);
	}

	/**
	 * Displays the given text in the console.
	 * The text is added to the current content
	 * in a new line.
	 * 
	 * @param mesg the string to be displayed.
	 */
	public void printSystemMessage( String mesg ) {
		_console.displaySystemMessage(mesg);
	}

	/**
	 * Displays the given warning text in the console.
	 * The text is added to the current content
	 * in a new line.
	 * 
	 * @param mesg the string to be displayed.
	 */
	public void printWarning( String mesg ) {
		_console.displayErrorMessage(Ariane.getString("warning") + mesg);
	}

	/**
	 * Displays the given error text in the console.
	 * The text is added to the current content
	 * in a new line.
	 * 
	 * @param mesg the string to be displayed.
	 */
	public void printErrorMessage( String mesg ) {
		_console.displayErrorMessage(Ariane.getString("error") + mesg);
	}

	/**
	 * Displays the given error text in the console.
	 * The text is added to the current content
	 * in a new line.
	 * 
	 * @param errorCode a conventional error code.
	 * @param mesg the string to be displayed.
	 */
	public void printErrorMessage( Ariane.ErrorCode errorCode, String item, int itemId, String mesg ) {
		_console.displayErrorMessage(item + " (" + itemId + "): " + Ariane.getString("error") + "#" + errorCode.ordinal() + ": " + mesg);
	}

	///// 
	/////  A C T I ON S
	///// 

	/**
	 * Displays the specified HTML 'page' as a manual page
	 * for the word 'title'. Pages are stored either
	 * <ol>
	 * <li>in the operators library package.
	 * <li>in the Ariane jar file.
	 * </ol>
	 * 
	 * @param title the title of the manual.
	 * @param page the URL address of the manual content.
	 * @param index index in the page as "#name".
	 */
	public void displayMan( String title, String page, String index ) {
		URI uri = HTMLDocumentViewer.getURI(page, index);
		if (uri == null) {
			return;
		}
		String command = PreferenceChooser.get(PreferenceChooser.WEB_BROWSER);
		if (command != null && command.length() > 0) {
			try {
				Runtime.getRuntime().exec(command + " " + uri);
				return;
			} catch (IOException e1) {
				Ariane.getWorkspace().printErrorMessage(Ariane.getString("Workspace.browser.error") + e1.getMessage());
			}
		}
		new HTMLDocumentViewer(title, uri);
	}

	/**
	 * Displays the HTML page for the
	 * Ariane manual.
	 */
	private void helpAction( ) {
		if (_tree.getIndexDoc() != null) {
			displayMan("Ariane", "manual.html", null);
		}
	}

	///// 
	/////  H I S T O R Y 
	/////

	/**
	 * Undoes the last action on the current graph.
	 */
	public void undo( ) {
		if (_currentGraph != null) {
			_currentGraph.undo();
		}
	}

	/**
	 * Redoes the last undo action (i.e., undoes undo).
	 */
	public void redo( ) {
		if (_currentGraph != null) {
			_currentGraph.redo();
		}
	}

	/**
	 * Updates the contents of the workspace.
	 */
	public void update( ) {
		_saveA.setEnabled(_currentGraph.isModified());
		_undoA.setEnabled(_currentGraph.canUndo());
		_redoA.setEnabled(_currentGraph.canRedo());
	}

	/**
	 * Changes the flag enabled of the save button
	 * to the value given in parameter.
	 * 
	 * @param status the status wanted for the flag of the button.
	 */
	public void setModifiedSaveB( boolean status ) {
		_saveA.setEnabled(status);
	}

	/**
	 * Sets the new current graph with the specified graph.
	 * 
	 * @param graph the graph.
	 */
	public void setCurrentGraph( Graph graph ) {
		_currentGraph = graph;
		if (_currentGraph == null) {
			_closeA.setEnabled(false);
			_closeAllA.setEnabled(false);
			_saveA.setEnabled(false);
			_undoA.setEnabled(false);
			_redoA.setEnabled(false);
			_saveAsA.setEnabled(false);
			_printA.setEnabled(false);
			_copyA.setEnabled(false);
			_cutA.setEnabled(false);
			_pasteA.setEnabled(false);
			_eraseA.setEnabled(false);
			_generateA.setEnabled(false);
			_selectA.setEnabled(false);
		} else {
			_closeA.setEnabled(true);
			_closeAllA.setEnabled(true);
			_saveAsA.setEnabled(true);
			if (!_stopA.isEnabled()) {
				_execA.setEnabled(true);
			}
			_printA.setEnabled(true);
			_copyA.setEnabled(true);
			_cutA.setEnabled(true);
			_pasteA.setEnabled(true);
			_eraseA.setEnabled(true);
			_generateA.setEnabled(true);
			_selectA.setEnabled(true);
			update();
		}
	}

	/**
	 * Selects all items in the current graph.
	 */
	public void selectAll( ) {
		if (_currentGraph != null) {
			_currentGraph.selectAll();
		}
	}

	///// 
	///// C O P Y  /  P A S T E
	///// 

	/**
	 * Returns the current session identifier.
	 * A session is either a loading or a pasting.
	 * @return the current session identifier.
	 */
	public static int getCurrentReadPasteSession() {
		return _currentReadPasteSession;
	}

	@Override
	public void lostOwnership( Clipboard clipboard, Transferable contents ) { /* empty */ }

	/**
	 * Copies the selected items in the clipboard. Each item
	 * are copied in the clipboard.
	 * <b>Warning</b>: connections between copied item
	 * are kept but connection between other items are deleted.
	 */
	public void copy( ) {
		if (_currentGraph != null) {
			Clipboard clipboard = Toolkit.getDefaultToolkit().getSystemClipboard();
			clipboard.setContents(_currentGraph.copySelectedItems(), this);
		}
	}

	/**
	 * Copies the selected items in the clipboard.
	 * <b>Warning</b>: connections between copied item
	 * are kept but connection between other items are deleted.
	 */
	public void cut( ) {
		copy();
		deleteSelectedItems();
	}

	/**
	 * @return true if the clipboard is not empty.
	 */
	public boolean isPastable( ) {
		Clipboard clipboard = Toolkit.getDefaultToolkit().getSystemClipboard();
		Transferable contents = clipboard.getContents(null);
		return _currentGraph != null && contents != null && contents.isDataFlavorSupported(XMLElement.FLAVOR);
	}

	/**
	 * Pastes all items in the current clipboard.
	 */
	public void paste( ) {
		Clipboard clipboard = Toolkit.getDefaultToolkit().getSystemClipboard();
		Transferable contents = clipboard.getContents(null);
		if (_currentGraph != null && contents != null && contents.isDataFlavorSupported(XMLElement.FLAVOR)) {
			_currentReadPasteSession++;
			XMLElement result;
			try {
				result = (XMLElement)contents.getTransferData(XMLElement.FLAVOR);
				_currentGraph.pasteSelectedItems(result);
			} catch (UnsupportedFlavorException e) {
			} catch (IOException e) {
			}
		}
	}

	public void deleteSelectedItems( ) {
		if (_currentGraph != null) {
			_currentGraph.deleteSelectedItems();
		}
	}

	/**
	 * Adds the specified item in the current graph.
	 * 
	 * @param treeItem the item to add.
	 * @param coordinate the coordinate of the future item.
	 * @throws InstantiationException if there is no current graph.
	 */
	public void addNewItem( TreeItem treeItem, Point coordinate ) throws InstantiationException {
		if (_currentGraph == null) {
			newGraph();
		}
		GraphItem graphItem = treeItem.createGraphItem(_currentGraph);
		Point pt = _currentGraph.getVisibleRect().getLocation();
		pt.translate(coordinate.x, coordinate.y);
		_currentGraph.startActionEdition();
		_currentGraph.unSelectAll();
		_currentGraph.addItemInSelection(graphItem, false);
		_currentGraph.addItem(graphItem, pt, true);
		_currentGraph.endActionEdition();
		_currentGraph.validate();
	}

	///// 
	/////  P R I NT I N G
	/////

	/**
	 * Prints the current graph.
	 * First of all displays a dialog box to 
	 * set users's preferences.
	 */
	private void print( ) {
		if (_currentGraph != null) {
			// Unselect all items.
			_currentGraph.unSelectAll();

			// Set up defaults preferences.
			PrintRequestAttributeSet aset = new HashPrintRequestAttributeSet();
			aset.add(new JobName("Ariane", PreferenceChooser.getCurrentLocale()));
			aset.add(OrientationRequested.LANDSCAPE);
			aset.add(MediaSizeName.ISO_A4);

			// Print it
			PrinterJob printJob = PrinterJob.getPrinterJob();
			printJob.setPrintable(_currentGraph);
			if (printJob.printDialog(aset)) {
				try {
					printJob.print(aset);
				} catch (PrinterException e) {
					printErrorMessage(Ariane.getString("Workspace.printing.error1") + e);
				}
			}
		}
	}

	/**
	 * Repaints all graphs.
	 * This method is invoked after by preferences setting
	 * and operators configuration.
	 */
	protected void repaintAllGraphs( ) {
		_editor.repaintAllGraphs();
	}

	/**
	 * Adds the given graph in the workspace.
	 * 
	 * @param graph the graph to be added.
	 */
	public void addGraph( Graph graph ) {
		_editor.add(graph);
	}

	/**
	 * Changes the title of the frame that
	 * supports the given graph.
	 * 
	 * @param graph the graph to be renamed.
	 * @param name the new name.
	 */
	public void renameGraph( Graph graph, String name ) {
		_editor.setTitle(graph, name);
	}

	/**
	 * Removes the given graph in the workspace.
	 * 
	 * @param graph the graph to be added.
	 */
	public void removeGraph( Graph graph ) {
		_editor.remove(graph);
	}

	/**
	 * Creates a new empty graph.
	 */
	public Graph newGraph( ) {
		Graph graph = new Graph(Ariane.getString("defaultfilename"), true);
		addGraph(graph);
		return graph;
	}

	/**
	 * Closes the current graph.
	 */
	private void close( ) {
		if (_currentGraph != null) {
			_currentGraph.close();
			updateRecentFilesListMenu();
		}
	}

	/**
	 * Closes all the graphs.
	 */
	private void closeAll( ) {
		_editor.closeAllGraphs();
	}

	/**
	 * Closes the application.
	 */
	private void quitAction( ) {
		if (quitConfirmed()) {
			System.exit(0);
		}
	}

	/**
	 * Generates the related code for
	 * the current root graph.
	 */
	private void generate( ) {
		FadingEffects.fadeIn(Ariane.getFrame());
		CodeGenerator generator = CodeGenerator.selectCodeGenerator(_currentGraph.getRoot());
		FadingEffects.fadeOut(Ariane.getFrame());
		if (generator != null) {
			generator.generate();
		}
	}

	/**
	 * Executes each selected items.
	 */
	private void execute( ) {
		if (_currentGraph != null) {
			_currentGraph.execute();
		}
	}

	/**
	 * Executes each selected items.
	 */
	private void reset( ) {
		if (_currentGraph != null) {
			_currentGraph.resetSelectedItems();
		}
	}

	/**
	 * Stops the current execution.
	 */
	private void stopExecution( ) {
		if (_currentGraph != null && !Graph.isExecutionStopped()) {
			_currentGraph.stopExecution();
			printSystemMessage(Ariane.getString("Workspace.execution.stop"));
		}
	}

	/**
	 * Sets the execution button to enabled
	 * or disabled from the running state.
	 * 
	 * @param isRunning the execution state.
	 */
	public void setExecution( boolean isRunning ) {
		if (isRunning) {
			_console.clear();
			_execA.setEnabled(false);
			_stopA.setEnabled(true);
		} else {
			_execA.setEnabled(true);
			_stopA.setEnabled(false);
		}
	}

	/**
	 * Deletes all temporary files.
	 * 
	 * @return the number of deleted files.
	 */
	public static int deleteTemporaryFiles() {
		File directory = new File(getTempDir());
		String[] files = directory.list();
		if (files == null) {
			return 0;
		}

		int numberOfErasedFiles = 0;
		for (int i = 0; i < files.length; i++) {
			if (files[i].startsWith(PreferenceChooser.getTemporaryFilePrefix()) && files[i].endsWith(PreferenceChooser.getTemporaryFileExtension())) {
				File file = new File(Workspace.getTempDir() + files[i]);
				if (file.delete()) {
					numberOfErasedFiles++;
				} else {
					Ariane.getWorkspace().printErrorMessage(Ariane.getString("PreferenceChooser.environment.error1") + " " + files[i]);
				}
			}
		}
		directory.delete();
		return numberOfErasedFiles;
	}

	/**
	 * Checks that all graphs have been saved
	 * before quitting.
	 * 
	 * @return true if all graphs have been deleted.
	 */
	public boolean quitConfirmed( ) {
		boolean quit = _editor.closeAllGraphs();
		if (quit) {
			deleteTemporaryFiles();
			PreferenceChooser.save();
		}
		return quit;
	}

	/**
	 * Loads a graph from an XML file.
	 * 
	 * @param file the file to be load.
	 * @return a graph or null.
	 */
	public Graph loadGraph( File file ) {
		_currentReadPasteSession++;
		try {
			XMLFileInputStream convert = new XMLFileInputStream(file);
			convert.parse();
			convert.close();
			Graph graph = new Graph(file);
			graph.setXML(convert.getElements());
			addRecentProjects(file.getAbsolutePath());
			return graph;
		} catch (IOException | SAXException | ParserConfigurationException ex) {
			printErrorMessage(Ariane.getString("Workspace.loading.error1") + file.getName() + " " + ex.getLocalizedMessage());
//			ex.printStackTrace();
			return null;
		}
	}

	/**
	 * Opens a new graph or multiple graphs.
	 * Opens a dialog box for choosing the graph(s).
	 */
	private void openFileAction( ) {
		File currentFileName = new File(getCurrentDir());

		JFileChooser fileChooser = new JFileChooser(currentFileName);
		fileChooser.setDialogType(JFileChooser.OPEN_DIALOG);
		fileChooser.setDialogTitle(Ariane.getString("FileChooser.titleOpenGraph"));
		fileChooser.setMultiSelectionEnabled(true);
		fileChooser.setFileFilter(new javax.swing.filechooser.FileFilter() {
			@Override
			public boolean accept( File file ) {
				String filename = file.getName();
				return filename.endsWith(Ariane.DEFAULT_FILE_EXT) || file.isDirectory();
			}
			@Override
			public String getDescription() {
				return "Ariane (*" + Ariane.DEFAULT_FILE_EXT + ")";
			}
		});

		FadingEffects.fadeIn(Ariane.getFrame());
		if (fileChooser.showOpenDialog(this) == JFileChooser.APPROVE_OPTION) {
			for ( File file : fileChooser.getSelectedFiles()) {
				openFile(file.getAbsolutePath());
			}
			setCurrentDir(fileChooser.getSelectedFiles()[0].getParent());
		}
		FadingEffects.fadeOut(Ariane.getFrame());
	}

	/**
	 * Loads the specified graph.
	 * 
	 * @param filename a regular filename.
	 */
	private void openFile( String filename ) {
		if (!_editor.exists(filename)) {
			Graph graph = loadGraph(new File(filename));
			if (graph != null) {
				addGraph(graph);
			}
		}
	}

	/**
	 * Gets the current directory.
	 * 
	 * @return the current directory name.
	 */
	public static String getCurrentDir( ) { 
		return _currentDir;
	}

	/**
	 * Sets the new current directory.
	 * 
	 * @param directory a regular directory.
	 */
	public static void setCurrentDir( String directory ) { 
		_currentDir = directory;
	}

	/**
	 * Saves the graph.
	 */
	private void saveFileAction( ) {
		if (_currentGraph != null) {
			_currentGraph.getRoot().save();
		}
	}

	/**
	 * Saves the graph with a new name.
	 */
	private void saveAsFileAction( ) {
		if (_currentGraph != null) {
			_currentGraph.getRoot().saveAs();
		}
	}

	/**
	 * Opens the dialog box for setting configuration.
	 */
	public void configurationAction( ) {
		FadingEffects.fadeIn(Ariane.getFrame());
		new LibraryPathSetting().setVisible(true);
		FadingEffects.fadeOut(Ariane.getFrame());
	}

	/**
	 * Opens the dialog box for setting preferences.
	 */
	private void preferencesAction( ) {
		FadingEffects.fadeIn(Ariane.getFrame());
		new PreferenceChooser(this).setVisible(true);
		FadingEffects.fadeOut(Ariane.getFrame());
	}

	/**
	 * Opens the dialog box for managing personal operators.
	 */
	private void treeOperatorAction( ) {
		FadingEffects.fadeIn(Ariane.getFrame());
		new OperatorTreeEditor(getTree().getUserOperatorTree()).setVisible(true);
		FadingEffects.fadeOut(Ariane.getFrame());
	}

	/**
	 * Opens the dialog box for managing subroutine tree.
	 */
	private void treeSubroutineAction( ) {
		FadingEffects.fadeIn(Ariane.getFrame());
		new SubroutineTreeEditor(getTree().getSubroutineTree()).setVisible(true);
		FadingEffects.fadeOut(Ariane.getFrame());
	}

	/**
	 * Opens the "About Box".
	 */
	private void aboutAction( ) {
		FadingEffects.fadeIn(Ariane.getFrame());
		new AboutBox(Ariane.getFrame()).setVisible(true);
		FadingEffects.fadeOut(Ariane.getFrame());
	}

	/**
	 * Imports new routines from a file
	 * asked to the user.
	 */
	private void importSubroutinesAction() {
		JFileChooser fileChooser = new JFileChooser();
		fileChooser.setDialogType(JFileChooser.OPEN_DIALOG);
		fileChooser.setDialogTitle(Ariane.getString("FileChooser.titleOpenFile"));
		fileChooser.setFileSelectionMode(JFileChooser.FILES_ONLY);
		fileChooser.setSelectedFile(new File(Ariane.getLibraryPath()));
		int returnVal = fileChooser.showOpenDialog(this);
		if (returnVal == JFileChooser.APPROVE_OPTION) {
			File file = fileChooser.getSelectedFile();
			try {
				MutableTreeItem tree = _tree.addSubroutines(Ariane.readXMLFile(file));
				SubroutineTreeEditor.save(tree);
				Ariane.getWorkspace().printSystemMessage(Ariane.getString("Workspace.addsubroutine.message"));
			} catch (IOException e) {
				Ariane.getWorkspace().printErrorMessage(Ariane.getString("Workspace.addsubroutine.error") + " " + file.getName() + ". " + e.getLocalizedMessage());
			}
		}
	}

	/**
	 * Exports new routines from a file.
	 */
	private void exportSubroutinesAction() {
		JFileChooser fileChooser = new JFileChooser();
		fileChooser.setDialogType(JFileChooser.SAVE_DIALOG);
		fileChooser.setDialogTitle(Ariane.getString("FileChooser.titleSaveFile"));
		fileChooser.setFileSelectionMode(JFileChooser.FILES_ONLY);
		fileChooser.setSelectedFile(new File(Ariane.getLibraryPath()));
		if (fileChooser.showSaveDialog(this) == JFileChooser.APPROVE_OPTION) {
			File file = fileChooser.getSelectedFile();
			if (file.exists() && JOptionPane.showConfirmDialog(Ariane.getWorkspace(),
					Ariane.getString("Workspace.dialog.message.part1")
					+ file.getName()
					+ Ariane.getString("Workspace.dialog.message.part2"),
					Ariane.getString("Workspace.dialog.title"),
					JOptionPane.YES_NO_OPTION) == JOptionPane.NO_OPTION) {
				return;
			}
			try {
				DataFile.copy(Ariane.getSubroutineFilename(), file.getAbsolutePath());
				Ariane.getWorkspace().printSystemMessage(file.getName()
						+ " " + Ariane.getString("Workspace.saving.success"));
			} catch ( IOException e ) {
				Ariane.getWorkspace().printErrorMessage(Ariane.getString("PreferenceChooser.error.saved")
						+ " " + Ariane.getPreferencesFilename() + " " + e.getLocalizedMessage());
			}
		}
	}
}
