/* -*- c-basic-offset: 3 -*-
 *
 * Copyright (c) 2013, GREYC.
 * All rights reserved
 *
 * You may use this file under the terms of the BSD license as follows:
 *
 * "Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of the GREYC, nor the name of its
 *     contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
 * 
 *   
 * For more information, refer to:
 * https://clouard.users.greyc.fr/Pandore
 */

package fr.greyc.pantheon.ariane.utils;

import java.awt.Graphics2D;
import java.awt.Image;
import java.awt.RenderingHints;
import java.awt.Toolkit;
import java.awt.geom.AffineTransform;
import java.awt.image.BufferedImage;
import java.io.File;
import java.io.IOException;

import javax.imageio.ImageIO;
import javax.swing.ImageIcon;

/**
 * The class <code>ImageDecoder</code> provides facilities
 * to decode image files.
 * Recognized image formats are gif, png, jpg and Pandore.
 * ?? tif, bmp, ppm, pgm, pbm
 * 
 * @author Régis CLOUARD - Jan 11, 2007
 */
public final class ImageDecoder {
	
	private ImageDecoder() { }
	
	/**
	 * Decodes the image in the given filename and return a thumbnail
	 * of the image with the specified maximum size.
	 * 
	 * @param filename the filename.
	 * @param maxSize the maximum size both in width and height.
	 * @return the image icon or null of any.
	 */
	public static ImageIcon getThumbnail( String filename, int maxSize ) {
		Image image;

		// may be a regular image file (bmp, png, ...)
		try {
			image = ImageIO.read(new File(filename));
		} catch (IOException e) { 
			image = null;
		}

		// May be a Pandore image??
		if (image == null) {
			PandoreImageDecoder pan = new PandoreImageDecoder(maxSize);
			if (pan.readImage(filename)) {
				image = Toolkit.getDefaultToolkit().createImage(pan);
			}
		}

		// Not a regular image
		if (image == null) {
			return null;
		}

		// Resize image.
		if (image.getWidth(null) > image.getHeight(null)) {
			if (image.getWidth(null) > maxSize) {
				// Optimization for custom images (otherwise resampling is very very slow
				if (image instanceof BufferedImage && ((BufferedImage)image).getType() == BufferedImage.TYPE_CUSTOM) {
					BufferedImage image1 = new BufferedImage(image.getWidth(null), image.getHeight(null), BufferedImage.TYPE_INT_RGB);
					image1.setData(((BufferedImage)image).getData());
					image = scaleImage(image1, maxSize, -1);
				} else {
					image = scaleImage(image, maxSize, -1);
				}
			}
		} else {
			if (image.getHeight(null) > maxSize) {
				// Optimization for custom images (otherwise resampling is very very slow
				if (image instanceof BufferedImage && ((BufferedImage)image).getType() == BufferedImage.TYPE_CUSTOM) {
					BufferedImage image1 = new BufferedImage(image.getWidth(null), image.getHeight(null), BufferedImage.TYPE_INT_RGB);
					image1.setData(((BufferedImage)image).getData());
					image = scaleImage(image1, -1, maxSize);
				} else {
					image = scaleImage(image, -1, maxSize);
				}
			}
		}
		ImageIcon icon = new ImageIcon(image);

		return icon;
	}
	
	private static Image scaleImage( Image image, int width, int height ) {
		int requiredWidth;
		int requiredHeight;

		if (width < 0) {
			requiredWidth = (int)((double)height / image.getHeight(null) * image.getWidth(null));
			requiredHeight = height;
		} else if (height < 0) {
			requiredHeight = (int)((double)width / image.getWidth(null) * image.getHeight(null));
			requiredWidth = width;
		} else {
			requiredWidth = width;
			requiredHeight = height;
		}
		double scaleFactor = (double)requiredHeight / image.getHeight(null);

		BufferedImage image2 = new BufferedImage(requiredWidth, requiredHeight, BufferedImage.TYPE_INT_RGB);
		Graphics2D imgG2d = (Graphics2D)image2.getGraphics();
		AffineTransform scaleTransform = AffineTransform.getScaleInstance(scaleFactor, scaleFactor);
		imgG2d.setRenderingHint(RenderingHints.KEY_INTERPOLATION, RenderingHints.VALUE_INTERPOLATION_BILINEAR);
		imgG2d.setTransform(scaleTransform);
		imgG2d.drawImage(image, 0, 0, null);
		imgG2d.dispose();

		return image2;
	}
}