/* -*- c-basic-offset: 3 -*-
 *
 * Copyright (c) 2013, GREYC.
 * All rights reserved
 *
 * You may use this file under the terms of the BSD license as follows:
 *
 * "Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of the GREYC, nor the name of its
 *     contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
 * 
 *   
 * For more information, refer to:
 * https://clouard.users.greyc.fr/Pandore
 */

package fr.greyc.pantheon.ariane.utils;

import java.awt.AWTException;
import java.awt.Dimension;
import java.awt.image.ColorModel;
import java.awt.image.ImageConsumer;
import java.awt.image.ImageProducer;
import java.io.BufferedInputStream;
import java.io.ByteArrayInputStream;
import java.io.DataInputStream;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.Random;

/**
 * Extends ImageProducer to decode Pandore images. This code is based
 * on the book "Java AWT references" by John Zukowski.
 * 
 * Custom usage: PandoreImageDecoder pan= new PandoreImageDecoder();
 * pan.readImage(imagebyte); img = Toolkit.getDefaultToolkit().createImage(pan);
 * 
 * @author Régis CLOUARD - Jan 03, 2001
 * @version Régis CLOUARD - Nov 07, 2002 (Pandore images).
 * @version Régis CLOUARD - Jan 16, 2003 (Pandore 4).
 */
public final class PandoreImageDecoder implements ImageProducer {
	/* List of the Pandore types. */
	/* Do not change the values, they are conventional. */

	//	private static final int Po_Collection = 1;
	//	private static final int Po_Img1duc = 2;
	//	private static final int Po_Img1dsl = 3;
	//	private static final int Po_Img1dsf = 4;

	private static final int PO_IMG2DUC = 5;
	private static final int PO_IMG2DSL = 6;
	private static final int PO_IMG2DSF = 7;

	private static final int PO_IMG3DUC = 8;
	private static final int PO_IMG3DSL = 9;
	private static final int PO_IMG3DSF = 10;

	//	private static final int Po_Reg1d = 11;
	private static final int PO_REG2D = 12;
	private static final int PO_REG3D = 13;

	//	private static final int Po_Graph2d = 14;
	//	private static final int Po_Graph3d = 15;

	private static final int PO_IMC2DUC = 16;
	private static final int PO_IMC2DSL = 17;
	private static final int PO_IMC2DSF = 18;

	private static final int PO_IMC3DUC = 19;
	private static final int PO_IMC3DSL = 20;
	private static final int PO_IMC3DSF = 21;

	//	private static final int Po_Imx1duc = 22;
	//	private static final int Po_Imx1dsl = 23;
	//	private static final int Po_Imx1dsf = 24;
	//	private static final int Po_Imx2duc = 25;
	//	private static final int Po_Imx2dsl = 26;
	//	private static final int Po_Imx2dsf = 27;
	//	private static final int Po_Imx3duc = 28;
	//	private static final int Po_Imx3dsl = 29;
	//	private static final int Po_Imx3dsf = 30;

	private static final String PANDOREMAGIC = "PANDORE04";
	/** object type */
	private int _potype;
	/** Is it little or big endian encoding. */
	private boolean _reverseByteOrder;
	private final int _maxSize;

	/** Since done in-memory; only one consumer */
	private ImageConsumer _consumer;

	/**
	 * Format of Pandore file is single pass/frame, w/ complete scan lines in
	 * order
	 */
	private static final int PANDORE_HINTS =
			ImageConsumer.TOPDOWNLEFTRIGHT
			| ImageConsumer.COMPLETESCANLINES
			| ImageConsumer.SINGLEPASS
			| ImageConsumer.SINGLEFRAME;

	/**
	 * Problem with big-endian and little-endian:
	 * direct reading or reverse byte order?
	 */
	private boolean _loadError;

	private Dimension _thumbnailDimension;

	/** Array of pixels. */
	private int[][] _store;

	/**
	 * Build a Pandore image decoder in order to build
	 * an image of maxSize from a file.
	 * @param maxSize max size for both width and height.
	 */
	public PandoreImageDecoder( int maxSize ) {
		_maxSize = maxSize;
	}

	/**
	 * Reads Pandore file header and checks consistency.
	 * 
	 * @param dis buffered image data;
	 * @return identifier of the Pandore objet if 'dis' is a regular Pandore
	 *         file or -1 otherwise.
	 * @exception AWTException when file is not a regular Pandore file.
	 * @exception IOException in case of io exception.
	 */
	private int readHeader( DataInputStream dis ) throws AWTException, IOException {
		byte[] buf = new byte[10];

		// Magic number.
		dis.read(buf, 0, 9);
		/** Magic number : PO_MAGIC */
		String magicNumber = new String(buf, 0, 9);
		dis.skipBytes(1); // skip '\n'
		if (!(PANDOREMAGIC.equals(magicNumber))) {
			throw new AWTException("Invalid Pandore Format " + magicNumber);
		}

		// skip unused bytes.
		dis.skipBytes(2);

		// Hint: When typeoj is wrong -> reverse byte order.
		_potype = dis.readInt();
		if (_potype > 65535) {
			_reverseByteOrder = true;
			_potype = swap(_potype);
		}

		// Creator name
		dis.read(buf, 0, 8);
		dis.skipBytes(1); // skip '\0'

		// Date
		dis.read(buf, 0, 9);
		dis.skipBytes(1); // '\0';

		// Skip unused bytes
		dis.skipBytes(1);

		return _potype;
	}

	/**
	 * There is only a single consumer. When it registers, produce image. On
	 * error, notify consumer.
	 * 
	 * @param ic new consumer to be added.
	 */
	@Override
	public void addConsumer( ImageConsumer imageConsumer ) {
		_consumer = imageConsumer;
		try {
			produce();
		} catch (Exception e) {
			if (_consumer != null) {
				_consumer.imageComplete(ImageConsumer.IMAGEERROR);
			}
		}
		_consumer = null;
	}

	/**
	 * If consumer passed to routine is single consumer, return true, else false.
	 * 
	 * @param ic consumer.
	 * @return true if it is a consumer.
	 */
	@Override
	public boolean isConsumer( ImageConsumer imageConsumer ) {
		return imageConsumer.equals(_consumer);
	}

	/**
	 * Disables consumer if currently consuming.
	 * 
	 * @param ic consumer.
	 */
	@Override
	public void removeConsumer( ImageConsumer imageConsumer ) {
		if (imageConsumer.equals(_consumer)) {
			_consumer = null;
		}
	}

	/**
	 * Production is done by adding consumer.
	 * 
	 * @param ic consumer.
	 */
	@Override
	public void startProduction( ImageConsumer imageConsumer ) {
		addConsumer(imageConsumer);
	}

	/**
	 * Requests, on behalf of the <code>ImageConsumer</code>, 
	 * that the <code>ImageProducer</code> attempt to resend 
	 * the image data one more time in TOPDOWNLEFTRIGHT order 
	 * so that higher quality conversion algorithms which 
	 * depend on receiving pixels in order can be used to 
	 * produce a better output version of the image.
	 * 
	 * @param ic consumer.
	 */
	@Override
	public void requestTopDownLeftRightResend( ImageConsumer imageConsumer ) {
		// Not needed. The data is always in this format.
	}

	/**
	 * Production Process: Prerequisite: Image already read into store array.
	 * (readImage) props / width / height already set (readImage) Assumes RGB
	 * Color Model - would need to filter to change. Sends Pandore Image data to
	 * consumer. Pixels sent one row at a time.
	 */
	private void produce( ) {
		ColorModel colorModel = ColorModel.getRGBdefault();
		if (_consumer != null) {
			if (_loadError) {
				_consumer.imageComplete(ImageConsumer.IMAGEERROR);
			} else {
				_consumer.setDimensions(_thumbnailDimension.width, _thumbnailDimension.height);
				_consumer.setColorModel(colorModel);

				_consumer.setHints(PANDORE_HINTS);
				for (int y = 0; y < _thumbnailDimension.height; y++) {
					_consumer.setPixels(0, y, _thumbnailDimension.width, 1, colorModel, _store[y], 0, _thumbnailDimension.width);
				}
				_consumer.imageComplete(ImageConsumer.STATICIMAGEDONE);
			}
		}
	}

	/**
	 * Swaps integer.
	 * @param value the integer to be swapped.
	 * @return the swapped integer.
	 */
	public int swap( int value ) {
		int byte0 = value & 0xff;
		int byte1 = (value >> 8) & 0xff;
		int byte2 = (value >> 16) & 0xff;
		int byte3 = (value >> 24) & 0xff;
		// swap the byte order
		return (byte0 << 24) | (byte1 << 16) | (byte2 << 8) | byte3;
	}

	/**
	 * Allows reading from a filename.
	 * 
	 * @param filename complete name of the image file.
	 */
	public boolean readImage( String filename ) {
		try {
			FileInputStream fileImage = new FileInputStream(filename);
			if (fileImage.available() > 0) {
				readImage(fileImage);
			}
			fileImage.close();
			return true;
		} catch (Exception e) {
			return false;
		}
	}


	/**
	 * Reads an integer format the input stream.
	 * If needed inverts the bytes.
	 * 
	 * @param dis the input stream.
	 * @return the data array.
	 * @throws IOException in case if IO error.
	 */
	private int readInt( DataInputStream dis ) throws IOException {
		byte[] row = new byte[4];
		for (int b = 0; b < 4; b++) {
			row[b] = dis.readByte();
		}
		return parseBytes(row, 0, 4);
	}

	/**
	 * Converts Pandore pixel to color that can be used by Java.
	 * 
	 * @param row line of pixels
	 * @param column column number of the pixel.
	 * @param bytes bytes per pixel.
	 * @return color to be displayed
	 */
	private int parseBytes( byte[] row, int column, int bytes ) {
		int[] colorByte = new int[4];

		for (int b = 0; b < bytes; b++) {
			colorByte[b] = row[column * bytes + b];
			// Problem signed -> unsigned.
			if (colorByte[b] < 0) {
				colorByte[b] += 256;
			}
		}
		int color = 0;

		if (_reverseByteOrder) {
			for (int b = 0; b < bytes; b++) {
				int shift = b * 8;
				color |= colorByte[b] << shift;
			}
		} else {
			for (int b = 0; b < bytes; b++) {
				int shift = (bytes - 1 - b) * 8;
				color |= colorByte[b] << shift;
			}
		}

		return color;
	}

	/**
	 * Allows reading to be from internal byte array, in addition to
	 * disk/socket.
	 * 
	 * @param data the byte array.
	 */
	public void readImage( byte[] data ) throws AWTException {
		readImage(new ByteArrayInputStream(data));
	}

	private Dimension computeThumbnailSize( int height, int width ) {
		Dimension dimension = new Dimension();
		if (width >= height) {
			if (width > _maxSize) {
				int ratio = 1 + width / _maxSize;
				dimension.width = width / ratio;
				dimension.height = height / ratio;
			}
		} else {
			if (height > _maxSize) {
				int ratio = 1 + height / _maxSize;
				dimension.width = width / ratio;
				dimension.height = height / ratio;
			}
		}
		return dimension;
	}

	/**
	 * Reads gray data image. First of all read the image size.
	 * 
	 * @param dis buffered data to parse.
	 * @return the data array.
	 * @throws IOException in case of IO errors.
	 */
	public int[][] read2dGrayData( DataInputStream dis ) throws IOException {
		// read image size.
		readInt(dis); // bands = 1 !
		int height = readInt(dis);
		int width = readInt(dis);

		// read data
		_thumbnailDimension = computeThumbnailSize(height, width);
		int[][] store = new int[_thumbnailDimension.height][_thumbnailDimension.width];
		int stepX = width / _thumbnailDimension.width;
		int stepY = height / _thumbnailDimension.height;
		byte[] row;
		switch (_potype) {
		case PO_IMG2DUC:
			row = new byte[width];
			for (int y = 0; y < _thumbnailDimension.height; y++) {
				dis.readFully(row);
				dis.skipBytes(width * (stepY - 1));
				for (int x = 0; x < _thumbnailDimension.width; x++) {
					int color = parseBytes(row, x * stepX, 1);

					store[y][x] = (0xff << 24) | (color << 16);
					store[y][x] |= color << 8;
					store[y][x] |= color;
				}
			}
			break;

		case PO_IMG2DSL:
			row = new byte[width * 4];
			for (int y = 0; y < _thumbnailDimension.height; y++) {
				dis.readFully(row);
				dis.skipBytes(width * 4 * (stepY - 1));
				for (int x = 0; x < _thumbnailDimension.width; x++) {
					int color = parseBytes(row, x * stepX, 4);
					store[y][x] = (0xff << 24) | (color << 16) | (color << 8) | color;
				}
			}
			break;

		case PO_REG2D:
			Random rand = new Random(0);
			int nlabels = readInt(dis);
			int[] lut = new int[nlabels];
			int bytes;
			if (nlabels < 256) {
				bytes = 1;
			} else if (nlabels < 65536) {
				bytes = 2;
			} else {
				bytes = 4;
			}
			// Colors are set randomly except 0
			lut[0] = 0xff << 24;
			for (int i = 1; i < nlabels; i++) {
				lut[i] = (0xff << 24) | (rand.nextInt(256) << 16)
						| (rand.nextInt(256) << 8) | rand.nextInt(256);
			}
			row = new byte[width * bytes];
			for (int y = 0; y < _thumbnailDimension.height; y++) {
				dis.readFully(row);
				dis.skipBytes(width * bytes * (stepY - 1));
				for (int x = 0; x < _thumbnailDimension.width; x++) {
					int color = parseBytes(row, x * stepX, bytes);
					if (color < nlabels) {
						store[y][x] = lut[color];
					}
				}
			}
			break;

		case PO_IMG2DSF:
			row = new byte[width * 4];
			for (int y = 0; y < _thumbnailDimension.height; y++) {
				dis.readFully(row);
				dis.skipBytes(width * 4 * (stepY - 1));
				for (int x = 0; x < _thumbnailDimension.width; x++) {
					int color = parseBytes(row, x * stepX, 4);
					float toto = Float.intBitsToFloat(color);
					color = (int)toto;
					store[y][x] = (0xff << 24) | (color << 16) | (color << 8) | color;
				}
			}
			break;
		default:
			break;
		}
		return store;
	}

	/**
	 * Reads color data image.
	 * 
	 * @param dis the input stream.
	 * @return the data array.
	 * @throws IOException in case of IO error.
	 */
	public int[][] read2dColorData( DataInputStream dis ) throws IOException {
		// read image size.
		readInt(dis); // bands = 3;
		int height = readInt(dis);
		int width = readInt(dis);

		// ColorSpace.
		dis.readInt();

		// read data
		_thumbnailDimension = computeThumbnailSize(height, width);
		int[][] store = new int[_thumbnailDimension.height][_thumbnailDimension.width];
		int stepX = width / _thumbnailDimension.width;
		int stepY = height / _thumbnailDimension.height;
		byte[] row;

		switch (_potype) {
		case PO_IMC2DUC:
			row = new byte[width];
			// red band
			for (int y = 0; y < _thumbnailDimension.height; y++) {
				dis.readFully(row);
				dis.skipBytes(width * (stepY - 1));
				for (int x = 0; x < _thumbnailDimension.width; x++) {
					int color = parseBytes(row, x * stepX, 1);
					store[y][x] = (0xff << 24) | (color << 16);
				}
			}
			dis.skipBytes(width * (height - stepY * _thumbnailDimension.height));
			// green band
			for (int y = 0; y < _thumbnailDimension.height; y++) {
				dis.readFully(row);
				dis.skipBytes(width * (stepY - 1));
				for (int x = 0; x < _thumbnailDimension.width; x++) {
					int color = parseBytes(row, x * stepX, 1);
					store[y][x] |= color << 8;
				}
			}
			dis.skipBytes(width * (height - stepY * _thumbnailDimension.height));
			// blue band
			for (int y = 0; y < _thumbnailDimension.height; y++) {
				dis.readFully(row);
				dis.skipBytes(width * (stepY - 1));
				for (int x = 0; x < _thumbnailDimension.width; x++) {
					int color = parseBytes(row, x * stepX, 1);
					store[y][x] |= color;
				}
			}
			break;

		case PO_IMC2DSL:
			row = new byte[width * 4];
			for (int y = 0; y < _thumbnailDimension.height; y++) {
				dis.readFully(row);
				dis.skipBytes(width * 4 * (stepY - 1));
				for (int x = 0; x < _thumbnailDimension.width; x++) {
					int color = parseBytes(row, x * stepX, 4);
					store[y][x] = (0xff << 24) | (color << 16);
				}
			}
			dis.skipBytes(4 * width * (height - stepY * _thumbnailDimension.height));

			for (int y = 0; y < _thumbnailDimension.height; y++) {
				dis.readFully(row);
				dis.skipBytes(width * 4 * (stepY - 1));
				for (int x = 0; x < _thumbnailDimension.width; x++) {
					int color = parseBytes(row, x * stepX, 4);
					store[y][x] |= color << 8;
				}
			}
			dis.skipBytes(4 * width * (height - stepY * _thumbnailDimension.height));

			for (int y = 0; y < _thumbnailDimension.height; y++) {
				dis.readFully(row);
				dis.skipBytes(width * 4 * (stepY - 1));
				for (int x = 0; x < _thumbnailDimension.width; x++) {
					int color = parseBytes(row, x * stepX, 4);
					store[y][x] |= color;
				}
			}
			break;

		case PO_IMC2DSF:
			row = new byte[width * 4];
			for (int y = 0; y < _thumbnailDimension.height; y++) {
				dis.readFully(row);
				dis.skipBytes(width * 4 * (stepY - 1));
				for (int x = 0; x < _thumbnailDimension.width; x++) {
					int color = parseBytes(row, x * stepX, 4);
					float toto = Float.intBitsToFloat(color);
					color = (int)toto;
					store[y][x] = (0xff << 24) | (color << 16);
				}
			}
			dis.skipBytes(4 * width * (height - stepY * _thumbnailDimension.height));

			for (int y = 0; y < _thumbnailDimension.height; y++) {
				dis.readFully(row);
				dis.skipBytes(width * 4 * (stepY - 1));
				for (int x = 0; x < _thumbnailDimension.width; x++) {
					int color = parseBytes(row, x * stepX, 4);
					float toto = Float.intBitsToFloat(color);
					color = (int)toto;
					store[y][x] |= color << 8;
				}
			}
			dis.skipBytes(4 * width * (height - stepY * _thumbnailDimension.height));

			for (int y = 0; y < _thumbnailDimension.height; y++) {
				dis.readFully(row);
				dis.skipBytes(width * 4 * (stepY - 1));
				for (int x = 0; x < _thumbnailDimension.width; x++) {
					int color = parseBytes(row, x * stepX, 4);
					float toto = Float.intBitsToFloat(color);
					color = (int)toto;
					store[y][x] |= color;
				}
			}
			break;
		default:
			break;
		}
		return store;
	}

	/**
	 * Reads gray data image.
	 * Note that this method is used for displaying
	 * image thumbnail so it does not read all the data but
	 * only the slice in the middle of the image..
	 * 
	 * @param dis buffered data to parse.
	 * @return the data array.
	 * @throws IOException in case of IO errors.
	 */
	public int[][] read3dGrayData( DataInputStream dis ) throws IOException {
		// read image size.
		readInt(dis); // read the number of bands : always 1 !
		final int depth = readInt(dis);
		final int height = readInt(dis);
		final int width = readInt(dis);

		_thumbnailDimension = computeThumbnailSize(height, width);
		int[][] store = new int[_thumbnailDimension.height][_thumbnailDimension.width];
		int stepX = width / _thumbnailDimension.width;
		int stepY = height / _thumbnailDimension.height;
		byte[] row;

		// read data
		switch (_potype) {
		case PO_IMG3DUC:
			row = new byte[width];
			dis.skipBytes((depth / 2) * width * height);
			for (int y = 0; y < _thumbnailDimension.height; y++) {
				dis.readFully(row);
				dis.skipBytes(width * (stepY - 1));
				for (int x = 0; x < _thumbnailDimension.width; x++) {
					int color = parseBytes(row, x * stepX, 1);

					store[y][x] = (0xff << 24) | (color << 16);
					store[y][x] |= color << 8;
					store[y][x] |= color;
				}
			}
			break;

		case PO_IMG3DSL:
			row = new byte[width * 4];
			dis.skipBytes((depth / 2) * width * 4 * height);
			for (int y = 0; y < _thumbnailDimension.height; y++) {
				dis.readFully(row);
				dis.skipBytes(width * 4 * (stepY - 1));
				for (int x = 0; x < _thumbnailDimension.width; x++) {
					int color = parseBytes(row, x * stepX, 4);

					store[y][x] = (0xff << 24) | (color << 16);
					store[y][x] |= color << 8;
					store[y][x] |= color;
				}
			}
			break;

		case PO_IMG3DSF:
			row = new byte[width * 4];
			dis.skipBytes((depth / 2) * width * 4 * height);
			for (int y = 0; y < _thumbnailDimension.height; y++) {
				dis.readFully(row);
				dis.skipBytes(width * 4 * (stepY - 1));
				for (int x = 0; x < _thumbnailDimension.width; x++) {
					int color = parseBytes(row, x * stepX, 4);
					float toto = Float.intBitsToFloat(color);
					color = (int)toto;

					store[y][x] = (0xff << 24) | (color << 16);
					store[y][x] |= color << 8;
					store[y][x] |= color;
				}
			}
			break;

		case PO_REG3D:
			Random rand = new Random(0);
			int nlabels = readInt(dis);
			int[] lut = new int[nlabels];
			int bytes;
			if (nlabels < 256) {
				bytes = 1;
			} else if (nlabels < 65536) {
				bytes = 2;
			} else {
				bytes = 4;
			}
			// Colors are set randomly except 0
			lut[0] = 0xff << 24;
			for (int i = 1; i < nlabels; i++) {
				lut[i] = (0xff << 24) | (rand.nextInt(256) << 16)
						| (rand.nextInt(256) << 8) | rand.nextInt(256);
			}
			row = new byte[width * bytes];
			dis.skipBytes((depth / 2) * width * height);
			for (int y = 0; y < _thumbnailDimension.height; y++) {
				dis.readFully(row);
				dis.skipBytes(width * bytes * (stepY - 1));
				for (int x = 0; x < _thumbnailDimension.width; x++) {
					int color = parseBytes(row, x * stepX, bytes);
					if (color < nlabels) {
						store[y][x] = lut[color];
					}
				}
			}
			break;
		default:
			break;

		}
		return store;
	}

	/**
	 * Reads color data image.
	 * 
	 * @param dis the input stream.
	 * @return the data array.
	 * @throws IOException in case of IO error.
	 */
	public int[][] read3dColorData( DataInputStream dis ) throws IOException {
		// read image size.
		readInt(dis); // bands = 1 !
		int depth = readInt(dis);
		int height = readInt(dis);
		int width = readInt(dis);

		// read data
		_thumbnailDimension = computeThumbnailSize(height, width);
		int[][] store = new int[_thumbnailDimension.height][_thumbnailDimension.width];
		int stepX = width / _thumbnailDimension.width;
		int stepY = height / _thumbnailDimension.height;
		byte[] row;
		switch (_potype) {
		case PO_IMC3DUC:
			row = new byte[width];
			dis.skipBytes((depth / 2) * width * height);
			for (int y = 0; y < _thumbnailDimension.height; y++) {
				dis.readFully(row);
				dis.skipBytes(width * (stepY - 1));
				for (int x = 0; x < _thumbnailDimension.width; x++) {
					int color = parseBytes(row, x * stepX, 1);
					store[y][x] = (0xff << 24) | (color << 16);
				}
			}
			dis.skipBytes(width * (height - stepY * _thumbnailDimension.height));

			dis.skipBytes((depth - 1) * width * height);
			for (int y = 0; y < _thumbnailDimension.height; y++) {
				dis.readFully(row);
				dis.skipBytes(width * (stepY - 1));
				for (int x = 0; x < _thumbnailDimension.width; x++) {
					int color = parseBytes(row, x * stepX, 1);
					store[y][x] |= color << 8;
				}
			}
			dis.skipBytes(width * (height - stepY * _thumbnailDimension.height));

			dis.skipBytes((depth - 1) * width * height);
			for (int y = 0; y < _thumbnailDimension.height; y++) {
				dis.readFully(row);
				dis.skipBytes(width * (stepY - 1));
				for (int x = 0; x < _thumbnailDimension.width; x++) {
					int color = parseBytes(row, x * stepX, 1);
					store[y][x] |= color;
				}
			}
			break;

		case PO_IMC3DSL:
			break;

		case PO_IMC3DSF:
			break;
		default:
			break;

		}

		return store;
	}

	/**
	 * Reads image data from Stream.
	 * Parses data for Pandore format Close input
	 * stream when done.
	 * 
	 * @param is input stream.
	 */
	public void readImage( InputStream imageStream ) throws AWTException {
		DataInputStream dis = null;
		BufferedInputStream bis = null;

		try {
			// Try a supported image.
			bis = new BufferedInputStream(imageStream);
			dis = new DataInputStream(bis);

			readHeader(dis);

			switch (_potype) {
			case PO_IMG2DUC:
			case PO_IMG2DSL:
			case PO_IMG2DSF:
			case PO_REG2D:
				_store = read2dGrayData(dis);
				break;

			case PO_IMC2DUC:
			case PO_IMC2DSL:
			case PO_IMC2DSF:
				_store = read2dColorData(dis);
				break;

			case PO_IMG3DUC:
			case PO_IMG3DSL:
			case PO_IMG3DSF:
			case PO_REG3D:
				_store = read3dGrayData(dis);
				break;

			case PO_IMC3DUC:
			case PO_IMC3DSL:
			case PO_IMC3DSF:
				_store = read3dColorData(dis);
				break;
			default:
				break;
			}
		} catch (IOException io) {
			_loadError = true;
		} finally {
			try {
				if (dis != null) {
					dis.close();
				}
				if (bis != null) {
					bis.close();
				}
				if (imageStream != null) {
					imageStream.close();
				}
			} catch (IOException io) {
			}
		}
	}
}