/* -*- c-basic-offset: 3 -*-
 *
 * Copyright (c) 2013, GREYC.
 * All rights reserved
 *
 * You may use this file under the terms of the BSD license as follows:
 *
 * "Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of the GREYC, nor the name of its
 *     contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
 * 
 *   
 * For more information, refer to:
 * https://clouard.users.greyc.fr/Pandore
 */

package fr.greyc.pantheon.ariane.utils;

import fr.greyc.pantheon.ariane.Ariane;
import fr.greyc.pantheon.ariane.graph.item.GraphItem;
import fr.greyc.pantheon.ariane.ui.ILocalizable;
import fr.greyc.pantheon.ariane.ui.PreferenceChooser;
import fr.greyc.pantheon.ariane.xml.XMLElement;

import java.awt.Cursor;
import java.awt.GridLayout;
import java.awt.Point;
import java.awt.Toolkit;
import java.awt.dnd.DnDConstants;
import java.awt.dnd.DragGestureEvent;
import java.awt.dnd.DragGestureListener;
import java.awt.dnd.DragSource;
import java.awt.dnd.DragSourceDragEvent;
import java.awt.dnd.DragSourceDropEvent;
import java.awt.dnd.DragSourceEvent;
import java.awt.dnd.DragSourceListener;
import java.awt.dnd.DropTargetDragEvent;
import java.awt.event.KeyEvent;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JTree;
import javax.swing.KeyStroke;
import javax.swing.tree.DefaultTreeCellRenderer;
import javax.swing.tree.DefaultTreeModel;
import javax.swing.tree.TreePath;
import javax.swing.tree.TreeSelectionModel;

/**
 * The class <code>Tree</code> contains all the available graph items
 * and organized them in a tree.
 * This tree is displayed in the explorer from where the user can drag and
 * drop items into the graph editor.
 * 
 * @author Régis CLOUARD - Oct 01, 1998
 * @version Sébastien LEFÈVRE, Richard LOBRÉAU - Feb 21, 1999
 * @version Adrien SCHRÈQUE, Hicham AMADI - Sep 05, 2000
 * @version Mathieu JOETS - Jul 09, 2001
 * @version Régis CLOUARD - Oct 16, 2001
 * @version Régis CLOUARD - Dec 03, 2004
 * @version Régis CLOUARD - Dec 12, 2008
 * @version Régis CLOUARD - July 17, 2010 (add the macro-operator library)
 */
public final class Tree extends JPanel implements DragSourceListener, DragGestureListener, ILocalizable {
	/** The displayed tree. */
	private JTree _jtree;

	/** The root of the tree item. */
	private MutableTreeItem _rootNode;

	/** The model of the tree. */
	private DefaultTreeModel _treeModel;

	/** The root of the user-defined operators subtree. */
	private MutableTreeItem _userOperatorSubtree;

	/** The root of the common operators subtree. */
	private MutableTreeItem _commonOperatorSubtree;

	/** The root of the user-defined subroutines subtree. */
	private MutableTreeItem _subRoutineSubtree;

	/** The root of the built-in item subtree. */
	private MutableTreeItem _builtinSubtree;

	/** The drag site. */
	private DragSource _dragSource;

	/** The name of the file that contains the documentation index. */
	private String _indexDoc;

	/** The cursor for drag and drop. */
	private static final Cursor ITEM_CURSOR;

	static {
		ITEM_CURSOR = Toolkit.getDefaultToolkit().createCustomCursor(Ariane.getIcon("itemcursor.png").getImage(), new Point(0, 0), "itemCursor");
	}

	/**
	 * Creates the tree of available graph items from
	 * the given XML tree.
	 * 
	 * @param readItems the XML tree of graph items.
	 */
	public Tree( final List<XMLElement> readItems ) {
		initContents();
		if (readItems != null) {
			setCommonOperators(readItems);
		}
		setInternationalText();
	}

	private void addBuiltInItems( MutableTreeItem tree) {
		MutableTreeItem fileItem;
		
		// Add the various built-in items.
		TreeItem treeItem;
		treeItem = new TreeItem(TreeItem.Item.LOADER, "loader", "loader.html", null, null, 0, 0, 1, "");
		fileItem = new MutableTreeItem(treeItem, false);
		tree.add(fileItem);

		treeItem = new TreeItem(TreeItem.Item.SAVER, "saver", "saver.html", null, null, 1, 0, 0, "");
		fileItem = new MutableTreeItem(treeItem, false);
		tree.add(fileItem);

		treeItem = new TreeItem(TreeItem.Item.VIEWER, "viewer", "viewer.html", null, null, 1, 0, 0, "");
		fileItem = new MutableTreeItem(treeItem, false);
		tree.add(fileItem);

		treeItem = new TreeItem(TreeItem.Item.FILENAME, "filename", "filename.html", null, null, 0, 0, 0, "");
		fileItem = new MutableTreeItem(treeItem, false);
		tree.add(fileItem);

		treeItem = new TreeItem(TreeItem.Item.CONSTANT, "constant", "constant.html", null, null, 0, 0, 0, "");
		fileItem = new MutableTreeItem(treeItem, false);
		tree.add(fileItem);

		treeItem = new TreeItem(TreeItem.Item.SCRIPT, "script", "script.html", null, null, 0, 0, 0, "");
		fileItem = new MutableTreeItem(treeItem, false);
		tree.add(fileItem);

		treeItem = new TreeItem(TreeItem.Item.CONTROLEVAL, "eval", "eval.html", null, null, 0, 2, 0, "");
		fileItem = new MutableTreeItem(treeItem, false);
		tree.add(fileItem);

		treeItem = new TreeItem(TreeItem.Item.CONTROLFORMAT, "format", "format.html", null, null, 0, 2, 0, "");
		fileItem = new MutableTreeItem(treeItem, false);
		tree.add(fileItem);

		treeItem = new TreeItem(TreeItem.Item.SYSTEM, "system", "system.html", null, null, 0, 0, 0, "");
		fileItem = new MutableTreeItem(treeItem, false);
		tree.add(fileItem);

		treeItem = new TreeItem(TreeItem.Item.CONTROLIF, "if", "if.html", null, null, 0, 0, 0, "");
		fileItem = new MutableTreeItem(treeItem, false);
		tree.add(fileItem);

		treeItem = new TreeItem(TreeItem.Item.CONTROLSELECT, "select", "select.html", null, null, 0, 0, 1, "$0:index");
		fileItem = new MutableTreeItem(treeItem, false);
		tree.add(fileItem);

		treeItem = new TreeItem(TreeItem.Item.CONTROLMERGE, "merge", "merge.html", null, null, 4, 0, 0, "");
		fileItem = new MutableTreeItem(treeItem, false);
		tree.add(fileItem);

		treeItem = new TreeItem(TreeItem.Item.CONTROLFOR, "for", "for.html", null, null, 3, 0, 0, "$0:initial value;$1:final value;$2:increment");
		fileItem = new MutableTreeItem(treeItem, false);
		tree.add(fileItem);

		treeItem = new TreeItem(TreeItem.Item.CONTROLWHILE, "while", "while.html", null, null, 2, 0, 0, "");
		fileItem = new MutableTreeItem(treeItem, false);
		tree.add(fileItem);

		treeItem = new TreeItem(TreeItem.Item.MACROOPERATOR, "macro-operator", "macro.html", null, null, 0, 0, 0, "");
		fileItem = new MutableTreeItem(treeItem, false);
		tree.add(fileItem);

		treeItem = new TreeItem(TreeItem.Item.NOTE, "note", "note.html", null, null, 0, 0, 0, "");
		fileItem = new MutableTreeItem(treeItem, false);
		tree.add(fileItem);
	}

	/**
	 * Creates permanent items in the tree such as
	 * Constant, ControlStructure, MacroOperator ...).
	 * Other items are read from the configuration files.
	 */
	private void initContents( ) {
		_rootNode = new MutableTreeItem("Library", true);

		// Sub categories.
		_userOperatorSubtree = new MutableTreeItem(null, true);
		_subRoutineSubtree = new MutableTreeItem(null, true);
		_commonOperatorSubtree = new MutableTreeItem(null, true);
		_builtinSubtree = new MutableTreeItem(null, true);

		addBuiltInItems(_builtinSubtree);

		_rootNode.add(_builtinSubtree);
		_rootNode.add(_commonOperatorSubtree);
		_rootNode.add(_subRoutineSubtree);
		_rootNode.add(_userOperatorSubtree);

		// Creates explorer with a Jtree.
		_treeModel = new DefaultTreeModel(_rootNode);
		_jtree = new JTree(_treeModel);
		_jtree.setFont(PreferenceChooser.getDefaultFont());
		_jtree.getSelectionModel().setSelectionMode(TreeSelectionModel.DISCONTIGUOUS_TREE_SELECTION);
		_jtree.putClientProperty("JTree.lineStyle", "Angled");
		final DefaultTreeCellRenderer renderer = new DefaultTreeCellRenderer();
		renderer.setLeafIcon(Ariane.getIcon("item.png"));
		renderer.setOpenIcon(null);
		renderer.setClosedIcon(null);
		_jtree.setCellRenderer(renderer);

		_dragSource = DragSource.getDefaultDragSource();
		_dragSource.createDefaultDragGestureRecognizer(_jtree, DnDConstants.ACTION_MOVE, this);

		setLayout(new GridLayout());
		setBackground(_jtree.getBackground());

		// Listeners event from the tree.
		final MouseListener listeners = new MouseAdapter() {
			// Case of double-click -> display manual.
			@Override
			public void mousePressed( MouseEvent event ) {
				if (event.getClickCount() > 1) {
					TreePath selectedPath = _jtree.getPathForLocation(event.getX(), event.getY());
					if (selectedPath != null) {
						MutableTreeItem selectedNode = (MutableTreeItem)selectedPath.getLastPathComponent();
						if (selectedNode.isLeaf() && selectedNode.getUserObject() instanceof TreeItem) {
							String doc = ((TreeItem)selectedNode.getUserObject()).getDocURL();
							if (doc != null) {
								Ariane.getWorkspace().displayMan(Ariane.getString("HTMLViewer.manual") + " " + selectedNode.toString(), doc, null);
							}
						}
					}
				}
			}
		};
		_jtree.addMouseListener(listeners);

		// Discard key CTRL-A (used by the Graph).
		_jtree.getInputMap().put(KeyStroke.getKeyStroke(KeyEvent.VK_A, KeyEvent.CTRL_MASK), "none");

		add(_jtree);
	}

	/**
	 * Highlights all operators whose name
	 * contains the specified pattern.
	 * 
	 * @param pattern a part of operator name.
	 * @return true if some operators has been found.
	 */
	public boolean search( final String pattern ) {
		List<TreePath> paths = search(_rootNode, pattern);
		if (!paths.isEmpty()) {
			TreePath[] pathsTab = new TreePath[paths.size()];
			for (int i = 0; i < paths.size(); i++) {
				pathsTab[i] = paths.get(i);
			}
			// why paths.toArray(); doesn't work ?
			_jtree.setSelectionPaths(pathsTab);
			return true;
		}
		return false;
	}

	/**
	 * Returns the tree item with the specified name
	 * or null.
	 * 
	 * @param name a tree item name.
	 * @param category the related category (or null).
	 * @return the tree item or null.
	 */
	public TreeItem getTreeItem( String name, String category ) {
		return getNode(_rootNode, name, category);
	}

	/**
	 * Returns the node with the specified name.
	 * 
	 * @param node the root node or the search.
	 * @param name the name to search.
	 * @param category the related category.
	 * @return the path to the node or null.
	 */
	private TreeItem getNode( MutableTreeItem node, String name, String category ) {
		for (final Enumeration<?> e = node.children(); e.hasMoreElements();) {
			MutableTreeItem subnode = (MutableTreeItem)e.nextElement();
			if (subnode.isLeaf()) {
				String operatorName = subnode.getUserObject().toString();
				if (operatorName.equals(name)) {
					TreeItem treeItem = (TreeItem)subnode.getUserObject();
					if (category == null || category.equals(treeItem.getCategory())) {
						return treeItem;
					}
				}
			} else {
				TreeItem item = getNode(subnode, name, category);
				if (item != null) {
					return item;
				}
			}
		}
		return null;
	}

	/**
	 * Returns the list of path that locate all
	 * node whose name contains the specified
	 * pattern.
	 * 
	 * @param node the root node or the search.
	 * @param pattern a string.
	 * @return the list of paths.
	 */
	public List<TreePath> search( MutableTreeItem node, String pattern ) {
		ArrayList<TreePath> paths = new ArrayList<TreePath>();
		for (final Enumeration<?> e = node.children(); e.hasMoreElements();) {
			MutableTreeItem subnode = (MutableTreeItem)e.nextElement();
			if (subnode.isLeaf()) {
				String operatorName = subnode.getUserObject().toString();
				if (operatorName.contains(pattern)) {
					paths.add(new TreePath(subnode.getPath()));
				}
			} else {
				paths.addAll(search(subnode, pattern));
			}
		}
		return paths;
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void setInternationalText( ) {
		_subRoutineSubtree.setUserObject(Ariane.getString("Tree.subroutines"));
		_userOperatorSubtree.setUserObject(Ariane.getString("Tree.personalOperators"));
		_commonOperatorSubtree.setUserObject(Ariane.getString("Tree.operators"));
		_builtinSubtree.setUserObject(Ariane.getString("Tree.builtinItems"));
		_treeModel.reload();
	}

	/**
	 * Gets the documentation index page.
	 * 
	 * @return the relative path to the page.
	 */
	public String getIndexDoc( ) {
		return _indexDoc;
	}

	/**
	 * Returns the list of personal operators
	 * as an XML tree.
	 * 
	 * @return a LinkedList of operators.
	 */
	public List<TreeItem> getUserOperatorList( ) {
		LinkedList<TreeItem> operatorList = new LinkedList<TreeItem>();
		Enumeration<?> node = _userOperatorSubtree.depthFirstEnumeration();
		while (node.hasMoreElements()) {
			MutableTreeItem leaf = (MutableTreeItem)node.nextElement();
			if (leaf.getUserObject() instanceof TreeItem) {
				operatorList.add((TreeItem)leaf.getUserObject());
			}
		}
		return operatorList;
	}

	/**
	 * Returns the tree of personal operators.
	 * 
	 * @return the root of the tree.
	 */
	public MutableTreeItem getUserOperatorTree( ) {
		return _userOperatorSubtree;
	}

	/**
	 * Returns the tree of subroutines.
	 * 
	 * @return the root of the tree.
	 */
	public MutableTreeItem getSubroutineTree( ) {
		return _subRoutineSubtree;
	}

	/**
	 * Returns the tree of subroutines.
	 * 
	 * @return the root of the tree.
	 */
	public MutableTreeItem getCommonOperatorTree( ) {
		return _commonOperatorSubtree;
	}

	/**
	 * Sets a new version for the tree of personal operators.
	 * Deletes the previous one first.
	 * 
	 * @param newTree the new tree.
	 */
	public void setUserOperatorTree( MutableTreeItem newTree ) {
		int index = _treeModel.getIndexOfChild(_rootNode, _userOperatorSubtree);
		_treeModel.removeNodeFromParent(_userOperatorSubtree);
		_userOperatorSubtree = newTree;
		_treeModel.insertNodeInto(_userOperatorSubtree, _rootNode, index);
	}

	/**
	 * Sets a new version for the tree of personal operators.
	 * Deletes the previous one first.
	 * 
	 * @param newTree the new tree.
	 */
	public void setSubroutineTree( MutableTreeItem newTree ) {
		int index = _treeModel.getIndexOfChild(_rootNode, _subRoutineSubtree);
		_treeModel.removeNodeFromParent(_subRoutineSubtree);
		_subRoutineSubtree = newTree;
		_treeModel.insertNodeInto(_subRoutineSubtree, _rootNode, index);
	}

	/**
	 * Sets the specified tree as the operator tree.
	 * 
	 * @param readItems the XML tree of graph items.
	 */
	public void setCommonOperators( List<XMLElement> readItems ) {
		List<TreeItem> treeItems = load(readItems);
		addCommonOperators(treeItems);
	}

	/**
	 * Adds the list of common operators stored in an XML tree
	 * in the explorer.
	 * 
	 * @param operatorList the XML list of operators.
	 */
	private void addCommonOperators( List<TreeItem> operatorList ) {
		_commonOperatorSubtree.removeAllChildren();
		addItem(operatorList, _commonOperatorSubtree);
	}

	/**
	 * Adds the list of personal operators stored in an XML tree
	 * in the explorer.
	 * 
	 * @param operatorList the XML list of operators.
	 */
	public void addUserOperators( List<XMLElement> operatorList ) {
		addItem(load(operatorList), _userOperatorSubtree);
	}

	/**
	 * Adds the list of subroutines stored in an XML tree
	 * in the explorer.
	 * 
	 * @param operatorList the XML list of subroutines.
	 * @return the root of the tree items.
	 */
	public MutableTreeItem addSubroutines( List<XMLElement> operatorList ) {
		addItem(load(operatorList), _subRoutineSubtree);
		return _subRoutineSubtree;
	}

	/**
	 * Adds a new subroutine stored in the explorer.
	 * 
	 * @param item the item to add in the tree.
	 * @throws Exception problem.
	 */
	public boolean addSubroutine( TreeItem item ) {
		LinkedList<TreeItem> treeItems = new LinkedList<TreeItem>();
		treeItems.add(item);
		return addItem(treeItems, _subRoutineSubtree);
	}

	/**
	 * Returns the list of subroutines
	 * as an XML tree.
	 * 
	 * @return a LinkedList of subroutines.
	 */
	public List<TreeItem> getSubroutineList( ) {
		List<TreeItem> subRoutineList = new LinkedList<TreeItem>();
		Enumeration<?> node = _subRoutineSubtree.depthFirstEnumeration();
		while (node.hasMoreElements()) {
			MutableTreeItem leaf = (MutableTreeItem)node.nextElement();
			if (leaf.getUserObject() instanceof TreeItem) {
				subRoutineList.add((TreeItem)leaf.getUserObject());
			}
		}
		return subRoutineList;
	}

	/**
	 * Returns the list of subroutine categories.
	 * 
	 * @return a list of category name.
	 */
//	@SuppressWarnings("unchecked")
	public String[] getSubroutineCategories( ) {
		String[] categories = new String[_subRoutineSubtree.getChildCount()];
		for (int i = 0; i < _subRoutineSubtree.getChildCount(); i++) {
			categories[i] =  _subRoutineSubtree.getChildAt(i).toString();
		}
		return categories;
	}

	/**
	 * Adds the list of items <i>itemList</i> in the tree
	 * under the category <i>node</i>.
	 * 
	 * @param itemList the list of items.
	 * @param node the root of the future subtree (category).
	 */
	private boolean addItem( List<TreeItem> itemList, MutableTreeItem node ) {
		for ( TreeItem treeItem : itemList ) {
			MutableTreeItem item = new MutableTreeItem(treeItem, false);
			String categoryName = treeItem.getCategory();
			MutableTreeItem category = node.getNode(categoryName);
			if (category == null) { // create a new category.
				category = new MutableTreeItem( categoryName, true);
				node.add(category); // new subroutine
			}

			if (category.getNode(treeItem.getName()) != null
					&& Ariane.getLanguage() != null
					&& Ariane.getFrame() != null
					&& Ariane.getFrame().isVisible()) { // Name Conflict
				if (JOptionPane.showConfirmDialog(null,
						treeItem.getName() + " " + Ariane.getString("Tree.dialog.overwriteMessage"),
						Ariane.getString("Tree.dialog.overWriteTitle"),
						JOptionPane.YES_NO_OPTION) == JOptionPane.YES_OPTION) {
					category.remove(category.getNode(treeItem.getName()));
					category.add(item); // replace existing subroutine
					_treeModel.reload(node);
					Ariane.getWorkspace().printWarning(Ariane.getString("CompositeItem.message2"));
					return true;
				} else {
					return false;
				}
			} else {
				category.add(item);
				_treeModel.reload(node);
			}
		}
		return true;
	}

	///// 
	///// D R A G   &   D R O P
	/////

	/**
	 * Called when the user has modified the drop gesture.
	 * 
	 * @param event the DragSourceDragEvent.
	 */
	public void dropActionChanged( DropTargetDragEvent event ) {
		// Do nothing
	}

	/**
	 * Invoked to signify that the Drag and Drop operation is complete.
	 * Sets the border to normal looking.
	 * 
	 * @param event the DragSourceDropEvent
	 */
	@Override
	public void dragDropEnd( DragSourceDropEvent event ) {
		// Do nothing
	}

	/**
	 * Called as the cursor's hotspot enters a platform-dependent drop site.
	 * Changes the cursor.
	 * 
	 * @param event the DragSourceDragEvent.
	 */
	@Override
	public void dragEnter( DragSourceDragEvent event ) {
		// Do nothing
	}

	/**
	 * Called as the cursor's hot spot exits a platform-dependent drop site.
	 * Do nothing.
	 * 
	 * @param event the DragSourceEvent.
	 */
	@Override
	public void dragExit( DragSourceEvent event ) {
		// Do nothing
	}

	/**
	 * Called as the cursor's hot spot moves over a platform-dependent
	 * drop site (a Graph or a Slot).
	 * 
	 * @param event the DragSourceDragEvent.
	 */
	@Override
	public void dragOver( DragSourceDragEvent event ) {
		// Do nothing
	}

	/**
	 * Called when the user has modified the drop gesture.
	 * 
	 * @param event the DragSourceDragEvent.
	 */
	@Override
	public void dropActionChanged( DragSourceDragEvent event ) {
		// Do nothing
	}

	/**
	 * A DragGestureRecognizer has detected a platform-dependent drag
	 * initiating gesture and is notifying this listener in order for
	 * it to initiate the action for the user.
	 * 
	 * @param event the gesture that has just occurred.
	 */
	@Override
	public void dragGestureRecognized( DragGestureEvent event ) {
		// Detect the select tree item.
		TreePath selectedPath = _jtree.getSelectionPath();
		if (selectedPath != null) {
			MutableTreeItem selectedNode = (MutableTreeItem)selectedPath.getLastPathComponent();
			// Is a valid drag only if it is a leaf.
			if (selectedNode.isLeaf() && selectedNode.getUserObject() instanceof TreeItem) {
				TreeItem selected = (TreeItem)selectedNode.getUserObject();
				_dragSource.startDrag(event, ITEM_CURSOR, selected, this);
			}
		}
	}

	///// 
	///// I N P U T   /   O U P U T
	/////

	/**
	 * Loads the tree from an XML tree.
	 * 
	 * @param xmlTree the XNML tree of items.
	 * @return the list of all operators.
	 */
	private List<TreeItem> load( List<XMLElement> xmlTree ) {
		List<TreeItem> treeItems = new LinkedList<TreeItem>();

		// Discard tag <ARIANE>.
		List<XMLElement> xmlContents = xmlTree.get(0).getChildren();

		if (xmlContents == null) {
			return treeItems;
		}

		Iterator<XMLElement> it = xmlContents.iterator();
		XMLElement element;
		while (it.hasNext()) {
			element = it.next();
			if (element.getName().equalsIgnoreCase("OperatorLibrary")
					|| element.getName().equalsIgnoreCase("Library")
					|| element.getName().equalsIgnoreCase("Bibliotheque")) {
				xmlContents = element.getChildren();
				it = xmlContents.iterator();
				while (it.hasNext()) {
					element = it.next();
					// kept for sake of compatibility
					if (element.getName().equalsIgnoreCase("DocPath")) {
						// kept for sake of compatibility
						Ariane.setDocPath(element.getValue());
					} else if (element.getName().equalsIgnoreCase("CommandStatus")) {
						GraphItem.setCommandStatus(element.getValue());
					} else if (element.getName().equalsIgnoreCase("CommandVisu")) {
						GraphItem.setCommandVisu(element.getValue());
					} else if (element.getName().equalsIgnoreCase("Version")) {
						// kept for sake of compatibility
						// Gets the version number of library.
						_rootNode.setUserObject(element.getValue());
						_treeModel.reload(_rootNode);
						if (element.getValue().startsWith("Pandore")) {
							GraphItem.setCommandStatus("pstatus");
							GraphItem.setCommandVisu("pvisu");
						}
					} else if (element.getName().equalsIgnoreCase("Index")) {
						// kept for sake of compatibility
						// Gets the documentation index page.
						_indexDoc = element.getValue();
					} else if (element.getName().equalsIgnoreCase("Operator") || element.getName().equalsIgnoreCase("Operateur")) {
						treeItems.add(new TreeItem(element.getChildren()));
					}
				}
				// subroutines
			} else if (element.getName().equalsIgnoreCase("SubroutineLibrary")) {
				xmlContents = element.getChildren();
				it = xmlContents.iterator();
				while (it.hasNext()) {
					element = it.next();
					// kept for sake of compatibility
					if (element.getName().equalsIgnoreCase("Subroutine")) {
						treeItems.add(new SubroutineTreeItem(element.getChildren()));
					}
				}
			}
		}
		return treeItems;
	}
}