/* -*- c-basic-offset: 3 -*-
 *
 * Copyright (c) 2013, GREYC.
 * All rights reserved
 *
 * You may use this file under the terms of the BSD license as follows:
 *
 * "Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of the GREYC, nor the name of its
 *     contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
 * 
 *   
 * For more information, refer to:
 * https://clouard.users.greyc.fr/Pandore
 */

package fr.greyc.pantheon.ariane.utils;

import java.awt.datatransfer.DataFlavor;
import java.awt.datatransfer.Transferable;
import java.awt.datatransfer.UnsupportedFlavorException;
import java.io.IOException;
import java.util.List;

import fr.greyc.pantheon.ariane.graph.Graph;
import fr.greyc.pantheon.ariane.graph.item.ConstantItem;
import fr.greyc.pantheon.ariane.graph.item.DataFile;
import fr.greyc.pantheon.ariane.graph.item.EvalItem;
import fr.greyc.pantheon.ariane.graph.item.FormatItem;
import fr.greyc.pantheon.ariane.graph.item.ControlStructureForItem;
import fr.greyc.pantheon.ariane.graph.item.ControlStructureIfItem;
import fr.greyc.pantheon.ariane.graph.item.ControlStructureMergeItem;
import fr.greyc.pantheon.ariane.graph.item.ControlStructureSelectItem;
import fr.greyc.pantheon.ariane.graph.item.ControlStructureWhileItem;
import fr.greyc.pantheon.ariane.graph.item.FileNameItem;
import fr.greyc.pantheon.ariane.graph.item.LoaderItem;
import fr.greyc.pantheon.ariane.graph.item.SaverItem;
import fr.greyc.pantheon.ariane.graph.item.GraphItem;
import fr.greyc.pantheon.ariane.graph.item.MacroOperatorItem;
import fr.greyc.pantheon.ariane.graph.item.NoteItem;
import fr.greyc.pantheon.ariane.graph.item.OperatorItem;
import fr.greyc.pantheon.ariane.graph.item.ScriptItem;
import fr.greyc.pantheon.ariane.graph.item.SystemItem;
import fr.greyc.pantheon.ariane.graph.item.ViewerItem;
import fr.greyc.pantheon.ariane.ui.PreferenceChooser;
import fr.greyc.pantheon.ariane.xml.XMLElement;

/**
 * A <code>TreeItem</code> is an item that can be inserted in
 * the tree of the explorer.
 *
 * @author Régis CLOUARD - Oct 01, 1998
 * @version Sébastien LEFÈVRE, Richard LOBRÉAU - Feb 21, 1999
 * @version Adrien SCHRÈQUE, Hicham AMADI - Nov 05, 2000
 * @version Mathieu JOETS - Sep 09, 2001
 * @version Régis CLOUARD - Oct 16, 2001 (add datafile item)
 * @version Régis CLOUARD - Dec 27, 2011 (add default values)
 * @version Régis CLOUARD - Jul 07, 2012 (add new item script)
 */
public class TreeItem implements Transferable {
	/** The type of the item (see magic numbers). */
	private Item _typeItem;
	/** The name of the tree item (language independent). */
	private String _name;
	/** The path to the related documentation page. */
	private String _docURL;
	/** The path to the binary. */
	private String _binary;
	/** The related category. */
	private String _category;
	/** The signature is the list of parameters names and default values, eg. $0:toto=1;$1:titi=2. */
	private String _signature;
	private int _inputDataSlots;
	private int _inputParameterSlots;
	private int _outputDataSlots;

	/** The magic number for transferable item. */
	public static final DataFlavor[] FLAVORS = { new DataFlavor("application/x-java-serialized-item-tree", "TreeItem Flavor") };

	/** The item magic numbers */ 
	public enum Item {
		/** Operator */
		OPERATOR,
		/** A file loader. */
		LOADER,
		/** A file saver. */
		SAVER,
		/** A file viewer. */
		VIEWER,
		/** A file name. */
		FILENAME,
		/** A constant. */
		CONSTANT,
		/** A macro operator. */
		MACROOPERATOR,
		/** An if control structure. */
		CONTROLIF,
		/** An input selector. */
		CONTROLSELECT,
		/** An iterative control structure. */
		CONTROLFOR,
		/** A repetitive control structure. */
		CONTROLWHILE,
		/** Merge two inputs into one output. */
		CONTROLMERGE,
		/** An calculus unit. */
		CONTROLEVAL,
		/** A string manipulator unit. */
		CONTROLFORMAT,
		/** A comment. */
		NOTE,
		/** Execute an operating system command. */
		SYSTEM,
		/** A script. */
		SCRIPT,
	}

	/**
	 * Creates a new tree item with the specified name and URL.
	 */
	public TreeItem( Item typeItem, String name, String docURL, String binary, String category,
			int inputDataSlots, int inputParameterSlots, int outputDataSlots, String signature ) {
		_typeItem = typeItem;
		_name = name;
		_docURL = docURL;
		_binary = binary;
		_category = category;
		_inputParameterSlots = inputParameterSlots;
		_inputDataSlots = inputDataSlots;
		_outputDataSlots = outputDataSlots;
		_signature = signature;
	}

	/**
	 * Creates a new TreeItem from values in the given XML tree.
	 * 
	 * @param xmlTree the XML tree of properties.
	 * @throws Exception any Exception.
	 */
	public TreeItem( List<XMLElement> xmlTree ) {
		load(xmlTree);
	}

	/**
	 * @return the name of the item.
	 */
	public final String getName( ) {
		return _name;
	}

	/**
	 * Sets the name of the item.
	 * 
	 * @param text the new text.
	 */
	public final void setName( String text ) {
		_name = text;
	}

	/**
	 * Sets the type to the tree item.
	 * 
	 * @param type the new type.
	 */
	protected final void setType( Item type ) {
		_typeItem = type;
	}

	/**
	 * @return the tree item type.
	 */
	protected final Item getType( ) {
		return _typeItem;
	}

	/**
	 * @return the address of the documentation page.
	 */
	public final String getDocURL( ) {
		return _docURL;
	}

	/**
	 * Sets the address of the documentation page.
	 * 
	 * @param text the new text.
	 */
	public final void setDocURL( String text ) {
		_docURL = text;
	}

	/**
	 * @return the category.
	 */
	public final String getCategory( ) {
		return _category;
	}

	/**
	 * Sets the category.
	 * 
	 * @param text the new name.
	 */
	public final void setCategory( String text ) {
		_category = text;
	}

	/**
	 * @return the path to the binary.
	 */
	public final String getBinary( ) {
		return _binary;
	}

	/**
	 * Sets the path to the binary.
	 * 
	 * @param text the new binary.
	 */
	public final void setBinary( String text ) {
		String directory = PreferenceChooser.get(PreferenceChooser.BINARY_DIRECTORY);
		if (directory != null && directory.trim().isEmpty()) {
			String[] commandLine = text.split(OperatorItem.SEPARATOR);
			String binary = DataFile.getRelativePath(commandLine[0], directory);
			if (commandLine.length > 1) {
				_binary = binary + OperatorItem.SEPARATOR + commandLine[1];
			} else {
				_binary = binary + OperatorItem.SEPARATOR;
			}
		} else {
			_binary = text;
		}
	}

	/**
	 * @return the number of parameters.
	 */
	public final int getParameters( ) {
		return _inputParameterSlots;
	}

	/**
	 * Sets the number of parameters.
	 * 
	 * @param count the parameters count.
	 */
	public final void setParameters( int count ) {
		_inputParameterSlots = count;
	}

	public final String getSignature( ) {
		return _signature;
	}

	/**
	 * Sets the default values for the parameters.
	 * 
	 * @param signature (eg. $p0:toto=0; $p1:titi=1;).
	 */
	public final void setSignature( String signature ) {
		_signature = signature;
	}

	public final int getInputDataSlots( ) {
		return _inputDataSlots;
	}

	public final void setInputDataSlots( int count ) {
		_inputDataSlots = count;
	}

	public final int getOutputDataSlots( ) {
		return _outputDataSlots;
	}

	public final void setOutputDataSlots( int count ) {
		_outputDataSlots = count;
	}

	/**
	 * Returns the name of the item.
	 * Warning! this method is used to display the name in the tree.
	 * 
	 * @return the description of the node.
	 */
	@Override
	public final String toString( ) {
		return _name;
	}

	/**
	 * Gets the transfered object.
	 * 
	 * @return the list of flavors.
	 */
	@Override
	public DataFlavor[] getTransferDataFlavors( ) {
		return FLAVORS.clone();
	}

	/**
	 * Gets the transfered data.
	 * 
	 * @param flavor the flavor of the transfered object.
	 * @return the data.
	 * @throws UnsupportedFlavorException unknown flavors.
	 * @throws IOException in case of IO error.
	 */
	@Override
	public Object getTransferData( DataFlavor flavor ) throws UnsupportedFlavorException, IOException {
		if (isDataFlavorSupported(flavor)) {
			return this;
		} else {
			throw new UnsupportedFlavorException(flavor);
		}
	}

	/**
	 * Is the object the transferable object?
	 * 
	 * @param flavor the flavor to be tested.
	 * @return true if the flavor is transferable.
	 */
	@Override
	public boolean isDataFlavorSupported( DataFlavor flavor ) {
		return flavor.equals(FLAVORS[0]);
	}

	/**
	 * Instantiates an item in the given graph.
	 * 
	 * @param graph the graph where to add the item.
	 * @return the graph item.
	 */
	public GraphItem createGraphItem( Graph graph ) {
		try {
			switch(_typeItem) {
			case OPERATOR: return new OperatorItem(this, graph, _name, _category, _inputDataSlots, _inputParameterSlots, _outputDataSlots, _signature);
			case LOADER: return new LoaderItem(this, graph, _name, _category );
			case SAVER: return new SaverItem(this, graph, _name, _category );
			case VIEWER: return new ViewerItem(this, graph, _name, _category );
			case FILENAME: return new FileNameItem(this, graph, _name);
			case CONSTANT: return new ConstantItem(this, graph, _name);
			case MACROOPERATOR: return new MacroOperatorItem(this, graph);
			case CONTROLEVAL: return new EvalItem(this, graph, _name);
			case CONTROLFORMAT: return new FormatItem(this, graph, _name);
			case CONTROLIF: return new ControlStructureIfItem(this, graph, _name);
			case CONTROLSELECT: return new ControlStructureSelectItem(this, graph, _name);
			case CONTROLFOR: return new ControlStructureForItem(this, graph, _name);
			case CONTROLWHILE: return new ControlStructureWhileItem(this, graph, _name);
			case CONTROLMERGE: return new ControlStructureMergeItem(this, graph, _name);
			case NOTE: return new NoteItem(this, graph, _name);
			case SYSTEM: return new SystemItem(this, graph, _name);
			case SCRIPT: return new ScriptItem(this, graph, _name);
			default: return null;
			}
		} catch (InterruptedException e) {
			return null;
		}
	}

	///// 
	///// I N P U T  /  O U T P U T
	/////

	/**
	 * Saves the tree item as an XML Element.
	 * 
	 * @return the resulted XML tree.
	 * @throws IOException any Exception.
	 */
	public XMLElement save( ) throws IOException {
		XMLElement element = null;
		if (_typeItem == Item.OPERATOR) {
			element = new XMLElement("Operator");
			element.addChildren(new XMLElement("Category", getCategory()));
			element.addChildren(new XMLElement("Name", getName()));
			element.addChildren(new XMLElement("Documentation", getDocURL()));
			element.addChildren(new XMLElement("Binary", getBinary()));
			element.addChildren(new XMLElement("Parameters", Integer.toString(getParameters())));
			element.addChildren(new XMLElement("Signature", getSignature()));
			element.addChildren(new XMLElement("Inputs", Integer.toString(getInputDataSlots())));
			element.addChildren(new XMLElement("Outputs", Integer.toString(getOutputDataSlots())));
		}
		return element;
	}

	/**
	 * Loads properties of a tree item from an XML tree.
	 * 
	 * @param xmlelements the XML Tree of properties.
	 * @throws Exception any Exception.
	 */
	protected void load( List<XMLElement> xmlelements ) {
		setType(Item.OPERATOR);

		for (XMLElement item : xmlelements ) {
			if (item.getName().equalsIgnoreCase("Category") || item.getName().equalsIgnoreCase("Categorie")) {
				_category = item.getValue();
			} else if (item.getName().equalsIgnoreCase("Name") || item.getName().equalsIgnoreCase("Nom")) {
				_name = item.getValue();
			} else if (item.getName().equalsIgnoreCase("Documentation")) {
				_docURL = item.getValue();
			} else if (item.getName().equalsIgnoreCase("Binary") || item.getName().equalsIgnoreCase("Binaire")) {
				_binary = item.getValue();
			} else if (item.getName().equalsIgnoreCase("Parameters") || item.getName().equalsIgnoreCase("Parametre")) {
				_inputParameterSlots = Integer.parseInt(item.getValue());
			} else if (item.getName().equalsIgnoreCase("Signature")) {
				setSignature(item.getValue());
			} else if (item.getName().equalsIgnoreCase("Inputs") || item.getName().equalsIgnoreCase("Entree")) {
				_inputDataSlots = Integer.parseInt(item.getValue());
			} else if (item.getName().equalsIgnoreCase("Outputs") || item.getName().equalsIgnoreCase("Sortie")) {
				_outputDataSlots = Integer.parseInt(item.getValue());
			}
		}
	}
}